theory hobisim
  imports bisimulation bisimStructCong
begin

context env begin

nominal_primrec processFrame :: "'b frame \<Rightarrow> ('a,'b,'c) psi"
where
  "processFrame(FAssert \<Psi>) = \<lbrace>\<Psi>\<rbrace>"
| "processFrame(FRes n F) = \<lparr>\<nu>n\<rparr>(processFrame F)"
apply(finite_guess)+
apply(rule TrueI)+
apply(simp add: fresh_abs_funI2[OF pt_name_inst,OF at_name_inst])
by(fresh_guess+)

lemma processFrameEqvt[eqvt]:
  fixes p::"name prm"
  shows "(p\<bullet>processFrame F) = processFrame (p\<bullet>F)"
  apply(nominal_induct F avoiding: p rule: frame.strong_inducts)
  by(simp add: eqvts)+

lemma processFrameFresh[simp]:
  fixes a::name
  assumes "a \<sharp> F"
  shows   "a \<sharp> processFrame F"
  using assms
proof(nominal_induct F avoiding: a rule: frame.strong_inducts)
  case FAssert thus ?case by simp
next
  case(FRes n F)
  hence "a \<sharp> F"
    apply(rule_tac fresh_abs_funE[OF pt_name_inst,OF at_name_inst])
    by(finite_guess) simp+
  thus ?case using FRes
    apply simp
    apply(rule fresh_abs_funI1[OF pt_name_inst, OF at_name_inst])
    apply(finite_guess)
    by simp+
qed

lemma processFrameSupp: "supp(processFrame F) \<subseteq> (supp F::name set)"
  using processFrameFresh unfolding fresh_def by auto

lemma processFrameFreshChain[simp]:
  fixes xvec::"name list"
  assumes "xvec \<sharp>* F"
  shows   "xvec \<sharp>* processFrame F"
using assms
by(induct xvec) simp+

lemma processFrameChain[simp]:
  "processFrame(\<langle>A,\<Psi>\<rangle>) = \<lparr>\<nu>*A\<rparr>(\<lbrace>\<Psi>\<rbrace>)"
  by(induct A) simp+

lemma extractProcessFrame[simp]:
  shows "extractFrame(processFrame F) = F"
  by(nominal_induct F rule: frame.strong_inducts) simp+

definition \<F>::"('a,'b,'c) psi \<Rightarrow> ('a,'b,'c) psi" where
"\<F> P = processFrame(extractFrame P)"

lemma \<F>eqvt[eqvt]:
  fixes p::"name prm"
  shows "(p\<bullet>\<F> P) = \<F>(p\<bullet>P)"
  unfolding \<F>_def by(simp add: eqvts)

lemma \<F>fresh[simp]:
  fixes a::name
  assumes "a \<sharp> P"
  shows   "a \<sharp> \<F> P"
  using assms unfolding \<F>_def
  by(simp add: extractFrameFresh)

lemma \<F>freshChain[simp]:
  fixes xvec::"name list"
  assumes "xvec \<sharp>* P"
  shows   "xvec \<sharp>* \<F> P"
  using assms
  by(induct xvec) (simp)+

lemma \<F>supp: "supp(\<F> P) \<subseteq> (supp P::name set)"
  using \<F>fresh unfolding fresh_def by auto

lemma \<F>frame[simp]:
  "extractFrame(\<F> P) = extractFrame P"
  unfolding \<F>_def by simp

lemma \<F>simps[simp]:
  assumes "extractFrame P = \<langle>A,\<Psi>\<rangle>"
  shows "(\<F>(P)) = \<lparr>\<nu>*A\<rparr>\<lbrace>\<Psi>\<rbrace>"
  using assms
  unfolding \<F>_def
  by simp

lemma resNilTransAux:
  assumes "\<Psi> \<rhd> \<lparr>\<nu>a\<rparr>P \<longmapsto> \<alpha> \<prec> P'"
  and "bn \<alpha> \<sharp>* subject \<alpha>"
  and elm: "\<And> (p::name prm) R. \<Psi> \<rhd> (p\<bullet>P) \<longmapsto> R \<Longrightarrow> False"

  shows False
proof -
  obtain b::name where "b \<sharp> \<Psi>" and "b \<noteq> a" and "b \<sharp> \<Psi>" and "b \<sharp> \<alpha>" and "b \<sharp> P'" and "b \<sharp> P"
    by(generate_fresh "name") auto
  moreover obtain p::"name prm" where "set p \<subseteq> set(bn \<alpha>) \<times> set(bn(p\<bullet>\<alpha>))" and "(bn(p\<bullet>\<alpha>)) \<sharp>* \<Psi>" and "(bn(p\<bullet>\<alpha>)) \<sharp>* \<alpha>" and "(bn(p\<bullet>\<alpha>)) \<sharp>* P'" and "(bn(p\<bullet>\<alpha>)) \<sharp>* P" and "(bn(p\<bullet>\<alpha>)) \<sharp>* b" and "(bn(p\<bullet>\<alpha>)) \<sharp>* ([(a,b)]\<bullet>P)" and "(bn(p\<bullet>\<alpha>)) \<sharp>* a" and "distinctPerm p"
    by(rule_tac name_list_avoiding[where c="(\<Psi>,\<alpha>,P',P,b,a)" and xvec="bn \<alpha>"]) (auto simp add: eqvts)
  ultimately have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>([(a,b)]\<bullet>P) \<longmapsto> (p\<bullet>\<alpha>) \<prec> p\<bullet>P'" using `\<Psi> \<rhd> \<lparr>\<nu>a\<rparr>P \<longmapsto> \<alpha> \<prec> P'` `bn \<alpha> \<sharp>* subject \<alpha>`
    apply(subst (asm) alphaRes[where y=b])
    apply assumption
    apply(subst residualAlpha[where p=p,symmetric])
    by simp+
  moreover note `b \<sharp> \<Psi>`
  moreover have "b \<sharp> (p\<bullet>\<alpha>)" and "b \<sharp> (p\<bullet>P')"
  proof -
    have "p\<bullet>b = b" using `set p \<subseteq> set(bn \<alpha>) \<times> set(bn(p\<bullet>\<alpha>))` `distinctPerm p` `b \<sharp> \<alpha>` `(bn(p\<bullet>\<alpha>)) \<sharp>* b`
      by simp
    thus "b \<sharp> (p\<bullet>\<alpha>)" and "b \<sharp> (p\<bullet>P')" using `distinctPerm p` `b \<sharp> \<alpha>` `b \<sharp> P'`
      by(subst perm_bool[where pi=p,symmetric],simp add: eqvts)+
  qed
  moreover note `bn(p\<bullet>\<alpha>) \<sharp>* \<Psi>` `bn(p\<bullet>\<alpha>) \<sharp>* ([(a,b)]\<bullet>P)`
  moreover have "bn(p\<bullet>\<alpha>) \<sharp>* subject(p\<bullet>\<alpha>)" using `bn \<alpha> \<sharp>* subject \<alpha>`
    by(subst (asm) perm_bool[where pi=p,symmetric]) (simp add: eqvts)
  ultimately show ?thesis
  proof(induct rule: resCases)
    case cRes thus ?case
      by(rule_tac elm)
  next
    case cOpen thus ?case
      by(rule_tac elm)
  qed
qed

lemma resTransition:
  fixes C::"'d::fs_name"
  assumes "\<Psi> \<rhd> \<lparr>\<nu>a\<rparr>P \<longmapsto> \<alpha> \<prec> P'"
  and "bn \<alpha> \<sharp>* subject \<alpha>"
  obtains y \<beta> P'' where "\<Psi> \<rhd> ([(a,y)]\<bullet>P) \<longmapsto> \<beta> \<prec> P''" and "bn \<beta> \<sharp>* subject \<beta>" and "y \<sharp> C"
proof -
  assume AS: "\<And>y \<beta> P''. \<lbrakk>\<Psi> \<rhd> [(a, y)] \<bullet> P \<longmapsto> \<beta> \<prec> P''; bn \<beta> \<sharp>* subject \<beta>; y \<sharp> C\<rbrakk> \<Longrightarrow> thesis"
  obtain b::name where "b \<sharp> \<Psi>" and "b \<noteq> a" and "b \<sharp> P" and "b \<sharp> \<alpha>" and "b \<sharp> P'" and "b \<sharp> C"
    by(generate_fresh "name") auto
  moreover obtain p::"name prm" where "set p \<subseteq> set(bn \<alpha>) \<times> set(bn(p\<bullet>\<alpha>))" and "(bn(p\<bullet>\<alpha>)) \<sharp>* \<Psi>" and "(bn(p\<bullet>\<alpha>)) \<sharp>* \<alpha>" and "(bn(p\<bullet>\<alpha>)) \<sharp>* P'" and "(bn(p\<bullet>\<alpha>)) \<sharp>* P" and "(bn(p\<bullet>\<alpha>)) \<sharp>* b" and "(bn(p\<bullet>\<alpha>)) \<sharp>* ([(a,b)]\<bullet>P)" and "(bn(p\<bullet>\<alpha>)) \<sharp>* a" and "(bn(p\<bullet>\<alpha>)) \<sharp>* C" and "distinctPerm p"
    by(rule_tac name_list_avoiding[where c="(\<Psi>,\<alpha>,P',P,b,a,C)" and xvec="bn \<alpha>"]) (auto simp add: eqvts)
  ultimately have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>([(a,b)]\<bullet>P) \<longmapsto> (p\<bullet>\<alpha>) \<prec> p\<bullet>P'" using `\<Psi> \<rhd> \<lparr>\<nu>a\<rparr>P \<longmapsto> \<alpha> \<prec> P'` `bn \<alpha> \<sharp>* subject \<alpha>`
    apply(subst (asm) alphaRes[where y=b])
    apply assumption
    apply(subst residualAlpha[where p=p,symmetric])
    by simp+
  moreover note `b \<sharp> \<Psi>`
  moreover have "b \<sharp> (p\<bullet>\<alpha>)" and "b \<sharp> (p\<bullet>P')"
  proof -
    have "p\<bullet>b = b" using `set p \<subseteq> set(bn \<alpha>) \<times> set(bn(p\<bullet>\<alpha>))` `distinctPerm p` `b \<sharp> \<alpha>` `(bn(p\<bullet>\<alpha>)) \<sharp>* b`
      by simp
    thus "b \<sharp> (p\<bullet>\<alpha>)" and "b \<sharp> (p\<bullet>P')" using `distinctPerm p` `b \<sharp> \<alpha>` `b \<sharp> P'`
      by(subst perm_bool[where pi=p,symmetric],simp add: eqvts)+
  qed
  moreover note `bn(p\<bullet>\<alpha>) \<sharp>* \<Psi>` `bn(p\<bullet>\<alpha>) \<sharp>* ([(a,b)]\<bullet>P)`
  moreover have "bn(p\<bullet>\<alpha>) \<sharp>* subject(p\<bullet>\<alpha>)" using `bn \<alpha> \<sharp>* subject \<alpha>`
    by(subst (asm) perm_bool[where pi=p,symmetric]) (simp add: eqvts)
  ultimately have "\<exists> y \<beta> P''. \<Psi> \<rhd> ([(a,y)]\<bullet>P) \<longmapsto> \<beta> \<prec> P'' \<and> bn \<beta> \<sharp>* subject \<beta> \<and> y \<sharp> C"
  proof(induct rule: resCases)
    case(cRes P') thus ?case using `bn(p\<bullet>\<alpha>) \<sharp>* subject(p\<bullet>\<alpha>)` `b \<sharp> C`
      by blast
  next
    case(cOpen M xvec yvec y N P') thus ?case using `b \<sharp> C`
      apply(rule_tac exI)+
      apply(rule conjI)
      apply assumption
      apply(rule conjI)
      apply simp
      apply assumption
      done
  qed
  thus ?thesis using AS
    by blast
qed

lemma resNilTransAux':
  assumes "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>P \<longmapsto> \<alpha> \<prec> P'"
  and "bn \<alpha> \<sharp>* subject \<alpha>"
  and "distinct xvec"
  and elm: "\<And> (p::name prm) R. \<Psi> \<rhd> (p\<bullet>P) \<longmapsto> R \<Longrightarrow> False"

  shows False
  using assms  
proof(induct xvec arbitrary: P P' \<alpha>)
  case Nil thus ?case
    apply(drule_tac x="[]" in meta_spec)
    by simp
next
  case(Cons x xvec)
  from `\<Psi> \<rhd> \<lparr>\<nu>*(x # xvec)\<rparr>P \<longmapsto> \<alpha> \<prec> P'` `bn \<alpha> \<sharp>* subject \<alpha>`
  obtain y \<beta> P'' where "\<Psi> \<rhd> ([(x,y)]\<bullet>(\<lparr>\<nu>*xvec\<rparr>P)) \<longmapsto> \<beta> \<prec> P''" and "bn \<beta> \<sharp>* subject \<beta>" and "y \<sharp> xvec"
    by(rule_tac resTransition) auto
  from `\<Psi> \<rhd> ([(x,y)]\<bullet>(\<lparr>\<nu>*xvec\<rparr>P)) \<longmapsto> \<beta> \<prec> P''` have "\<Psi> \<rhd> (\<lparr>\<nu>*xvec\<rparr>([(x,y)]\<bullet>P)) \<longmapsto> \<beta> \<prec> P''" using `distinct(x#xvec)` `y \<sharp> xvec`
    by(simp add: eqvts)
  moreover have "\<And>(p::name prm) R. \<Psi> \<rhd> p\<bullet>([(x,y)]\<bullet>P) \<longmapsto> R \<Longrightarrow> False"
  proof -
    fix p::"name prm" and R
    assume "\<Psi> \<rhd> p\<bullet>([(x,y)]\<bullet>P) \<longmapsto> R"
    thus False
      apply(rule_tac Cons(5)[where p="p@[(x,y)]" and R=R])
      by(subst pt2[OF pt_name_inst])
  qed        
  ultimately show ?case using `bn \<beta> \<sharp>* subject \<beta>` `distinct(x # xvec)`
    by(rule_tac Cons) auto
qed

lemma \<F>nilTrans:
  assumes "\<Psi> \<rhd> \<F> P \<longmapsto> \<alpha> \<prec> P'"
  and "bn \<alpha> \<sharp>* subject \<alpha>"
  shows "False"
proof -  
  obtain A\<^isub>P \<Psi>\<^isub>P where "extractFrame P = \<langle>A\<^isub>P,\<Psi>\<^isub>P\<rangle>" and "A\<^isub>P \<sharp>* P" and "A\<^isub>P \<sharp>* \<alpha>" and "A\<^isub>P \<sharp>* P'" and "A\<^isub>P \<sharp>* \<Psi>" and "distinct A\<^isub>P"
    by(rule freshFrame[where C="(P,\<alpha>,P',\<Psi>)"]) auto
  hence "\<Psi> \<rhd> \<lparr>\<nu>*A\<^isub>P\<rparr>\<lbrace>\<Psi>\<^isub>P\<rbrace> \<longmapsto> \<alpha> \<prec> P'" using assms
    by simp
  moreover have "\<And> (p::name prm) R. \<Psi> \<rhd> (p\<bullet>\<lbrace>\<Psi>\<^isub>P\<rbrace>) \<longmapsto> R \<Longrightarrow> False"
  proof -
    fix p::"name prm" and R
    assume "\<Psi> \<rhd> p\<bullet>\<lbrace>\<Psi>\<^isub>P\<rbrace> \<longmapsto> R"
    thus False unfolding eqvts
      by(rule nilTrans)
  qed
  ultimately show ?thesis using `distinct(A\<^isub>P)` `bn \<alpha> \<sharp>* subject \<alpha>` 
    by(rule_tac resNilTransAux') auto
qed

definition legalClause :: "'a \<Rightarrow> ('a,'b,'c) psi \<Rightarrow> bool"
where
"legalClause M P = ((supp P::name set) \<subseteq> supp M \<and> (guarded P))"

lemma legalClauseEqvt[eqvt]:
  fixes p::"name prm"
  shows "(p\<bullet>legalClause M P) = legalClause (p\<bullet>M) (p\<bullet>P)"
by(simp add: legalClause_def eqvts)

coinductive_set hobisim :: "('b \<times> ('a, 'b, 'c) psi \<times> ('a, 'b, 'c) psi) set" 
where
  step: "\<lbrakk>\<forall> \<phi>. (\<not>clauseDom \<phi> \<and> (insertAssertion (extractFrame P)) \<Psi> \<turnstile>\<^sub>F \<phi>) \<longrightarrow> (insertAssertion (extractFrame Q) \<Psi>) \<turnstile>\<^sub>F \<phi>;
          \<forall> M P'. (legalClause M P' \<and> insertAssertion (extractFrame P) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> P') \<longrightarrow> (\<exists>Q'. legalClause M Q' \<and> insertAssertion (extractFrame Q) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> Q' \<and> (\<one>,P',Q') \<in> hobisim);
          \<Psi> \<rhd> P \<leadsto>[hobisim] Q;
          \<forall>\<Psi>'. (\<Psi> \<otimes> \<Psi>',  P, Q) \<in> hobisim; (\<Psi>, Q, P) \<in> hobisim\<rbrakk> \<Longrightarrow> (\<Psi>, P, Q) \<in> hobisim"
monos monoCoinduct

abbreviation
  hobisimJudge ("_ \<rhd> _ \<sim>\<^sub>H _" [70, 70, 70] 65) where "\<Psi> \<rhd> P \<sim>\<^sub>H Q \<equiv> (\<Psi>, P, Q) \<in> hobisim"
abbreviation
  hobisimNilJudge ("_ \<sim>\<^sub>H _" [70, 70] 65) where "P \<sim>\<^sub>H Q \<equiv> SBottom' \<rhd> P \<sim>\<^sub>H Q"

lemma hobisimCoinductAux[consumes 1]:
  fixes \<Psi> :: 'b
  and   P :: "('a, 'b, 'c) psi"
  and   Q :: "('a, 'b, 'c) psi"
  and   X :: "('b \<times> ('a, 'b, 'c) psi \<times> ('a, 'b, 'c) psi) set"

  assumes "(\<Psi>, P, Q) \<in> X"
  and    2: "\<And>\<Psi> P Q. (\<Psi>, P, Q) \<in> X \<Longrightarrow>
                                    (\<forall> \<phi>. (\<not>clauseDom \<phi> \<and> (insertAssertion (extractFrame P)) \<Psi> \<turnstile>\<^sub>F \<phi>) \<longrightarrow> (insertAssertion (extractFrame Q) \<Psi>) \<turnstile>\<^sub>F \<phi>) \<and>
(\<forall> M P'. ((legalClause M P' \<and> insertAssertion (extractFrame P) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> P') \<longrightarrow> (\<exists>Q'. legalClause M Q' \<and> insertAssertion (extractFrame Q) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> Q' \<and> ((\<one>,P',Q') \<in> X \<or> (\<one>,P',Q') \<in> hobisim))))
 \<and>
                                    (\<Psi> \<rhd> P \<leadsto>[(X \<union> hobisim)] Q) \<and>
                                    (\<forall>\<Psi>'. (\<Psi> \<otimes> \<Psi>', P, Q) \<in> X \<or> (\<Psi> \<otimes> \<Psi>', P, Q) \<in> hobisim) \<and>
                                    ((\<Psi>, Q, P) \<in> X \<or> (\<Psi>, Q, P) \<in> hobisim)"

  shows "(\<Psi>, P, Q) \<in> hobisim"
proof -
  note `(\<Psi>, P, Q) \<in> X`
  moreover have "X \<union> hobisim = {(\<Psi>, P, Q). (\<Psi>, P, Q) \<in> X \<or> (\<Psi>, P, Q) \<in> hobisim}" by auto
  moreover note 2
  ultimately show ?thesis
    by coinduct auto
qed

lemma hobisimCoinduct[consumes 1, case_names cStatEq cClauseEq cSim cExt cSym]:
  fixes F :: 'b
  and   P :: "('a, 'b, 'c) psi"
  and   Q :: "('a, 'b, 'c) psi"
  and   X :: "('b \<times> ('a, 'b, 'c) psi \<times> ('a, 'b, 'c) psi) set"

  assumes "(\<Psi>, P, Q) \<in> X"
  and     1: "\<And> \<Psi>' R S \<phi>. \<lbrakk>(\<Psi>', R, S) \<in> X; \<not>clauseDom \<phi>; (insertAssertion (extractFrame R)) \<Psi>' \<turnstile>\<^sub>F \<phi>\<rbrakk> \<Longrightarrow> (insertAssertion (extractFrame S)) \<Psi>' \<turnstile>\<^sub>F \<phi>"
  and     2: "\<And>\<Psi>' R S M P. \<lbrakk>(\<Psi>', R, S) \<in> X; legalClause M P; insertAssertion (extractFrame R) \<Psi>' \<turnstile>\<^sub>F M \<Leftarrow> P\<rbrakk> \<Longrightarrow> \<exists>Q. legalClause M Q \<and> insertAssertion (extractFrame S) \<Psi>' \<turnstile>\<^sub>F M \<Leftarrow> Q \<and> ((\<one>,P,Q) \<in> X \<or> (\<one>,P,Q) \<in> hobisim)"
  and     3: "\<And>\<Psi>' R S. (\<Psi>', R, S) \<in> X \<Longrightarrow> \<Psi>' \<rhd> R \<leadsto>[(X \<union> hobisim)] S"
  and     4: "\<And>\<Psi>' R S \<Psi>''. (\<Psi>', R, S) \<in> X \<Longrightarrow> (\<Psi>' \<otimes> \<Psi>'', R, S) \<in> X \<or> (\<Psi>' \<otimes> \<Psi>'', R, S) \<in> hobisim"
  and     "\<And>\<Psi>' R S. (\<Psi>', R, S) \<in> X \<Longrightarrow> (\<Psi>', S, R) \<in> X \<or> (\<Psi>', S, R) \<in> hobisim"

  shows "(\<Psi>, P, Q) \<in> hobisim"
proof -
  have "X \<union> hobisim = {(\<Psi>, P, Q). (\<Psi>, P, Q) \<in> X \<or> (\<Psi>, P, Q) \<in> hobisim}" by auto
  with assms show ?thesis
    by(coinduct rule: hobisimCoinductAux) auto
qed

lemma hobisimWeakCoinductAux[consumes 1]:
  fixes F :: 'b
  and   P :: "('a, 'b, 'c) psi"
  and   Q :: "('a, 'b, 'c) psi"
  and   X :: "('b \<times> ('a, 'b, 'c) psi \<times> ('a, 'b, 'c) psi) set"

  assumes "(\<Psi>, P, Q) \<in> X"
  and     "\<And>\<Psi> P Q. (\<Psi>, P, Q) \<in> X \<Longrightarrow>
                                    (\<forall> \<phi>. (\<not>clauseDom \<phi> \<and> (insertAssertion (extractFrame P)) \<Psi> \<turnstile>\<^sub>F \<phi>) \<longrightarrow> (insertAssertion (extractFrame Q) \<Psi>) \<turnstile>\<^sub>F \<phi>) \<and>
                                    (\<forall> M P'. ((legalClause M P' \<and> insertAssertion (extractFrame P) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> P' \<longrightarrow> (\<exists>Q'. legalClause M Q' \<and> insertAssertion (extractFrame Q) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> Q' \<and> (\<one>,P',Q') \<in> X)))) \<and>
                                    (\<Psi> \<rhd> P \<leadsto>[X] Q) \<and>
                                    (\<forall>\<Psi>'. (\<Psi> \<otimes> \<Psi>', P, Q) \<in> X) \<and>
                                    (\<Psi>, Q, P) \<in> X"

  shows "(\<Psi>, P, Q) \<in> hobisim"
using assms
by(coinduct rule: hobisimCoinductAux) (blast intro: monotonic)

lemma hobisimWeakCoinductAuxSC[consumes 1]:
  fixes F :: 'b
  and   P :: "('a, 'b, 'c) psi"
  and   Q :: "('a, 'b, 'c) psi"
  and   X :: "('b \<times> ('a, 'b, 'c) psi \<times> ('a, 'b, 'c) psi) set"

  assumes "(\<Psi>, P, Q) \<in> X"
  and     "\<And>\<Psi> P Q. (\<Psi>, P, Q) \<in> X \<Longrightarrow>
                                    (\<forall> \<phi>. (\<not>clauseDom \<phi> \<and> (insertAssertion (extractFrame P)) \<Psi> \<turnstile>\<^sub>F \<phi>) \<longrightarrow> (insertAssertion (extractFrame Q) \<Psi>) \<turnstile>\<^sub>F \<phi>) \<and>
                                    (\<forall> M P'. (legalClause M P' \<and> insertAssertion (extractFrame P) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> P') \<longrightarrow> (\<exists>Q'. legalClause M Q' \<and> insertAssertion (extractFrame Q) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> Q' \<and> ((\<one>,P',Q') \<in> X \<or> (\<one>,P',Q') \<in> hobisim))) \<and>
                                    (\<Psi> \<rhd> P \<leadsto>[X] Q) \<and>
                                    (\<forall>\<Psi>'. (\<Psi> \<otimes> \<Psi>', P, Q) \<in> X) \<and>
                                    (\<Psi>, Q, P) \<in> X"

  shows "(\<Psi>, P, Q) \<in> hobisim"
using assms
by(coinduct rule: hobisimCoinductAux) (blast intro: monotonic)

lemma hobisimWeakCoinductSC[consumes 1, case_names cStatEq cClauseEq cSim cExt cSym]:
  fixes F :: 'b
  and   P :: "('a, 'b, 'c) psi"
  and   Q :: "('a, 'b, 'c) psi"
  and   X :: "('b \<times> ('a, 'b, 'c) psi \<times> ('a, 'b, 'c) psi) set"

  assumes "(\<Psi>, P, Q) \<in> X"
  and     rStatEq: "\<And> \<Psi>' R S \<phi>. \<lbrakk>(\<Psi>', R, S) \<in> X; \<not>clauseDom \<phi>; (insertAssertion (extractFrame R)) \<Psi>' \<turnstile>\<^sub>F \<phi>\<rbrakk> \<Longrightarrow> (insertAssertion (extractFrame S)) \<Psi>' \<turnstile>\<^sub>F \<phi>"
  and     rClauseEq: "\<And>\<Psi>' R S M P. \<lbrakk>(\<Psi>', R, S) \<in> X; legalClause M P; insertAssertion (extractFrame R) \<Psi>' \<turnstile>\<^sub>F M \<Leftarrow> P\<rbrakk> \<Longrightarrow> \<exists>Q. legalClause M Q \<and> insertAssertion (extractFrame S) \<Psi>' \<turnstile>\<^sub>F M \<Leftarrow> Q \<and> ((\<one>,P,Q) \<in> X \<or> (\<one>,P,Q) \<in> hobisim)"
  and     rSim: "\<And>\<Psi>' R S. (\<Psi>', R, S) \<in> X \<Longrightarrow> \<Psi>' \<rhd> R \<leadsto>[X] S"
  and     rExt: "\<And>\<Psi>' R S \<Psi>''. (\<Psi>', R, S) \<in> X \<Longrightarrow> (\<Psi>' \<otimes> \<Psi>'', R, S) \<in> X"
  and     rSym: "\<And>\<Psi>' R S. (\<Psi>', R, S) \<in> X \<Longrightarrow> (\<Psi>', S, R) \<in> X"

  shows "(\<Psi>, P, Q) \<in> hobisim"
proof -
  note `(\<Psi>, P, Q) \<in> X`
  moreover have "X \<union> hobisim = {(\<Psi>, P, Q). (\<Psi>, P, Q) \<in> X \<or> (\<Psi>, P, Q) \<in> hobisim}" by auto
  ultimately show ?thesis
  proof(coinduct rule: hobisimCoinduct)
    case cStatEq thus ?case by(blast intro: rStatEq)
  next
    case (cClauseEq \<Psi> R S M P)
    thus ?case by(blast intro: rClauseEq)
  next
    case cSim thus ?case by(blast intro: rSim monotonic)
  next
    case cExt thus ?case by(blast intro: rExt)
  next
    case cSym thus ?case by(blast intro: rSym)
  qed
qed

lemma hobisimWeakCoinduct[consumes 1, case_names cStatEq cClauseEq cSim cExt cSym]:
  fixes F :: 'b
  and   P :: "('a, 'b, 'c) psi"
  and   Q :: "('a, 'b, 'c) psi"
  and   X :: "('b \<times> ('a, 'b, 'c) psi \<times> ('a, 'b, 'c) psi) set"

  assumes "(\<Psi>, P, Q) \<in> X"
  and     rStatEq: "\<And> \<Psi>' R S \<phi>. \<lbrakk>(\<Psi>', R, S) \<in> X; \<not>clauseDom \<phi>; (insertAssertion (extractFrame R)) \<Psi>' \<turnstile>\<^sub>F \<phi>\<rbrakk> \<Longrightarrow> (insertAssertion (extractFrame S)) \<Psi>' \<turnstile>\<^sub>F \<phi>"
  and     rClauseEq: "\<And>\<Psi>' R S M P. \<lbrakk>(\<Psi>', R, S) \<in> X; legalClause M P; insertAssertion (extractFrame R) \<Psi>' \<turnstile>\<^sub>F M \<Leftarrow> P\<rbrakk> \<Longrightarrow> \<exists>Q. legalClause M Q \<and> insertAssertion (extractFrame S) \<Psi>' \<turnstile>\<^sub>F M \<Leftarrow> Q \<and> (\<one>,P,Q) \<in> X"
  and     rSim: "\<And>\<Psi>' R S. (\<Psi>', R, S) \<in> X \<Longrightarrow> \<Psi>' \<rhd> R \<leadsto>[X] S"
  and     rExt: "\<And>\<Psi>' R S \<Psi>''. (\<Psi>', R, S) \<in> X \<Longrightarrow> (\<Psi>' \<otimes> \<Psi>'', R, S) \<in> X"
  and     rSym: "\<And>\<Psi>' R S. (\<Psi>', R, S) \<in> X \<Longrightarrow> (\<Psi>', S, R) \<in> X"

  shows "(\<Psi>, P, Q) \<in> hobisim"
proof -
  note `(\<Psi>, P, Q) \<in> X`
  moreover have "X \<union> hobisim = {(\<Psi>, P, Q). (\<Psi>, P, Q) \<in> X \<or> (\<Psi>, P, Q) \<in> hobisim}" by auto
  ultimately show ?thesis
  proof(coinduct rule: hobisimCoinduct)
    case cStatEq thus ?case by(blast intro: rStatEq)
  next
    case (cClauseEq \<Psi> R S M P)
    hence "\<exists>Q. legalClause M Q \<and>
        assertionAux.FrameImp SImp'
         (insertAssertion (extractFrame S) \<Psi>) (M\<Leftarrow>Q) \<and>
        (\<one>, P, Q) \<in> X"
      by(rule_tac rClauseEq)
    thus ?case by blast
  next
    case cSim thus ?case by(blast intro: rSim monotonic)
  next
    case cExt thus ?case by(blast intro: rExt)
  next
    case cSym thus ?case by(blast intro: rSym)
  qed
qed

lemma hobisimE:
  fixes P  :: "('a, 'b, 'c) psi"
  and   Q  :: "('a, 'b, 'c) psi"
  and   \<Psi>  :: 'b
  and   \<Psi>' :: 'b

  assumes "(\<Psi>, P, Q) \<in> hobisim"

  shows "(\<not>clauseDom \<phi> \<and> (insertAssertion (extractFrame P)) \<Psi> \<turnstile>\<^sub>F \<phi>) \<longrightarrow> (insertAssertion (extractFrame Q) \<Psi>) \<turnstile>\<^sub>F \<phi>"
  and "(legalClause M P' \<and> insertAssertion (extractFrame P) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> P') \<longrightarrow> (\<exists>Q'. legalClause M Q' \<and> insertAssertion (extractFrame Q) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> Q' \<and> (\<one>,P',Q') \<in> hobisim)"
  and   "\<Psi> \<rhd> P \<leadsto>[hobisim] Q"
  and   "(\<Psi> \<otimes> \<Psi>', P, Q) \<in> hobisim"
  and   "(\<Psi>, Q, P) \<in> hobisim"
using assms
apply -
by(rule hobisim.cases,assumption,blast)+

lemma hobisimI:
  fixes P :: "('a, 'b, 'c) psi"
  and   Q :: "('a, 'b, 'c) psi"
  and   \<Psi> :: 'b

  assumes "(\<forall> \<phi>. (\<not>clauseDom \<phi> \<and> (insertAssertion (extractFrame P)) \<Psi> \<turnstile>\<^sub>F \<phi>) \<longrightarrow> (insertAssertion (extractFrame Q) \<Psi>) \<turnstile>\<^sub>F \<phi>)"
  and     "(\<forall> M P'. (legalClause M P' \<and> insertAssertion (extractFrame P) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> P') \<longrightarrow> (\<exists>Q'. legalClause M Q' \<and> insertAssertion (extractFrame Q) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> Q' \<and> (\<one>,P',Q') \<in> hobisim))"
  and     "\<Psi> \<rhd> P \<leadsto>[hobisim] Q"
  and     "\<forall>\<Psi>'. (\<Psi> \<otimes> \<Psi>', P, Q) \<in> hobisim"
  and     "(\<Psi>, Q, P) \<in> hobisim"

  shows "(\<Psi>, P, Q) \<in> hobisim"
using assms
by(auto intro: hobisim.step)

lemma hobisimReflexive:
  fixes \<Psi> :: 'b
  and   P :: "('a, 'b, 'c) psi"


  shows "\<Psi> \<rhd> P \<sim>\<^sub>H P"
proof -
  let ?X = "{(\<Psi>, P, P) | \<Psi> P. True}"
  have "(\<Psi>, P, P) \<in> ?X" by simp
  thus ?thesis
    apply(coinduct rule: hobisimWeakCoinduct)
    by(auto intro: reflexive)
qed

lemma frameBisim:
  assumes "\<Psi> \<rhd> P \<sim> Q"
  shows   "\<Psi> \<rhd> \<F> P \<sim> \<F> Q"
proof -
  let ?X = "{(\<Psi>, \<F> P, \<F> Q) | \<Psi> P Q. \<Psi> \<rhd> P \<sim> Q}"
  from assms have "(\<Psi>,\<F> P, \<F> Q) \<in> ?X"
    by blast
  thus ?thesis
  proof(coinduct rule: bisimCoinduct)
    case(cStatEq \<Psi> R S)
    thus ?case
      by(auto intro: bisimE)
  next
    case cSym thus ?case
      by(auto intro: bisimE)
  next
    case cExt thus ?case
      by(auto intro: bisimE)
  next
    case(cSim \<Psi> R S)
    have "eqvt ?X"
      apply(auto simp add: eqvt_def)
      apply(rule_tac x="p\<bullet>P" in exI)
      apply(auto simp add: eqvts)
      apply(rule_tac x="p\<bullet>Q" in exI)
      by(auto simp add: eqvts bisimClosed)
    hence "eqvt(?X \<union> bisim)" using bisimEqvt
      by auto
    thus ?case
      apply(rule simI)
      using cSim
      apply(rule_tac exI)
      by(auto dest: \<F>nilTrans)
  qed
qed    

lemma hobisimBisim:
  assumes "\<Psi> \<rhd> P \<sim> Q"
  shows    "\<Psi> \<rhd> P \<sim>\<^sub>H Q"
using assms
proof(coinduct rule: hobisimWeakCoinduct)
  case (cStatEq \<Psi> R S \<phi>)
  from `\<Psi> \<rhd> R \<sim> S` have "insertAssertion (extractFrame R) \<Psi> \<simeq>\<^sub>F insertAssertion (extractFrame S) \<Psi>" by(drule_tac bisimE)
  with `insertAssertion (extractFrame R) \<Psi> \<turnstile>\<^sub>F \<phi>` show ?case by(rule_tac frameStatEqEnt)
next
  case (cClauseEq \<Psi> R S M P)
  from `\<Psi> \<rhd> R \<sim> S` have "insertAssertion (extractFrame R) \<Psi> \<simeq>\<^sub>F insertAssertion (extractFrame S) \<Psi>" by(drule_tac bisimE)
  with `insertAssertion (extractFrame R) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> P` have "insertAssertion (extractFrame S) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> P" by(rule_tac frameStatEqEnt)
  moreover note `legalClause M P`
  moreover have "\<one> \<rhd> P \<sim> P"
    by(rule bisimReflexive)
  ultimately show ?case by blast
next
  case cSim thus ?case by(drule_tac bisimE)
next
  case cExt thus ?case by(drule_tac bisimE)
next
  case cSym thus ?case by(drule_tac bisimE)
qed

lemma hobisimClosed:
  fixes \<Psi> :: 'b
  and   P :: "('a, 'b, 'c) psi"
  and   Q :: "('a, 'b, 'c) psi"
  and   p :: "name prm"
  
  assumes PBisimQ: "\<Psi> \<rhd> P \<sim>\<^sub>H Q"

  shows "(p \<bullet> \<Psi>) \<rhd>  (p \<bullet> P) \<sim>\<^sub>H (p \<bullet> Q)"
proof -
  let ?X = "{(p \<bullet> \<Psi>, p \<bullet> P, p \<bullet> Q) | (p::name prm) \<Psi>  P Q. \<Psi> \<rhd> P \<sim>\<^sub>H Q}"
  have "eqvt ?X"
    apply(auto simp add: eqvt_def)
    apply(rule_tac x="pa@p" in exI)
    by(auto simp add: pt2[OF pt_name_inst])
  have "hobisim \<subseteq> ?X"
  proof -
    {
      fix x
      assume "x \<in> hobisim"
      hence "x \<in> ?X"
        apply(induct x)
        apply auto
        apply(rule exI[where x="[]"])
        by simp
    }
    thus ?thesis by force
  qed
  from PBisimQ have "(p \<bullet> \<Psi>, p \<bullet> P, p \<bullet> Q) \<in> ?X" by blast
  thus ?thesis
  proof(coinduct rule: hobisimWeakCoinduct)
    case(cStatEq \<Psi> P Q \<phi>)
    then obtain \<Psi>' P' Q' and p::"name prm" where "\<Psi>=p\<bullet>\<Psi>'" and "P=p\<bullet>P'" and "Q=p\<bullet>Q'" and "\<Psi>' \<rhd> P' \<sim>\<^sub>H Q'"
      by blast
    have "\<forall> \<phi>. (\<not>clauseDom \<phi> \<and> (insertAssertion (extractFrame P')) \<Psi>' \<turnstile>\<^sub>F \<phi>) \<longrightarrow> (insertAssertion (extractFrame Q') \<Psi>') \<turnstile>\<^sub>F \<phi>"
    proof(rule allI)
      fix \<phi>
      from `\<Psi>' \<rhd> P' \<sim>\<^sub>H Q'` show "(\<not>clauseDom \<phi> \<and> (insertAssertion (extractFrame P')) \<Psi>' \<turnstile>\<^sub>F \<phi>) \<longrightarrow> (insertAssertion (extractFrame Q') \<Psi>') \<turnstile>\<^sub>F \<phi>"
        by(rule hobisimE)
    qed
    hence "(\<not>clauseDom \<phi> \<and> (insertAssertion (extractFrame P)) \<Psi> \<turnstile>\<^sub>F \<phi>) \<longrightarrow> (insertAssertion (extractFrame Q) \<Psi>) \<turnstile>\<^sub>F \<phi>" unfolding `\<Psi>=p\<bullet>\<Psi>'` and `P=p\<bullet>P'` and `Q=p\<bullet>Q'`
      by(subst (asm) perm_bool[where pi=p,symmetric]) (simp add: eqvts)
    with cStatEq show ?case
      by simp
  next
    case (cClauseEq \<Psi> P Q M R)
    then obtain \<Psi>' P' Q' and p::"name prm" where "\<Psi>=p\<bullet>\<Psi>'" and "P=p\<bullet>P'" and "Q=p\<bullet>Q'" and "\<Psi>' \<rhd> P' \<sim>\<^sub>H Q'"
      by auto
    have "\<forall> M R. (legalClause M R \<and> insertAssertion (extractFrame P') \<Psi>' \<turnstile>\<^sub>F M \<Leftarrow> R) \<longrightarrow> (\<exists>R'. legalClause M R' \<and> insertAssertion (extractFrame Q') \<Psi>' \<turnstile>\<^sub>F M \<Leftarrow> R' \<and> (\<one>,R,R') \<in> ?X)"

    proof((rule allI)+,rule impI)
      fix M R
      assume "legalClause M R \<and> insertAssertion (extractFrame P') \<Psi>' \<turnstile>\<^sub>F M \<Leftarrow> R"
      with `\<Psi>' \<rhd> P' \<sim>\<^sub>H Q'` 
      have "(\<exists>R'. legalClause M R' \<and> insertAssertion (extractFrame Q') \<Psi>'  \<turnstile>\<^sub>F M \<Leftarrow> R' \<and> (\<one>,R,R') \<in> hobisim)" using hobisimE(2)
        by blast
      then obtain R' where 1: "legalClause M R'" and 2: "insertAssertion (extractFrame Q') \<Psi>' \<turnstile>\<^sub>F M \<Leftarrow> R'" and "(\<one>,R,R') \<in> hobisim"
        by blast
      note 1 2
      moreover from `(\<one>,R,R') \<in> hobisim` have "(\<one>,R,R') \<in> ?X" using `hobisim \<subseteq> ?X`
        by blast
      ultimately show "\<exists>R'. legalClause M R' \<and> insertAssertion (extractFrame Q') \<Psi>' \<turnstile>\<^sub>F M \<Leftarrow> R' \<and> (\<one>,R,R') \<in> ?X"
        by blast
    qed
    hence "(legalClause M R \<and> insertAssertion (extractFrame P) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> R) \<longrightarrow> (\<exists>R'. legalClause M R' \<and> insertAssertion (extractFrame Q) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> R' \<and> (\<one>,R,R') \<in> p\<bullet>?X)" unfolding `\<Psi>=p\<bullet>\<Psi>'` and `P=p\<bullet>P'` and `Q=p\<bullet>Q'`
      apply(subst (asm) perm_bool[where pi=p, symmetric])
      unfolding eqvts pt_pi_rev[OF pt_name_inst, OF at_name_inst]
      by blast
    with cClauseEq obtain R' where "legalClause M R'" and "insertAssertion (extractFrame Q) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> R'" and "(\<one>,R,R') \<in> p\<bullet>?X" by auto
    moreover have "(\<one>,R,R') \<in> ?X"
    proof -
      from `eqvt ?X` have "eqvt(p\<bullet>?X)"
      proof(induct p)
        case Nil thus ?case by simp
      next
        case (Cons a p) thus ?case
        proof(induct a)
          case(Pair x y) thus ?case by simp
        qed
      qed
      from `(\<one>,R,R') \<in> p\<bullet>?X` have "(p\<bullet>(\<one>,R,R') \<in> p\<bullet>?X)" using `eqvt(p\<bullet>?X)`
        by(rule_tac eqvtI)
      thus ?thesis
        by(subst perm_bool[where pi=p,symmetric], simp only: eqvts)
    qed
    ultimately show ?case by blast
  next
    case(cSim \<Psi> P Q)
    {
      fix p :: "name prm"
      fix \<Psi> P Q
      have "eqvt ?X"
	apply(auto simp add: eqvt_def)
	apply(rule_tac x="pa@p" in exI)
	by(auto simp add: pt2[OF pt_name_inst])
      moreover assume "\<Psi> \<rhd> P \<leadsto>[hobisim] Q"
      hence "\<Psi> \<rhd> P \<leadsto>[?X] Q"
	apply(rule_tac A=hobisim in monotonic, auto)
	by(rule_tac x="[]::name prm" in exI) auto
      ultimately have "((p::name prm) \<bullet> \<Psi>) \<rhd> (p \<bullet> P) \<leadsto>[?X] (p \<bullet> Q)"
	by(rule_tac simClosed)
    }
    with `(\<Psi>, P, Q) \<in> ?X` show ?case
      by(blast dest: hobisimE)
  next
    case(cExt \<Psi> P Q \<Psi>')
    {
      fix p :: "name prm"
      fix \<Psi> P Q \<Psi>'
      assume "\<forall>\<Psi>'. (\<Psi> \<otimes> \<Psi>', P, Q) \<in> hobisim"
      hence "((p \<bullet> \<Psi>) \<otimes> \<Psi>', p \<bullet> P, p \<bullet> Q) \<in> ?X"  
	apply(auto, rule_tac x=p in exI)
	apply(rule_tac x="\<Psi> \<otimes> (rev p \<bullet> \<Psi>')" in exI)
	by(auto simp add: eqvts)
    }
    with `(\<Psi>, P, Q) \<in> ?X` show ?case
      by(blast dest: hobisimE)
  next
    case(cSym \<Psi> P Q)
    thus ?case
      by(blast dest: hobisimE)
  qed
qed

lemma hobisimEqvt[simp]:
  shows "eqvt hobisim"
using assms
by(auto simp add: eqvt_def hobisimClosed)

lemma legalClauseFresh[simp]:
  fixes xvec::"name list"
  assumes "xvec \<sharp>* M"
  and     "legalClause M P"
  shows   "xvec \<sharp>* P"
using assms
unfolding legalClause_def fresh_star_def fresh_def
by auto

lemma assertionFrameCompose:
  assumes "\<Psi> \<simeq> \<Psi>'"
  shows "insertAssertion F \<Psi> \<simeq>\<^sub>F insertAssertion F \<Psi>'"
proof -
  obtain A\<^isub>F \<Psi>\<^isub>F where "F = \<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>" and "A\<^isub>F \<sharp>* \<Psi>" and "A\<^isub>F \<sharp>* \<Psi>'"
    by(rule_tac freshFrame[where C="(\<Psi>,\<Psi>')"]) auto
  with assms show ?thesis
    by(auto intro: frameIntComposition)
qed

lemma hostatEqBisim:
  fixes \<Psi>  :: 'b
  and   P  :: "('a, 'b, 'c) psi"
  and   Q  :: "('a, 'b, 'c) psi"
  and   \<Psi>' :: 'b
  
  assumes "\<Psi> \<rhd> P \<sim>\<^sub>H Q"
  and     "\<Psi> \<simeq> \<Psi>'"

  shows "\<Psi>' \<rhd> P \<sim>\<^sub>H Q"
proof -
  let ?X = "{(\<Psi>', P, Q) | \<Psi> P Q \<Psi>'. \<Psi> \<rhd> P \<sim>\<^sub>H Q \<and> \<Psi> \<simeq> \<Psi>'}"
  have "eqvt ?X"
    apply(auto simp add: eqvt_def)
    apply(rule exI)
    apply(rule conjI)
    apply(rule hobisimClosed)
    apply assumption
    by(rule AssertionStatEqClosed)
  from `\<Psi> \<rhd> P \<sim>\<^sub>H Q` `\<Psi> \<simeq> \<Psi>'` have "(\<Psi>', P, Q) \<in> ?X" by auto
  thus ?thesis
  proof(coinduct rule: hobisimCoinduct)
    case(cStatEq \<Psi>' P Q \<phi>)
    from `(\<Psi>', P, Q) \<in> ?X` obtain \<Psi> where "\<Psi> \<rhd> P \<sim>\<^sub>H Q" and "\<Psi> \<simeq> \<Psi>'"
      by auto
    obtain A\<^isub>P \<Psi>\<^isub>P where FrP: "extractFrame P = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>" and "A\<^isub>P \<sharp>* \<Psi>" and "A\<^isub>P \<sharp>* \<Psi>'"
      by(rule_tac C="(\<Psi>, \<Psi>')" in freshFrame) auto
    obtain A\<^isub>Q \<Psi>\<^isub>Q where FrQ: "extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>" and "A\<^isub>Q \<sharp>* \<Psi>" and "A\<^isub>Q \<sharp>* \<Psi>'"
      by(rule_tac C="(\<Psi>, \<Psi>')" in freshFrame) auto
    have "insertAssertion (extractFrame P) \<Psi>' \<simeq>\<^sub>F insertAssertion (extractFrame P) \<Psi>" using `A\<^isub>P \<sharp>* \<Psi>` `A\<^isub>P \<sharp>* \<Psi>'` `\<Psi> \<simeq> \<Psi>'`
      unfolding FrP
      by simp (metis frameIntComposition FrameStatEqTrans FrameStatEqSym)
    hence "insertAssertion (extractFrame P) \<Psi> \<turnstile>\<^sub>F \<phi>" using `insertAssertion (extractFrame P) \<Psi>' \<turnstile>\<^sub>F \<phi>`
      by(rule frameStatEqEnt)
    from `\<Psi> \<rhd> P \<sim>\<^sub>H Q` `insertAssertion (extractFrame P) \<Psi> \<turnstile>\<^sub>F \<phi>` `\<not> clauseDom \<phi>` have "insertAssertion (extractFrame Q) \<Psi> \<turnstile>\<^sub>F \<phi>"
      by(auto dest: hobisimE)
    moreover have "insertAssertion (extractFrame Q) \<Psi> \<simeq>\<^sub>F insertAssertion (extractFrame Q) \<Psi>'" using `A\<^isub>Q \<sharp>* \<Psi>` `A\<^isub>Q \<sharp>* \<Psi>'` `\<Psi> \<simeq> \<Psi>'`
      unfolding FrQ
      by(simp add: frameIntComposition)
    ultimately show ?case
      by(rule_tac frameStatEqEnt)
  next
    case(cClauseEq \<Psi>' P Q M R)
    from `(\<Psi>', P, Q) \<in> ?X` obtain \<Psi> where "\<Psi> \<rhd> P \<sim>\<^sub>H Q" and "\<Psi> \<simeq> \<Psi>'"
      by auto
(*    obtain A\<^isub>P \<Psi>\<^isub>P where FrP: "extractFrame P = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>" and "A\<^isub>P \<sharp>* \<Psi>" and "A\<^isub>P \<sharp>* \<Psi>'" and "A\<^isub>P \<sharp>* M" and "A\<^isub>P \<sharp>* R" and "A\<^isub>P \<sharp>* P" and "A\<^isub>P \<sharp>* Q"
      by(rule_tac C="(\<Psi>, \<Psi>',M,R,P,Q)" in freshFrame) auto
    obtain A\<^isub>Q \<Psi>\<^isub>Q where FrQ: "extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>" and "A\<^isub>Q \<sharp>* \<Psi>" and "A\<^isub>Q \<sharp>* \<Psi>'" and "A\<^isub>Q \<sharp>* M" and "A\<^isub>Q \<sharp>* R" and "A\<^isub>Q \<sharp>* P" and "A\<^isub>Q \<sharp>* Q" and "A\<^isub>Q \<sharp>* A\<^isub>S"
      by(rule_tac C="(\<Psi>, \<Psi>',M,R,P,Q,A\<^isub>S)" in freshFrame) auto *)
    have "insertAssertion (extractFrame P) \<Psi>' \<simeq>\<^sub>F insertAssertion (extractFrame P) \<Psi>" using `\<Psi> \<simeq> \<Psi>'`
      by(metis AssertionStatEqSym assertionFrameCompose)
    hence "insertAssertion (extractFrame P) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> R" using `insertAssertion (extractFrame P) \<Psi>' \<turnstile>\<^sub>F M \<Leftarrow> R`
      by(metis frameStatEqEnt)
    from `\<Psi> \<rhd> P \<sim>\<^sub>H Q` `insertAssertion (extractFrame P) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> R` `legalClause M R` have "(\<exists>Q'. legalClause M Q' \<and> insertAssertion (extractFrame Q) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> Q' \<and> \<one> \<rhd> R \<sim>\<^sub>H Q')"
      by(blast dest: hobisimE)
    then obtain R' where "legalClause M R'" and "insertAssertion (extractFrame Q) \<Psi> \<turnstile>\<^sub>F (M\<Leftarrow>R')" and "R \<sim>\<^sub>H R'"
      by auto
    have "insertAssertion (extractFrame Q) \<Psi> \<simeq>\<^sub>F insertAssertion (extractFrame Q) \<Psi>'" using `\<Psi> \<simeq> \<Psi>'`
      by(metis assertionFrameCompose)
    with `insertAssertion (extractFrame Q) \<Psi> \<turnstile>\<^sub>F (M\<Leftarrow>R')` have "(insertAssertion (extractFrame Q) \<Psi>') \<turnstile>\<^sub>F (M\<Leftarrow>R')"
      by(metis frameStatEqEnt)
    moreover from `\<Psi> \<simeq> \<Psi>'` `R \<sim>\<^sub>H R'` have "(\<one>,R,R') \<in> ?X"
      by(auto simp add: exI[where x="\<Psi>\<otimes>\<Psi>\<^isub>Q"] Composition)
    ultimately show ?case using `legalClause M R'`
      by blast
  next
    case(cSim \<Psi>' P Q)
    from `(\<Psi>', P, Q) \<in> ?X` obtain \<Psi> where "\<Psi> \<rhd> P \<sim>\<^sub>H Q" and "\<Psi> \<simeq> \<Psi>'"
      by auto
    from `\<Psi> \<rhd> P \<sim>\<^sub>H Q` have "\<Psi> \<rhd> P \<leadsto>[hobisim] Q" by(blast dest: hobisimE)
    moreover have "eqvt ?X"
      by(auto simp add: eqvt_def) (metis hobisimClosed AssertionStatEqClosed)
    hence "eqvt(?X \<union> hobisim)" by auto
    moreover note `\<Psi> \<simeq> \<Psi>'`
    moreover have "\<And>\<Psi> P Q \<Psi>'. \<lbrakk>\<Psi> \<rhd> P \<sim>\<^sub>H Q; \<Psi> \<simeq> \<Psi>'\<rbrakk> \<Longrightarrow> (\<Psi>', P, Q) \<in> ?X \<union> hobisim"
      by auto
    ultimately show ?case
      by(rule statEqSim)
  next
    case(cExt \<Psi>' P Q \<Psi>'')
    from `(\<Psi>', P, Q) \<in> ?X` obtain \<Psi> where "\<Psi> \<rhd> P \<sim>\<^sub>H Q" and "\<Psi> \<simeq> \<Psi>'"
      by auto
    from `\<Psi> \<rhd> P \<sim>\<^sub>H Q` have "\<Psi> \<otimes> \<Psi>'' \<rhd> P \<sim>\<^sub>H Q" by(rule hobisimE)
    moreover from `\<Psi> \<simeq> \<Psi>'` have "\<Psi> \<otimes> \<Psi>'' \<simeq> \<Psi>' \<otimes> \<Psi>''" by(rule Composition)
    ultimately show ?case by blast
  next
    case(cSym \<Psi>' P Q)
    from `(\<Psi>', P, Q) \<in> ?X` obtain \<Psi> where "\<Psi> \<rhd> P \<sim>\<^sub>H Q" and "\<Psi> \<simeq> \<Psi>'"
      by auto
    from `\<Psi> \<rhd> P \<sim>\<^sub>H Q` have "\<Psi> \<rhd> Q \<sim>\<^sub>H P" by(rule hobisimE)
    thus ?case using `\<Psi> \<simeq> \<Psi>'` by auto
  qed
qed

lemma hobisimTransitive:
  fixes \<Psi> :: 'b
  and   P :: "('a, 'b, 'c) psi"
  and   Q :: "('a, 'b, 'c) psi"
  and   R :: "('a, 'b, 'c) psi"

  assumes PQ: "\<Psi> \<rhd> P \<sim>\<^sub>H Q"
  and     QR: "\<Psi> \<rhd> Q \<sim>\<^sub>H R"

  shows "\<Psi> \<rhd> P \<sim>\<^sub>H R"
proof -
  let ?X = "{(\<Psi>, P, R) | \<Psi> P Q R. \<Psi> \<rhd> P \<sim>\<^sub>H Q \<and> \<Psi> \<rhd> Q \<sim>\<^sub>H R}" 
  from PQ QR have "(\<Psi>, P, R) \<in> ?X" by auto
  thus ?thesis
  proof(coinduct rule: hobisimCoinduct)
    case(cStatEq \<Psi> P R \<phi>)
    thus ?case
      by(auto dest: hobisimE frameStatEqEnt)
  next
    case(cClauseEq \<Psi> P R M Q)
    from `(\<Psi>,P,R) \<in> ?X` obtain S where "\<Psi> \<rhd> P \<sim>\<^sub>H S" and "\<Psi> \<rhd> S \<sim>\<^sub>H R" by auto
    from `legalClause M Q` `insertAssertion (extractFrame P) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> Q` `\<Psi> \<rhd> P \<sim>\<^sub>H S`
    have "\<exists> Q'. legalClause M Q' \<and> insertAssertion (extractFrame S) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> Q' \<and> (\<one>,Q,Q') \<in> hobisim"
      by(blast dest: hobisimE)
    then obtain Q' where "legalClause M Q'" "insertAssertion (extractFrame S) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> Q'" "(\<one>,Q,Q') \<in> hobisim"
      by auto
    from `\<Psi> \<rhd> S \<sim>\<^sub>H R` `legalClause M Q'` `insertAssertion (extractFrame S) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> Q'`
    have "\<exists> Q''. legalClause M Q'' \<and> insertAssertion (extractFrame R) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> Q'' \<and> (\<one>,Q',Q'') \<in> hobisim"
      by(blast dest: hobisimE)
    then obtain Q'' where "legalClause M Q''" "insertAssertion (extractFrame R) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> Q''" "(\<one>,Q',Q'') \<in> hobisim"
      by auto
    note `legalClause M Q''` `insertAssertion (extractFrame R) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> Q''`
    moreover from `(\<one>,Q,Q') \<in> hobisim` `(\<one>,Q',Q'') \<in> hobisim` have "(\<one>,Q,Q'') \<in> ?X"
      by blast
    ultimately show ?case
      by blast
  next
    case(cSim \<Psi> P R)
    {
      fix \<Psi> P Q R
      assume "\<Psi> \<rhd> P \<leadsto>[hobisim] Q" and "\<Psi> \<rhd> Q \<leadsto>[hobisim] R"
      moreover have "eqvt ?X"
	by(force simp add: eqvt_def dest: hobisimClosed)
      with hobisimEqvt have "eqvt (?X \<union> hobisim)" by blast
      moreover have "?X \<subseteq> ?X \<union> hobisim" by auto
      ultimately have "\<Psi> \<rhd> P \<leadsto>[(?X \<union> hobisim)] R"
	by(force intro: transitive)
    }
    with `(\<Psi>, P, R) \<in> ?X` show ?case
      by(blast dest: hobisimE)
  next
    case(cExt \<Psi> P R \<Psi>')
    thus ?case by(blast dest: hobisimE)
  next
    case(cSym \<Psi> P R)
    thus ?case by(blast dest: hobisimE)
  qed
qed

(*
lemma hobisimAssertionParClosed:
  assumes "\<Psi>\<otimes>\<Psi>' \<rhd> P \<sim>\<^sub>H Q"
  shows "\<Psi> \<rhd> P \<parallel> \<lbrace>\<Psi>'\<rbrace> \<sim>\<^sub>H Q \<parallel> \<lbrace>\<Psi>'\<rbrace>"
proof -
  let ?X = "{(\<Psi>,P \<parallel> \<lbrace>\<Psi>'\<rbrace>,Q \<parallel> \<lbrace>\<Psi>'\<rbrace>)|\<Psi> P Q \<Psi>'. \<Psi>\<otimes>\<Psi>' \<rhd> P \<sim>\<^sub>H Q}"
  have "eqvt ?X"
    apply(auto simp add: eqvt_def)
    apply(rule_tac x = "p\<bullet>P" in exI)
    apply(rule_tac x = "p\<bullet>Q" in exI)
    apply(rule_tac x = "p\<bullet>\<Psi>'" in exI)
    apply simp
    apply(subst statEqvt'[symmetric])
    by(rule hobisimClosed)
  have "(\<Psi>,P \<parallel> \<lbrace>\<Psi>'\<rbrace>, Q \<parallel> \<lbrace>\<Psi>'\<rbrace>) \<in> ?X"
    using assms by blast
  thus ?thesis
  proof(coinduct rule: hobisimCoinduct)
    case(cExt \<Psi> R S \<Psi>'')
    then obtain \<Psi>' P Q where "R = P \<parallel> \<lbrace>\<Psi>'\<rbrace>" and "S = Q \<parallel> \<lbrace>\<Psi>'\<rbrace>" and "\<Psi>\<otimes>\<Psi>' \<rhd> P \<sim>\<^sub>H Q"
      by blast
    from `\<Psi>\<otimes>\<Psi>' \<rhd> P \<sim>\<^sub>H Q` have "(\<Psi>\<otimes>\<Psi>'')\<otimes>\<Psi>' \<rhd> P \<sim>\<^sub>H Q"
      apply(drule_tac hobisimE(4)[where \<Psi>'=\<Psi>''])
      apply(rule hostatEqBisim)
      apply assumption
      by(rule associativitySym)
    thus ?case unfolding `R = P \<parallel> \<lbrace>\<Psi>'\<rbrace>` `S = Q \<parallel> \<lbrace>\<Psi>'\<rbrace>`
      by auto
  next
    case cSym thus ?case by(auto intro: hobisimE)
  next
    case(cStatEq \<Psi> R S \<phi>)
    then obtain \<Psi>' P Q where "R = P \<parallel> \<lbrace>\<Psi>'\<rbrace>" and "S = Q \<parallel> \<lbrace>\<Psi>'\<rbrace>" and "\<Psi>\<otimes>\<Psi>' \<rhd> P \<sim>\<^sub>H Q"
      by blast
    obtain A\<^isub>P \<Psi>\<^isub>P where "extractFrame P = \<langle>A\<^isub>P,\<Psi>\<^isub>P\<rangle>" and "distinct A\<^isub>P" and "A\<^isub>P \<sharp>* \<Psi>" and "A\<^isub>P \<sharp>* R" and "A\<^isub>P \<sharp>* S" and "A\<^isub>P \<sharp>* \<phi>" and "A\<^isub>P \<sharp>* \<Psi>'" and "A\<^isub>P \<sharp>* P" and "A\<^isub>P \<sharp>* Q"
      by(rule_tac freshFrame[where C="(\<Psi>, R, S, \<phi>, \<Psi>', P, Q)"]) auto
    hence "A\<^isub>P \<sharp>* (\<Psi> \<otimes> \<Psi>')" by auto
    from `insertAssertion (extractFrame R) \<Psi> \<turnstile>\<^sub>F \<phi>`
    have "insertAssertion (extractFrame P) (\<Psi>\<otimes>\<Psi>') \<turnstile>\<^sub>F \<phi>"
      unfolding `R = P \<parallel> \<lbrace>\<Psi>'\<rbrace>`
      apply(rule_tac frameStatEqEnt[where F="insertAssertion (extractFrame (P \<parallel> \<lbrace>\<Psi>'\<rbrace>)) \<Psi>"])
      unfolding extractFrameSimps `extractFrame P = \<langle>A\<^isub>P,\<Psi>\<^isub>P\<rangle>` mergeFrameSimps[OF `A\<^isub>P \<sharp>* \<Psi>'`] insertAssertionSimps[OF `A\<^isub>P \<sharp>* (\<Psi> \<otimes> \<Psi>')`] insertAssertionSimps[OF `A\<^isub>P \<sharp>* \<Psi>`]
      by(metis AssertionStatEqTrans Associativity FrameStatEqSym associativitySym frameNilStatEq frameResChainPres)
    moreover note `\<Psi>\<otimes>\<Psi>' \<rhd> P \<sim>\<^sub>H Q`
    ultimately have "insertAssertion (extractFrame Q) (\<Psi>\<otimes>\<Psi>') \<turnstile>\<^sub>F \<phi>" using `\<not> clauseDom \<phi>`
      by(blast dest: hobisimE)
    obtain A\<^isub>Q \<Psi>\<^isub>Q where "extractFrame Q = \<langle>A\<^isub>Q,\<Psi>\<^isub>Q\<rangle>" and "distinct A\<^isub>Q" and "A\<^isub>Q \<sharp>* \<Psi>" and "A\<^isub>Q \<sharp>* R" and "A\<^isub>Q \<sharp>* S" and "A\<^isub>Q \<sharp>* \<phi>" and "A\<^isub>Q \<sharp>* \<Psi>'" and "A\<^isub>Q \<sharp>* P" and "A\<^isub>Q \<sharp>* Q"
      by(rule_tac freshFrame[where C="(\<Psi>, R, S, \<phi>, \<Psi>', P, Q)"]) auto
    hence "A\<^isub>Q \<sharp>* (\<Psi> \<otimes> \<Psi>')" by auto
    from `insertAssertion (extractFrame Q) (\<Psi>\<otimes>\<Psi>') \<turnstile>\<^sub>F \<phi>`
    show ?case
      unfolding `S = Q \<parallel> \<lbrace>\<Psi>'\<rbrace>`
      apply(rule_tac frameStatEqEnt[where F="insertAssertion (extractFrame Q) (\<Psi>\<otimes>\<Psi>')"])
      unfolding extractFrameSimps `extractFrame Q = \<langle>A\<^isub>Q,\<Psi>\<^isub>Q\<rangle>` mergeFrameSimps[OF `A\<^isub>Q \<sharp>* \<Psi>'`] insertAssertionSimps[OF `A\<^isub>Q \<sharp>* (\<Psi> \<otimes> \<Psi>')`] insertAssertionSimps[OF `A\<^isub>Q \<sharp>* \<Psi>`]
      by(metis AssertionStatEqTrans Associativity FrameStatEqSym associativitySym frameNilStatEq frameResChainPres)
  next
    case(cClauseEq \<Psi> R S M P')
    thus ?case
    then obtain \<Psi>' P Q where "R = P \<parallel> \<lbrace>\<Psi>'\<rbrace>" and "S = Q \<parallel> \<lbrace>\<Psi>'\<rbrace>" and "\<Psi>\<otimes>\<Psi>' \<rhd> P \<sim>\<^sub>H Q"
      by blast
    obtain A\<^isub>P \<Psi>\<^isub>P where "extractFrame P = \<langle>A\<^isub>P,\<Psi>\<^isub>P\<rangle>" and "distinct A\<^isub>P" and "A\<^isub>P \<sharp>* \<Psi>" and "A\<^isub>P \<sharp>* R" and "A\<^isub>P \<sharp>* S" and "A\<^isub>P \<sharp>* M" and "A\<^isub>P \<sharp>* P'" and "A\<^isub>P \<sharp>* \<Psi>'" and "A\<^isub>P \<sharp>* P" and "A\<^isub>P \<sharp>* Q" and "A\<^isub>P \<sharp>* A\<^isub>S" and "A\<^isub>P \<sharp>* \<Psi>\<^isub>S"
      by(rule_tac freshFrame[where C="(\<Psi>, R, S, M, P', \<Psi>', P, Q, A\<^isub>S, \<Psi>\<^isub>S)"]) auto
    hence "A\<^isub>P \<sharp>* (\<Psi> \<otimes> \<Psi>')" by auto
    from `insertAssertion (extractFrame R) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> P'`
    have "insertAssertion (extractFrame P) (\<Psi>\<otimes>\<Psi>') \<turnstile>\<^sub>F M \<Leftarrow> P'"
      unfolding `R = P \<parallel> \<lbrace>\<Psi>'\<rbrace>`
      apply(rule_tac frameStatEqEnt[where F="insertAssertion (extractFrame (P \<parallel> \<lbrace>\<Psi>'\<rbrace>)) \<Psi>"])
      unfolding extractFrameSimps `extractFrame P = \<langle>A\<^isub>P,\<Psi>\<^isub>P\<rangle>` mergeFrameSimps[OF `A\<^isub>P \<sharp>* \<Psi>'`] insertAssertionSimps[OF `A\<^isub>P \<sharp>* (\<Psi> \<otimes> \<Psi>')`] insertAssertionSimps[OF `A\<^isub>P \<sharp>* \<Psi>`]
      by(metis AssertionStatEqTrans Associativity FrameStatEqSym associativitySym frameNilStatEq frameResChainPres)
    moreover note `\<Psi>\<otimes>\<Psi>' \<rhd> P \<sim>\<^sub>H Q`
    moreover obtain A\<^isub>Q \<Psi>\<^isub>Q where "extractFrame Q = \<langle>A\<^isub>Q,\<Psi>\<^isub>Q\<rangle>" and "distinct A\<^isub>Q" and "A\<^isub>Q \<sharp>* \<Psi>" and "A\<^isub>Q \<sharp>* R" and "A\<^isub>Q \<sharp>* S" and "A\<^isub>Q \<sharp>* M" and "A\<^isub>Q \<sharp>* P'" and "A\<^isub>Q \<sharp>* \<Psi>'" and "A\<^isub>Q \<sharp>* P" and "A\<^isub>Q \<sharp>* Q" and "A\<^isub>Q \<sharp>* A\<^isub>S" and "A\<^isub>Q \<sharp>* \<Psi>\<^isub>S"
      by(rule_tac freshFrame[where C="(\<Psi>, R, S, M, P', \<Psi>', P, Q, A\<^isub>S, \<Psi>\<^isub>S)"]) auto
    ultimately obtain Q' where "(\<langle>A\<^isub>Q,(\<Psi>\<otimes>\<Psi>')\<otimes>\<Psi>\<^isub>Q\<rangle>) \<turnstile>\<^sub>F M \<Leftarrow> Q'" and "legalClause M Q'" and "(\<Psi>\<otimes>\<Psi>')\<otimes>\<Psi>\<^isub>Q \<rhd> P' \<sim>\<^sub>H Q'" using `legalClause M P'`
      apply(drule_tac hobisimE(2))
      by auto
    from `(\<langle>A\<^isub>Q,(\<Psi>\<otimes>\<Psi>')\<otimes>\<Psi>\<^isub>Q\<rangle>) \<turnstile>\<^sub>F M \<Leftarrow> Q'`
    have "insertAssertion (extractFrame S) \<Psi> \<turnstile>\<^sub>F M \<Leftarrow> Q'"
      unfolding `S = Q \<parallel> \<lbrace>\<Psi>'\<rbrace>` `extractFrame Q = \<langle>A\<^isub>Q,\<Psi>\<^isub>Q\<rangle>` extractFrameSimps
      mergeFrameSimps[OF `A\<^isub>Q \<sharp>* \<Psi>'`]
      unfolding insertAssertionSimps[OF `A\<^isub>Q \<sharp>* \<Psi>`]
      by (metis Associativity FrameImp_def `A\<^isub>Q \<sharp>* M` `legalClause M Q'` associativitySym frameImpE freshClauseChain(1) legalClauseFresh statEqEnt)
    hence "(\<langle>A\<^isub>S,\<Psi>\<otimes>\<Psi>\<^isub>S\<rangle>) \<turnstile>\<^sub>F M \<Leftarrow> Q'"
      by(simp add: cClauseEq insertAssertionSimps)
    moreover have "\<Psi> \<otimes> \<Psi>\<^isub>S \<rhd> P' \<sim>\<^sub>H Q'"
    proof -
      from `(\<Psi>\<otimes>\<Psi>')\<otimes>\<Psi>\<^isub>Q \<rhd> P' \<sim>\<^sub>H Q'` have "\<Psi>\<otimes>(\<Psi>\<^isub>Q\<otimes>\<Psi>') \<rhd> P' \<sim>\<^sub>H Q'"
        apply(rule hostatEqBisim)
        by (metis AssertionStatEqTrans Associativity associativitySym)
      have "extractFrame S = \<langle>A\<^isub>Q,\<Psi>\<^isub>Q \<otimes> \<Psi>'\<rangle>"
        unfolding `S = Q \<parallel> \<lbrace>\<Psi>'\<rbrace>` `extractFrame Q = \<langle>A\<^isub>Q,\<Psi>\<^isub>Q\<rangle>` extractFrameSimps
        mergeFrameSimps[OF `A\<^isub>Q \<sharp>* \<Psi>'`]
        by(rule refl)
      hence "\<langle>A\<^isub>S,\<Psi>\<^isub>S\<rangle> = \<langle>A\<^isub>Q,\<Psi>\<^isub>Q \<otimes> \<Psi>'\<rangle>"
        unfolding `extractFrame S = \<langle>A\<^isub>S,\<Psi>\<^isub>S\<rangle>`
        by simp
      with `A\<^isub>Q \<sharp>* A\<^isub>S` obtain p where "set p \<subseteq> set A\<^isub>S \<times> set A\<^isub>Q" and "distinctPerm p" and "\<Psi>\<^isub>Q \<otimes> \<Psi>' = p \<bullet> \<Psi>\<^isub>S"
        by(rule_tac frameChainEq) (blast|simp)+
      thus ?thesis using `\<Psi>\<otimes>(\<Psi>\<^isub>Q\<otimes>\<Psi>') \<rhd> P' \<sim>\<^sub>H Q'` `legalClause M P'` `legalClause M Q'` `A\<^isub>S \<sharp>* \<Psi>` `A\<^isub>Q \<sharp>* \<Psi>` `A\<^isub>S \<sharp>* M` `A\<^isub>Q \<sharp>* M`
        apply simp
        apply(drule_tac hobisimClosed[where p=p])
        by(simp add: eqvts)
    qed
    moreover note `legalClause M Q'`
    ultimately show ?case
      by blast
  next
    case(cSim \<Psi> R S)
    then obtain \<Psi>' P Q where "R = P \<parallel> \<lbrace>\<Psi>'\<rbrace>" and "S = Q \<parallel> \<lbrace>\<Psi>'\<rbrace>" and "\<Psi>\<otimes>\<Psi>' \<rhd> P \<sim>\<^sub>H Q"
      by blast
    have "\<Psi> \<rhd> P \<parallel> \<lbrace>\<Psi>'\<rbrace> \<leadsto>[?X] Q \<parallel> \<lbrace>\<Psi>'\<rbrace>"
    proof(rule simI[where C="(\<Psi>,R,S,\<Psi>',P,Q)", OF `eqvt ?X`])
      fix \<alpha> Q'
      assume "\<Psi> \<rhd> Q \<parallel> \<lbrace>\<Psi>'\<rbrace> \<longmapsto> \<alpha> \<prec> Q'"
        and  "bn \<alpha> \<sharp>* (P \<parallel> \<lbrace>\<Psi>'\<rbrace>)"
        and  "bn \<alpha> \<sharp>* (Q \<parallel> \<lbrace>\<Psi>'\<rbrace>)"
        and  "bn \<alpha> \<sharp>* \<Psi>"
        and  "distinct (bn \<alpha>)"
        and  "bn \<alpha> \<sharp>* subject \<alpha>"
        and  "bn \<alpha> \<sharp>* (\<Psi>, R, S, \<Psi>', P, Q)"
      note `\<Psi> \<rhd> Q \<parallel> \<lbrace>\<Psi>'\<rbrace> \<longmapsto> \<alpha> \<prec> Q'`
      moreover from `bn \<alpha> \<sharp>* (\<Psi>, R, S, \<Psi>', P, Q)` have "bn \<alpha> \<sharp>* \<Psi>" and "bn \<alpha> \<sharp>* Q" and "bn \<alpha> \<sharp>* \<lbrace>\<Psi>'\<rbrace>"
        by simp+
      moreover note `bn \<alpha> \<sharp>* subject \<alpha>`
      ultimately have "\<exists>Q''. Q' = Q'' \<parallel> \<lbrace>\<Psi>'\<rbrace> \<and> \<Psi>\<otimes>\<Psi>' \<rhd> Q \<longmapsto> \<alpha> \<prec> Q''"
      proof(induct rule: parCases[where C="()"])
        case(cPar1 P' A\<^isub>Q \<Psi>\<^isub>Q)
        thus ?case
          by auto
      next
        case cPar2 thus ?case by(auto dest: nilTrans)
      next
        case cComm1 thus ?case by(auto dest: nilTrans)
      next
        case cComm2 thus ?case by(auto dest: nilTrans)
      qed
      then obtain Q'' where "Q' = Q'' \<parallel> \<lbrace>\<Psi>'\<rbrace>" "\<Psi>\<otimes>\<Psi>' \<rhd> Q \<longmapsto> \<alpha> \<prec> Q''"
        by auto
      note `\<Psi>\<otimes>\<Psi>' \<rhd> P \<sim>\<^sub>H Q`
      with `\<Psi>\<otimes>\<Psi>' \<rhd> Q \<longmapsto> \<alpha> \<prec> Q''`
      have "\<exists>P'. \<Psi>\<otimes>\<Psi>' \<rhd> P \<longmapsto>\<alpha> \<prec> P' \<and> (\<Psi>\<otimes>\<Psi>', P', Q'') \<in> hobisim" using `bn \<alpha> \<sharp>* (\<Psi>, R, S, \<Psi>', P, Q)`
        apply(drule_tac hobisimE(3))
        apply(rule_tac simE)
        apply assumption
        apply assumption
        by auto
      then obtain P' where "\<Psi>\<otimes>\<Psi>' \<rhd> P \<longmapsto>\<alpha> \<prec> P'" "(\<Psi>\<otimes>\<Psi>', P', Q'') \<in> hobisim"
        by auto
      thus "\<exists>P'. \<Psi> \<rhd> P \<parallel> \<lbrace>\<Psi>'\<rbrace> \<longmapsto> \<alpha> \<prec> P' \<and> (\<Psi>, P', Q') \<in> ?X" unfolding `Q' = Q'' \<parallel> \<lbrace>\<Psi>'\<rbrace>` using `bn \<alpha> \<sharp>* (\<Psi>, R, S, \<Psi>', P, Q)` `\<Psi>\<otimes>\<Psi>' \<rhd> P \<sim>\<^sub>H Q`
        apply(rule_tac exI[where x="P' \<parallel> \<lbrace>\<Psi>'\<rbrace>"])
        apply(rule conjI)
        apply(rule Par1[where A\<^isub>Q="[]"])
        apply assumption
        by(auto simp add: psi.inject)
    qed
    thus ?case unfolding `R = P \<parallel> \<lbrace>\<Psi>'\<rbrace>` `S = Q \<parallel> \<lbrace>\<Psi>'\<rbrace>`
      by(auto intro: monotonic)
  qed
qed
*)

(*
lemma hobisimAssertionParClosed':
  assumes "\<Psi> \<rhd> P \<sim>\<^sub>H Q"
  shows "\<Psi> \<rhd> P \<parallel> \<lbrace>\<Psi>'\<rbrace> \<sim>\<^sub>H Q \<parallel> \<lbrace>\<Psi>'\<rbrace>"
proof -
  from assms have "\<Psi>\<otimes>\<Psi>' \<rhd> P \<sim>\<^sub>H Q"
    by(rule hobisimE)
  thus ?thesis
    by(rule hobisimAssertionParClosed)
qed
*)

lemma howeakTransitiveCoinductSC[case_names cStatEq cClauseEq cSim cExt cSym, case_conclusion bisim step, consumes 2]:
  fixes \<Psi> :: 'b
  and   P :: "('a, 'b, 'c) psi"
  and   Q :: "('a, 'b, 'c) psi"
  and   X :: "('b \<times> ('a, 'b, 'c) psi \<times> ('a, 'b, 'c) psi) set"

  assumes p: "(\<Psi>, P, Q) \<in> X"
  and Eqvt: "eqvt X"
  and     rStatEq: "\<And> \<Psi>' R S \<phi>. \<lbrakk>(\<Psi>', R, S) \<in> X; \<not>clauseDom \<phi>; (insertAssertion (extractFrame R)) \<Psi>' \<turnstile>\<^sub>F \<phi>\<rbrakk> \<Longrightarrow> (insertAssertion (extractFrame S)) \<Psi>' \<turnstile>\<^sub>F \<phi>"
  and     rClauseEq: "\<And>\<Psi>' R S M P. \<lbrakk>(\<Psi>', R, S) \<in> X; legalClause M P; (insertAssertion (extractFrame R)) \<Psi>' \<turnstile>\<^sub>F M \<Leftarrow> P\<rbrakk> \<Longrightarrow> \<exists>Q. legalClause M Q \<and> (insertAssertion (extractFrame S) \<Psi>') \<turnstile>\<^sub>F M \<Leftarrow> Q \<and> ((\<one>,P,Q) \<in> X \<or> (\<one>,P,Q) \<in> hobisim)"
  and rSim: "\<And>\<Psi> P Q. (\<Psi>, P, Q) \<in> X \<Longrightarrow> \<Psi> \<rhd> P \<leadsto>[({(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and>
                                                                        (\<Psi>, P', Q') \<in> X \<and>
                                                                        \<Psi> \<rhd> Q' \<sim>\<^sub>H Q})] Q"
  and rExt: "\<And>\<Psi> P Q \<Psi>'. (\<Psi>, P, Q) \<in> X \<Longrightarrow> (\<Psi> \<otimes> \<Psi>', P, Q) \<in> X"
  and rSym: "\<And>\<Psi> P Q. (\<Psi>, P, Q) \<in> X \<Longrightarrow> (\<Psi>, Q, P) \<in> X"

  shows "\<Psi> \<rhd> P \<sim>\<^sub>H Q"
proof -
  let ?X = "{(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and> (\<Psi>, P', Q') \<in> X \<and> \<Psi> \<rhd> Q' \<sim>\<^sub>H Q}"
  from p have "(\<Psi>, P, Q) \<in> ?X"
    by(blast intro: hobisimReflexive)
  thus ?thesis
  proof(coinduct rule: hobisimWeakCoinductSC)
    case(cStatEq \<Psi> P Q \<phi>) thus ?case by(blast dest: hobisimE rStatEq)
  next
    case(cClauseEq \<Psi> P Q M R)
    then obtain P' Q' where "\<Psi> \<rhd> P \<sim>\<^sub>H P'" and "(\<Psi>, P', Q') \<in> X" and "\<Psi> \<rhd> Q' \<sim>\<^sub>H Q" by auto
    from `\<Psi> \<rhd> P \<sim>\<^sub>H P'` `legalClause M R` `insertAssertion (extractFrame P) \<Psi> \<turnstile>\<^sub>F (M \<Leftarrow> R)` obtain R' where "legalClause M R'" and "insertAssertion (extractFrame P') \<Psi> \<turnstile>\<^sub>F (M \<Leftarrow> R')" and "\<one> \<rhd> R \<sim>\<^sub>H R'"
      by(blast dest: hobisimE)
    with `(\<Psi>, P', Q') \<in> X` obtain R'' where "legalClause M R''" and "insertAssertion (extractFrame Q') \<Psi> \<turnstile>\<^sub>F (M \<Leftarrow> R'')" and "(\<one>,R',R'') \<in> X \<or> (\<one>,R',R'') \<in> hobisim"
      by (drule_tac rClauseEq) auto
    with `\<Psi> \<rhd> Q' \<sim>\<^sub>H Q` obtain R''' where "legalClause M R'''" and "insertAssertion (extractFrame Q) \<Psi> \<turnstile>\<^sub>F (M \<Leftarrow> R''')" and "\<one> \<rhd> R'' \<sim>\<^sub>H R'''"
      by(blast dest: hobisimE)
    with `(\<one>,R',R'') \<in> X \<or> (\<one>,R',R'') \<in> hobisim` `\<one> \<rhd> R \<sim>\<^sub>H R'` show ?case
      by(blast intro: hobisimTransitive)
  next
    case(cSim \<Psi> P Q)
    {
      fix \<Psi> P P' Q' Q
      assume "\<Psi> \<rhd> P \<leadsto>[hobisim] P'"
      moreover assume P'RelQ': "(\<Psi>, P', Q') \<in> X"
      hence "\<Psi> \<rhd> P' \<leadsto>[?X] Q'" by(rule rSim)
      moreover from `eqvt X` P'RelQ' have "eqvt ?X"
	apply(auto simp add: eqvt_def)
	apply(drule_tac p=p in hobisimClosed)
	apply(drule_tac p=p in hobisimClosed)
	apply(rule_tac x="p \<bullet> P'a" in exI, simp)
	by(rule_tac x="p \<bullet> Q'a" in exI, auto)
      ultimately have "\<Psi> \<rhd> P \<leadsto>[?X] Q'"
	by(force intro: transitive dest: hobisimTransitive)
      moreover assume "\<Psi> \<rhd> Q' \<leadsto>[hobisim] Q"
      ultimately have "\<Psi> \<rhd> P \<leadsto>[?X] Q" using `eqvt ?X`
	by(force intro: transitive dest: hobisimTransitive)
    }
    with `(\<Psi>, P, Q) \<in> ?X` show ?case
      by(blast dest: hobisimE)
  next
    case(cExt \<Psi> P Q \<Psi>')
    thus ?case by(blast dest: hobisimE intro: rExt)
  next
    case(cSym \<Psi> P Q)
    thus ?case by(blast dest: hobisimE intro: rSym)
  qed
qed

lemma howeakTransitiveCoinduct[case_names cStatEq cClauseEq cSim cExt cSym, case_conclusion bisim step, consumes 2]:
  fixes \<Psi> :: 'b
  and   P :: "('a, 'b, 'c) psi"
  and   Q :: "('a, 'b, 'c) psi"
  and   X :: "('b \<times> ('a, 'b, 'c) psi \<times> ('a, 'b, 'c) psi) set"

  assumes p: "(\<Psi>, P, Q) \<in> X"
  and Eqvt: "eqvt X"
  and     rStatEq: "\<And> \<Psi>' R S \<phi>. \<lbrakk>(\<Psi>', R, S) \<in> X; \<not>clauseDom \<phi>; (insertAssertion (extractFrame R)) \<Psi>' \<turnstile>\<^sub>F \<phi>\<rbrakk> \<Longrightarrow> (insertAssertion (extractFrame S)) \<Psi>' \<turnstile>\<^sub>F \<phi>"
  and     rClauseEq: "\<And>\<Psi>' R S M P. \<lbrakk>(\<Psi>', R, S) \<in> X; legalClause M P; (insertAssertion (extractFrame R)) \<Psi>' \<turnstile>\<^sub>F M \<Leftarrow> P\<rbrakk> \<Longrightarrow> \<exists>Q. legalClause M Q \<and> (insertAssertion (extractFrame S) \<Psi>') \<turnstile>\<^sub>F M \<Leftarrow> Q \<and> (\<one>,P,Q) \<in> X"
  and rSim: "\<And>\<Psi> P Q. (\<Psi>, P, Q) \<in> X \<Longrightarrow> \<Psi> \<rhd> P \<leadsto>[({(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and>
                                                                        (\<Psi>, P', Q') \<in> X \<and>
                                                                        \<Psi> \<rhd> Q' \<sim>\<^sub>H Q})] Q"
  and rExt: "\<And>\<Psi> P Q \<Psi>'. (\<Psi>, P, Q) \<in> X \<Longrightarrow> (\<Psi> \<otimes> \<Psi>', P, Q) \<in> X"
  and rSym: "\<And>\<Psi> P Q. (\<Psi>, P, Q) \<in> X \<Longrightarrow> (\<Psi>, Q, P) \<in> X"

  shows "\<Psi> \<rhd> P \<sim>\<^sub>H Q"
proof -
  let ?X = "{(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and> (\<Psi>, P', Q') \<in> X \<and> \<Psi> \<rhd> Q' \<sim>\<^sub>H Q}"
  from p have "(\<Psi>, P, Q) \<in> ?X"
    by(blast intro: hobisimReflexive)
  thus ?thesis
  proof(coinduct rule: hobisimWeakCoinduct)
    case(cStatEq \<Psi> P Q \<phi>) thus ?case by(blast dest: hobisimE rStatEq)
  next
    case(cClauseEq \<Psi> P Q M R)
    then obtain P' Q' where "\<Psi> \<rhd> P \<sim>\<^sub>H P'" and "(\<Psi>, P', Q') \<in> X" and "\<Psi> \<rhd> Q' \<sim>\<^sub>H Q" by auto
    from `\<Psi> \<rhd> P \<sim>\<^sub>H P'` `legalClause M R` `insertAssertion (extractFrame P) \<Psi> \<turnstile>\<^sub>F (M \<Leftarrow> R)` obtain R' where "legalClause M R'" and "insertAssertion (extractFrame P') \<Psi> \<turnstile>\<^sub>F (M \<Leftarrow> R')" and "\<one> \<rhd> R \<sim>\<^sub>H R'"
      by(blast dest: hobisimE)
    with `(\<Psi>, P', Q') \<in> X` obtain R'' where "legalClause M R''" and "insertAssertion (extractFrame Q') \<Psi> \<turnstile>\<^sub>F (M \<Leftarrow> R'')" and "(\<one>,R',R'') \<in> X"
      by(auto dest: rClauseEq)
    with `\<Psi> \<rhd> Q' \<sim>\<^sub>H Q` obtain R''' where "legalClause M R'''" and "insertAssertion (extractFrame Q) \<Psi> \<turnstile>\<^sub>F (M \<Leftarrow> R''')" and "R'' \<sim>\<^sub>H R'''"
      by(blast dest: hobisimE)
    with `(\<one>,R',R'') \<in> X` `\<one> \<rhd> R \<sim>\<^sub>H R'` show ?case
      by blast
  next
    case(cSim \<Psi> P Q)
    {
      fix \<Psi> P P' Q' Q
      assume "\<Psi> \<rhd> P \<leadsto>[hobisim] P'"
      moreover assume P'RelQ': "(\<Psi>, P', Q') \<in> X"
      hence "\<Psi> \<rhd> P' \<leadsto>[?X] Q'" by(rule rSim)
      moreover from `eqvt X` P'RelQ' have "eqvt ?X"
	apply(auto simp add: eqvt_def)
	apply(drule_tac p=p in hobisimClosed)
	apply(drule_tac p=p in hobisimClosed)
	apply(rule_tac x="p \<bullet> P'a" in exI, simp)
	by(rule_tac x="p \<bullet> Q'a" in exI, auto)
      ultimately have "\<Psi> \<rhd> P \<leadsto>[?X] Q'"
	by(force intro: transitive dest: hobisimTransitive)
      moreover assume "\<Psi> \<rhd> Q' \<leadsto>[hobisim] Q"
      ultimately have "\<Psi> \<rhd> P \<leadsto>[?X] Q" using `eqvt ?X`
	by(force intro: transitive dest: hobisimTransitive)
    }
    with `(\<Psi>, P, Q) \<in> ?X` show ?case
      by(blast dest: hobisimE)
  next
    case(cExt \<Psi> P Q \<Psi>')
    thus ?case by(blast dest: hobisimE intro: rExt)
  next
    case(cSym \<Psi> P Q)
    thus ?case by(blast dest: hobisimE intro: rSym)
  qed
qed

lemma howeakTransitiveCoinduct'[case_names cStatEq cClauseEq cSim cExt cSym, case_conclusion bisim step, consumes 2]:
  fixes \<Psi> :: 'b
  and   P :: "('a, 'b, 'c) psi"
  and   Q :: "('a, 'b, 'c) psi"
  and   X :: "('b \<times> ('a, 'b, 'c) psi \<times> ('a, 'b, 'c) psi) set"

  assumes p: "(\<Psi>, P, Q) \<in> X"
  and Eqvt: "eqvt X"
  and     rStatEq: "\<And> \<Psi>' R S \<phi>. \<lbrakk>(\<Psi>', R, S) \<in> X; \<not>clauseDom \<phi>; (insertAssertion (extractFrame R)) \<Psi>' \<turnstile>\<^sub>F \<phi>\<rbrakk> \<Longrightarrow> (insertAssertion (extractFrame S)) \<Psi>' \<turnstile>\<^sub>F \<phi>"
  and     rClauseEq: "\<And>\<Psi>' R S M P. \<lbrakk>(\<Psi>', R, S) \<in> X; legalClause M P; (insertAssertion (extractFrame R)) \<Psi>' \<turnstile>\<^sub>F M \<Leftarrow> P\<rbrakk> \<Longrightarrow> \<exists>Q. legalClause M Q \<and> (insertAssertion (extractFrame S) \<Psi>') \<turnstile>\<^sub>F M \<Leftarrow> Q \<and> (\<one>,P,Q) \<in> X"
  and rSim: "\<And>\<Psi> P Q. (\<Psi>, P, Q) \<in> X \<Longrightarrow> \<Psi> \<rhd> P \<leadsto>[({(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and>
                                                                        (\<Psi>, P', Q') \<in> X \<and>
                                                                        \<Psi> \<rhd> Q' \<sim>\<^sub>H Q})] Q"
  and rExt: "\<And>\<Psi> P Q \<Psi>'. (\<Psi>, P, Q) \<in> X \<Longrightarrow> (\<Psi> \<otimes> \<Psi>', P, Q) \<in> X"
  and rSym: "\<And>\<Psi> P Q. (\<Psi>, P, Q) \<in> X \<Longrightarrow>
                      (\<Psi>, Q, P) \<in> {(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and> (\<Psi>, P', Q') \<in> X \<and> \<Psi> \<rhd> Q' \<sim>\<^sub>H Q}"

  shows "\<Psi> \<rhd> P \<sim>\<^sub>H Q"
proof -
  let ?X = "{(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and> (\<Psi>, P', Q') \<in> X \<and> \<Psi> \<rhd> Q' \<sim>\<^sub>H Q}"
  from p have "(\<Psi>, P, Q) \<in> ?X"
    by(blast intro: hobisimReflexive)
  thus ?thesis
  proof(coinduct rule: hobisimWeakCoinduct)
    case(cStatEq \<Psi> P Q \<phi>) thus ?case by(blast dest: hobisimE rStatEq)
  next
    case(cClauseEq \<Psi> P Q M R)
    then obtain P' Q' where "\<Psi> \<rhd> P \<sim>\<^sub>H P'" and "(\<Psi>, P', Q') \<in> X" and "\<Psi> \<rhd> Q' \<sim>\<^sub>H Q" by auto
    from `\<Psi> \<rhd> P \<sim>\<^sub>H P'` `legalClause M R` `insertAssertion (extractFrame P) \<Psi> \<turnstile>\<^sub>F (M \<Leftarrow> R)` obtain R' where "legalClause M R'" and "insertAssertion (extractFrame P') \<Psi> \<turnstile>\<^sub>F (M \<Leftarrow> R')" and "R \<sim>\<^sub>H R'"
      by(blast dest: hobisimE)
    with `(\<Psi>, P', Q') \<in> X` obtain R'' where "legalClause M R''" and "insertAssertion (extractFrame Q') \<Psi> \<turnstile>\<^sub>F (M \<Leftarrow> R'')" and "(\<one>,R',R'') \<in> X"
      by(auto dest: rClauseEq)
    with `\<Psi> \<rhd> Q' \<sim>\<^sub>H Q` obtain R''' where "legalClause M R'''" and "insertAssertion (extractFrame Q) \<Psi> \<turnstile>\<^sub>F (M \<Leftarrow> R''')" and "R'' \<sim>\<^sub>H R'''"
      by(blast dest: hobisimE)
    with `(\<one>,R',R'') \<in> X` `R \<sim>\<^sub>H R'` show ?case
      by blast
  next
    case(cSim \<Psi> P Q)
    {
      fix \<Psi> P P' Q' Q
      assume "\<Psi> \<rhd> P \<leadsto>[hobisim] P'"
      moreover assume P'RelQ': "(\<Psi>, P', Q') \<in> X"
      hence "\<Psi> \<rhd> P' \<leadsto>[?X] Q'" by(rule rSim)
      moreover from `eqvt X` P'RelQ' have "eqvt ?X"
	apply(auto simp add: eqvt_def)
	apply(drule_tac p=p in hobisimClosed)
	apply(drule_tac p=p in hobisimClosed)
	apply(rule_tac x="p \<bullet> P'a" in exI, simp)
	by(rule_tac x="p \<bullet> Q'a" in exI, auto)
      ultimately have "\<Psi> \<rhd> P \<leadsto>[?X] Q'"
	by(force intro: transitive dest: hobisimTransitive)
      moreover assume "\<Psi> \<rhd> Q' \<leadsto>[hobisim] Q"
      ultimately have "\<Psi> \<rhd> P \<leadsto>[?X] Q" using `eqvt ?X`
	by(force intro: transitive dest: hobisimTransitive)
    }
    with `(\<Psi>, P, Q) \<in> ?X` show ?case
      by(blast dest: hobisimE)
  next
    case(cExt \<Psi> P Q \<Psi>')
    thus ?case by(blast dest: hobisimE intro: rExt)
  next
    case(cSym \<Psi> P Q)
    thus ?case
      apply auto
      apply(drule rSym)
      apply auto
      by(metis hobisimTransitive hobisimE(5))
  qed
qed

lemma howeakTransitiveCoinduct''[case_names cStatEq cClauseEq cSim cExt cSym, case_conclusion bisim step, consumes 2]:
  fixes \<Psi> :: 'b
  and   P :: "('a, 'b, 'c) psi"
  and   Q :: "('a, 'b, 'c) psi"
  and   X :: "('b \<times> ('a, 'b, 'c) psi \<times> ('a, 'b, 'c) psi) set"

  assumes p: "(\<Psi>, P, Q) \<in> X"
  and Eqvt: "eqvt X"
  and     rStatEq: "\<And> \<Psi>' R S \<phi>. \<lbrakk>(\<Psi>', R, S) \<in> X; \<not>clauseDom \<phi>; (insertAssertion (extractFrame R)) \<Psi>' \<turnstile>\<^sub>F \<phi>\<rbrakk> \<Longrightarrow> (insertAssertion (extractFrame S)) \<Psi>' \<turnstile>\<^sub>F \<phi>"
  and     rClauseEq: "\<And>\<Psi>' R S M P. \<lbrakk>(\<Psi>', R, S) \<in> X; legalClause M P; (insertAssertion (extractFrame R)) \<Psi>' \<turnstile>\<^sub>F M \<Leftarrow> P\<rbrakk> \<Longrightarrow> \<exists>Q. legalClause M Q \<and> (insertAssertion (extractFrame S) \<Psi>') \<turnstile>\<^sub>F M \<Leftarrow> Q \<and> (\<one>,P,Q) \<in> X"
  and rSim: "\<And>\<Psi> P Q. (\<Psi>, P, Q) \<in> X \<Longrightarrow> \<Psi> \<rhd> P \<leadsto>[({(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and>
                                                                        (\<Psi>, P', Q') \<in> X \<and>
                                                                        \<Psi> \<rhd> Q' \<sim>\<^sub>H Q})] Q"
  and rExt: "\<And>\<Psi> P Q \<Psi>'. (\<Psi>, P, Q) \<in> X \<Longrightarrow> (\<Psi> \<otimes> \<Psi>', P, Q) \<in> {(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and> (\<Psi>, P', Q') \<in> X \<and> \<Psi> \<rhd> Q' \<sim>\<^sub>H Q}"
  and rSym: "\<And>\<Psi> P Q. (\<Psi>, P, Q) \<in> X \<Longrightarrow>
                      (\<Psi>, Q, P) \<in> {(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and> (\<Psi>, P', Q') \<in> X \<and> \<Psi> \<rhd> Q' \<sim>\<^sub>H Q}"

  shows "\<Psi> \<rhd> P \<sim>\<^sub>H Q"
proof -
  let ?X = "{(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and> (\<Psi>, P', Q') \<in> X \<and> \<Psi> \<rhd> Q' \<sim>\<^sub>H Q}"
  from p have "(\<Psi>, P, Q) \<in> ?X"
    by(blast intro: hobisimReflexive)
  thus ?thesis
  proof(coinduct rule: hobisimWeakCoinduct)
    case(cStatEq \<Psi> P Q \<phi>) thus ?case by(blast dest: hobisimE rStatEq)
  next
    case(cClauseEq \<Psi> P Q M R)
    then obtain P' Q' where "\<Psi> \<rhd> P \<sim>\<^sub>H P'" and "(\<Psi>, P', Q') \<in> X" and "\<Psi> \<rhd> Q' \<sim>\<^sub>H Q" by auto
    from `\<Psi> \<rhd> P \<sim>\<^sub>H P'` `legalClause M R` `insertAssertion (extractFrame P) \<Psi> \<turnstile>\<^sub>F (M \<Leftarrow> R)` obtain R' where "legalClause M R'" and "insertAssertion (extractFrame P') \<Psi> \<turnstile>\<^sub>F (M \<Leftarrow> R')" and "R \<sim>\<^sub>H R'"
      by(blast dest: hobisimE)
    with `(\<Psi>, P', Q') \<in> X` obtain R'' where "legalClause M R''" and "insertAssertion (extractFrame Q') \<Psi> \<turnstile>\<^sub>F (M \<Leftarrow> R'')" and "(\<one>,R',R'') \<in> X"
      by(auto dest: rClauseEq)
    with `\<Psi> \<rhd> Q' \<sim>\<^sub>H Q` obtain R''' where "legalClause M R'''" and "insertAssertion (extractFrame Q) \<Psi> \<turnstile>\<^sub>F (M \<Leftarrow> R''')" and "R'' \<sim>\<^sub>H R'''"
      by(blast dest: hobisimE)
    with `(\<one>,R',R'') \<in> X` `R \<sim>\<^sub>H R'` show ?case
      by blast
  next
    case(cSim \<Psi> P Q)
    {
      fix \<Psi> P P' Q' Q
      assume "\<Psi> \<rhd> P \<leadsto>[hobisim] P'"
      moreover assume P'RelQ': "(\<Psi>, P', Q') \<in> X"
      hence "\<Psi> \<rhd> P' \<leadsto>[?X] Q'" by(rule rSim)
      moreover from `eqvt X` P'RelQ' have "eqvt ?X"
	apply(auto simp add: eqvt_def)
	apply(drule_tac p=p in hobisimClosed)
	apply(drule_tac p=p in hobisimClosed)
	apply(rule_tac x="p \<bullet> P'a" in exI, simp)
	by(rule_tac x="p \<bullet> Q'a" in exI, auto)
      ultimately have "\<Psi> \<rhd> P \<leadsto>[?X] Q'"
	by(force intro: transitive dest: hobisimTransitive)
      moreover assume "\<Psi> \<rhd> Q' \<leadsto>[hobisim] Q"
      ultimately have "\<Psi> \<rhd> P \<leadsto>[?X] Q" using `eqvt ?X`
	by(force intro: transitive dest: hobisimTransitive)
    }
    with `(\<Psi>, P, Q) \<in> ?X` show ?case
      by(blast dest: hobisimE)
  next
    case(cExt \<Psi> P Q \<Psi>')
    then obtain P' Q' where "\<Psi> \<rhd> P \<sim>\<^sub>H P'" and "(\<Psi>, P', Q') \<in> X" and "\<Psi> \<rhd> Q' \<sim>\<^sub>H Q"
      by auto
    have "\<Psi> \<otimes> \<Psi>'\<rhd> P \<sim>\<^sub>H P'" using `\<Psi> \<rhd> P \<sim>\<^sub>H P'`
      by(rule hobisimE)
    moreover have "\<Psi> \<otimes> \<Psi>' \<rhd> Q' \<sim>\<^sub>H Q" using `\<Psi> \<rhd> Q' \<sim>\<^sub>H Q`
      by(rule hobisimE)
    moreover have "(\<Psi>\<otimes>\<Psi>', P', Q') \<in> ?X" using `(\<Psi>, P', Q') \<in> X`
      by(rule rExt)
    ultimately show ?case
      by(blast intro: hobisimTransitive)
  next
    case(cSym \<Psi> P Q)
    thus ?case
      apply auto
      apply(drule rSym)
      apply auto
      by(metis hobisimTransitive hobisimE(5))
  qed
qed

lemma howeakTransitiveCoinduct'''[case_names cStatEq cClauseEq cSim cExt cSym, case_conclusion bisim step, consumes 2]:
  fixes \<Psi> :: 'b
  and   P :: "('a, 'b, 'c) psi"
  and   Q :: "('a, 'b, 'c) psi"
  and   X :: "('b \<times> ('a, 'b, 'c) psi \<times> ('a, 'b, 'c) psi) set"

  assumes p: "(\<Psi>, P, Q) \<in> X"
  and Eqvt: "eqvt X"
  and     rStatEq: "\<And> \<Psi>' R S \<phi>. \<lbrakk>(\<Psi>', R, S) \<in> X; \<not>clauseDom \<phi>; (insertAssertion (extractFrame R)) \<Psi>' \<turnstile>\<^sub>F \<phi>\<rbrakk> \<Longrightarrow> (insertAssertion (extractFrame S)) \<Psi>' \<turnstile>\<^sub>F \<phi>"
  and     rClauseEq: "\<And>\<Psi>' R S M P. \<lbrakk>(\<Psi>', R, S) \<in> X; legalClause M P; (insertAssertion (extractFrame R)) \<Psi>' \<turnstile>\<^sub>F M \<Leftarrow> P\<rbrakk> \<Longrightarrow> \<exists>Q. legalClause M Q \<and> (insertAssertion (extractFrame S) \<Psi>') \<turnstile>\<^sub>F M \<Leftarrow> Q \<and> (\<one>,P,Q) \<in> {(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and> (\<Psi>, P', Q') \<in> X \<and> \<Psi> \<rhd> Q' \<sim>\<^sub>H Q}"
  and rSim: "\<And>\<Psi> P Q. (\<Psi>, P, Q) \<in> X \<Longrightarrow> \<Psi> \<rhd> P \<leadsto>[({(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and>
                                                                        (\<Psi>, P', Q') \<in> X \<and>
                                                                        \<Psi> \<rhd> Q' \<sim>\<^sub>H Q})] Q"
  and rExt: "\<And>\<Psi> P Q \<Psi>'. (\<Psi>, P, Q) \<in> X \<Longrightarrow> (\<Psi> \<otimes> \<Psi>', P, Q) \<in> {(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and> (\<Psi>, P', Q') \<in> X \<and> \<Psi> \<rhd> Q' \<sim>\<^sub>H Q}"
  and rSym: "\<And>\<Psi> P Q. (\<Psi>, P, Q) \<in> X \<Longrightarrow>
                      (\<Psi>, Q, P) \<in> {(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and> (\<Psi>, P', Q') \<in> X \<and> \<Psi> \<rhd> Q' \<sim>\<^sub>H Q}"

  shows "\<Psi> \<rhd> P \<sim>\<^sub>H Q"
proof -
  let ?X = "{(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and> (\<Psi>, P', Q') \<in> X \<and> \<Psi> \<rhd> Q' \<sim>\<^sub>H Q}"
  from p have "(\<Psi>, P, Q) \<in> ?X"
    by(blast intro: hobisimReflexive)
  thus ?thesis
  proof(coinduct rule: hobisimWeakCoinduct)
    case(cStatEq \<Psi> P Q \<phi>) thus ?case by(blast dest: hobisimE rStatEq)
  next
    case(cClauseEq \<Psi> P Q M R)
    then obtain P' Q' where "\<Psi> \<rhd> P \<sim>\<^sub>H P'" and "(\<Psi>, P', Q') \<in> X" and "\<Psi> \<rhd> Q' \<sim>\<^sub>H Q" by auto
    from `\<Psi> \<rhd> P \<sim>\<^sub>H P'` `legalClause M R` `insertAssertion (extractFrame P) \<Psi> \<turnstile>\<^sub>F (M \<Leftarrow> R)` obtain R' where "legalClause M R'" and "insertAssertion (extractFrame P') \<Psi> \<turnstile>\<^sub>F (M \<Leftarrow> R')" and "R \<sim>\<^sub>H R'"
      by(blast dest: hobisimE)
    with `(\<Psi>, P', Q') \<in> X` obtain R'' where "legalClause M R''" and "insertAssertion (extractFrame Q') \<Psi> \<turnstile>\<^sub>F (M \<Leftarrow> R'')" and "(\<one>,R',R'') \<in> ?X"
      by(drule_tac rClauseEq) blast+
    with `\<Psi> \<rhd> Q' \<sim>\<^sub>H Q` obtain R''' where "legalClause M R'''" and "insertAssertion (extractFrame Q) \<Psi> \<turnstile>\<^sub>F (M \<Leftarrow> R''')" and "R'' \<sim>\<^sub>H R'''"
      by(auto dest: hobisimE)
    have "(\<one>,R,R''') \<in> ?X" using `\<one> \<rhd> R'' \<sim>\<^sub>H R'''` `(\<one>,R',R'') \<in> ?X` `R \<sim>\<^sub>H R'`
      by(blast dest: hobisimTransitive)
    thus ?case using `legalClause M R'''` and `insertAssertion (extractFrame Q) \<Psi> \<turnstile>\<^sub>F (M \<Leftarrow> R''')`
      by blast
  next
    case(cSim \<Psi> P Q)
    {
      fix \<Psi> P P' Q' Q
      assume "\<Psi> \<rhd> P \<leadsto>[hobisim] P'"
      moreover assume P'RelQ': "(\<Psi>, P', Q') \<in> X"
      hence "\<Psi> \<rhd> P' \<leadsto>[?X] Q'" by(rule rSim)
      moreover from `eqvt X` P'RelQ' have "eqvt ?X"
	apply(auto simp add: eqvt_def)
	apply(drule_tac p=p in hobisimClosed)
	apply(drule_tac p=p in hobisimClosed)
	apply(rule_tac x="p \<bullet> P'a" in exI, simp)
	by(rule_tac x="p \<bullet> Q'a" in exI, auto)
      ultimately have "\<Psi> \<rhd> P \<leadsto>[?X] Q'"
	by(force intro: transitive dest: hobisimTransitive)
      moreover assume "\<Psi> \<rhd> Q' \<leadsto>[hobisim] Q"
      ultimately have "\<Psi> \<rhd> P \<leadsto>[?X] Q" using `eqvt ?X`
	by(force intro: transitive dest: hobisimTransitive)
    }
    with `(\<Psi>, P, Q) \<in> ?X` show ?case
      by(blast dest: hobisimE)
  next
    case(cExt \<Psi> P Q \<Psi>')
    then obtain P' Q' where "\<Psi> \<rhd> P \<sim>\<^sub>H P'" and "(\<Psi>, P', Q') \<in> X" and "\<Psi> \<rhd> Q' \<sim>\<^sub>H Q"
      by auto
    have "\<Psi> \<otimes> \<Psi>'\<rhd> P \<sim>\<^sub>H P'" using `\<Psi> \<rhd> P \<sim>\<^sub>H P'`
      by(rule hobisimE)
    moreover have "\<Psi> \<otimes> \<Psi>' \<rhd> Q' \<sim>\<^sub>H Q" using `\<Psi> \<rhd> Q' \<sim>\<^sub>H Q`
      by(rule hobisimE)
    moreover have "(\<Psi>\<otimes>\<Psi>', P', Q') \<in> ?X" using `(\<Psi>, P', Q') \<in> X`
      by(rule rExt)
    ultimately show ?case
      by(blast intro: hobisimTransitive)
  next
    case(cSym \<Psi> P Q)
    thus ?case
      apply auto
      apply(drule rSym)
      apply auto
      by(metis hobisimTransitive hobisimE(5))
  qed
qed

lemma hotransitiveCoinduct[case_names cStatEq cClauseEq cSim cExt cSym, case_conclusion bisim step, consumes 2]:
  fixes \<Psi> :: 'b
  and   P :: "('a, 'b, 'c) psi"
  and   Q :: "('a, 'b, 'c) psi"
  and   X :: "('b \<times> ('a, 'b, 'c) psi \<times> ('a, 'b, 'c) psi) set"

  assumes p: "(\<Psi>, P, Q) \<in> X"
  and Eqvt: "eqvt X"
  and     rStatEq: "\<And> \<Psi>' R S \<phi>. \<lbrakk>(\<Psi>', R, S) \<in> X; \<not>clauseDom \<phi>; (insertAssertion (extractFrame R)) \<Psi>' \<turnstile>\<^sub>F \<phi>\<rbrakk> \<Longrightarrow> (insertAssertion (extractFrame S)) \<Psi>' \<turnstile>\<^sub>F \<phi>"
  and     rClauseEq: "\<And>\<Psi>' R S M P. \<lbrakk>(\<Psi>', R, S) \<in> X; legalClause M P; (insertAssertion (extractFrame R)) \<Psi>' \<turnstile>\<^sub>F M \<Leftarrow> P\<rbrakk> \<Longrightarrow> \<exists>Q. legalClause M Q \<and> (insertAssertion (extractFrame S) \<Psi>') \<turnstile>\<^sub>F M \<Leftarrow> Q \<and> ((\<one>,P,Q) \<in> X \<or> P \<sim>\<^sub>H Q)"
  and rSim: "\<And>\<Psi>' R S. (\<Psi>', R, S) \<in> X \<Longrightarrow> \<Psi>' \<rhd> R \<leadsto>[({(\<Psi>', R, S) | \<Psi>' R R' S' S. \<Psi>' \<rhd> R \<sim>\<^sub>H R' \<and>
                                                                        ((\<Psi>', R', S') \<in> X \<or> \<Psi>' \<rhd> R' \<sim>\<^sub>H S') \<and>
                                                                        \<Psi>' \<rhd> S' \<sim>\<^sub>H S})] S"
  and rExt: "\<And>\<Psi>' R S \<Psi>''. (\<Psi>', R, S) \<in> X \<Longrightarrow> (\<Psi>' \<otimes> \<Psi>'', R, S) \<in> X \<or> \<Psi>' \<otimes> \<Psi>'' \<rhd> R \<sim>\<^sub>H S"
  and rSym: "\<And>\<Psi>' R S. (\<Psi>', R, S) \<in> X \<Longrightarrow> (\<Psi>', S, R) \<in> X \<or> \<Psi>' \<rhd> S \<sim>\<^sub>H R"


  shows "\<Psi> \<rhd> P \<sim>\<^sub>H Q"
proof -
  from p have "(\<Psi>, P, Q) \<in> (X \<union> hobisim)"
    by blast
  moreover from `eqvt X` hobisimEqvt have "eqvt (X \<union> hobisim)"
    by auto
  ultimately show ?thesis
  proof(coinduct rule: howeakTransitiveCoinduct')
    case(cStatEq \<Psi> P Q \<phi>)
    from `(\<Psi>, P, Q) \<in> X \<union> hobisim` have "(\<Psi>, P, Q) \<in> X \<or> (\<Psi>, P, Q) \<in> hobisim"
      by simp
    thus ?case
    proof(rule disjE)
      assume "(\<Psi>, P, Q) \<in> X"
      show ?case
        by(rule rStatEq) fact+
    next
      assume "\<Psi> \<rhd> P \<sim>\<^sub>H Q"
      with `insertAssertion (extractFrame P) \<Psi> \<turnstile>\<^sub>F \<phi>` `\<not> clauseDom \<phi>`
      show ?case by(blast dest: hobisimE)
    qed
  next
    case(cClauseEq \<Psi> P Q M R)
    thus ?case by(auto dest: rClauseEq hobisimE)
  next
    case(cSim \<Psi> P Q)
    thus ?case
      apply auto
      apply(blast intro: rSim)
      apply(drule hobisimE(3))
      apply(rule_tac A=hobisim in monotonic, simp)
      by(force intro: hobisimReflexive)
  next
    case(cExt \<Psi> P Q \<Psi>')
    thus ?case
      by(blast dest: hobisimE rExt)
  next
    case(cSym \<Psi> P Q)
    thus ?case by(blast dest: hobisimE rSym intro: hobisimReflexive)
  qed
qed

lemma hotransitiveCoinduct'[case_names cStatEq cClauseEq cSim cExt cSym, case_conclusion bisim step, consumes 2]:
  fixes \<Psi> :: 'b
  and   P :: "('a, 'b, 'c) psi"
  and   Q :: "('a, 'b, 'c) psi"
  and   X :: "('b \<times> ('a, 'b, 'c) psi \<times> ('a, 'b, 'c) psi) set"

  assumes p: "(\<Psi>, P, Q) \<in> X"
  and Eqvt: "eqvt X"
  and     rStatEq: "\<And> \<Psi>' R S \<phi>. \<lbrakk>(\<Psi>', R, S) \<in> X; \<not>clauseDom \<phi>; (insertAssertion (extractFrame R)) \<Psi>' \<turnstile>\<^sub>F \<phi>\<rbrakk> \<Longrightarrow> (insertAssertion (extractFrame S)) \<Psi>' \<turnstile>\<^sub>F \<phi>"
  and     rClauseEq: "\<And>\<Psi>' R S M P. \<lbrakk>(\<Psi>', R, S) \<in> X; legalClause M P; (insertAssertion (extractFrame R)) \<Psi>' \<turnstile>\<^sub>F M \<Leftarrow> P\<rbrakk> \<Longrightarrow> \<exists>Q. legalClause M Q \<and> (insertAssertion (extractFrame S) \<Psi>') \<turnstile>\<^sub>F M \<Leftarrow> Q \<and> ((\<one>,P,Q) \<in> X \<or> P \<sim>\<^sub>H Q)"
  and rSim: "\<And>\<Psi> P Q. (\<Psi>, P, Q) \<in> X \<Longrightarrow> \<Psi> \<rhd> P \<leadsto>[({(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and>
                                                                        (\<Psi>, P', Q') \<in> (X \<union> hobisim) \<and>
                                                                        \<Psi> \<rhd> Q' \<sim>\<^sub>H Q})] Q"
  and rExt: "\<And>\<Psi> P Q \<Psi>'. (\<Psi>, P, Q) \<in> X \<Longrightarrow> (\<Psi> \<otimes> \<Psi>', P, Q) \<in> X \<or> \<Psi> \<otimes> \<Psi>' \<rhd> P \<sim>\<^sub>H Q"
  and rSym: "\<And>\<Psi> P Q. (\<Psi>, P, Q) \<in> X \<Longrightarrow>
                      (\<Psi>, Q, P) \<in> {(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and> ((\<Psi>, P', Q') \<in> (X \<union> hobisim)) \<and> \<Psi> \<rhd> Q' \<sim>\<^sub>H Q}"

  shows "\<Psi> \<rhd> P \<sim>\<^sub>H Q"
proof -
  from p have "(\<Psi>, P, Q) \<in> (X \<union> hobisim)"
    by blast
  moreover from `eqvt X` hobisimEqvt have "eqvt (X \<union> hobisim)"
    by auto
  ultimately show ?thesis
  proof(coinduct rule: howeakTransitiveCoinduct')
    case(cStatEq \<Psi> P Q \<phi>)
    thus ?case
      by(blast intro: rStatEq dest: hobisimE)
  next
    case(cClauseEq \<Psi> R S M P)
    thus ?case
      by(auto dest: rClauseEq hobisimE)
  next
    case(cSim \<Psi> P Q)
    thus ?case
      apply -
      apply(case_tac "(\<Psi>, P, Q) \<in> ?X")
      apply(rule_tac rSim)
      apply simp
      apply(clarify)
      apply(drule hobisimE(3))
      apply(rule_tac A=hobisim in monotonic, simp)
      by(force intro: hobisimReflexive)
  next
    case(cExt \<Psi> P Q \<Psi>')
    thus ?case
      by(blast dest: hobisimE rExt)
  next
    case(cSym \<Psi> P Q)
    thus ?case
      apply auto
      apply(drule rSym)
      apply auto
      apply(rule_tac x=Q in exI)
      apply(auto intro: hobisimReflexive)
      apply(rule_tac x=P in exI)
      by(auto intro: hobisimReflexive dest: hobisimE(5))
  qed
qed

lemma hotransitiveCoinduct''[case_names cStatEq cClauseEq cSim cExt cSym, case_conclusion bisim step, consumes 2]:
  fixes \<Psi> :: 'b
  and   P :: "('a, 'b, 'c) psi"
  and   Q :: "('a, 'b, 'c) psi"
  and   X :: "('b \<times> ('a, 'b, 'c) psi \<times> ('a, 'b, 'c) psi) set"

  assumes p: "(\<Psi>, P, Q) \<in> X"
  and Eqvt: "eqvt X"
  and     rStatEq: "\<And> \<Psi>' R S \<phi>. \<lbrakk>(\<Psi>', R, S) \<in> X; \<not>clauseDom \<phi>; (insertAssertion (extractFrame R)) \<Psi>' \<turnstile>\<^sub>F \<phi>\<rbrakk> \<Longrightarrow> (insertAssertion (extractFrame S)) \<Psi>' \<turnstile>\<^sub>F \<phi>"
  and     rClauseEq: "\<And>\<Psi>' R S M P. \<lbrakk>(\<Psi>', R, S) \<in> X; legalClause M P; (insertAssertion (extractFrame R)) \<Psi>' \<turnstile>\<^sub>F M \<Leftarrow> P\<rbrakk> \<Longrightarrow> \<exists>Q. (legalClause M Q \<and> (insertAssertion (extractFrame S) \<Psi>') \<turnstile>\<^sub>F M \<Leftarrow> Q \<and> (\<one>,P,Q) \<in> {(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and> ((\<Psi>, P', Q') \<in> (X \<union> hobisim)) \<and> \<Psi> \<rhd> Q' \<sim>\<^sub>H Q})"
  and rSim: "\<And>\<Psi> P Q. (\<Psi>, P, Q) \<in> X \<Longrightarrow> \<Psi> \<rhd> P \<leadsto>[({(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and>
                                                                        (\<Psi>, P', Q') \<in> (X \<union> hobisim) \<and>
                                                                        \<Psi> \<rhd> Q' \<sim>\<^sub>H Q})] Q"
  and rExt: "\<And>\<Psi> P Q \<Psi>'. (\<Psi>, P, Q) \<in> X \<Longrightarrow> (\<Psi> \<otimes> \<Psi>', P, Q) \<in> X \<or> \<Psi> \<otimes> \<Psi>' \<rhd> P \<sim>\<^sub>H Q"
  and rSym: "\<And>\<Psi> P Q. (\<Psi>, P, Q) \<in> X \<Longrightarrow>
                      (\<Psi>, Q, P) \<in> {(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and> ((\<Psi>, P', Q') \<in> (X \<union> hobisim)) \<and> \<Psi> \<rhd> Q' \<sim>\<^sub>H Q}"

  shows "\<Psi> \<rhd> P \<sim>\<^sub>H Q"
proof -
  from p have "(\<Psi>, P, Q) \<in> (X \<union> hobisim)"
    by blast
  moreover from `eqvt X` hobisimEqvt have "eqvt (X \<union> hobisim)"
    by auto
  ultimately show ?thesis
  proof(coinduct rule: howeakTransitiveCoinduct''')
    case(cStatEq \<Psi> P Q \<phi>)
    thus ?case
      by(blast intro: rStatEq dest: hobisimE)
  next
    case(cClauseEq \<Psi> R S M P)
    thus ?case
      apply(auto dest: rClauseEq hobisimE)
      apply(drule_tac rClauseEq)
      apply assumption
      apply assumption
      apply blast
      apply(drule_tac hobisimE(2))
      by(blast dest: hobisimTransitive intro: hobisimReflexive)
  next
    case(cSim \<Psi> P Q)
    thus ?case
      apply -
      apply(case_tac "(\<Psi>, P, Q) \<in> ?X")
      apply(rule_tac rSim)
      apply simp
      apply(clarify)
      apply(drule hobisimE(3))
      apply(rule_tac A=hobisim in monotonic, simp)
      by(force intro: hobisimReflexive)
  next
    case(cExt \<Psi> P Q \<Psi>')
    thus ?case
      by(auto dest: hobisimE rExt intro: hobisimReflexive)
  next
    case(cSym \<Psi> P Q)
    thus ?case
      apply auto
      apply(drule rSym)
      apply auto
      apply(rule_tac x=Q in exI)
      apply(auto intro: hobisimReflexive)
      apply(rule_tac x=P in exI)
      by(auto intro: hobisimReflexive dest: hobisimE(5))
  qed
qed

lemma hobisimSymmetric:
  fixes \<Psi> :: 'b
  and   P :: "('a, 'b, 'c) psi"
  and   Q :: "('a, 'b, 'c) psi"

  assumes "\<Psi> \<rhd> P \<sim>\<^sub>H Q"
  
  shows "\<Psi> \<rhd> Q \<sim>\<^sub>H P"
using assms
by(rule hobisimE)

lemma hoeqvtTrans[intro]:
  assumes "eqvt X"

  shows "eqvt {(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and> ((\<Psi>, P', Q') \<in> X \<or> \<Psi> \<rhd> P' \<sim>\<^sub>H Q') \<and> \<Psi> \<rhd> Q' \<sim>\<^sub>H Q}"
using assms
apply(auto simp add: eqvt_def eqvts)
apply(erule_tac x="(a, P', Q')" in ballE, auto)
by(blast dest: hobisimClosed)+

lemma hoeqvtWeakTrans[intro]:
  assumes "eqvt X"

  shows "eqvt {(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim>\<^sub>H P' \<and> (\<Psi>, P', Q') \<in> X \<and> \<Psi> \<rhd> Q' \<sim>\<^sub>H Q}"
using assms
apply(auto simp add: eqvt_def eqvts)
apply(erule_tac x="(a, P', Q')" in ballE, auto)
by(blast dest: hobisimClosed)+

lemma hobisimStructCong:
  fixes P :: "('a, 'b, 'c) psi"
  and   Q :: "('a, 'b, 'c) psi"

  assumes "P \<equiv>\<^sub>s Q"

  shows "P \<sim>\<^sub>H Q"
using assms
by(rule_tac hobisimBisim) (rule structCongBisim)

end
end