theory replication
  imports hosum hocase
begin

context env begin


lemma assertionStatEqImp:
  assumes "\<Psi> \<simeq> \<Psi>'"
  shows "\<Psi> \<hookrightarrow> \<Psi>'"
using assms
unfolding AssertionStatEq_def
by simp

inductive copyPred :: "'a \<Rightarrow> ('a,'b,'c) psi \<Rightarrow> ('a,'b,'c) psi \<Rightarrow> bool"
where
  caux1: "copyPred M P (\<langle>M\<rangle>)"
| caux2: "copyPred M P (P \<parallel> \<langle>M\<rangle>)"

lemma eqvtI':
  fixes X :: "'d::fs_name set"
  and   x :: 'd
  and   p :: "name prm"
  
  assumes "eqvt X"
  and     "x \<in> X"
 
  shows "x \<in> (p\<bullet>X)"
proof -
  from assms have "(rev p\<bullet>x \<in> X)"
    by(unfold eqvt_def) auto
  thus ?thesis
    apply(subst (asm) perm_bool[where pi=p,symmetric])
    by(simp add: eqvts)
qed

lemma eqvtI'':
  fixes X :: "'d::fs_name set"
  and   x :: 'd
  and   p :: "name prm"
  
  assumes "eqvt X"
  and     "x \<in> (p\<bullet>X)"
 
  shows "x \<in> X"
proof -
  from `x \<in> (p\<bullet>X)` have "(rev p\<bullet>x) \<in> X"
    apply(subst perm_bool[where pi=p,symmetric])
    by(simp add: eqvts)
  with `eqvt X` have "(p\<bullet>rev p\<bullet>x) \<in> X"
    unfolding eqvt_def
    by blast
  thus ?thesis by simp
qed

lemma bangDerivative':
  fixes \<Psi>   :: 'b
  and   P    :: "('a, 'b, 'c) psi"
  and   \<alpha>    :: "'a action"
  and   P'   :: "('a, 'b, 'c) psi"

  assumes "\<Psi> \<rhd> !P \<longmapsto>\<alpha> \<prec> P'"
  and     "bn \<alpha> \<sharp>* \<Psi>"
  and     "bn \<alpha> \<sharp>* \<Psi>'"
  and     "bn \<alpha> \<sharp>* P"
  and     "bn \<alpha> \<sharp>* subject \<alpha>"

  obtains R where "\<Psi>' \<rhd> P' \<sim> R \<parallel> !P" and "((supp R)::name set) \<subseteq> supp P'"
proof -
  from `\<Psi> \<rhd> !P \<longmapsto>\<alpha> \<prec> P'` have "guarded P" apply - by(ind_cases "\<Psi> \<rhd> !P \<longmapsto>\<alpha> \<prec> P'") (auto simp add: psi.inject)
  assume "\<And>R. \<lbrakk>\<Psi>' \<rhd> P' \<sim> R \<parallel> !P; ((supp R)::name set) \<subseteq> supp P'\<rbrakk> \<Longrightarrow> thesis"
  moreover from `\<Psi> \<rhd> !P \<longmapsto>\<alpha> \<prec> P'` `bn \<alpha> \<sharp>* subject \<alpha>` `bn \<alpha> \<sharp>* \<Psi>` `bn \<alpha> \<sharp>* \<Psi>'` `bn \<alpha> \<sharp>* P`
  have "\<exists>R . \<Psi>' \<rhd> P' \<sim> R \<parallel> !P \<and> ((supp R)::name set) \<subseteq> supp P'"
  proof(nominal_induct avoiding: \<Psi>' rule: bangInduct')
    case(cAlpha \<alpha> P' p)
    then obtain R where "\<Psi>' \<rhd> P' \<sim> R \<parallel> (P \<parallel> !P)" and suppR: "((supp R)::name set) \<subseteq> supp P'"
      by blast
    have S: "set p \<subseteq> set(bn \<alpha>) \<times> set(bn(p \<bullet> \<alpha>))" by fact
    from `\<Psi>' \<rhd> P' \<sim> R \<parallel> (P \<parallel> !P)` have "(p\<bullet>\<Psi>') \<rhd> (p \<bullet> P') \<sim> (p \<bullet> (R \<parallel> (P \<parallel> !P)))"
      by(rule bisimClosed)
    thm cAlpha
    with `bn \<alpha> \<sharp>* \<Psi>'` `bn \<alpha> \<sharp>* P` `bn(p \<bullet> \<alpha>) \<sharp>* \<Psi>'` `bn(p \<bullet> \<alpha>) \<sharp>* P` S have "\<Psi>' \<rhd> (p \<bullet> P') \<sim> (p \<bullet> R) \<parallel> (P \<parallel> !P)"
      by(simp add: eqvts)
    moreover from suppR have "((supp(p \<bullet> R))::name set) \<subseteq> supp(p \<bullet> P')"
      apply(erule_tac rev_mp)
      by(subst subsetClosed[of p, symmetric]) (simp add: eqvts)
    ultimately show ?case by blast
  next
    case(cPar1 \<alpha> P')
    show ?case using `guarded P`
      apply(rule_tac x=P' in exI)
      apply(rule conjI)
      apply(metis bisimTransitive bisimParComm bisimParPres bangExt)
      by(simp add: psi.supp)
  next
    case(cPar2 \<alpha> P')
    from `bn \<alpha> \<sharp>* \<Psi>` `bn \<alpha> \<sharp>* \<Psi>'` `bn \<alpha> \<sharp>* P` have "\<exists>R. (\<Psi>' \<rhd> P' \<sim> R \<parallel> !P \<and> (supp R::name set) \<subseteq> supp P')"
      by(rule_tac cPar2)
    then obtain R where p'bis: "\<Psi>' \<rhd> P' \<sim> R \<parallel> !P" and suppR: "(supp R::name set) \<subseteq> supp P'"
      by blast
    have "\<Psi>' \<rhd> P \<parallel> P' \<sim> R \<parallel> (P \<parallel> !P)"
      by(metis bisimTransitive bisimSymmetric bisimParComm bisimParPres bisimParAssoc p'bis)
    moreover have "(supp R::name set) \<subseteq> supp (P \<parallel> P')" using suppR
      by(force simp add: psi.supp)
    ultimately show ?case
      by blast
  next
    case(cComm1 M N P' K xvec P'')
    from `xvec \<sharp>* \<Psi>` `xvec \<sharp>* \<Psi>'` `xvec \<sharp>* P` have "\<exists>R. (\<Psi>' \<rhd> P'' \<sim> R \<parallel> !P \<and> (supp R::name set) \<subseteq> supp P'')"
      by(rule_tac cComm1) simp+
    then obtain R where p'bis: "\<Psi>' \<rhd> P'' \<sim> R \<parallel> !P" and suppR: "(supp R::name set) \<subseteq> supp P''"
      by blast
    have "\<Psi>' \<rhd> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'' \<sim> (\<lparr>\<nu>*xvec\<rparr>P'\<parallel>R) \<parallel> (P \<parallel> !P)"
      apply(rule bisimSymmetric)
      apply(rule bisimTransitive[OF bisimParComm])
      apply(rule bisimTransitive[OF bisimSymmetric[OF bisimScopeExtChain]])
      apply fact
      apply(simp add: `xvec \<sharp>* P`)
      apply(rule bisimResChainPres)
      apply(rule bisimTransitive[OF bisimParComm])
      apply(rule bisimTransitive[OF bisimParAssoc])
      apply(rule bisimParPresSym)
      apply(rule bisimSymmetric[OF bisimTransitive[OF p'bis]])
      apply(rule bisimParPresSym)
      by(rule bangExt[OF `guarded P`], fact)
    moreover have "(supp(\<lparr>\<nu>*xvec\<rparr>P'\<parallel>R)::name set) \<subseteq> supp (\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'')" using suppR
      by(auto simp add: resChainSupp psi.supp)
    ultimately show ?case by blast
  next
    case(cComm2 M xvec N P' K P'')
    from `xvec \<sharp>* \<Psi>` `xvec \<sharp>* \<Psi>'` `xvec \<sharp>* P` have "\<exists>R. (\<Psi>' \<rhd> P'' \<sim> R \<parallel> !P \<and> (supp R::name set) \<subseteq> supp P'')"
      by(rule_tac cComm2) simp+
    then obtain R where p'bis: "\<Psi>' \<rhd> P'' \<sim> R \<parallel> !P" and suppR: "(supp R::name set) \<subseteq> supp P''"
      by blast
    have "\<Psi>' \<rhd> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'' \<sim> (\<lparr>\<nu>*xvec\<rparr>P'\<parallel>R) \<parallel> (P \<parallel> !P)"
      apply(rule bisimSymmetric)
      apply(rule bisimTransitive[OF bisimParComm])
      apply(rule bisimTransitive[OF bisimSymmetric[OF bisimScopeExtChain]])
      apply fact
      apply(simp add: `xvec \<sharp>* P`)
      apply(rule bisimResChainPres)
      apply(rule bisimTransitive[OF bisimParComm])
      apply(rule bisimTransitive[OF bisimParAssoc])
      apply(rule bisimParPresSym)
      apply(rule bisimSymmetric[OF bisimTransitive[OF p'bis]])
      apply(rule bisimParPresSym)
      by(rule bangExt[OF `guarded P`], fact)
    moreover have "(supp(\<lparr>\<nu>*xvec\<rparr>P'\<parallel>R)::name set) \<subseteq> supp (\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'')" using suppR
      by(auto simp add: resChainSupp psi.supp)
    ultimately show ?case by blast
  next
    case(cBang \<alpha> P')
    from `bn \<alpha> \<sharp>* \<Psi>` `bn \<alpha> \<sharp>* \<Psi>'` `bn \<alpha> \<sharp>* P`
    have "\<exists>R. \<Psi>' \<rhd> P' \<sim> R \<parallel> (P \<parallel> !P) \<and> (supp R::name set) \<subseteq> supp P'"
      by(rule cBang)
    then obtain R where "\<Psi>' \<rhd> P' \<sim> R \<parallel> (P \<parallel> !P)" and "(supp R::name set) \<subseteq> supp P'"
      by blast
    from `\<Psi>' \<rhd> P' \<sim> R \<parallel> (P \<parallel> !P)` have "\<Psi>' \<rhd> P' \<sim> R \<parallel> !P"
      apply(rule bisimTransitive)
      apply(rule bisimParPresSym)
      by(rule bisimSymmetric[OF bangExt[OF `guarded P`]])
    thus ?case using `(supp R::name set) \<subseteq> supp P'`
      by blast
  qed
  ultimately show ?thesis
    by blast
qed

lemma replInduct[consumes 4, case_names cPar1 cPar2 cComm1 cComm2 cInv]:
  fixes \<Psi>   :: 'b
  and   P    :: "('a, 'b, 'c) psi"
  and   Rs   :: "('a, 'b, 'c) residual"
  and   C    :: "'d::fs_name"
  and   a    :: name
  assumes Trans: "\<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto> Rs"
  and     "guarded P"
  and     EntCl: "\<forall> \<Psi>'. \<Psi> \<otimes> \<Psi>' \<turnstile> M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle>"
  and     EntExcl: "\<forall> \<Psi>' Q. ((a \<sharp> \<Psi>' \<and> \<Psi> \<otimes> \<Psi>' \<turnstile> M\<^isub>a \<Leftarrow> Q) \<longrightarrow> Q=P \<parallel> \<langle>M\<^isub>a\<rangle>)"
  and     rPar1: "\<And>\<alpha> P' C.
                   \<lbrakk>\<Psi> \<rhd> P \<longmapsto>\<alpha> \<prec> P'; bn \<alpha> \<sharp>* \<Psi>; bn \<alpha> \<sharp>* P; bn \<alpha> \<sharp>* M\<^isub>a; bn \<alpha> \<sharp>* subject \<alpha>; bn \<alpha> \<sharp>* C; distinct(bn \<alpha>)\<rbrakk> \<Longrightarrow>
                    Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) (\<alpha> \<prec> (P' \<parallel> \<langle>M\<^isub>a\<rangle>))"
  and     rPar2: "\<And>\<alpha> P' C. \<lbrakk>\<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto>\<alpha> \<prec> P'; bn \<alpha> \<sharp>* \<Psi>; bn \<alpha> \<sharp>* P; bn \<alpha> \<sharp>* M\<^isub>a; bn \<alpha> \<sharp>* subject \<alpha>; bn \<alpha> \<sharp>* C; distinct(bn \<alpha>);
                             \<And>C. Prop C \<Psi> (\<langle>M\<^isub>a\<rangle>) (\<alpha> \<prec> P')\<rbrakk> \<Longrightarrow> Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) (\<alpha> \<prec> (P \<parallel> P'))"
  and     rComm1: "\<And>M N P' K xvec P'' C. \<lbrakk>\<Psi> \<rhd> P \<longmapsto>M\<lparr>N\<rparr> \<prec> P'; \<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto>K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P''; \<And>C. Prop C \<Psi> (\<langle>M\<^isub>a\<rangle>) (K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P''); \<Psi> \<turnstile> M \<leftrightarrow> K; 
                                             xvec \<sharp>* \<Psi>; xvec \<sharp>* P; xvec \<sharp>* M\<^isub>a; xvec \<sharp>* M; xvec \<sharp>* K; xvec \<sharp>* C; distinct xvec\<rbrakk> \<Longrightarrow> Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) (\<tau> \<prec> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> P''))"
  and     rComm2: "\<And>M xvec N P' K P'' C. \<lbrakk>\<Psi> \<rhd> P \<longmapsto>M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'; \<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto>K\<lparr>N\<rparr> \<prec> P''; \<And>C. Prop C \<Psi> (\<langle>M\<^isub>a\<rangle>) (K\<lparr>N\<rparr> \<prec> P''); \<Psi> \<turnstile> M \<leftrightarrow> K; 
                                             xvec \<sharp>* \<Psi>; xvec \<sharp>* P; xvec \<sharp>* M\<^isub>a; xvec \<sharp>* M; xvec \<sharp>* K; xvec \<sharp>* C; distinct xvec\<rbrakk> \<Longrightarrow> Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) (\<tau> \<prec> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> P''))"
  and     rInv:   "\<And>Rs C. \<lbrakk>\<Psi> \<rhd> P \<parallel> \<langle>M\<^isub>a\<rangle> \<longmapsto> Rs; \<And>C. Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) Rs; guarded P\<rbrakk> \<Longrightarrow> Prop C \<Psi> (\<langle>M\<^isub>a\<rangle>) Rs"
  shows "Prop C \<Psi> (\<langle>M\<^isub>a\<rangle>) Rs"
proof -
  {
    fix Q :: "('a,'b,'c) psi"
    and \<Psi>' :: 'b
    assume Trans': "\<Psi>' \<rhd> Q \<longmapsto> Rs"
    and    "guarded Q"
    and    "copyPred M\<^isub>a P Q"
    and    "\<Psi> \<simeq> \<Psi>'"
    from Trans' have "Prop C \<Psi> Q Rs" using `copyPred M\<^isub>a P Q` `guarded Q` `\<Psi> \<simeq> \<Psi>'` rPar1 rPar2 rComm1 rComm2 rInv EntExcl EntCl
    proof(nominal_induct avoiding: C a \<Psi> rule: semantics.strong_induct)
      case(cInput \<Psi>' M K xvec N Tvec Q)
      thus ?case by - (ind_cases "copyPred M\<^isub>a P (M\<lparr>\<lambda>*xvec N\<rparr>.Q)")
    next
      case(Output \<Psi>' M K N Q)
      thus ?case by - (ind_cases "copyPred M\<^isub>a P (M\<langle>N\<rangle>.Q)")
    next
      case(Case \<Psi>' Q Rs \<phi> Cs)
      thus ?case by - (ind_cases "copyPred M\<^isub>a P (Cases Cs)")
    next
      case(cPar1 \<Psi>' \<Psi>\<^isub>R Q \<alpha> P' R A\<^isub>R)
      from `copyPred M\<^isub>a P (Q \<parallel> R)` have "Q=P" and "R=\<langle>M\<^isub>a\<rangle>"
        by - (ind_cases "copyPred M\<^isub>a P (Q \<parallel> R)",simp add: psi.inject)+
      have "bn \<alpha> \<sharp>* P" unfolding `Q=P`[symmetric] by fact
      have "bn \<alpha> \<sharp>* M\<^isub>a" using `bn \<alpha> \<sharp>* R` unfolding `R=\<langle>M\<^isub>a\<rangle>` by simp
      from `extractFrame R = \<langle>A\<^isub>R,\<Psi>\<^isub>R\<rangle>` have "\<Psi>\<^isub>R = \<one>" unfolding `R=\<langle>M\<^isub>a\<rangle>` by simp 
      have "\<Psi> \<rhd> P \<longmapsto> \<alpha> \<prec> P'" unfolding `Q=P`[symmetric] using `\<Psi> \<simeq> \<Psi>'`
        apply -
        apply(rule statEqTransition[OF `\<Psi>' \<otimes> \<Psi>\<^isub>R \<rhd> Q \<longmapsto> \<alpha> \<prec> P'`])
        apply(subst `\<Psi>\<^isub>R = \<one>`)
        apply(rule AssertionStatEqTrans[OF Identity])
        by(rule AssertionStatEqSym)
      show ?case
        unfolding `Q=P` `R=\<langle>M\<^isub>a\<rangle>`
        by(rule cPar1) fact+
    next
      case(cPar2 \<Psi>' \<Psi>\<^isub>P R \<alpha> P' Q A\<^isub>P)
      from `copyPred M\<^isub>a P (Q \<parallel> R)` have "Q=P" and "R=\<langle>M\<^isub>a\<rangle>"
        by - (ind_cases "copyPred M\<^isub>a P (Q \<parallel> R)",simp add: psi.inject)+
      from `guarded(Q\<parallel>R)` have "guarded Q" by simp
      have "\<Psi>\<^isub>P \<simeq> \<one>" using `extractFrame Q = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>` `guarded Q`
        by(blast dest: guardedStatEq)
      have "\<Psi> \<simeq> \<Psi>' \<otimes> \<Psi>\<^isub>P"
        apply(rule AssertionStatEqTrans[OF `\<Psi> \<simeq> \<Psi>'`])
        apply(rule AssertionStatEqSym)
        apply(rule AssertionStatEqTrans[OF compositionSym[OF `\<Psi>\<^isub>P \<simeq> \<one>`]])
        by(rule Identity)
      {
        fix C
        have "Prop C \<Psi> R (\<alpha> \<prec> P')"
          by(rule cPar2) (fact | simp add: `R=\<langle>M\<^isub>a\<rangle>` caux1)+
      }
      have "\<And> C. Prop C \<Psi> (\<langle>M\<^isub>a\<rangle>) (\<alpha> \<prec> P')" unfolding `R=\<langle>M\<^isub>a\<rangle>`[symmetric] by fact
      have "bn \<alpha> \<sharp>* P" unfolding `Q=P`[symmetric] by fact
      have "bn \<alpha> \<sharp>* M\<^isub>a" using `bn \<alpha> \<sharp>* R` unfolding `R=\<langle>M\<^isub>a\<rangle>` by simp
      have "\<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto> \<alpha> \<prec> P'" unfolding `R=\<langle>M\<^isub>a\<rangle>`[symmetric]
        apply(rule statEqTransition[OF `\<Psi>' \<otimes> \<Psi>\<^isub>P \<rhd> R \<longmapsto> \<alpha> \<prec> P'`])
        apply(rule AssertionStatEqSym)
        by fact
      have "Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) (\<alpha> \<prec> P \<parallel> P')"
        by(rule cPar2) fact+
      thus ?case unfolding `Q = P` `R = \<langle>M\<^isub>a\<rangle>`
        by simp
    next
      case(cComm1 \<Psi>' \<Psi>\<^isub>R Q M N P' A\<^isub>P \<Psi>\<^isub>P R K xvec P'' A\<^isub>R)
      from `copyPred M\<^isub>a P (Q \<parallel> R)` have "Q=P" and "R=\<langle>M\<^isub>a\<rangle>"
        by - (ind_cases "copyPred M\<^isub>a P (Q \<parallel> R)",simp add: psi.inject)+
      from `guarded(Q\<parallel>R)` have "guarded Q" by simp
      from `extractFrame R = \<langle>A\<^isub>R,\<Psi>\<^isub>R\<rangle>` have "\<Psi>\<^isub>R = \<one>" unfolding `R=\<langle>M\<^isub>a\<rangle>` by simp 
      have "\<Psi>\<^isub>P \<simeq> \<one>" using `extractFrame Q = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>` `guarded Q`
        by(blast dest: guardedStatEq)
      have "\<Psi> \<simeq> \<Psi>' \<otimes> \<Psi>\<^isub>P"
        apply(rule AssertionStatEqTrans[OF `\<Psi> \<simeq> \<Psi>'`])
        apply(rule AssertionStatEqSym)
        apply(rule AssertionStatEqTrans[OF compositionSym[OF `\<Psi>\<^isub>P \<simeq> \<one>`]])
        by(rule Identity)
      have "\<Psi> \<simeq> \<Psi>' \<otimes> \<Psi>\<^isub>P \<otimes> \<Psi>\<^isub>R" unfolding `\<Psi>\<^isub>R = \<one>`
        apply(rule AssertionStatEqTrans[OF `\<Psi> \<simeq> \<Psi>' \<otimes> \<Psi>\<^isub>P`])
        apply(rule AssertionStatEqSym)
        apply(rule AssertionStatEqTrans[OF AssertionStatEqSym[OF Associativity]])
        by(rule Identity)
      {
        fix C
        have "Prop C \<Psi> R (K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'')"
          by(rule cComm1) (fact | simp add: `R=\<langle>M\<^isub>a\<rangle>` caux1)+
      }
      have "\<And> C. Prop C \<Psi> (\<langle>M\<^isub>a\<rangle>) (K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'')"
        unfolding `R=\<langle>M\<^isub>a\<rangle>`[symmetric] by fact
      have "\<Psi> \<rhd> P \<longmapsto> M\<lparr>N\<rparr> \<prec> P'" unfolding `Q=P`[symmetric]
        apply(rule statEqTransition[OF `\<Psi>' \<otimes> \<Psi>\<^isub>R \<rhd> Q \<longmapsto> M\<lparr>N\<rparr> \<prec> P'`])
        apply(subst `\<Psi>\<^isub>R = \<one>`)
        apply(rule AssertionStatEqTrans[OF Identity])
        apply(rule AssertionStatEqSym)
        by fact
      have "\<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P''" unfolding `R=\<langle>M\<^isub>a\<rangle>`[symmetric]
        apply(rule statEqTransition[OF `\<Psi>' \<otimes> \<Psi>\<^isub>P \<rhd> R \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P''`])
        apply(rule AssertionStatEqSym)
        by fact
      have "\<Psi> \<turnstile> M \<leftrightarrow> K" using `\<Psi>' \<otimes> \<Psi>\<^isub>P \<otimes> \<Psi>\<^isub>R \<turnstile> M \<leftrightarrow> K` `\<Psi> \<simeq> \<Psi>' \<otimes> \<Psi>\<^isub>P \<otimes> \<Psi>\<^isub>R`
        apply -
        apply(rule statEqEnt)
        apply(rule AssertionStatEqSym)
        by assumption+
      have "xvec \<sharp>* P" unfolding `Q=P`[symmetric] by fact
      have "xvec \<sharp>* M\<^isub>a" using `xvec \<sharp>* R` unfolding `R=\<langle>M\<^isub>a\<rangle>` by simp
      have "Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) (\<tau> \<prec> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'')"
        by(rule cComm1) fact+
      thus ?case unfolding `Q=P` `R=\<langle>M\<^isub>a\<rangle>` by assumption
    next
      case(cComm2 \<Psi>' \<Psi>\<^isub>R Q M xvec N P' A\<^isub>P \<Psi>\<^isub>P R K P'' A\<^isub>R)
      from `copyPred M\<^isub>a P (Q \<parallel> R)` have "Q=P" and "R=\<langle>M\<^isub>a\<rangle>"
        by - (ind_cases "copyPred M\<^isub>a P (Q \<parallel> R)",simp add: psi.inject)+
      from `guarded(Q\<parallel>R)` have "guarded Q" by simp
      from `extractFrame R = \<langle>A\<^isub>R,\<Psi>\<^isub>R\<rangle>` have "\<Psi>\<^isub>R = \<one>" unfolding `R=\<langle>M\<^isub>a\<rangle>` by simp 
      have "\<Psi>\<^isub>P \<simeq> \<one>" using `extractFrame Q = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>` `guarded Q`
        by(blast dest: guardedStatEq)
      have "\<Psi> \<simeq> \<Psi>' \<otimes> \<Psi>\<^isub>P"
        apply(rule AssertionStatEqTrans[OF `\<Psi> \<simeq> \<Psi>'`])
        apply(rule AssertionStatEqSym)
        apply(rule AssertionStatEqTrans[OF compositionSym[OF `\<Psi>\<^isub>P \<simeq> \<one>`]])
        by(rule Identity)
      have "\<Psi> \<simeq> \<Psi>' \<otimes> \<Psi>\<^isub>P \<otimes> \<Psi>\<^isub>R" unfolding `\<Psi>\<^isub>R = \<one>`
        apply(rule AssertionStatEqTrans[OF `\<Psi> \<simeq> \<Psi>' \<otimes> \<Psi>\<^isub>P`])
        apply(rule AssertionStatEqSym)
        apply(rule AssertionStatEqTrans[OF AssertionStatEqSym[OF Associativity]])
        by(rule Identity)
      {
        fix C
        have "Prop C \<Psi> R (K\<lparr>N\<rparr> \<prec> P'')"
          by(rule cComm2) (fact | simp add: `R=\<langle>M\<^isub>a\<rangle>` caux1)+
      }
      have "\<And> C. Prop C \<Psi> (\<langle>M\<^isub>a\<rangle>) (K\<lparr>N\<rparr> \<prec> P'')"
        unfolding `R=\<langle>M\<^isub>a\<rangle>`[symmetric] by fact
      have "\<Psi> \<rhd> P \<longmapsto> M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'" unfolding `Q=P`[symmetric]
        apply(rule statEqTransition[OF `\<Psi>' \<otimes> \<Psi>\<^isub>R \<rhd> Q \<longmapsto> M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'`])
        apply(subst `\<Psi>\<^isub>R = \<one>`)
        apply(rule AssertionStatEqTrans[OF Identity])
        apply(rule AssertionStatEqSym)
        by fact
      have "\<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto> K\<lparr>N\<rparr> \<prec> P''" unfolding `R=\<langle>M\<^isub>a\<rangle>`[symmetric]
        apply(rule statEqTransition[OF `\<Psi>' \<otimes> \<Psi>\<^isub>P \<rhd> R \<longmapsto> K\<lparr>N\<rparr> \<prec> P''`])
        apply(rule AssertionStatEqSym)
        by fact
      have "\<Psi> \<turnstile> M \<leftrightarrow> K" using `\<Psi>' \<otimes> \<Psi>\<^isub>P \<otimes> \<Psi>\<^isub>R \<turnstile> M \<leftrightarrow> K` `\<Psi> \<simeq> \<Psi>' \<otimes> \<Psi>\<^isub>P \<otimes> \<Psi>\<^isub>R`
        apply -
        apply(rule statEqEnt)
        apply(rule AssertionStatEqSym)
        by assumption+
      have "xvec \<sharp>* P" unfolding `Q=P`[symmetric] by fact
      have "xvec \<sharp>* M\<^isub>a" using `xvec \<sharp>* R` unfolding `R=\<langle>M\<^isub>a\<rangle>` by simp
      have "Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) (\<tau> \<prec> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'')"
        by(rule cComm2) fact+
      thus ?case unfolding `Q=P` `R=\<langle>M\<^isub>a\<rangle>` by assumption
    next
      case(cOpen \<Psi> Q M xvec yvec N P' x)
      thus ?case by - (ind_cases "copyPred M\<^isub>a P (\<lparr>\<nu>x\<rparr>Q)")
    next
      case(cScope \<Psi> Q \<alpha> P' x C)
      thus ?case by - (ind_cases "copyPred M\<^isub>a P (\<lparr>\<nu>x\<rparr>Q)")
    next
      case(Bang \<Psi>' Q Rs \<Psi> C)
      thus ?case by - (ind_cases "copyPred M\<^isub>a P (!Q)")
    next
      case(cInv \<Psi>' M Q \<alpha> P')
      have "M = M\<^isub>a" using `copyPred M\<^isub>a P (\<langle>M\<rangle>)`
        apply -
        by(ind_cases "copyPred M\<^isub>a P (\<langle>M\<rangle>)") (simp add: psi.inject)
      have "\<Psi>' \<simeq> \<Psi> \<otimes> \<one>"
        apply(rule AssertionStatEqSym)
        apply(rule AssertionStatEqTrans[OF Identity])
        by fact
      have EntCl: "\<And> \<Psi>'. \<Psi> \<otimes> \<Psi>' \<turnstile> M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle>" and EntExcl: "\<And> \<Psi>' Q. \<lbrakk>a \<sharp> \<Psi>'; \<Psi> \<otimes> \<Psi>' \<turnstile> M\<^isub>a \<Leftarrow> Q\<rbrakk> \<Longrightarrow> Q=P \<parallel> \<langle>M\<^isub>a\<rangle>" using cInv
        by auto
      have "\<Psi> \<otimes> \<one> \<turnstile> M\<^isub>a \<Leftarrow> Q" unfolding `M=M\<^isub>a`[symmetric]
        by(rule statEqEnt[OF `\<Psi>' \<simeq> \<Psi> \<otimes> \<one>`]) fact
      hence "\<Psi> \<turnstile> M\<^isub>a \<Leftarrow> Q" using Identity
        by - (rule statEqEnt)
      note `\<Psi> \<otimes> \<one> \<turnstile> M\<^isub>a \<Leftarrow> Q`
      moreover have "a \<sharp> \<one>" by simp
      ultimately have "Q=P \<parallel> \<langle>M\<^isub>a\<rangle>"
        by - (rule EntExcl)
      {
        fix C
        have "Prop C \<Psi> Q (\<alpha> \<prec> P')"
          by(rule cInv) (fact|simp add: `Q=P \<parallel> \<langle>M\<^isub>a\<rangle>` caux2)+
      }
      have "\<And> C. Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) (\<alpha> \<prec> P')"
        unfolding `Q=P \<parallel> \<langle>M\<^isub>a\<rangle>`[symmetric] by fact
      have "\<Psi> \<rhd> P \<parallel> \<langle>M\<^isub>a\<rangle> \<longmapsto> \<alpha> \<prec> P'" unfolding `Q=P \<parallel> \<langle>M\<^isub>a\<rangle>`[symmetric]
        apply(rule statEqTransition[where \<Psi>=\<Psi>'])
        apply fact
        by(rule AssertionStatEqSym) fact
      have "Prop C \<Psi> (\<langle>M\<^isub>a\<rangle>) (\<alpha> \<prec> P')"
        by(rule cInv) fact+
      thus ?case unfolding `M=M\<^isub>a` by assumption
    qed
  }
  note r=this
  show ?thesis
    by(rule r) (fact|rule caux1|simp)+
qed

lemma replInputInduct[consumes 4, case_names cPar1 cPar2 cInv]:
  fixes \<Psi>    :: 'b
  and   C    :: "'d::fs_name"
  and   a    :: name
  assumes Trans: "\<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto> M\<lparr>N\<rparr> \<prec> P'"
  and     "guarded P"
  and     EntCl: "\<forall> \<Psi>'. \<Psi> \<otimes> \<Psi>' \<turnstile> M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle>"
  and     EntExcl: "\<forall> \<Psi>' Q. ((a \<sharp> \<Psi>' \<and> \<Psi> \<otimes> \<Psi>' \<turnstile> M\<^isub>a \<Leftarrow> Q) \<longrightarrow> Q=P \<parallel> \<langle>M\<^isub>a\<rangle>)"
  and     rPar1: "\<And>P'.
                   \<lbrakk>\<Psi> \<rhd> P \<longmapsto>M\<lparr>N\<rparr> \<prec> P'\<rbrakk> \<Longrightarrow>
                    Prop \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) M N (P' \<parallel> \<langle>M\<^isub>a\<rangle>)"
  and     rPar2: "\<And>P'. \<lbrakk>\<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto>M\<lparr>N\<rparr> \<prec> P'; Prop \<Psi> (\<langle>M\<^isub>a\<rangle>) M N P'\<rbrakk> \<Longrightarrow> Prop \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) M N (P \<parallel> P')"
  and     rInv:   "\<And>P'. \<lbrakk>\<Psi> \<rhd> P \<parallel> \<langle>M\<^isub>a\<rangle> \<longmapsto> M\<lparr>N\<rparr> \<prec> P'; Prop \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) M N P'; guarded P\<rbrakk> \<Longrightarrow> Prop \<Psi> (\<langle>M\<^isub>a\<rangle>) M N P'"
  shows "Prop \<Psi> (\<langle>M\<^isub>a\<rangle>) M N P'"
using Trans `guarded P` EntCl EntExcl
by(nominal_induct \<Psi> M\<^isub>a Rs=="M\<lparr>N\<rparr> \<prec> P'" arbitrary: P' rule: replInduct) (auto simp add: residualInject intro: rPar1 rPar2 rInv)

lemma replOutputInduct[consumes 4, case_names cPar1 cPar2 cInv]:
  fixes \<Psi>   :: 'b
  and   P    :: "('a, 'b, 'c) psi"
  and   Rs   :: "('a, 'b, 'c) residual"
  and   C    :: "'d::fs_name"
  and   a    :: name
  assumes Trans: "\<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto> M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'"
  and     "guarded P"
  and     EntCl: "\<forall> \<Psi>'. \<Psi> \<otimes> \<Psi>' \<turnstile> M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle>"
  and     EntExcl: "\<forall> \<Psi>' Q. ((a \<sharp> \<Psi>' \<and> \<Psi> \<otimes> \<Psi>' \<turnstile> M\<^isub>a \<Leftarrow> Q) \<longrightarrow> Q=P \<parallel> \<langle>M\<^isub>a\<rangle>)"
  and     rPar1: "\<And>xvec N P' C.
                   \<lbrakk>\<Psi> \<rhd> P \<longmapsto>M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'; xvec \<sharp>* \<Psi>; xvec \<sharp>* P; xvec \<sharp>* M\<^isub>a; xvec \<sharp>* M; xvec \<sharp>* C; distinct xvec\<rbrakk> \<Longrightarrow>
                    Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) M (\<lparr>\<nu>*xvec\<rparr>N \<prec>' (P' \<parallel> \<langle>M\<^isub>a\<rangle>))"
  and     rPar2: "\<And>xvec N P' C. \<lbrakk>\<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto>M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'; xvec \<sharp>* \<Psi>; xvec \<sharp>* P; xvec \<sharp>* M\<^isub>a; xvec \<sharp>* M; xvec \<sharp>* C; distinct xvec;
                             \<And>C. Prop C \<Psi> (\<langle>M\<^isub>a\<rangle>) M (\<lparr>\<nu>*xvec\<rparr>N \<prec>' P')\<rbrakk> \<Longrightarrow> Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) M (\<lparr>\<nu>*xvec\<rparr>N \<prec>' (P \<parallel> P'))"
  and     rInv:   "\<And>B C. \<lbrakk>\<Psi> \<rhd> P \<parallel> \<langle>M\<^isub>a\<rangle> \<longmapsto> ROut M B; \<And>C. Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) M B; guarded P\<rbrakk> \<Longrightarrow> Prop C \<Psi> (\<langle>M\<^isub>a\<rangle>) M B"
  shows "Prop C \<Psi> (\<langle>M\<^isub>a\<rangle>) M (\<lparr>\<nu>*xvec\<rparr>N \<prec>' P')"
proof -
  from Trans have Trans': "\<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto> ROut M (\<lparr>\<nu>*xvec\<rparr>N \<prec>' P')" by(auto simp add: residualInject)
  from Trans' `guarded P` EntCl EntExcl show ?thesis
  proof(nominal_induct \<Psi> M\<^isub>a Rs=="ROut M (\<lparr>\<nu>*xvec\<rparr>N \<prec>' P')" avoiding: C arbitrary: xvec N P' rule: replInduct)
    case(cPar1 \<alpha> P' C xvec N P'')
    thus ?case
      by(auto simp add: residualInject dest: rPar1)
  next
    case(cPar2 \<alpha> P' C xvec N P'')
    thus ?case
      apply(auto simp add: residualInject)
      apply(rule rPar2)
      by(auto simp add: residualInject) blast
  next
    case cComm1 thus ?case by(auto simp add: residualInject)
  next
    case cComm2 thus ?case by(auto simp add: residualInject)
  next
    case cInv thus ?case by(auto simp add: residualInject dest: rInv)
  qed
qed

lemma replTauInduct[consumes 4, case_names cPar1 cPar2 cComm1 cComm2 cInv]:
  fixes \<Psi>   :: 'b
  and   P    :: "('a, 'b, 'c) psi"
  and   C    :: "'d::fs_name"
  and   a    :: name
  assumes Trans: "\<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto> \<tau> \<prec> P'"
  and     "guarded P"
  and     EntCl: "\<forall> \<Psi>'. \<Psi> \<otimes> \<Psi>' \<turnstile> M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle>"
  and     EntExcl: "\<forall> \<Psi>' Q. ((a \<sharp> \<Psi>' \<and> \<Psi> \<otimes> \<Psi>' \<turnstile> M\<^isub>a \<Leftarrow> Q) \<longrightarrow> Q=P \<parallel> \<langle>M\<^isub>a\<rangle>)"
  and     rPar1: "\<And>P' C.
                   \<lbrakk>\<Psi> \<rhd> P \<longmapsto>\<tau> \<prec> P'\<rbrakk> \<Longrightarrow>
                    Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) (P' \<parallel> \<langle>M\<^isub>a\<rangle>)"
  and     rPar2: "\<And>P' C. \<lbrakk>\<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto>\<tau> \<prec> P'; 
                             \<And>C. Prop C \<Psi> (\<langle>M\<^isub>a\<rangle>) P'\<rbrakk> \<Longrightarrow> Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) (P \<parallel> P')"
  and     rComm1: "\<And>M N P' K xvec P'' C. \<lbrakk>\<Psi> \<rhd> P \<longmapsto>M\<lparr>N\<rparr> \<prec> P'; \<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto>K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P''; \<Psi> \<turnstile> M \<leftrightarrow> K; 
                                             xvec \<sharp>* \<Psi>; xvec \<sharp>* P; xvec \<sharp>* M\<^isub>a; xvec \<sharp>* M; xvec \<sharp>* K; xvec \<sharp>* C; distinct xvec\<rbrakk> \<Longrightarrow> Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) (\<lparr>\<nu>*xvec\<rparr>(P' \<parallel> P''))"
  and     rComm2: "\<And>M xvec N P' K P'' C. \<lbrakk>\<Psi> \<rhd> P \<longmapsto>M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'; \<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto>K\<lparr>N\<rparr> \<prec> P''; \<Psi> \<turnstile> M \<leftrightarrow> K; 
                                             xvec \<sharp>* \<Psi>; xvec \<sharp>* P; xvec \<sharp>* M\<^isub>a; xvec \<sharp>* M; xvec \<sharp>* K; xvec \<sharp>* C; distinct xvec\<rbrakk> \<Longrightarrow> Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) (\<lparr>\<nu>*xvec\<rparr>(P' \<parallel> P''))"
  and     rInv:   "\<And>P' C. \<lbrakk>\<Psi> \<rhd> P \<parallel> \<langle>M\<^isub>a\<rangle> \<longmapsto> \<tau> \<prec> P'; \<And>C. Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) P'; guarded P\<rbrakk> \<Longrightarrow> Prop C \<Psi> (\<langle>M\<^isub>a\<rangle>) P'"
  shows "Prop C \<Psi> (\<langle>M\<^isub>a\<rangle>) P'"
using Trans `guarded P` EntCl EntExcl
by(nominal_induct \<Psi> M\<^isub>a Rs=="\<tau> \<prec> P'" avoiding: C arbitrary: P' rule: replInduct) (auto simp add: residualInject intro: rPar1 rPar2 rComm1 rComm2 rInv)

lemma replInduct'[consumes 5, case_names cAlpha cPar1 cPar2 cComm1 cComm2 cInv]:
  fixes \<Psi>   :: 'b
  and   P    :: "('a, 'b, 'c) psi"
  and   P'   :: "('a, 'b, 'c) psi"
  and   C    :: "'d::fs_name"
  and   a    :: name
  assumes Trans: "\<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto> \<alpha> \<prec> P'"
  and     "bn \<alpha> \<sharp>* subject \<alpha>"
  and     "guarded P"
  and     EntCl: "\<forall> \<Psi>'. \<Psi> \<otimes> \<Psi>' \<turnstile> M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle>"
  and     EntExcl: "\<forall> \<Psi>' Q. ((a \<sharp> \<Psi>' \<and> \<Psi> \<otimes> \<Psi>' \<turnstile> M\<^isub>a \<Leftarrow> Q) \<longrightarrow> Q=P \<parallel> \<langle>M\<^isub>a\<rangle>)"
  and     rAlpha: "\<And>\<alpha> P' p C. \<lbrakk>bn \<alpha> \<sharp>* \<Psi>; bn \<alpha> \<sharp>* P; bn \<alpha> \<sharp>* M\<^isub>a; bn \<alpha> \<sharp>* subject \<alpha>;  bn \<alpha> \<sharp>* C;
                                set p \<subseteq> set(bn \<alpha>) \<times> set(bn(p \<bullet> \<alpha>)); distinctPerm p;
                                bn(p \<bullet> \<alpha>) \<sharp>* \<alpha>; bn(p \<bullet> \<alpha>) \<sharp>* P'; Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) \<alpha> P'\<rbrakk> \<Longrightarrow>
                                Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) (p \<bullet> \<alpha>) (p \<bullet> P')"
  and     rPar1: "\<And>\<alpha> P' C.
                   \<lbrakk>\<Psi> \<rhd> P \<longmapsto>\<alpha> \<prec> P'; bn \<alpha> \<sharp>* \<Psi>; bn \<alpha> \<sharp>* P; bn \<alpha> \<sharp>* M\<^isub>a; bn \<alpha> \<sharp>* subject \<alpha>; bn \<alpha> \<sharp>* C; distinct(bn \<alpha>)\<rbrakk> \<Longrightarrow>
                    Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) \<alpha> (P' \<parallel> \<langle>M\<^isub>a\<rangle>)"
  and     rPar2: "\<And>\<alpha> P' C. \<lbrakk>\<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto>\<alpha> \<prec> P'; bn \<alpha> \<sharp>* \<Psi>; bn \<alpha> \<sharp>* P; bn \<alpha> \<sharp>* M\<^isub>a; bn \<alpha> \<sharp>* subject \<alpha>; bn \<alpha> \<sharp>* C; distinct(bn \<alpha>);
                             \<And>C. Prop C \<Psi> (\<langle>M\<^isub>a\<rangle>) \<alpha> P'\<rbrakk> \<Longrightarrow> Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) \<alpha> (P \<parallel> P')"
  and     rComm1: "\<And>M N P' K xvec P'' C. \<lbrakk>\<Psi> \<rhd> P \<longmapsto>M\<lparr>N\<rparr> \<prec> P'; \<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto>K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P''; \<And>C. Prop C \<Psi> (\<langle>M\<^isub>a\<rangle>) (K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) P''; \<Psi> \<turnstile> M \<leftrightarrow> K; 
                                             xvec \<sharp>* \<Psi>; xvec \<sharp>* P; xvec \<sharp>* M\<^isub>a; xvec \<sharp>* M; xvec \<sharp>* K; xvec \<sharp>* C; distinct xvec\<rbrakk> \<Longrightarrow> Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) (\<tau>) (\<lparr>\<nu>*xvec\<rparr>(P' \<parallel> P''))"
  and     rComm2: "\<And>M xvec N P' K P'' C. \<lbrakk>\<Psi> \<rhd> P \<longmapsto>M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'; \<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto>K\<lparr>N\<rparr> \<prec> P''; \<And>C. Prop C \<Psi> (\<langle>M\<^isub>a\<rangle>) (K\<lparr>N\<rparr>) P''; \<Psi> \<turnstile> M \<leftrightarrow> K; 
                                             xvec \<sharp>* \<Psi>; xvec \<sharp>* P; xvec \<sharp>* M\<^isub>a; xvec \<sharp>* M; xvec \<sharp>* K; xvec \<sharp>* C; distinct xvec\<rbrakk> \<Longrightarrow> Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) (\<tau>) (\<lparr>\<nu>*xvec\<rparr>(P' \<parallel> P''))"
  and     rInv:   "\<And>\<alpha> P' C. \<lbrakk>\<Psi> \<rhd> P \<parallel> \<langle>M\<^isub>a\<rangle> \<longmapsto> \<alpha> \<prec> P'; \<And>C. Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) \<alpha> P'; guarded P; distinct(bn \<alpha>)\<rbrakk> \<Longrightarrow> Prop C \<Psi> (\<langle>M\<^isub>a\<rangle>) \<alpha> P'"
  shows "Prop C \<Psi> (\<langle>M\<^isub>a\<rangle>) \<alpha> P'"
proof -
  from `\<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto>\<alpha> \<prec> P'` have "distinct(bn \<alpha>)" by(rule boundOutputDistinct)
  from `\<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto>\<alpha> \<prec> P'` `guarded P` EntCl EntExcl `bn \<alpha> \<sharp>* subject \<alpha>` `distinct(bn \<alpha>)` show ?thesis
  proof(nominal_induct \<Psi> M\<^isub>a Rs=="\<alpha> \<prec> P'" avoiding: C \<alpha> P' rule: replInduct)
    case(cPar1 \<alpha> P' C \<alpha>' P'')
    note  `\<alpha> \<prec> (P' \<parallel> \<langle>M\<^isub>a\<rangle>) = \<alpha>' \<prec> P''`
    moreover from `bn \<alpha> \<sharp>* \<alpha>'` have "bn \<alpha> \<sharp>* bn \<alpha>'" by simp
    moreover note `distinct(bn \<alpha>)` `distinct(bn \<alpha>')`
    moreover from `bn \<alpha> \<sharp>* subject \<alpha>` have "bn \<alpha> \<sharp>* (\<alpha> \<prec> P' \<parallel> \<langle>M\<^isub>a\<rangle>)" by simp
    moreover from `bn \<alpha>' \<sharp>* subject \<alpha>'` have "bn \<alpha>' \<sharp>* (\<alpha>' \<prec> P'')" by simp
    ultimately obtain p where S: "set p \<subseteq> set(bn \<alpha>) \<times> set(bn(p \<bullet> \<alpha>))" and "distinctPerm p" and "\<alpha>' = p \<bullet> \<alpha>"
      and  P'eq: "P'' = p \<bullet> (P' \<parallel> \<langle>M\<^isub>a\<rangle>)" and "bn(p \<bullet> \<alpha>) \<sharp>* \<alpha>" and "bn(p \<bullet> \<alpha>) \<sharp>* (P' \<parallel> \<langle>M\<^isub>a\<rangle>)"
      by(rule_tac residualEq)
    from `\<Psi> \<rhd> P \<longmapsto>\<alpha> \<prec> P'` `bn \<alpha> \<sharp>* \<Psi>` `bn \<alpha> \<sharp>* P` `bn \<alpha> \<sharp>* subject \<alpha>` `bn \<alpha> \<sharp>* C` `distinct(bn \<alpha>)` `bn \<alpha> \<sharp>* M\<^isub>a` 
    have "Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) \<alpha> (P' \<parallel> \<langle>M\<^isub>a\<rangle>)"
      by - (rule rPar1)
    with `bn \<alpha> \<sharp>* \<Psi>` `bn \<alpha> \<sharp>* P` `bn \<alpha> \<sharp>* subject \<alpha>` `bn \<alpha> \<sharp>* C` S `distinctPerm p` `bn \<alpha> \<sharp>* \<alpha>'` `bn \<alpha> \<sharp>* P''` `\<alpha>' = (p \<bullet> \<alpha>)` P'eq `bn(p \<bullet> \<alpha>) \<sharp>* \<alpha>` `bn(p \<bullet> \<alpha>) \<sharp>* (P' \<parallel> \<langle>M\<^isub>a\<rangle>)` `bn \<alpha> \<sharp>* M\<^isub>a`
    have "Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) (p \<bullet> \<alpha>) (p \<bullet> (P' \<parallel> \<langle>M\<^isub>a\<rangle>))"
      by(rule_tac rAlpha) 
    with P'eq `\<alpha>' = p \<bullet> \<alpha>` `distinctPerm p` show ?case  by simp
  next
    case(cPar2 \<alpha> P' C \<alpha>' P'')
    note  `\<alpha> \<prec> (P \<parallel> P') = \<alpha>' \<prec> P''`
    moreover from `bn \<alpha> \<sharp>* \<alpha>'` have "bn \<alpha> \<sharp>* bn \<alpha>'" by simp
    moreover note `distinct(bn \<alpha>)` `distinct(bn \<alpha>')`
    moreover from `bn \<alpha> \<sharp>* subject \<alpha>` have "bn \<alpha> \<sharp>* (\<alpha> \<prec> P \<parallel> P')" by simp
    moreover from `bn \<alpha>' \<sharp>* subject \<alpha>'` have "bn \<alpha>' \<sharp>* (\<alpha>' \<prec> P'')" by simp
    ultimately obtain p where S: "set p \<subseteq> set(bn \<alpha>) \<times> set(bn(p \<bullet> \<alpha>))" and "distinctPerm p" and "\<alpha>' = p \<bullet> \<alpha>"
      and  P'eq: "P'' = p \<bullet> (P \<parallel> P')" and "bn(p \<bullet> \<alpha>) \<sharp>* \<alpha>" and "bn(p \<bullet> \<alpha>) \<sharp>* (P \<parallel> P')"
      by(rule_tac residualEq)
    
    note `\<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto>\<alpha> \<prec> P'`
    moreover from `bn \<alpha> \<sharp>* subject \<alpha>` `distinct(bn \<alpha>)` have "\<And>C. Prop C \<Psi> (\<langle>M\<^isub>a\<rangle>) \<alpha> P'" by(rule_tac cPar2) auto
    moreover note `bn \<alpha> \<sharp>* \<Psi>` `bn \<alpha>  \<sharp>* P` `bn \<alpha>  \<sharp>* subject \<alpha>` `bn \<alpha>  \<sharp>* C` `distinct(bn \<alpha>)` `bn \<alpha> \<sharp>* M\<^isub>a`
    ultimately have "Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) \<alpha> (P \<parallel> P')"
      by - (rule rPar2)
    with `bn \<alpha> \<sharp>* \<Psi>` `bn \<alpha> \<sharp>* P` `bn \<alpha> \<sharp>* subject \<alpha>` `bn \<alpha> \<sharp>* C` S `distinctPerm p` `bn \<alpha> \<sharp>* \<alpha>'` `bn \<alpha> \<sharp>* P''` `\<alpha>' = (p \<bullet> \<alpha>)` P'eq `bn(p \<bullet> \<alpha>) \<sharp>* \<alpha>` `bn(p \<bullet> \<alpha>) \<sharp>* (P \<parallel> P')` `bn \<alpha> \<sharp>* M\<^isub>a`
    have "Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) (p \<bullet> \<alpha>) (p \<bullet> (P \<parallel> P'))"
      by(rule_tac rAlpha)
    with P'eq `\<alpha>' = p \<bullet> \<alpha>` show ?case by simp
  next
    case(cComm1 M N P' K xvec P'' C \<alpha> P''')
    hence "Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) (\<tau>) (\<lparr>\<nu>*xvec\<rparr>(P' \<parallel> P''))"
      by(rule_tac rComm1) (assumption | simp)+
    thus ?case using `\<tau> \<prec> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> P'') = \<alpha> \<prec> P'''`
      by(simp add: residualInject)
  next
    case(cComm2 M xvec N P' K P'' C \<alpha> P''')
    hence "Prop C \<Psi> (P \<parallel> \<langle>M\<^isub>a\<rangle>) (\<tau>) (\<lparr>\<nu>*xvec\<rparr>(P' \<parallel> P''))"
      by(rule_tac rComm2) (assumption | simp)+
    thus ?case using `\<tau> \<prec> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> P'') = \<alpha> \<prec> P'''`
      by(simp add: residualInject)
  next
    case(cInv C \<alpha> P') thus ?case by(auto intro: rInv)
  qed
qed


(*
lemma resReplInduct:
  fixes \<Psi>   :: 'b
  and   P    :: "('a, 'b, 'c) psi"
  and   C    :: "'d::fs_name"
  and   a    :: name
  assumes Trans: "\<Psi> \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle>\<parallel>\<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> P'"
  and     "a \<sharp> \<Psi>"
  and     "a \<sharp> \<alpha>"
  and     "a \<sharp> P'"
  and     "bn \<alpha> \<sharp>* \<Psi>"
  and     "bn \<alpha> \<sharp>* M\<^isub>a"
  and     "bn \<alpha> \<sharp>* \<Psi>\<^isub>c"
  and     "bn \<alpha> \<sharp>* subject \<alpha>"
  and     "a \<sharp> P"
  and     "guarded P"
  and     EntCl: "\<forall> \<Psi>'. \<Psi>' \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle>"
  and     EntExcl: "\<forall> \<Psi>' Q. ((a \<sharp> \<Psi>' \<and> \<Psi>' \<otimes> \<Psi>\<^isub>c\<turnstile> M\<^isub>a \<Leftarrow> Q) \<longrightarrow> Q=P \<parallel> \<langle>M\<^isub>a\<rangle>)"
  shows "Prop C (\<alpha>) (\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle>\<parallel>\<lbrace>\<Psi>\<^isub>c\<rbrace>)) P'"
proof -
  note Trans `a \<sharp> \<Psi>` `a \<sharp> \<alpha>` `a \<sharp> P'` `bn \<alpha> \<sharp>* \<Psi>`
  moreover from `bn \<alpha> \<sharp>* M\<^isub>a` `bn \<alpha> \<sharp>* \<Psi>\<^isub>c` have "bn \<alpha> \<sharp>* (\<langle>M\<^isub>a\<rangle>\<parallel>\<lbrace>\<Psi>\<^isub>c\<rbrace>)"
    by simp
  moreover note `bn \<alpha> \<sharp>* subject \<alpha>`
  ultimately show ?thesis using `a \<sharp> P` EntCl EntExcl
  proof(induct rule: resCases)
    case(cRes P')
    note `\<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace> \<longmapsto> \<alpha> \<prec> P'` `bn \<alpha> \<sharp>* \<Psi>`
    moreover have "bn \<alpha> \<sharp>* \<langle>M\<^isub>a\<rangle>" using `bn \<alpha> \<sharp>* M\<^isub>a` by simp
    moreover have "bn \<alpha> \<sharp>* \<lbrace>\<Psi>\<^isub>c\<rbrace>" using `bn \<alpha> \<sharp>* \<Psi>\<^isub>c` by simp
    moreover note `bn \<alpha> \<sharp>* subject \<alpha>`
    ultimately show ?case
    proof(induct rule: parCases[where C="()"])
      case cPar2 thus ?case by auto
    next
      case cComm1 thus ?case by auto
    next
      case cComm2 thus ?case by auto
    next
      case(cPar1 P' A\<^isub>Q \<Psi>\<^isub>Q)
      from `extractFrame (\<lbrace>\<Psi>\<^isub>c\<rbrace>) = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>` have "\<Psi>\<^isub>Q = \<Psi>\<^isub>c" and "A\<^isub>Q=[]" by auto
      note `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto> \<alpha> \<prec> P'` `bn \<alpha> \<sharp>* subject \<alpha>` `guarded P`
      moreover have "\<forall>\<Psi>'. (\<Psi>\<otimes>\<Psi>\<^isub>c) \<otimes> \<Psi>' \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>"
      proof -
        {
          fix \<Psi>'
          have "(\<Psi>\<otimes>\<Psi>\<^isub>c) \<otimes> \<Psi>' \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>" using `\<forall>\<Psi>'. \<Psi>' \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>`
            apply(rule_tac statEqEnt)
            apply(rule associativitySym)
            by simp
        }
        thus ?thesis by simp
      qed
      moreover have "\<forall>\<Psi>' Q. a \<sharp> \<Psi>' \<and> (\<Psi> \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>' \<turnstile> M\<^isub>a\<Leftarrow>Q \<longrightarrow> Q = P \<parallel> \<langle>M\<^isub>a\<rangle>"
      proof -
        {
          fix \<Psi>' Q
          assume "a \<sharp> \<Psi>'" and "(\<Psi> \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>' \<turnstile> M\<^isub>a\<Leftarrow>Q"
          have "(\<Psi> \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>' \<simeq> (\<Psi> \<otimes> \<Psi>') \<otimes> \<Psi>\<^isub>c"
            by(rule associativitySym)
          have "a \<sharp> (\<Psi> \<otimes> \<Psi>')" using `a \<sharp> \<Psi>` `a \<sharp> \<Psi>'` by force
          moreover have "(\<Psi> \<otimes> \<Psi>') \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>Q" using `(\<Psi> \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>' \<simeq> (\<Psi> \<otimes> \<Psi>') \<otimes> \<Psi>\<^isub>c` `(\<Psi> \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>' \<turnstile> M\<^isub>a\<Leftarrow>Q`
            by(rule statEqEnt)
          ultimately have "Q = P \<parallel> \<langle>M\<^isub>a\<rangle>" using `\<forall> \<Psi>' Q. ((a \<sharp> \<Psi>' \<and> \<Psi>' \<otimes> \<Psi>\<^isub>c\<turnstile> M\<^isub>a \<Leftarrow> Q) \<longrightarrow> Q=P \<parallel> \<langle>M\<^isub>a\<rangle>)` by auto
        }
        thus ?thesis by auto
      qed
      ultimately show ?case unfolding `\<Psi>\<^isub>Q=\<Psi>\<^isub>c`
      proof(nominal_induct rule: replInduct')
        case cInv thus ?case by -
      next
        
      thm parCases
    proof(induct rule: replInduct')
*)

nominal_primrec isOutput :: "'a \<Rightarrow> 'a action \<Rightarrow> bool"
  where
  "isOutput L (M\<lparr>N\<rparr>) = False"
| "isOutput L (M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) = (L = M)"
| "isOutput L (\<tau>) = False"
by(rule TrueI)+

lemma isOutputEqvt[eqvt]:
  fixes p::"name prm"
  shows "(p\<bullet>(isOutput L \<alpha>)) = (isOutput (p\<bullet>L) (p\<bullet>\<alpha>))"
by(nominal_induct \<alpha> rule: action.strong_induct) (simp add: eqvts)+

lemma isOutputGetOut:
  assumes "isOutput M \<alpha>"
  obtains N xvec where "\<alpha> = (M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>)"
using assms
proof(nominal_induct \<alpha> rule: action.strong_inducts)
  case In thus ?case by simp
next
  case(Out M xvec N) thus ?case by simp
next
  case Tau thus ?case by simp
qed

lemma isOutputSubject:
  assumes "isOutput M \<alpha>"
  shows   "(subject \<alpha>) = Some M"
  using assms
  by(nominal_induct \<alpha> rule: action.strong_inducts) simp+

lemma bangOutputInduct'[consumes 2, case_names cAlpha cPar1 cPar2 cBang]:
  fixes   C::"('d::fs_name)"
  assumes Trans: "\<Psi> \<rhd> !P \<longmapsto>M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'"
  and     fresh: "xvec \<sharp>* M"
  assumes rAlpha: "\<And>xvec N P' p C. \<lbrakk>xvec \<sharp>* \<Psi>; xvec \<sharp>* P; xvec \<sharp>* M;  xvec \<sharp>* C;
                                set p \<subseteq> set(xvec) \<times> set(p \<bullet> xvec); distinctPerm p;
                                (p \<bullet> xvec) \<sharp>* xvec; (p \<bullet> xvec) \<sharp>* M; (p \<bullet> xvec) \<sharp>* N; (p \<bullet> xvec) \<sharp>* P'; Prop C \<Psi> (P \<parallel> !P) (M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) P'\<rbrakk> \<Longrightarrow>
                                Prop C \<Psi> (P \<parallel> !P) (M\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>\<langle>(p\<bullet>N)\<rangle>) (p \<bullet> P')"
  and rPar1: "\<And>xvec N P' C. \<lbrakk>\<Psi> \<rhd> P \<longmapsto>M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'; xvec \<sharp>* \<Psi>; xvec \<sharp>* P; xvec \<sharp>* M; xvec \<sharp>* C; distinct xvec\<rbrakk> \<Longrightarrow> Prop C \<Psi> (P \<parallel> !P) (M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) (P' \<parallel> !P)"
  and     rPar2: "\<And>xvec N P' C. \<lbrakk>\<Psi> \<rhd> !P \<longmapsto>M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'; \<And>C. Prop C \<Psi> (!P) (M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) P'; xvec \<sharp>* \<Psi>; xvec \<sharp>* P; xvec \<sharp>* M; xvec \<sharp>* C; distinct xvec\<rbrakk> \<Longrightarrow> Prop C \<Psi> (P \<parallel> !P) (M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) (P \<parallel> P')"
  and     rBang: "\<And>xvec N P' C. \<lbrakk>\<Psi> \<rhd> P \<parallel> !P \<longmapsto>M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'; \<And>C. Prop C \<Psi> (P \<parallel> !P) (M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) P'; guarded P\<rbrakk> \<Longrightarrow> Prop C \<Psi> (!P) (M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) P'"
  shows "Prop C \<Psi> (!P) (M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) P'"
proof -
  note Trans
  moreover have "bn(M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* subject(M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>)" using fresh
    by simp
  moreover have "isOutput M (M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>)"
    by simp
  ultimately show ?thesis
  proof(nominal_induct avoiding: C rule: bangInduct'[of _ _ "(M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>)" _])
    case(cAlpha \<alpha> P' p)
    have "bn(p\<bullet>\<alpha>) \<sharp>* subject(p\<bullet>\<alpha>)" using `bn \<alpha> \<sharp>* subject \<alpha>`
      by(subst (asm) perm_bool[where pi=p,symmetric]) (simp add: eqvts)
    hence "bn(p\<bullet>\<alpha>) \<sharp>* Some M" using `isOutput M (p \<bullet> \<alpha>)`
      by(simp add: isOutputSubject)
    hence "bn(p\<bullet>\<alpha>) \<sharp>* M" by simp
    moreover have "bn \<alpha> \<sharp>* M"
    proof -
      from `bn (p \<bullet> \<alpha>) \<sharp>* \<alpha>` `distinctPerm p` have "bn(\<alpha>) \<sharp>* (p\<bullet>\<alpha>)"
        by(subst perm_bool[where pi=p,symmetric]) (simp add: eqvts)
      hence "bn(\<alpha>) \<sharp>* subject(p\<bullet>\<alpha>)"
        by simp
      hence "bn(\<alpha>) \<sharp>* Some M" using `isOutput M (p \<bullet> \<alpha>)`
        by(simp add: isOutputSubject)
      thus ?thesis by simp
    qed
    moreover note `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))`
    ultimately have "(p\<bullet>M) = M" by simp
    from `isOutput M (p \<bullet> \<alpha>)` have "isOutput M \<alpha>"
      apply(subst (asm) `(p\<bullet>M) = M`[symmetric])
      apply(subst perm_bool[where pi=p,symmetric])
      by(simp add: eqvts)
    from `isOutput M \<alpha>` obtain N xvec where "\<alpha> = M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>"
      by(rule isOutputGetOut)
    have "Prop C \<Psi> (P \<parallel> !P) (M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) P'" unfolding `\<alpha> = M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>`[symmetric]
      by(rule cAlpha) fact
    hence "Prop C \<Psi> (P \<parallel> !P) (M\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>\<langle>(p\<bullet>N)\<rangle>) (p\<bullet>P')" using cAlpha unfolding `\<alpha> = M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>`
      by(rule_tac rAlpha) simp+
    thus ?case unfolding `\<alpha> = M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>` using `(p\<bullet>M) = M`
      by(simp add: eqvts)
  next
    case(cPar1 \<alpha> P')
    from `isOutput M \<alpha>` obtain N xvec where "\<alpha> = M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>"
      by(rule isOutputGetOut)
    from cPar1 show ?case unfolding `\<alpha> = M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>`
      by(rule_tac rPar1) simp+
  next
    case(cPar2 \<alpha> P')
    from `isOutput M \<alpha>` obtain N xvec where "\<alpha> = M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>"
      by(rule isOutputGetOut)
    from cPar2 show ?case unfolding `\<alpha> = M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>`
      by(rule_tac rPar2) simp+
  next
    case cComm1 thus ?case by simp
  next
    case cComm2 thus ?case by simp
  next
    case(cBang \<alpha> P')
    from `isOutput M \<alpha>` obtain N xvec where "\<alpha> = M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>"
      by(rule isOutputGetOut)
    from cBang show ?case unfolding `\<alpha> = M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>`
      by(rule_tac rBang) simp+
  qed
qed

(*lemma bangOutputInduct''[consumes 2, case_names cAlpha cPar1 cPar2 cBang]:
  fixes   C::"('d::fs_name)"
  assumes Trans: "\<Psi> \<otimes> \<Psi>' \<rhd> !P \<longmapsto>M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'"
  and     fresh: "xvec \<sharp>* M"
  assumes rAlpha: "\<And>xvec N P' p C. \<lbrakk>xvec \<sharp>* \<Psi>; xvec \<sharp>* \<Psi>'; xvec \<sharp>* P; xvec \<sharp>* M;  xvec \<sharp>* C;
                                set p \<subseteq> set(xvec) \<times> set(p \<bullet> xvec); distinctPerm p;
                                (p \<bullet> xvec) \<sharp>* xvec; (p \<bullet> xvec) \<sharp>* M; (p \<bullet> xvec) \<sharp>* N; (p \<bullet> xvec) \<sharp>* P'; Prop C \<Psi> \<Psi>' (P \<parallel> !P) (M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) P'\<rbrakk> \<Longrightarrow>
                                Prop C \<Psi> \<Psi>' (P \<parallel> !P) (M\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>\<langle>(p\<bullet>N)\<rangle>) (p \<bullet> P')"
  and rPar1: "\<And>xvec N P' C. \<lbrakk>\<Psi> \<otimes> \<Psi>' \<rhd> P \<longmapsto>M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'; xvec \<sharp>* \<Psi>; xvec \<sharp>* \<Psi>'; xvec \<sharp>* P; xvec \<sharp>* M; xvec \<sharp>* C; distinct xvec\<rbrakk> \<Longrightarrow> Prop C \<Psi> \<Psi>' (P \<parallel> !P) (M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) (P' \<parallel> !P)"
  and     rPar2: "\<And>xvec N P' C. \<lbrakk>\<Psi> \<otimes> \<Psi>' \<rhd> !P \<longmapsto>M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'; \<And>C. Prop C \<Psi> \<Psi>' (!P) (M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) P'; xvec \<sharp>* \<Psi>'; xvec \<sharp>* P; xvec \<sharp>* M; xvec \<sharp>* C; distinct xvec\<rbrakk> \<Longrightarrow> Prop C \<Psi> \<Psi>' (P \<parallel> !P) (M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) (P \<parallel> P')"
  and     rBang: "\<And>xvec N P' C. \<lbrakk>\<Psi> \<rhd> P \<parallel> !P \<longmapsto>M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'; \<And>C. Prop C \<Psi> \<Psi>' (P \<parallel> !P) (M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) P'; guarded P\<rbrakk> \<Longrightarrow> Prop C \<Psi> \<Psi>' (!P) (M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) P'"
  shows "Prop C \<Psi> \<Psi>' (!P) (M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) P'"
proof -
  note Trans fresh
  thus ?thesis
  proof(nominal_induct rule: bangOutputInduct')
    case cAlpha
  moreover have "bn(M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* subject(M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>)" using fresh
    by simp
  moreover have "isOutput M (M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>)"
    by simp
  ultimately show ?thesis
  proof(nominal_induct avoiding: C rule: bangInduct'[of _ _ "(M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>)" _])
    case(cAlpha \<alpha> P' p)
    have "bn(p\<bullet>\<alpha>) \<sharp>* subject(p\<bullet>\<alpha>)" using `bn \<alpha> \<sharp>* subject \<alpha>`
      by(subst (asm) perm_bool[where pi=p,symmetric]) (simp add: eqvts)
    hence "bn(p\<bullet>\<alpha>) \<sharp>* Some M" using `isOutput M (p \<bullet> \<alpha>)`
      by(simp add: isOutputSubject)
    hence "bn(p\<bullet>\<alpha>) \<sharp>* M" by simp
    moreover have "bn \<alpha> \<sharp>* M"
    proof -
      from `bn (p \<bullet> \<alpha>) \<sharp>* \<alpha>` `distinctPerm p` have "bn(\<alpha>) \<sharp>* (p\<bullet>\<alpha>)"
        by(subst perm_bool[where pi=p,symmetric]) (simp add: eqvts)
      hence "bn(\<alpha>) \<sharp>* subject(p\<bullet>\<alpha>)"
        by simp
      hence "bn(\<alpha>) \<sharp>* Some M" using `isOutput M (p \<bullet> \<alpha>)`
        by(simp add: isOutputSubject)
      thus ?thesis by simp
    qed
    moreover note `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))`
    ultimately have "(p\<bullet>M) = M" by simp
    from `isOutput M (p \<bullet> \<alpha>)` have "isOutput M \<alpha>"
      apply(subst (asm) `(p\<bullet>M) = M`[symmetric])
      apply(subst perm_bool[where pi=p,symmetric])
      by(simp add: eqvts)
    from `isOutput M \<alpha>` obtain N xvec where "\<alpha> = M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>"
      by(rule isOutputGetOut)
    have "Prop C \<Psi> (P \<parallel> !P) (M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) P'" unfolding `\<alpha> = M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>`[symmetric]
      by(rule cAlpha) fact
    hence "Prop C \<Psi> (P \<parallel> !P) (M\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>\<langle>(p\<bullet>N)\<rangle>) (p\<bullet>P')" using cAlpha unfolding `\<alpha> = M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>`
      by(rule_tac rAlpha) simp+
    thus ?case unfolding `\<alpha> = M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>` using `(p\<bullet>M) = M`
      by(simp add: eqvts)
  next
    case(cPar1 \<alpha> P')
    from `isOutput M \<alpha>` obtain N xvec where "\<alpha> = M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>"
      by(rule isOutputGetOut)
    from cPar1 show ?case unfolding `\<alpha> = M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>`
      by(rule_tac rPar1) simp+
  next
    case(cPar2 \<alpha> P')
    from `isOutput M \<alpha>` obtain N xvec where "\<alpha> = M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>"
      by(rule isOutputGetOut)
    from cPar2 show ?case unfolding `\<alpha> = M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>`
      by(rule_tac rPar2) simp+
  next
    case cComm1 thus ?case by simp
  next
    case cComm2 thus ?case by simp
  next
    case(cBang \<alpha> P')
    from `isOutput M \<alpha>` obtain N xvec where "\<alpha> = M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>"
      by(rule isOutputGetOut)
    from cBang show ?case unfolding `\<alpha> = M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>`
      by(rule_tac rBang) simp+
  qed
qed
*)

lemma residualFresh:
  fixes C :: "'d::fs_name"
  assumes "distinct(bn \<alpha>)"
  obtains \<beta> Q where "(\<alpha> \<prec> P) = (\<beta> \<prec> Q)" and "distinct(bn \<beta>)" and "bn(\<beta>) \<sharp>* C"
  using assms
proof(nominal_induct \<alpha> avoiding: C rule: action.strong_induct)
  case In
  thus ?case
    by simp
next
  case (Out M xvec N)
  obtain p where "(p \<bullet> xvec) \<sharp>* (C,N,P)" and "set p \<subseteq> set xvec \<times> set(p \<bullet> xvec)" and "distinctPerm p"
    by(rule name_list_avoiding)
  hence "(p \<bullet> xvec) \<sharp>* C" and "(p \<bullet> xvec) \<sharp>* N" and "(p \<bullet> xvec) \<sharp>* P" by simp+
  have reseq: "M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P = M\<lparr>\<nu>*(p \<bullet> xvec)\<rparr>\<langle>(p \<bullet> N)\<rangle> \<prec> (p \<bullet> P)"
    apply(rule alphaOutputResidual[where xvec=xvec])
    apply fact
    apply fact
    apply fact
    by simp
  from `distinct (bn (M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>))` have "distinct (bn (M\<lparr>\<nu>*(p \<bullet> xvec)\<rparr>\<langle>(p \<bullet> N)\<rangle>))"
    by simp
  show ?case
    apply(rule Out(1)[OF reseq])
    apply fact
    by(simp add: `(p \<bullet> xvec) \<sharp>* C`)
next
  case Tau
  thus ?case
    by simp
qed

lemma residualEq':
  fixes \<alpha> :: "'a action"
  and   P :: "('a, 'b, 'c) psi"
  and   \<beta> :: "'a action"
  and   Q :: "('a, 'b, 'c) psi"
  and   C :: "'d::fs_name"

  assumes "\<alpha> \<prec> P = \<beta> \<prec> Q"
  and     "bn \<alpha> \<sharp>* (bn \<beta>)"
  and     "distinct(bn \<alpha>)"
  and     "distinct(bn \<beta>)"
  and     "bn \<alpha> \<sharp>* (\<alpha> \<prec> P)"
  and     "bn \<beta> \<sharp>* (\<beta> \<prec> Q)"
  and     "bn \<alpha> \<sharp>* C"
  and     "bn \<beta> \<sharp>* C"

  obtains p where "set p \<subseteq> set(bn \<alpha>) \<times> set(bn(p \<bullet> \<alpha>))" and "distinctPerm p" and "\<beta> = p \<bullet> \<alpha>" and "Q = p \<bullet> P" and "bn \<alpha> \<sharp>* \<beta>" and "bn \<alpha> \<sharp>* Q" and "bn(p \<bullet> \<alpha>) \<sharp>* \<alpha>" and "bn(p \<bullet> \<alpha>) \<sharp>* P" and "bn(p \<bullet> \<alpha>) \<sharp>* C" and "bn(p \<bullet> \<beta>) \<sharp>* C"
using assms
proof(nominal_induct \<alpha> rule: action.strong_induct)
  case(In M N)
  thus ?case by(simp add: residualInject)
next
  case(Out M xvec N)
  thus ?case 
    by(auto simp add: residualInject)
      (drule_tac boundOutputChainEq'', auto) 
next
  case Tau
  thus ?case by(simp add: residualInject)
qed

lemma frameStatEqEnt:
  assumes "F \<simeq>\<^sub>F G"
  and "F \<turnstile>\<^sub>F \<phi>"
  shows "G \<turnstile>\<^sub>F \<phi>"
using assms
unfolding FrameStatEq_def FrameStatImp_def
by simp

lemma frameNilEqEnt:
  fixes A :: "name list"
  assumes "\<langle>\<epsilon>,\<Psi>\<rangle> \<simeq>\<^sub>F \<langle>A,\<Psi>'\<rangle>"
  and "A \<sharp>* \<phi>"
  and "\<Psi> \<turnstile> \<phi>"
  shows "\<Psi>' \<turnstile> \<phi>"
using assms
proof - 
  have "(\<langle>\<epsilon>, \<Psi>\<rangle>) \<turnstile>\<^sub>F \<phi>" using `\<Psi> \<turnstile> \<phi>`
    by simp
  hence "(\<langle>A,\<Psi>'\<rangle>) \<turnstile>\<^sub>F \<phi>" using `\<langle>\<epsilon>,\<Psi>\<rangle> \<simeq>\<^sub>F \<langle>A,\<Psi>'\<rangle>`
    apply -
    by(rule frameStatEqEnt)
  show ?thesis
    apply(rule frameImpE)
    by fact+
qed

lemma lol:
  assumes "P x y z z' z'' z'''"
  and "Q x y z z' z'' z'''"
  and "R x y z z' z'' z'''"
  and "S x y z z' z'' z'''"
  shows "x \<in> {x. \<exists> a b c d e. (P x a b c d e \<and> Q x a b c d e \<and> R x a b c d e \<and> S x a b c d e)}"
using assms
by blast

lemma lol':
  assumes "P x y z z' z'' z''' z''''"
  and "Q x y z z' z'' z''' z''''"
  and "R x y z z' z'' z''' z''''"
  and "S x y z z' z'' z''' z''''"
  and "T x y z z' z'' z''' z''''"
  shows "x \<in> {x. \<exists> a b c d e f. (P x a b c d e f \<and> Q x a b c d e f \<and> R x a b c d e f \<and> S x a b c d e f \<and> T x a b c d e f)}"
using assms
by blast

lemma insertAssertionMerge: "insertAssertion F \<Psi> = (\<langle>\<epsilon>, \<Psi>\<rangle>) \<otimes>\<^sub>F F"
apply(nominal_induct F avoiding: \<Psi> rule: frame.strong_induct)
by simp+

lemma frameBottomExpand: "\<bottom>\<^sub>F = \<langle>[],\<one>\<rangle>"
  by simp

lemma mergeFrameSimps'[simp]:
  fixes A\<^isub>F :: "name list"
  and   \<Psi>\<^isub>F :: 'b
  and   \<Psi>  :: 'b

  assumes "A\<^isub>F \<sharp>* \<Psi>"

  shows "(\<langle>\<epsilon>, \<Psi>\<rangle>) \<otimes>\<^sub>F \<langle>A\<^isub>F, \<Psi>\<^isub>F\<rangle>= \<langle>A\<^isub>F, \<Psi> \<otimes> \<Psi>\<^isub>F\<rangle>"
using assms
by(induct A\<^isub>F arbitrary: F) auto

lemma frameAssociativity:
  fixes F :: "'b frame"
  and   G :: "'b frame"
  and   H :: "'b frame"


  shows "F \<otimes>\<^sub>F (G \<otimes>\<^sub>F H) \<simeq>\<^sub>F (F \<otimes>\<^sub>F G) \<otimes>\<^sub>F H"
proof -
  obtain A\<^isub>F \<Psi>\<^isub>F where "F = \<langle>A\<^isub>F, \<Psi>\<^isub>F\<rangle>" and "A\<^isub>F \<sharp>* G" and "A\<^isub>F \<sharp>* H"
    apply(rule_tac C="(G,H)" in freshFrame)
    by auto
  moreover obtain A\<^isub>G \<Psi>\<^isub>G where "G = \<langle>A\<^isub>G, \<Psi>\<^isub>G\<rangle>" and "A\<^isub>G \<sharp>* \<Psi>\<^isub>F" and "A\<^isub>G \<sharp>* A\<^isub>F" "A\<^isub>G \<sharp>* H"
    by(rule_tac C="(A\<^isub>F, \<Psi>\<^isub>F,H)" in freshFrame) auto
  moreover obtain A\<^isub>H \<Psi>\<^isub>H where "H = \<langle>A\<^isub>H, \<Psi>\<^isub>H\<rangle>" and "A\<^isub>H \<sharp>* \<Psi>\<^isub>F" and "A\<^isub>H \<sharp>* A\<^isub>F"  and "A\<^isub>H \<sharp>* \<Psi>\<^isub>G" and "A\<^isub>H \<sharp>* A\<^isub>G"
    by(rule_tac C="(A\<^isub>F, \<Psi>\<^isub>F,A\<^isub>G, \<Psi>\<^isub>G)" in freshFrame) auto
  moreover from `A\<^isub>F \<sharp>* G` `A\<^isub>F \<sharp>* H` `A\<^isub>G \<sharp>* H` `G = \<langle>A\<^isub>G, \<Psi>\<^isub>G\<rangle>` `H = \<langle>A\<^isub>H, \<Psi>\<^isub>H\<rangle>` `A\<^isub>G \<sharp>* A\<^isub>F` `A\<^isub>H \<sharp>* A\<^isub>F` `A\<^isub>H \<sharp>* A\<^isub>G`have "A\<^isub>F \<sharp>* \<Psi>\<^isub>G" "A\<^isub>F \<sharp>* \<Psi>\<^isub>H" "A\<^isub>G \<sharp>* \<Psi>\<^isub>H"
    by auto
  ultimately show ?thesis
    apply -
    unfolding `F = \<langle>A\<^isub>F, \<Psi>\<^isub>F\<rangle>` `G = \<langle>A\<^isub>G, \<Psi>\<^isub>G\<rangle>` `H = \<langle>A\<^isub>H, \<Psi>\<^isub>H\<rangle>`
    apply(subst mergeFrames, simp, simp, simp)
    apply(subst mergeFrames, simp, simp, simp)
    apply(subst mergeFrames, simp, simp, simp, simp add: freshCompChain)
    apply(subst mergeFrames, simp, simp, simp add: freshCompChain, simp)
    apply(subst append_assoc)
    apply(rule frameResChainPres)
    apply(subst frameNilStatEq)
    by(rule AssertionStatEqSym[OF Associativity])
qed

lemma clauseClosed':
  fixes \<Psi> :: 'b
  and   M :: 'a
  and   P :: "('a,'b,'c) psi"
  and   p :: "name prm"
 
  shows "\<Psi> \<turnstile> M \<Leftarrow> P = (p \<bullet> \<Psi>) \<turnstile> (p \<bullet> M) \<Leftarrow> (p \<bullet> P)"
apply(rule_tac t = "\<Psi> \<turnstile> M \<Leftarrow> P" and pi1 = p in subst[OF perm_bool])
by(simp add: eqvts)

term "extractFrame P = \<bottom>\<^sub>F"

lemma extractFrameNilClosed:
  fixes p :: "name prm"
  shows "(extractFrame P = \<bottom>\<^sub>F)= (extractFrame (p \<bullet> P) = \<bottom>\<^sub>F)"
  apply(rule iffI)
  apply(subst perm_bool[where pi="rev p",symmetric], simp add: eqvts)
  apply(subst (asm) perm_bool[where pi="rev p",symmetric], simp add: eqvts)
  done

lemma clauseClosed'':
  fixes \<Psi> :: 'b
  and   M :: 'a
  and   P :: "('a,'b,'c) psi"
  and   Q :: "'a"
  and   p :: "name prm"
 
  shows "\<Psi> \<turnstile> M \<Leftarrow> P\<parallel>\<langle>Q\<rangle> = (p \<bullet> \<Psi>) \<turnstile> (p \<bullet> M) \<Leftarrow> (p \<bullet> P)\<parallel> \<langle>(p\<bullet>Q)\<rangle>"
apply(rule_tac t = "\<Psi> \<turnstile> M \<Leftarrow> P\<parallel>\<langle>Q\<rangle>" and pi1 = p in subst[OF perm_bool])
by(simp add: eqvts)

lemma "finite(X::'d set) \<Longrightarrow> finite(X - e)"
oops

(* lemma
  fixes y1::"name set"
  and x1::name
  shows "finite(supp (y1)::name set)\<Longrightarrow> supp(y1 - {x1}) = (supp(y1) - {x1})"
proof(rule ccontr)
apply auto *)


nominal_primrec bnlNames :: "('a,'b,'c) boundOutput \<Rightarrow> name set"
  where
  "bnlNames (\<lparr>\<nu>a\<rparr>M) = (supp(bnlNames M)) - {a}"
| "bnlNames (N \<prec>' P) = supp N"
apply(finite_guess)+
apply(rule TrueI)+
apply(subst fresh_def)
apply(subst name_set_supp)
apply(rule finite_Diff)
apply assumption
by(simp|fresh_guess)+


lemma fsbnl: "finite(bnlNames BO)"
proof(nominal_induct BO rule: boundOutput.strong_induct)
  case(BOut a P)
  thus ?case
    by simp finite_guess
next
  case(BStep n BO)
  thus ?case
    by(simp add: supports_finite[OF at_fin_set_supports, OF at_name_inst])
qed

lemma "bnlNames BO \<subseteq> supp BO"
proof(nominal_induct BO rule: boundOutput.strong_induct)
  case(BOut a P)
  thus ?case
    by(simp add: boundOutput.supp)
next
  case(BStep n BO)
  thus ?case
    apply(simp add: boundOutput.supp)
    apply(subst abs_fun_supp[OF pt_name_inst, OF at_name_inst])
    apply(finite_guess)
    apply(subst name_set_supp[OF fsbnl])
    by auto
qed

lemma bnlNamesSimps: "bnlNames (\<lparr>\<nu>a\<rparr>M) = (bnlNames M) - {a}"
  by(simp add: name_set_supp[OF fsbnl])

lemma
  bnlNamesJudge: "bnlNames(\<lparr>\<nu>*xvec\<rparr>(M \<prec>' P)) = ((supp M) - set xvec)"
proof(induct xvec)
  case Nil thus ?case by simp
next
  case(Cons x xvec)
  show ?case
    unfolding bnlNamesSimps BOresChain.simps Cons
    by auto
qed

(*
lemma
  assumes Trans1: "\<Psi> \<rhd> !P \<longmapsto> M\<lparr>N\<rparr> \<prec> P'"
  and     Trans2: "\<Psi> \<otimes> \<Psi>\<^isub>c \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto> M\<lparr>N\<rparr> \<prec> Q'"
  and     "guarded P"
  and     "a \<sharp> P"
  and     "a \<sharp> N"
  and     Equiv1: "\<And>F. F \<otimes>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F F"
  and     Equiv2: "\<And>\<Psi> R. \<lbrakk>a \<sharp> \<Psi>\<rbrakk> \<Longrightarrow> (\<lbrakk>\<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> R\<rbrakk> \<Longrightarrow> R = (P \<parallel> \<langle>M\<^isub>a\<rangle>))"
  and     EntSupp: "\<And>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<Longrightarrow> (supp M\<^isub>a) \<subseteq> ((supp \<Psi>)::name set))"
  and     EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> (P \<parallel> \<langle>M\<^isub>a\<rangle>)"
  obtains R where "a \<sharp> R" and "\<Psi> \<rhd> Q' \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace> \<sim> R \<parallel> \<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>" and "\<Psi> \<rhd> P' \<sim> R \<parallel> !P"
proof -
  assume ass: "\<And>R. \<lbrakk>a \<sharp> R; \<Psi> \<rhd> Q' \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace> \<sim> R \<parallel> \<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>; \<Psi> \<rhd> P' \<sim> R \<parallel> !P\<rbrakk> \<Longrightarrow> thesis"
  from Trans1 ass show thesis
  proof(induct arbitrary: thesis rule: bangInputInduct)
    case(cPar1 P')
    have EntCl': "\<forall> \<Psi>'. (\<Psi> \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>' \<turnstile> M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle>" using EntCl
      sorry
    have EntExcl: "\<forall> \<Psi>' Q. ((a \<sharp> \<Psi>' \<and> (\<Psi> \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>' \<turnstile> M\<^isub>a \<Leftarrow> Q) \<longrightarrow> Q=P \<parallel> \<langle>M\<^isub>a\<rangle>)" using Equiv2
      sorry
    from `\<Psi> \<rhd> P \<longmapsto> M\<lparr>N\<rparr> \<prec> P'` `a \<sharp> P` `a \<sharp> N` have "a \<sharp> P'"
      by(rule_tac inputFreshDerivative)
    moreover from Trans2 `guarded P` EntCl' EntExcl have "\<Psi> \<rhd> Q' \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace> \<sim> P' \<parallel> \<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>"
    proof(induct rule: replInputInduct)
      case(cPar1 Q')
      thus ?case
    thm cPar1
    moreover have "\<Psi> \<rhd> P' \<parallel> !P \<sim> P' \<parallel> (P \<parallel> !P)"
      by(metis bisimParPres bisimTransitive bisimSymmetric bisimParComm bangExt[OF `guarded P`])
    ultimately show thesis
      by(rule_tac cPar1)
  next
    case(cPar2 P')
    obtain R where "a \<sharp> R" and "\<Psi> \<rhd> Q' \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace> \<sim> R \<parallel> \<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>" and "\<Psi> \<rhd> P' \<sim> R \<parallel> !P"
      by(rule cPar2)
    have "\<Psi> \<rhd> P \<parallel> P' \<sim> R \<parallel> (P \<parallel> !P)" using `\<Psi> \<rhd> P' \<sim> R \<parallel> !P`
      apply -
      apply(drule bisimParPres[where R=P])
      apply(rule bisimTransitive[OF bisimParComm])
      apply(rule bisimTransitive, assumption)
      apply(rule bisimTransitive[OF bisimParAssoc])
      apply(rule bisimTransitive[OF bisimParComm])
      apply(rule bisimSymmetric)
      apply(rule bisimTransitive[OF bisimParComm])
      apply(rule bisimParPres)
      by(rule bisimParComm)
    with `a \<sharp> R` `\<Psi> \<rhd> Q' \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace> \<sim> R \<parallel> \<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>` show thesis
      by(rule cPar2(3))
  next
    case(cBang P')
    obtain R where "a \<sharp> R" and "\<Psi> \<rhd> Q' \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace> \<sim> R \<parallel> \<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>" and "\<Psi> \<rhd> P' \<sim> R \<parallel> (P \<parallel> !P)"
      by(rule cBang)
    have "\<Psi> \<rhd> P' \<sim> R \<parallel> !P"
      by(rule bisimTransitive[OF `\<Psi> \<rhd> P' \<sim> R \<parallel> (P \<parallel> !P)`]) (metis bisimTransitive bisimSymmetric bisimParComm bisimParPres bangExt[OF `guarded P`])
    with `a \<sharp> R` `\<Psi> \<rhd> Q' \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace> \<sim> R \<parallel> \<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>` show thesis
      by(rule cBang(4))
*)

lemma bisimParPres':
  fixes \<Psi> :: 'b
  and   P :: "('a, 'b, 'c) psi"
  and   Q :: "('a, 'b, 'c) psi"
  and   R :: "('a, 'b, 'c) psi"
  
  assumes "\<Psi> \<rhd> P \<sim> Q"

  shows "\<Psi> \<rhd> R \<parallel> P \<sim> R \<parallel> Q"
apply(rule bisimTransitive[OF bisimParComm])
apply(rule bisimSymmetric[OF bisimTransitive[OF bisimParComm]])
apply(rule bisimParPres)
by(rule bisimSymmetric[OF assms])

lemma eqvtFreshTrans[intro]:
  assumes "eqvt X"
  shows "eqvt {(\<Psi>, P, Q, A::name list) | \<Psi> P P' Q' Q A. \<Psi> \<rhd> P \<sim> P' \<and> ((\<Psi>, P', Q') \<in> X \<or> \<Psi> \<rhd> P' \<sim> Q') \<and> \<Psi> \<rhd> Q' \<sim> Q \<and> A \<sharp>* (P,P',Q',Q)}"
using assms
apply(auto simp add: eqvt_def eqvts)
apply(erule_tac x="(a, P', Q')" in ballE, auto)
apply(blast dest: bisimClosed iffD2[OF fresh_star_bij(2)])+
done

(* chain.fin_supp *)

lemma resInputCases'[consumes 4, case_names cRes]:
  fixes \<Psi>    :: 'b
  and   x    :: name
  and   M    :: 'a
  and   N    :: 'a
  and   P    :: "('a, 'b, 'c) psi"
  and   R    :: "('a, 'b, 'c) psi"
  and   C    :: "'f::fs_name"

  assumes Trans: "\<Psi> \<rhd> \<lparr>\<nu>x\<rparr>P \<longmapsto>M\<lparr>N\<rparr> \<prec> R"
  and     2: "x \<sharp> \<Psi>"
  and     "x \<sharp> (M\<lparr>N\<rparr>)"
  and     4: "x \<sharp> R"
  and     rScope:  "\<And>P'. \<lbrakk>\<Psi> \<rhd> P \<longmapsto>M\<lparr>N\<rparr> \<prec> P'\<rbrakk> \<Longrightarrow> Prop (\<lparr>\<nu>x\<rparr>P')"

  shows "Prop R"
proof -
  from Trans obtain \<alpha> where 1: "\<Psi> \<rhd> \<lparr>\<nu>x\<rparr>P \<longmapsto>\<alpha> \<prec> R" and 5: "bn \<alpha> \<sharp>* \<Psi>" and 6: "bn \<alpha> \<sharp>* P" and 7: "bn \<alpha> \<sharp>* subject \<alpha>" and "\<alpha> = M\<lparr>N\<rparr>" by auto
  from `x \<sharp> (M\<lparr>N\<rparr>)` `\<alpha> = (M\<lparr>N\<rparr>)` have 3: "x \<sharp> \<alpha>" by simp
  show ?thesis using 1 2 3 4 5 6 7 `\<alpha> = M\<lparr>N\<rparr>` rScope
    by(induct rule: resCases') (auto simp add: residualInject)
qed

lemma resReplCases[consumes 13, case_names cRes]:
  fixes \<Psi>   :: 'b
  and   P    :: "('a, 'b, 'c) psi"
  and   C    :: "'d::fs_name"
  and   a    :: name
  assumes Trans: "\<Psi> \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle>\<parallel>\<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> P'"
  and     "a \<sharp> \<Psi>"
  and     "a \<sharp> \<alpha>"
  and     "a \<sharp> P'"
  and     "bn \<alpha> \<sharp>* \<Psi>"
  and     "bn \<alpha> \<sharp>* M\<^isub>a"
  and     "bn \<alpha> \<sharp>* \<Psi>\<^isub>c"
  and     "bn \<alpha> \<sharp>* subject \<alpha>"
  and     "a \<sharp> P"
  and     "guarded P"
  and     "\<not> a \<sharp> M\<^isub>a"
  and     EntCl: "\<forall> \<Psi>'. \<Psi>' \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle>"
  and     EntExcl: "\<forall> \<Psi>' Q. ((a \<sharp> \<Psi>' \<and> \<Psi>' \<otimes> \<Psi>\<^isub>c\<turnstile> M\<^isub>a \<Leftarrow> Q) \<longrightarrow> Q=P \<parallel> \<langle>M\<^isub>a\<rangle>)"
  and     rScope: "\<And>P'. \<lbrakk>\<Psi> \<rhd> (\<langle>M\<^isub>a\<rangle>\<parallel>\<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> P'\<rbrakk> \<Longrightarrow> Prop (\<lparr>\<nu>a\<rparr>P')"
  shows "Prop P'"
proof -
  have "a \<sharp> bn \<alpha>" using `a \<sharp> \<alpha>`
    by simp
  note Trans `a \<sharp> \<Psi>` `a \<sharp> \<alpha>` `a \<sharp> P'` `bn \<alpha> \<sharp>* \<Psi>`
  moreover from `bn \<alpha> \<sharp>* M\<^isub>a` `bn \<alpha> \<sharp>* \<Psi>\<^isub>c` have "bn \<alpha> \<sharp>* (\<langle>M\<^isub>a\<rangle>\<parallel>\<lbrace>\<Psi>\<^isub>c\<rbrace>)"
    by simp
  moreover note `bn \<alpha> \<sharp>* subject \<alpha>`
  ultimately show ?thesis using `a \<sharp> bn \<alpha>`
  proof(induct rule: resCases)
    case(cOpen M xvec yvec y N P')
    note `\<Psi> \<rhd> \<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace> \<longmapsto> M\<lparr>\<nu>*(xvec @ yvec)\<rparr>\<langle>([(a, y)] \<bullet> N)\<rangle> \<prec> [(a, y)] \<bullet> P'`
    moreover from cOpen have "(xvec @ yvec) \<sharp>* \<Psi>" and "(xvec @ yvec) \<sharp>* \<langle>M\<^isub>a\<rangle>" and "(xvec @ yvec) \<sharp>* \<lbrace>\<Psi>\<^isub>c\<rbrace>" and "(xvec @ yvec) \<sharp>* M"
      by auto
    ultimately have False
    proof(induct rule: parOutputCases[where C="()"])
      case cPar2 thus ?case by auto
    next
      case(cPar1 P' A\<^isub>R \<Psi>\<^isub>R)
      have "a \<in> supp ([(a, y)] \<bullet> N)" using `y \<in> supp N`
        apply(subst (asm) perm_bool[symmetric, where pi="[(a,y)]"])
        by(simp only: eqvts swap_simps)
      have psirdef: "\<Psi>\<^isub>R = \<Psi>\<^isub>c" using cPar1
        by simp
      have "a \<sharp> (xvec @ yvec)" using `a \<sharp> bn (M\<lparr>\<nu>*(xvec @ y # yvec)\<rparr>\<langle>N\<rangle>)`
        by simp
      hence inbnl: "a \<in> bnlNames (\<lparr>\<nu>*(xvec @ yvec)\<rparr>([(a, y)] \<bullet> N) \<prec>' P')"
        by(simp add: boundOutputFresh bnlNamesJudge `a \<in> supp ([(a, y)] \<bullet> N)`)
      note `\<Psi> \<otimes> \<Psi>\<^isub>R \<rhd> \<langle>M\<^isub>a\<rangle> \<longmapsto> M\<lparr>\<nu>*(xvec @ yvec)\<rparr>\<langle>([(a, y)] \<bullet> N)\<rangle> \<prec> P'` `guarded P`
      moreover have "\<forall> \<Psi>'. (\<Psi> \<otimes> \<Psi>\<^isub>R) \<otimes> \<Psi>' \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>"
      proof -
        {
          fix \<Psi>'
          have "(\<Psi> \<otimes> \<Psi>') \<otimes> \<Psi>\<^isub>R \<simeq> (\<Psi> \<otimes> \<Psi>\<^isub>R) \<otimes> \<Psi>'"
            by(rule associativitySym)
          have "(\<Psi> \<otimes> \<Psi>') \<otimes> \<Psi>\<^isub>R \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>" unfolding psirdef using EntCl
            by simp
          hence "(\<Psi> \<otimes> \<Psi>\<^isub>R) \<otimes> \<Psi>' \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>" using `(\<Psi> \<otimes> \<Psi>') \<otimes> \<Psi>\<^isub>R \<simeq> (\<Psi> \<otimes> \<Psi>\<^isub>R) \<otimes> \<Psi>'`
            by(rule_tac statEqEnt)
        }
        thus ?thesis by simp
      qed
      moreover have "\<forall>\<Psi>' Q. a \<sharp> \<Psi>' \<and> (\<Psi> \<otimes> \<Psi>\<^isub>R) \<otimes> \<Psi>' \<turnstile> M\<^isub>a\<Leftarrow>Q \<longrightarrow> Q = P \<parallel> \<langle>M\<^isub>a\<rangle>"
      proof -
        {
          fix \<Psi>' Q
          assume "a \<sharp> \<Psi>'"
            and    "(\<Psi> \<otimes> \<Psi>\<^isub>R) \<otimes> \<Psi>' \<turnstile> M\<^isub>a\<Leftarrow>Q"
          have "(\<Psi> \<otimes> \<Psi>\<^isub>R) \<otimes> \<Psi>' \<simeq> (\<Psi> \<otimes> \<Psi>') \<otimes> \<Psi>\<^isub>R"
            by(rule associativitySym)
          note EntExcl
          moreover have "(\<Psi> \<otimes> \<Psi>') \<otimes> \<Psi>\<^isub>R \<turnstile> M\<^isub>a\<Leftarrow>Q" using `(\<Psi> \<otimes> \<Psi>\<^isub>R) \<otimes> \<Psi>' \<turnstile> M\<^isub>a\<Leftarrow>Q` `(\<Psi> \<otimes> \<Psi>\<^isub>R) \<otimes> \<Psi>' \<simeq> (\<Psi> \<otimes> \<Psi>') \<otimes> \<Psi>\<^isub>R`
            by(rule_tac statEqEnt)
          moreover have "a \<sharp> (\<Psi> \<otimes> \<Psi>')"
            using `a \<sharp> \<Psi>` `a \<sharp> \<Psi>'`
            by(simp add: freshComp)
          ultimately have "Q = P \<parallel> \<langle>M\<^isub>a\<rangle>" unfolding psirdef
            by auto
        }
        thus ?thesis by auto
      qed
      ultimately show False using inbnl
      proof(nominal_induct rule: replOutputInduct)
        case(cPar1 xvec N P')
        have "a \<in> supp(M\<^isub>a)" using `\<not> a \<sharp> M\<^isub>a`
          unfolding fresh_def
          by simp
        hence "a \<sharp> xvec" using `xvec \<sharp>* M\<^isub>a` unfolding fresh_star_def fresh_def
          by(auto simp add: name_list_supp)
        have "a \<sharp> N"
          by(rule outputFreshDerivative) fact+
        moreover have "a \<in> supp N" using `a \<sharp> xvec` `a \<in> bnlNames (\<lparr>\<nu>*xvec\<rparr>N \<prec>' (P' \<parallel> \<langle>M\<^isub>a\<rangle>))`
          unfolding bnlNamesJudge fresh_def name_list_supp
          by simp
        ultimately show ?case unfolding fresh_def
          by simp                           
      next
        case(cPar2 B)
        thus ?case by(simp add: bnlNamesJudge)
      next
        case(cInv B)
        thus ?case by simp
      qed
    qed
    thus ?case by(rule FalseE)
  next
    case(cRes P')
    thus ?case by(rule_tac rScope)
  qed
qed

lemma freeFreshDerivative':
  fixes a:: name
  assumes "\<Psi> \<rhd> P \<longmapsto> \<alpha> \<prec> P'"
  and     "bn \<alpha> \<sharp>* subject \<alpha>"
  and     "a \<sharp> P"
  and     "a \<sharp> bn \<alpha>"
  and     "a \<sharp> object \<alpha>"
  shows   "a \<sharp> P'"
using assms
proof(nominal_induct rule: action.strong_inducts)
  case(In M N) thus ?case by(rule_tac inputFreshDerivative) simp+
next
  case(Out M xvec N) thus ?case
    by(rule_tac outputFreshDerivative) (auto dest: boundOutputDistinct)
next
  case Tau thus ?case by(rule_tac tauFreshDerivative)
qed

lemma aInClause:
  assumes Equiv1: "\<And>F. F \<otimes>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F F"
  assumes EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> Q"
  and     EntSupp: "\<And>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<Longrightarrow> (supp M\<^isub>a) \<subseteq> ((supp \<Psi>)::name set))"
  and     "\<not> a \<sharp> M\<^isub>a"
  shows   "a \<in> supp(M\<^isub>a \<Leftarrow> Q)"
proof(rule ccontr)
  assume "a \<notin> supp (M\<^isub>a\<Leftarrow>Q)"
  hence "a \<sharp> M\<^isub>a\<Leftarrow>Q" unfolding fresh_def by -
  have "\<one> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> Q" by fact
  hence "(\<langle>[a], \<one> \<otimes> \<Psi>\<^isub>c\<rangle>) \<turnstile>\<^sub>F M\<^isub>a \<Leftarrow> Q"
    apply(rule_tac frameImpI)
    apply(rule refl)
    by(simp add: `a \<sharp> M\<^isub>a\<Leftarrow>Q`)
  hence "((\<langle>\<epsilon>,\<one>\<rangle>) \<otimes>\<^sub>F (\<langle>[a], \<Psi>\<^isub>c\<rangle>)) \<turnstile>\<^sub>F M\<^isub>a \<Leftarrow> Q"
    by(rule_tac frameStatEqEnt[where F="(\<langle>[a], \<one> \<otimes> \<Psi>\<^isub>c\<rangle>)"]) simp
  moreover have "((\<langle>\<epsilon>,\<one>\<rangle>) \<otimes>\<^sub>F (\<langle>[a], \<Psi>\<^isub>c\<rangle>)) \<simeq>\<^sub>F (\<langle>\<epsilon>,\<one>\<rangle>)" by fact
  ultimately have "(\<langle>\<epsilon>,\<one>\<rangle>) \<turnstile>\<^sub>F M\<^isub>a \<Leftarrow> Q"
    by(rule_tac frameStatEqEnt)
  hence "\<one> \<turnstile> M\<^isub>a \<Leftarrow> Q"
    by simp
  hence "(supp M\<^isub>a::name set) \<subseteq> supp(\<one>)"
    by(rule EntSupp)
  hence "a \<sharp> M\<^isub>a" unfolding fresh_def
    by simp
  thus False using `\<not> a \<sharp> M\<^isub>a`
    by simp
qed

lemma replTrans:
  fixes   a::name
  assumes "\<Psi> \<otimes> \<Psi>\<^isub>c \<rhd> P'' \<longmapsto> \<alpha> \<prec> P'"
  and     "bn \<alpha> \<sharp>* subject \<alpha>"
  and     "distinct(bn \<alpha>)"
  and     Equiv1: "\<And>F. F \<otimes>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F F"
  and     EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle>"
  and     EntDisj: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> \<phi> \<Longrightarrow>  (\<phi> = M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
  and     EntSupp: "\<And> \<Psi> P. (\<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<Longrightarrow> (supp M\<^isub>a) \<subseteq> ((supp \<Psi>)::name set))"
  and     Inj: "\<And> P P'. ((M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P'))"
  and     "\<not> a \<sharp> M\<^isub>a"
  and     "a \<sharp> \<Psi>"
(*  and     "a \<sharp> P"*)
  and     "a \<sharp> P''"
  shows   "\<Psi> \<rhd> P'' \<longmapsto> \<alpha> \<prec> P'"
proof -
  obtain \<Psi>' where "\<Psi>'\<simeq>\<Psi>\<otimes>\<Psi>\<^isub>c"
    by(auto intro: AssertionStatEqRefl)
  have "\<Psi>' \<rhd> P'' \<longmapsto> \<alpha> \<prec> P'" using `\<Psi> \<otimes> \<Psi>\<^isub>c \<rhd> P'' \<longmapsto> \<alpha> \<prec> P'` `\<Psi>'\<simeq>\<Psi>\<otimes>\<Psi>\<^isub>c`
    by(rule_tac statEqTransition) (assumption|rule AssertionStatEqSym)+
  moreover note `bn \<alpha> \<sharp>* subject \<alpha>` `distinct (bn \<alpha>)` EntDisj EntSupp Inj Equiv1 EntCl `\<not> a \<sharp> M\<^isub>a` `a \<sharp> \<Psi>` (*`a \<sharp> P`*) `a \<sharp> P''`
  ultimately show ?thesis using `\<Psi>'\<simeq>\<Psi>\<otimes>\<Psi>\<^isub>c`
  proof(nominal_induct avoiding: \<Psi> \<Psi>\<^isub>c a M\<^isub>a rule: semanticsInduct)
    case(cAlpha \<Psi>' P \<alpha> P' p)
    thus ?case by(subst residualAlpha[symmetric]) auto
  next
    case(cInput \<Psi>' M K xvec N Tvec P'')
    have "\<Psi> \<turnstile> M \<leftrightarrow> K"
    proof -
      have "M \<leftrightarrow> K = M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> M \<leftrightarrow> K"
        apply(rule cInput)
        apply(rule statEqEnt)
        by fact+
      thus ?thesis using DisjointCCl by blast
    qed
    thus ?case
      by(rule Input) fact+
  next
    case(cOutput \<Psi>' M K N P'')
    have "\<Psi> \<turnstile> M \<leftrightarrow> K"
    proof -
      have "M \<leftrightarrow> K = M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> M \<leftrightarrow> K"
        apply(rule cOutput)
        apply(rule statEqEnt)
        by fact+
      thus ?thesis using DisjointCCl by blast
    qed
    thus ?case
      by(rule Output)
  next
    case(cCase \<Psi>' P'' \<alpha> P' \<phi> Cs)
    from `(\<phi>, P'') mem Cs` `a \<sharp> Cases Cs` have "a \<sharp> P''"
      by(induct Cs) auto
    have "\<phi> = M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
      apply(rule cCase)
      apply(rule statEqEnt)
      by fact+
    thus ?case
    proof(rule disjE)
      assume "\<phi> = M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>"
      hence "clauseDom \<phi>"
        by(simp add: clauseDom.intros)
      hence False using `\<not> clauseDom \<phi>` by simp
      thus ?thesis by(rule FalseE)
    next
      assume "\<Psi> \<turnstile> \<phi>"
      show ?thesis
        apply(rule Case[where P=P''])
        by(rule cCase) (fact+)
    qed
  next
    case(cPar1 \<Psi>' \<Psi>\<^isub>Q P'' \<alpha> P' A\<^isub>Q Q)
    from `a \<sharp> P'' \<parallel> Q` have "a \<sharp> Q" and "a \<sharp> P''" by auto
    with `extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>` `A\<^isub>Q \<sharp>* a` have "a \<sharp> \<Psi>\<^isub>Q"
      by(drule_tac extractFrameFresh) simp
    with `a \<sharp> \<Psi>` have "a \<sharp> \<Psi> \<otimes> \<Psi>\<^isub>Q" by auto
    have "(\<Psi>' \<otimes> \<Psi>\<^isub>Q) \<simeq> ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c)"
      apply(rule AssertionStatEqTrans[OF Composition[OF `\<Psi>' \<simeq> (\<Psi> \<otimes> \<Psi>\<^isub>c)`]])
      apply(rule AssertionStatEqTrans[OF Associativity])
      apply(rule AssertionStatEqSym[OF AssertionStatEqTrans[OF Associativity]])
      apply(rule compositionSym)
      by(rule Commutativity)
    show ?case
      apply(rule Par1)
      apply(rule cPar1)
      by fact+
  next
    case(cPar2 \<Psi>' \<Psi>\<^isub>Q P'' \<alpha> P' A\<^isub>Q Q)
    from `a \<sharp> Q \<parallel> P''` have "a \<sharp> Q" and "a \<sharp> P''" by auto
    with `extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>` `A\<^isub>Q \<sharp>* a` have "a \<sharp> \<Psi>\<^isub>Q"
      by(drule_tac extractFrameFresh) simp
    with `a \<sharp> \<Psi>` have "a \<sharp> \<Psi> \<otimes> \<Psi>\<^isub>Q" by auto
    have "(\<Psi>' \<otimes> \<Psi>\<^isub>Q) \<simeq> ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c)"
      apply(rule AssertionStatEqTrans[OF Composition[OF `\<Psi>' \<simeq> (\<Psi> \<otimes> \<Psi>\<^isub>c)`]])
      apply(rule AssertionStatEqTrans[OF Associativity])
      apply(rule AssertionStatEqSym[OF AssertionStatEqTrans[OF Associativity]])
      apply(rule compositionSym)
      by(rule Commutativity)
    show ?case
      apply(rule Par2)
      apply(rule cPar2)
      by fact+
  next
    case(cComm1 \<Psi>' \<Psi>\<^isub>Q P'' M N P' A\<^isub>P \<Psi>\<^isub>P Q K xvec Q' A\<^isub>Q)
    from `a \<sharp> P'' \<parallel> Q` have "a \<sharp> Q" and "a \<sharp> P''" by auto
    with `extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>` `A\<^isub>Q \<sharp>* a` have "a \<sharp> \<Psi>\<^isub>Q"
      by(drule_tac extractFrameFresh) simp
    with `a \<sharp> \<Psi>` have "a \<sharp> \<Psi> \<otimes> \<Psi>\<^isub>Q" by auto
    from `a \<sharp> P''` `extractFrame P'' = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>` `A\<^isub>P \<sharp>* a` have "a \<sharp> \<Psi>\<^isub>P"
      by(drule_tac extractFrameFresh) simp
    with `a \<sharp> \<Psi>` have "a \<sharp> \<Psi> \<otimes> \<Psi>\<^isub>P" by auto
    have "\<Psi> \<otimes> \<Psi>\<^isub>P \<otimes> \<Psi>\<^isub>Q \<turnstile> M \<leftrightarrow> K"
    proof -
      have "M \<leftrightarrow> K = M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<otimes> \<Psi>\<^isub>P \<otimes> \<Psi>\<^isub>Q \<turnstile> M \<leftrightarrow> K"
        apply(rule cComm1)
        apply(rule statEqEnt)
        apply(rule AssertionStatEqSym)
        apply(rule AssertionStatEqTrans[OF Commutativity])
        apply(rule AssertionStatEqTrans[OF AssertionStatEqSym[OF Associativity]])
        apply(rule Composition)
        apply(rule AssertionStatEqTrans[OF Commutativity])
        apply(rule AssertionStatEqSym)
        by fact+
      thus ?thesis using DisjointCCl by blast
    qed
    have "(\<Psi>' \<otimes> \<Psi>\<^isub>Q) \<simeq> ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c)"
      apply(rule AssertionStatEqTrans[OF Composition[OF `\<Psi>' \<simeq> (\<Psi> \<otimes> \<Psi>\<^isub>c)`]])
      apply(rule AssertionStatEqTrans[OF Associativity])
      apply(rule AssertionStatEqSym[OF AssertionStatEqTrans[OF Associativity]])
      apply(rule compositionSym)
      by(rule Commutativity)
    have "(\<Psi>' \<otimes> \<Psi>\<^isub>P) \<simeq> ((\<Psi> \<otimes> \<Psi>\<^isub>P) \<otimes> \<Psi>\<^isub>c)"
      apply(rule AssertionStatEqTrans[OF Composition[OF `\<Psi>' \<simeq> (\<Psi> \<otimes> \<Psi>\<^isub>c)`]])
      apply(rule AssertionStatEqTrans[OF Associativity])
      apply(rule AssertionStatEqSym[OF AssertionStatEqTrans[OF Associativity]])
      apply(rule compositionSym)
      by(rule Commutativity)
    show ?case
      apply(rule Comm1)
      apply(rule cComm1)
      apply fact+
      apply(rule cComm1)
      apply fact+
      done
  next
    case(cComm2 \<Psi>' \<Psi>\<^isub>Q P'' M xvec N P' A\<^isub>P \<Psi>\<^isub>P Q K Q' A\<^isub>Q)
    from `a \<sharp> P'' \<parallel> Q` have "a \<sharp> Q" and "a \<sharp> P''" by auto
    with `extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>` `A\<^isub>Q \<sharp>* a` have "a \<sharp> \<Psi>\<^isub>Q"
      by(drule_tac extractFrameFresh) simp
    with `a \<sharp> \<Psi>` have "a \<sharp> \<Psi> \<otimes> \<Psi>\<^isub>Q" by auto
    from `a \<sharp> P''` `extractFrame P'' = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>` `A\<^isub>P \<sharp>* a` have "a \<sharp> \<Psi>\<^isub>P"
      by(drule_tac extractFrameFresh) simp
    with `a \<sharp> \<Psi>` have "a \<sharp> \<Psi> \<otimes> \<Psi>\<^isub>P" by auto
    have "\<Psi> \<otimes> \<Psi>\<^isub>P \<otimes> \<Psi>\<^isub>Q \<turnstile> M \<leftrightarrow> K"
    proof -
      have "M \<leftrightarrow> K = M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<otimes> \<Psi>\<^isub>P \<otimes> \<Psi>\<^isub>Q \<turnstile> M \<leftrightarrow> K"
        apply(rule cComm2)
        apply(rule statEqEnt)
        apply(rule AssertionStatEqSym)
        apply(rule AssertionStatEqTrans[OF Commutativity])
        apply(rule AssertionStatEqTrans[OF AssertionStatEqSym[OF Associativity]])
        apply(rule Composition)
        apply(rule AssertionStatEqTrans[OF Commutativity])
        apply(rule AssertionStatEqSym)
        by fact+
      thus ?thesis using DisjointCCl by blast
    qed
    have "(\<Psi>' \<otimes> \<Psi>\<^isub>Q) \<simeq> ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c)"
      apply(rule AssertionStatEqTrans[OF Composition[OF `\<Psi>' \<simeq> (\<Psi> \<otimes> \<Psi>\<^isub>c)`]])
      apply(rule AssertionStatEqTrans[OF Associativity])
      apply(rule AssertionStatEqSym[OF AssertionStatEqTrans[OF Associativity]])
      apply(rule compositionSym)
      by(rule Commutativity)
    have "(\<Psi>' \<otimes> \<Psi>\<^isub>P) \<simeq> ((\<Psi> \<otimes> \<Psi>\<^isub>P) \<otimes> \<Psi>\<^isub>c)"
      apply(rule AssertionStatEqTrans[OF Composition[OF `\<Psi>' \<simeq> (\<Psi> \<otimes> \<Psi>\<^isub>c)`]])
      apply(rule AssertionStatEqTrans[OF Associativity])
      apply(rule AssertionStatEqSym[OF AssertionStatEqTrans[OF Associativity]])
      apply(rule compositionSym)
      by(rule Commutativity)
    show ?case
      apply(rule Comm2)
      apply(rule cComm2)
      apply fact+
      apply(rule cComm2)
      apply fact+
      done
  next
    case(cOpen \<Psi>' P M xvec yvec N P' x)
    from `a \<sharp> \<lparr>\<nu>x\<rparr>P` `x \<sharp> a` have "a \<sharp> P"
      by(simp add: fresh_abs_fun_iff[OF pt_name_inst,OF at_name_inst, OF fin_supp])
    show ?case
      apply(rule Open)
      apply(rule cOpen)
      by fact+
  next
    case(cScope \<Psi>' P \<alpha> P' x)
    from `a \<sharp> \<lparr>\<nu>x\<rparr>P` `x \<sharp> a` have "a \<sharp> P"
      by(simp add: fresh_abs_fun_iff[OF pt_name_inst,OF at_name_inst, OF fin_supp])    
    show ?case
      apply(rule Scope)
      apply(rule cScope)
      by fact+
  next
    case(cBang \<Psi>' P \<alpha> P')
    from `a \<sharp> !P` have "a \<sharp> P\<parallel>!P"
      by simp
    show ?case
      apply(rule Bang)
      apply(rule cBang)
      by fact+
  next
    case(cInv \<Psi>' M P'' \<alpha> P')
    have "a \<sharp> P''" using `supp P'' \<subseteq> supp M` `a \<sharp> \<langle>M\<rangle>`
      by(auto simp add: fresh_def psi.supp)
    have "M \<Leftarrow> P'' = M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> M \<Leftarrow> P''"
      apply(rule cInv)
      apply(rule statEqEnt)
      by fact+
    thus ?case
    proof(rule disjE)
      assume "\<Psi> \<turnstile> M\<Leftarrow>P''"
      thus ?case
        apply(rule Inv)
        apply(rule cInv)
        by fact+
    next
      assume "M\<Leftarrow>P'' = M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>"
      moreover have "a \<in> supp(M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>)"
        by(rule aInClause) fact+
      ultimately have "a \<in> supp(M\<Leftarrow>P'')" by simp
      with `a \<sharp> \<langle>M\<rangle>` `a \<sharp> P''` clauseSupp[where M=M and P=P''] have False
        by(auto simp add: fresh_def psi.supp)
      thus ?case by(rule FalseE)
  qed
qed
qed

lemma replication:
  fixes a  :: name
  and   M\<^isub>a :: "'a::fs_name"
  and   P  :: "('a, 'b::fs_name, 'c::fs_name) psi"
  and   R  :: "('a, 'b, 'c) psi"
  and   \<Psi>\<^isub>c :: 'b
  and   \<Psi>  :: 'b

  assumes EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> (P \<parallel> \<langle>M\<^isub>a\<rangle>)"
  and     Equiv1: "\<And>F. F \<otimes>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F F"
(*  and     Equiv2: "\<And>\<Psi> R. \<lbrakk>a \<sharp> \<Psi>\<rbrakk> \<Longrightarrow> (\<lbrakk>\<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> R\<rbrakk> \<Longrightarrow> R = (P \<parallel> \<langle>M\<^isub>a\<rangle>))"*)
  and     "guarded P"
  and     EntSupp: "\<And>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<Longrightarrow> (supp M\<^isub>a) \<subseteq> ((supp \<Psi>)::name set))"
  and     Inj: "\<And> P P'. ((M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P'))"
  and     EntDisj: "\<And> \<Psi> \<phi>. (\<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> \<phi>))"
  and     "a \<sharp> P"
  and     "\<not>(a \<sharp> M\<^isub>a)"
  and     "((supp P)::name set) \<subseteq> supp M\<^isub>a"
  and     "((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a"

  shows "\<Psi> \<rhd> (\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) \<sim> (!P)"
proof -
  {
    fix   \<Psi>::'b
      and R::"('a,'b,'c) psi"
    assume "a \<sharp> \<Psi>"
      and  "\<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> R"
    from `\<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> R` have Disj: "(M\<^isub>a \<Leftarrow> R = M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> M\<^isub>a \<Leftarrow> R)"
      by(rule EntDisj)
    from Disj have "R = (P \<parallel> \<langle>M\<^isub>a\<rangle>)"
    proof(rule disjE)
      assume "M\<^isub>a\<Leftarrow>R = M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>"
      thus ?thesis by(subst Inj[symmetric])
    next
      assume "\<Psi> \<turnstile> M\<^isub>a\<Leftarrow>R"
      hence "(supp M\<^isub>a::name set) \<subseteq> supp \<Psi>"
        by(rule EntSupp)
      hence False using `a \<sharp> \<Psi>` `\<not> a \<sharp> M\<^isub>a`
        by(auto simp add: fresh_def)
      thus ?thesis by(rule FalseE)
    qed
  }
  note Equiv2 = this
  {
    fix a::name and M\<^isub>a::'a and \<Psi>\<^isub>c::'b
    assume "a \<sharp> \<Psi>"
      and  EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> (P \<parallel> \<langle>M\<^isub>a\<rangle>)"
      and  Equiv1: "\<And>F. F \<otimes>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F F"
      and  Equiv2: "\<And>\<Psi> R. \<lbrakk>a \<sharp> \<Psi>\<rbrakk> \<Longrightarrow> (\<lbrakk>\<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> R\<rbrakk> \<Longrightarrow> R = (P \<parallel> \<langle>M\<^isub>a\<rangle>))"
      and  "guarded P"
      and  EntSupp: "\<And>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<Longrightarrow> (supp M\<^isub>a) \<subseteq> ((supp \<Psi>)::name set))"
      and  Inj: "\<And> P P'. ((M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P'))"
      and  EntEnt: "\<And> \<Psi> \<phi>. (\<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> \<phi>))"
      and  "a \<sharp> P"
      and  "\<not>(a \<sharp> M\<^isub>a)"
      and  "((supp P)::name set) \<subseteq> supp M\<^isub>a"
      and  "((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a"

    (* Build set ?X containing the candidate relation *)
    let ?X1 = "{((\<Psi>::'b), \<lparr>\<nu>*xvec\<rparr>((Q::('a,'b,'c) psi) \<parallel> \<lparr>\<nu>(a::name)\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)), \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> !P)) | \<Psi> a P Q M\<^isub>a \<Psi>\<^isub>c xvec. a \<sharp> \<Psi> \<and> a \<sharp> P \<and> a \<sharp> Q \<and> \<not>(a \<sharp> M\<^isub>a) \<and> (\<forall> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>) \<and> (\<forall>F. (F \<otimes>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F F)) \<and> (\<forall>\<Psi> R. a \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> R \<longrightarrow> (R = (P \<parallel> \<langle>M\<^isub>a\<rangle>))) \<and> guarded P \<and> (((supp P)::name set) \<subseteq> supp M\<^isub>a) \<and> (((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a) \<and> (\<forall> P P'. ((M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P'))) \<and> (\<forall> \<Psi> P. (\<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<longrightarrow> (supp M\<^isub>a) \<subseteq> ((supp \<Psi>)::name set))) \<and> xvec \<sharp>* \<Psi> \<and> (\<forall> \<Psi> \<phi>. (\<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> \<phi> \<longrightarrow> (\<phi> = M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> \<phi>)))}"
    let ?X2 = "{((\<Psi>::'b), \<lparr>\<nu>*xvec\<rparr>((Q::('a,'b,'c) psi) \<parallel> !P), \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> \<lparr>\<nu>(a::name)\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))) | \<Psi> a P Q M\<^isub>a \<Psi>\<^isub>c  xvec. a \<sharp> \<Psi> \<and> a \<sharp> P \<and> a \<sharp> Q \<and> \<not>(a \<sharp> M\<^isub>a) \<and> (\<forall> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>) \<and> (\<forall>F. (F \<otimes>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F F)) \<and> (\<forall>\<Psi> R. a \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> R \<longrightarrow> (R = (P \<parallel> \<langle>M\<^isub>a\<rangle>))) \<and> guarded P \<and> (((supp P)::name set) \<subseteq> supp M\<^isub>a) \<and> (((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a) \<and> (\<forall> P P'. ((M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P'))) \<and> (\<forall> \<Psi> P. (\<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<longrightarrow> (supp M\<^isub>a) \<subseteq> ((supp \<Psi>)::name set))) \<and> xvec \<sharp>* \<Psi> \<and> (\<forall> \<Psi> \<phi>. (\<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> \<phi> \<longrightarrow> (\<phi> = M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> \<phi>)))}"
    let ?X = "?X1 \<union> ?X2"
    {
      fix \<Psi> \<Psi>'::'b
        and S T
      assume "(\<Psi>,S,T) \<in> ?X1"
      then obtain a P Q M\<^isub>a \<Psi>\<^isub>c xvec where s_def: "S = \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))" and t_def: "T = \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> !P)" and "a \<sharp> P" and "a \<sharp> Q" and "a \<sharp> \<Psi>" and "\<not> a \<sharp> M\<^isub>a" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> R \<longrightarrow> (R = (P \<parallel> \<langle>M\<^isub>a\<rangle>)))" and "guarded P" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<longrightarrow> (supp M\<^isub>a) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P'))" and "((supp P)::name set) \<subseteq> supp M\<^isub>a" and "((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a" and "xvec \<sharp>* \<Psi>" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
        by auto
      obtain p::"name prm" where "set p \<subseteq> set xvec \<times> set(p\<bullet>xvec)" and "(p\<bullet>xvec) \<sharp>* a" and "distinctPerm p" and "(p\<bullet>xvec) \<sharp>* M\<^isub>a" and "(p\<bullet>xvec) \<sharp>* \<Psi>\<^isub>c" and "(p\<bullet>xvec) \<sharp>* P" and "(p\<bullet>xvec) \<sharp>* Q" and "(p\<bullet>xvec) \<sharp>* \<Psi>" and "(p\<bullet>xvec) \<sharp>* \<Psi>'"
        by(rule chain.name_list_avoiding[where xvec=xvec and c="(M\<^isub>a,\<Psi>\<^isub>c,a,P,Q,\<Psi>,\<Psi>')"]) auto
      obtain b::name where "b \<sharp> P" and "b \<sharp> Q" and "b \<sharp> \<Psi>" and "b \<sharp> M\<^isub>a" "b \<sharp> \<Psi>\<^isub>c" and "b \<sharp> \<Psi>'" and "b \<sharp> (p\<bullet>a)" and "b \<sharp> (p\<bullet>M\<^isub>a)" and "b \<sharp> (p\<bullet>\<Psi>\<^isub>c)" and "b \<sharp> (p\<bullet>P)" and "b \<sharp> p\<bullet>Q"
        apply(rule at_exists_fresh[OF at_name_inst, where x="(P,Q,\<Psi>,M\<^isub>a,\<Psi>\<^isub>c,\<Psi>',p\<bullet>a,p\<bullet>M\<^isub>a,p\<bullet>\<Psi>\<^isub>c,p\<bullet>P,p\<bullet>Q)"])
        unfolding supp_prod pt_perm_supp[OF pt_name_inst,OF at_name_inst, symmetric]
        apply(rule finite_UnI, (finite_guess | (subst pt_set_finite_ineq[OF pt_name_inst,OF at_name_inst], finite_guess)))+
        apply(subst pt_set_finite_ineq[OF pt_name_inst,OF at_name_inst], finite_guess)
        by force
      have "\<lparr>\<nu>(p\<bullet>a)\<rparr>(\<langle>(p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>c)\<rbrace>) = \<lparr>\<nu>b\<rparr>([(p\<bullet>a,b)] \<bullet> (\<langle>(p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
        apply(subst alphaRes[where y=b])
        apply(simp add: `b \<sharp> (p\<bullet>M\<^isub>a)` `b \<sharp> (p\<bullet>\<Psi>\<^isub>c)`)
        by(rule refl)
      have "\<lparr>\<nu>*xvec\<rparr>(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<lparr>\<nu>*p\<bullet>xvec\<rparr>((p\<bullet>Q) \<parallel> \<lparr>\<nu>(p\<bullet>a)\<rparr>(\<langle>(p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>c)\<rbrace>))" using `(p\<bullet>xvec) \<sharp>* a` `set p \<subseteq> set xvec \<times> set (p \<bullet> xvec)`
        by(subst resChainAlpha[where p=p]) (simp add: `(p\<bullet>xvec) \<sharp>* Q` `(p\<bullet>xvec) \<sharp>* M\<^isub>a` `(p\<bullet>xvec) \<sharp>* \<Psi>\<^isub>c` abs_fresh_list_star' eqvts)+
      have t_def: "T = \<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(p\<bullet>Q) \<parallel> !(p\<bullet>P)" unfolding t_def using `(p\<bullet>xvec) \<sharp>* a` `set p \<subseteq> set xvec \<times> set (p \<bullet> xvec)`
        by(subst resChainAlpha[where p=p]) (simp add: `(p\<bullet>xvec) \<sharp>* Q` `(p\<bullet>xvec) \<sharp>* P` abs_fresh_list_star' eqvts)+
      have s_def: "S = \<lparr>\<nu>*p\<bullet>xvec\<rparr>((p\<bullet>Q) \<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c))\<rbrace>))"
        unfolding `S = \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))` `\<lparr>\<nu>(p\<bullet>a)\<rparr>(\<langle>(p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>c)\<rbrace>) = \<lparr>\<nu>b\<rparr>([(p\<bullet>a,b)] \<bullet> (\<langle>(p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>c)\<rbrace>))` `\<lparr>\<nu>*xvec\<rparr>(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<lparr>\<nu>*p\<bullet>xvec\<rparr>((p\<bullet>Q) \<parallel> \<lparr>\<nu>(p\<bullet>a)\<rparr>(\<langle>(p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>c)\<rbrace>))`
        by simp
      have "(p\<bullet>a) \<sharp> (p\<bullet>P)" using `a \<sharp> P`
        by(subst pt_fresh_bij[OF pt_name_inst, OF at_name_inst])
      note `b \<sharp> \<Psi>'`
      moreover note `b \<sharp> (p\<bullet>P)` `b \<sharp> (p\<bullet>Q)`
      moreover have "guarded(p\<bullet>P)" using `guarded P` by simp
      moreover have "\<not> b \<sharp> ([(p\<bullet>a,b)] \<bullet> (p\<bullet>M\<^isub>a))" using `\<not> a \<sharp> M\<^isub>a`        
        apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
        apply(subst (asm) perm_bool[where pi="p",symmetric])
        by(simp add: eqvts swap_simps)
      moreover have "\<And> \<Psi>. \<Psi> \<otimes> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)) \<turnstile> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>(p\<bullet>P) \<parallel> \<langle>([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<rangle>" using EntCl
        apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
        apply(subst perm_bool[where pi="rev p",symmetric])
        by(simp add: eqvts `(p\<bullet>a) \<sharp> (p\<bullet>P)` `b \<sharp> (p\<bullet>P)`)
      moreover have "\<forall>F. (F \<otimes>\<^sub>F (\<langle>[b],[(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)\<rangle>) \<simeq>\<^sub>F F)"
        using Equiv1
        apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
        apply(subst perm_bool[where pi="rev p",symmetric])
        by(simp add: eqvts swap_simps)
      moreover have "(\<forall>\<Psi> R. b \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)) \<turnstile> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a)) \<Leftarrow> R \<longrightarrow> (R = ((p\<bullet>P) \<parallel> \<langle>([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<rangle>)))" using Equiv2
        apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
        apply(subst (asm) perm_bool[where pi="p",symmetric])
        by(simp add: eqvts `(p\<bullet>a) \<sharp> p\<bullet>P` `b \<sharp> p\<bullet>P` swap_simps)
      moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>P \<longrightarrow> (supp ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))) \<subseteq> ((supp \<Psi>)::name set))" using EntSupp
        apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
        apply(subst (asm) perm_bool[where pi="p",symmetric])
        by(simp add: eqvts swap_simps)
      moreover have "\<forall> P P'. ((([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>P = ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>P') = (P = P'))" using Inj
        apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
        apply(subst (asm) perm_bool[where pi="p",symmetric])
        by(simp add: eqvts swap_simps)
      moreover have "((supp(p\<bullet>P))::name set) \<subseteq> supp ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))" using `((supp P)::name set) \<subseteq> supp M\<^isub>a`
        apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
        apply(subst (asm) perm_bool[where pi="p",symmetric])
        by(simp add: eqvts swap_simps `(p\<bullet>a) \<sharp> p\<bullet>P` `b \<sharp> p\<bullet>P`)
      moreover have "((supp ([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)))::name set) = supp ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))" using `((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a`
        apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
        apply(subst (asm) perm_bool[where pi="p",symmetric])
        by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
      moreover note `(p\<bullet>xvec)\<sharp>*\<Psi>'`
      moreover have "(\<forall> \<Psi> \<phi>. (\<Psi> \<otimes> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)) \<turnstile> \<phi> \<longrightarrow> (\<phi> = ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a)) \<Leftarrow> (p\<bullet>P) \<parallel> \<langle>([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<rangle> \<or> \<Psi> \<turnstile> \<phi>)))" using EntEnt
        apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
        apply(subst perm_bool[where pi="p",symmetric])
        by(simp add: eqvts swap_simps `distinctPerm p` `(p\<bullet>a) \<sharp> p\<bullet>P` `b \<sharp> p\<bullet>P`)
      ultimately have "(\<Psi>',S,T) \<in> ?X1"
        unfolding s_def t_def
        by blast
    }
    note X1swapPsi = this
    {
      fix \<Psi> \<Psi>'::'b
        and S T
      assume "(\<Psi>,S,T) \<in> ?X2"
      then obtain a P Q M\<^isub>a \<Psi>\<^isub>c xvec where t_def: "T = \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))" and s_def: "S = \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> !P)" and "a \<sharp> P" and "a \<sharp> Q" and "a \<sharp> \<Psi>" and "\<not> a \<sharp> M\<^isub>a" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> R \<longrightarrow> (R = (P \<parallel> \<langle>M\<^isub>a\<rangle>)))" and "guarded P" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<longrightarrow> (supp M\<^isub>a) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P'))" and "((supp P)::name set) \<subseteq> supp M\<^isub>a" and "((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a" and "xvec \<sharp>* \<Psi>"  and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
        by auto
      obtain p::"name prm" where "set p \<subseteq> set xvec \<times> set(p\<bullet>xvec)" and "(p\<bullet>xvec) \<sharp>* a" and "distinctPerm p" and "(p\<bullet>xvec) \<sharp>* M\<^isub>a" and "(p\<bullet>xvec) \<sharp>* \<Psi>\<^isub>c" and "(p\<bullet>xvec) \<sharp>* P" and "(p\<bullet>xvec) \<sharp>* Q" and "(p\<bullet>xvec) \<sharp>* \<Psi>" and "(p\<bullet>xvec) \<sharp>* \<Psi>'"
        by(rule chain.name_list_avoiding[where xvec=xvec and c="(M\<^isub>a,\<Psi>\<^isub>c,a,P,Q,\<Psi>,\<Psi>')"]) auto
      obtain b::name where "b \<sharp> P" and "b \<sharp> Q" and "b \<sharp> \<Psi>" and "b \<sharp> M\<^isub>a" "b \<sharp> \<Psi>\<^isub>c" and "b \<sharp> \<Psi>'" and "b \<sharp> (p\<bullet>a)" and "b \<sharp> (p\<bullet>M\<^isub>a)" and "b \<sharp> (p\<bullet>\<Psi>\<^isub>c)" and "b \<sharp> (p\<bullet>P)" and "b \<sharp> p\<bullet>Q"
        apply(rule at_exists_fresh[OF at_name_inst, where x="(P,Q,\<Psi>,M\<^isub>a,\<Psi>\<^isub>c,\<Psi>',p\<bullet>a,p\<bullet>M\<^isub>a,p\<bullet>\<Psi>\<^isub>c,p\<bullet>P,p\<bullet>Q)"])
        unfolding supp_prod pt_perm_supp[OF pt_name_inst,OF at_name_inst, symmetric]
        apply(rule finite_UnI, (finite_guess | (subst pt_set_finite_ineq[OF pt_name_inst,OF at_name_inst], finite_guess)))+
        apply(subst pt_set_finite_ineq[OF pt_name_inst,OF at_name_inst], finite_guess)
        by force
      have "\<lparr>\<nu>(p\<bullet>a)\<rparr>(\<langle>(p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>c)\<rbrace>) = \<lparr>\<nu>b\<rparr>([(p\<bullet>a,b)] \<bullet> (\<langle>(p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
        apply(subst alphaRes[where y=b])
        apply(simp add: `b \<sharp> (p\<bullet>M\<^isub>a)` `b \<sharp> (p\<bullet>\<Psi>\<^isub>c)`)
        by(rule refl)
      have "\<lparr>\<nu>*xvec\<rparr>(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<lparr>\<nu>*p\<bullet>xvec\<rparr>((p\<bullet>Q) \<parallel> \<lparr>\<nu>(p\<bullet>a)\<rparr>(\<langle>(p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>c)\<rbrace>))" using `(p\<bullet>xvec) \<sharp>* a` `set p \<subseteq> set xvec \<times> set (p \<bullet> xvec)`
        by(subst resChainAlpha[where p=p]) (simp add: `(p\<bullet>xvec) \<sharp>* Q` `(p\<bullet>xvec) \<sharp>* M\<^isub>a` `(p\<bullet>xvec) \<sharp>* \<Psi>\<^isub>c` abs_fresh_list_star' eqvts)+
      have s_def: "S = \<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(p\<bullet>Q) \<parallel> !(p\<bullet>P)" unfolding s_def using `(p\<bullet>xvec) \<sharp>* a` `set p \<subseteq> set xvec \<times> set (p \<bullet> xvec)`
        by(subst resChainAlpha[where p=p]) (simp add: `(p\<bullet>xvec) \<sharp>* Q` `(p\<bullet>xvec) \<sharp>* P` abs_fresh_list_star' eqvts)+
      have t_def: "T = \<lparr>\<nu>*p\<bullet>xvec\<rparr>((p\<bullet>Q) \<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c))\<rbrace>))"
        unfolding `T = \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))` `\<lparr>\<nu>(p\<bullet>a)\<rparr>(\<langle>(p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>c)\<rbrace>) = \<lparr>\<nu>b\<rparr>([(p\<bullet>a,b)] \<bullet> (\<langle>(p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>c)\<rbrace>))` `\<lparr>\<nu>*xvec\<rparr>(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<lparr>\<nu>*p\<bullet>xvec\<rparr>((p\<bullet>Q) \<parallel> \<lparr>\<nu>(p\<bullet>a)\<rparr>(\<langle>(p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>c)\<rbrace>))`
        by simp
      have "(p\<bullet>a) \<sharp> (p\<bullet>P)" using `a \<sharp> P`
        by(subst pt_fresh_bij[OF pt_name_inst, OF at_name_inst])
      note `b \<sharp> \<Psi>'`
      moreover note `b \<sharp> (p\<bullet>P)` `b \<sharp> (p\<bullet>Q)`
      moreover have "guarded(p\<bullet>P)" using `guarded P` by simp
      moreover have "\<not> b \<sharp> ([(p\<bullet>a,b)] \<bullet> (p\<bullet>M\<^isub>a))" using `\<not> a \<sharp> M\<^isub>a`        
        apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
        apply(subst (asm) perm_bool[where pi="p",symmetric])
        by(simp add: eqvts swap_simps)
      moreover have "\<And> \<Psi>. \<Psi> \<otimes> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)) \<turnstile> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>(p\<bullet>P) \<parallel> \<langle>([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<rangle>" using EntCl
        apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
        apply(subst perm_bool[where pi="rev p",symmetric])
        by(simp add: eqvts `(p\<bullet>a) \<sharp> (p\<bullet>P)` `b \<sharp> (p\<bullet>P)`)
      moreover have "\<forall>F. (F \<otimes>\<^sub>F (\<langle>[b],[(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)\<rangle>) \<simeq>\<^sub>F F)"
        using Equiv1
        apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
        apply(subst perm_bool[where pi="rev p",symmetric])
        by(simp add: eqvts swap_simps)
      moreover have "(\<forall>\<Psi> R. b \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)) \<turnstile> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a)) \<Leftarrow> R \<longrightarrow> (R = ((p\<bullet>P) \<parallel> \<langle>([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<rangle>)))" using Equiv2
        apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
        apply(subst (asm) perm_bool[where pi="p",symmetric])
        by(simp add: eqvts `(p\<bullet>a) \<sharp> p\<bullet>P` `b \<sharp> p\<bullet>P` swap_simps)
      moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>P \<longrightarrow> (supp ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))) \<subseteq> ((supp \<Psi>)::name set))" using EntSupp
        apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
        apply(subst (asm) perm_bool[where pi="p",symmetric])
        by(simp add: eqvts swap_simps)
      moreover have "\<forall> P P'. ((([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>P = ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>P') = (P = P'))" using Inj
        apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
        apply(subst (asm) perm_bool[where pi="p",symmetric])
        by(simp add: eqvts swap_simps)
      moreover have "((supp(p\<bullet>P))::name set) \<subseteq> supp ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))" using `((supp P)::name set) \<subseteq> supp M\<^isub>a`
        apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
        apply(subst (asm) perm_bool[where pi="p",symmetric])
        by(simp add: eqvts swap_simps `(p\<bullet>a) \<sharp> p\<bullet>P` `b \<sharp> p\<bullet>P`)
      moreover have "((supp ([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)))::name set) = supp ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))" using `((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a`
        apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
        apply(subst (asm) perm_bool[where pi="p",symmetric])
        by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
      moreover note `(p\<bullet>xvec)\<sharp>*\<Psi>'`
      moreover have "(\<forall> \<Psi> \<phi>. (\<Psi> \<otimes> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)) \<turnstile> \<phi> \<longrightarrow> (\<phi> = ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a)) \<Leftarrow> (p\<bullet>P) \<parallel> \<langle>([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<rangle> \<or> \<Psi> \<turnstile> \<phi>)))" using EntEnt
        apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
        apply(subst perm_bool[where pi="p",symmetric])
        by(simp add: eqvts swap_simps `distinctPerm p` `(p\<bullet>a) \<sharp> p\<bullet>P` `b \<sharp> p\<bullet>P`)
      ultimately have "(\<Psi>',S,T) \<in> ?X2"
        unfolding s_def t_def
        by blast
    }
    note X2swapPsi = this
    {
      fix \<Psi> \<Psi>'::'b
        and S T
      assume "(\<Psi>,S,T) \<in> ?X"
      hence "(\<Psi>,S,T) \<in> ?X1 \<or> (\<Psi>,S,T) \<in> ?X2"
        by blast
      hence "(\<Psi>',S,T) \<in> ?X"
      proof(rule disjE)
        assume "(\<Psi>,S,T) \<in> ?X1"
        thus ?thesis
          by(rule_tac UnI1) (rule X1swapPsi)
      next
        assume "(\<Psi>,S,T) \<in> ?X2"
        thus ?thesis
          by(rule_tac UnI2) (rule X2swapPsi)
      qed
    }
    note XswapPsi = this
    from `a \<sharp> \<Psi>` `a \<sharp> P` `\<not>(a \<sharp> M\<^isub>a)` EntCl Equiv1 Equiv2 EntSupp EntEnt Inj `guarded P` `((supp P)::name set) \<subseteq> supp M\<^isub>a` `((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a` have "(\<Psi>, (\<zero> \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)), (\<zero> \<parallel> !P)) \<in> ?X"
      apply(auto)
      apply(rule_tac x="a" in exI)
      apply(rule_tac x="P" in exI)
      apply(rule_tac x="\<zero>" in exI)
      apply(rule_tac x="M\<^isub>a" in exI)
      apply(rule_tac x="\<Psi>\<^isub>c" in exI)
      apply(rule_tac x="[]" in exI)
      by(auto simp add: fresh_list_nil) blast
    moreover have "eqvt ?X" and "eqvt ?X1" and "eqvt ?X2"
    proof -
      have eqvt_lol: "\<And> (p::name prm) \<Psi>\<^isub>c a. ((\<forall>F. F \<otimes>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F F) = (\<forall>F. (F \<otimes>\<^sub>F \<langle>[(p\<bullet>a)], (p\<bullet>\<Psi>\<^isub>c)\<rangle> \<simeq>\<^sub>F F)))"
        apply(rule_tac pi1="rev p" and t = "(\<forall>F. assertionAux.FrameStatEq SImp'
              (assertionAux.mergeFrame op \<otimes> F (\<langle>[p \<bullet> a], p \<bullet> \<Psi>\<^isub>c\<rangle>))
              F)" in subst[OF perm_bool])
        by(simp add: eqvts)
      have eqvt_lol':
        "\<And> (p::name prm) \<Psi>\<^isub>c ab.
        (\<forall>F. assertionAux.FrameStatEq SImp'
             (assertionAux.mergeFrame op \<otimes> F (\<lparr>\<nu> ab \<rparr> \<langle>\<epsilon>, \<Psi>\<^isub>c\<rangle>)) F)=
        (\<forall>F. assertionAux.FrameStatEq SImp'
        (assertionAux.mergeFrame op \<otimes> F (\<lparr>\<nu>p \<bullet> ab \<rparr> \<langle>\<epsilon>, p \<bullet> \<Psi>\<^isub>c\<rangle>)) F)"
        apply(rule_tac pi1="rev p" and t = "(\<forall>F. assertionAux.FrameStatEq SImp'
          (assertionAux.mergeFrame op \<otimes> F (\<lparr>\<nu>p \<bullet> ab\<rparr> \<langle>\<epsilon>, p \<bullet> \<Psi>\<^isub>c\<rangle>)) F)" in subst[OF perm_bool])
        by(simp add: eqvts)
      have eqvt_lol'': "\<And> (p::name prm) \<Psi>\<^isub>c (ab::name) M\<^isub>a R P. ((\<forall>\<Psi>. ab \<sharp> \<Psi> \<longrightarrow> (\<forall>R. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>R \<longrightarrow> R = P \<parallel> \<langle>M\<^isub>a\<rangle>)) = (\<forall>\<Psi>. (p \<bullet> ab) \<sharp> \<Psi> \<longrightarrow>
              (\<forall>R. \<Psi> \<otimes> (p \<bullet> \<Psi>\<^isub>c) \<turnstile> (p \<bullet> M\<^isub>a)\<Leftarrow>R \<longrightarrow> R = (p \<bullet> P) \<parallel> \<langle>(p \<bullet> M\<^isub>a)\<rangle>)))"
        apply(rule_tac pi1="rev p" and t = "\<forall>\<Psi>. (p \<bullet> ab) \<sharp> \<Psi> \<longrightarrow>
              (\<forall>R. \<Psi> \<otimes> (p \<bullet> \<Psi>\<^isub>c) \<turnstile> (p \<bullet> M\<^isub>a)\<Leftarrow>R \<longrightarrow> R = (p \<bullet> P) \<parallel> \<langle>(p \<bullet> M\<^isub>a)\<rangle>)" in subst[OF perm_bool])
        by(simp add: eqvts)
      have eqvt_lol''': "\<And> (p::name prm) \<Psi>\<^isub>c M\<^isub>a P.
        (\<forall>\<Psi>. \<Psi> \<otimes> (\<Psi>\<^isub>c) \<turnstile> (M\<^isub>a)\<Leftarrow>(P) \<parallel> \<langle>(M\<^isub>a)\<rangle>)
        =
        (\<forall>\<Psi>. \<Psi> \<otimes> (p \<bullet> \<Psi>\<^isub>c) \<turnstile> (p \<bullet> M\<^isub>a)\<Leftarrow>(p \<bullet> P) \<parallel> \<langle>(p \<bullet> M\<^isub>a)\<rangle>)"
        apply(rule_tac pi1="p" and t="((\<forall>\<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>))" in subst[OF perm_bool])
        by(simp add: eqvts)
      have eqvt_lol'''': "\<And> (p::name prm) (P::('a,'b,'c) psi) M\<^isub>a::'a.
        (((supp P)::name set) \<subseteq> supp (M\<^isub>a)) = ((supp (p \<bullet> P)::name set) \<subseteq> supp (p \<bullet> M\<^isub>a))"
        apply(rule_tac pi1="p" and t="(supp P \<subseteq> supp M\<^isub>a)" in subst[OF perm_bool])
        by(simp add: eqvts)
      have eqvt_lol''''': "\<And> (p::name prm) (\<Psi>\<^isub>c::'b) (M\<^isub>a::'a).
        (((supp \<Psi>\<^isub>c)::name set) = supp (M\<^isub>a)) = ((supp (p \<bullet> \<Psi>\<^isub>c)::name set) = supp (p \<bullet> M\<^isub>a))"
        apply(rule_tac pi1="p" and t="(supp \<Psi>\<^isub>c = supp M\<^isub>a)" in subst[OF perm_bool])
        by(simp add: eqvts)
      have eqvt_lol'''''': "\<And> (p::name prm) M\<^isub>a.
        (\<forall>P P'. (M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P'))
        =
        (\<forall>P P'. ((p \<bullet> M\<^isub>a)\<Leftarrow>P = (p \<bullet> M\<^isub>a)\<Leftarrow>P') = (P = P'))"
        apply(rule_tac pi1="p" and t="(\<forall>P P'. (M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P'))" in subst[OF perm_bool])
        by(simp add: eqvts)
      have eqvt_lol''''''': "\<And> (p::name prm) M\<^isub>a.
        (\<forall>\<Psi>. (\<exists>P. \<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P) \<longrightarrow> (supp M\<^isub>a::name set) \<subseteq> supp \<Psi>)
        =
        (\<forall>\<Psi>. (\<exists>P. \<Psi> \<turnstile> (p \<bullet> M\<^isub>a)\<Leftarrow>P) \<longrightarrow> (supp (p \<bullet> M\<^isub>a)::name set) \<subseteq> supp \<Psi>)"
        apply(rule_tac pi1="p" and t="(\<forall>\<Psi>. (\<exists>P. \<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P) \<longrightarrow> (supp M\<^isub>a::name set) \<subseteq> supp \<Psi>)" in subst[OF perm_bool])
        by(simp add: eqvts)
      have eqvt_lol'''''''': "\<And> (p::name prm) M\<^isub>a P \<Psi>\<^isub>c.
        (\<forall>\<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> \<phi> \<longrightarrow> \<phi> = M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> \<phi>)
        =
        (\<forall>\<Psi> \<phi>. \<Psi> \<otimes> (p \<bullet> \<Psi>\<^isub>c) \<turnstile> \<phi> \<longrightarrow> \<phi> = (p \<bullet> M\<^isub>a)\<Leftarrow>(p \<bullet> P) \<parallel> \<langle>(p\<bullet>M\<^isub>a)\<rangle> \<or> \<Psi> \<turnstile> \<phi>)
        "
        apply(rule_tac pi1="p" and t="(\<forall>\<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> \<phi> \<longrightarrow> \<phi> = M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> \<phi>)" in subst[OF perm_bool])
        by(simp add: eqvts)
      show "eqvt ?X1"
        apply(auto simp add: eqvt_def eqvts)
        apply(
          rule_tac x = "p\<bullet>ab" in exI,
          rule_tac x = "p\<bullet>P" in exI,
          rule_tac x = "p\<bullet>Q" in exI,
          rule_tac x = "p\<bullet>M\<^isub>a" in exI,
          rule_tac x = "p\<bullet>\<Psi>\<^isub>c" in exI)
        apply(rule_tac x = "p\<bullet>xvec" in exI, simp)
        apply(simp add: pt_fresh_bij[OF pt_name_inst, OF at_name_inst] pt_fresh_star_bij[OF pt_name_inst, OF at_name_inst] clauseClosed''[symmetric] eqvt_lol'[symmetric] eqvt_lol''[symmetric] eqvt_lol'''[symmetric] eqvt_lol''''[symmetric] eqvt_lol'''''[symmetric] eqvt_lol''''''[symmetric] eqvt_lol'''''''[symmetric] eqvt_lol''''''''[symmetric])
        done
      moreover show "eqvt ?X2"
        apply(auto simp add: eqvt_def eqvts)
        apply(rule_tac x = "p\<bullet>ab" in exI)
        apply(rule_tac x = "p\<bullet>P" in exI)
        apply(rule_tac x = "p\<bullet>Q" in exI)
        apply simp
        apply(rule_tac x = "p\<bullet>M\<^isub>a" in exI)
        apply(rule_tac x = "p\<bullet>\<Psi>\<^isub>c" in exI)
        apply(rule_tac x = "p\<bullet>xvec" in exI, simp)        
        by(simp add: eqvts pt_fresh_bij[OF pt_name_inst, OF at_name_inst] pt_fresh_star_bij[OF pt_name_inst, OF at_name_inst] clauseClosed''[symmetric] eqvt_lol'[symmetric] eqvt_lol''[symmetric] eqvt_lol'''[symmetric] eqvt_lol''''[symmetric] eqvt_lol'''''[symmetric] eqvt_lol''''''[symmetric] eqvt_lol'''''''[symmetric] eqvt_lol''''''''[symmetric])
      ultimately show "eqvt ?X"
        by(rule eqvtUnion)
  qed
  ultimately have "\<Psi> \<rhd> (\<zero> \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) \<sim> (\<zero> \<parallel> !P)" using `a \<sharp> \<Psi>` `a \<sharp> P` `\<not>(a \<sharp> M\<^isub>a)`
    proof(coinduct rule: transitiveCoinduct) (* transitiveCoinduct' might work as well. *)
      case(cStatEq \<Psi> S T)
      show ?case
      proof(case_tac "(\<Psi>, S, T) \<in> ?X1")
	assume "(\<Psi>, S, T) \<in> ?X1"
	then obtain a P Q M\<^isub>a \<Psi>\<^isub>c xvec where "S = \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))" and "T = \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> !P)" and "a \<sharp> P" and "a \<sharp> Q" and "a \<sharp> \<Psi>" and "\<not> a \<sharp> M\<^isub>a" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> R \<longrightarrow> (R = (P \<parallel> \<langle>M\<^isub>a\<rangle>)))" and "guarded P" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<longrightarrow> (supp M\<^isub>a) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P'))" and "((supp P)::name set) \<subseteq> supp M\<^isub>a" and "((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a" and "xvec \<sharp>* \<Psi>" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
          by auto
        have Equiv2: "\<And>\<Psi> R. \<lbrakk>a \<sharp> \<Psi>\<rbrakk> \<Longrightarrow> (\<lbrakk>\<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> R\<rbrakk> \<Longrightarrow> R = (P \<parallel> \<langle>M\<^isub>a\<rangle>))" using Equiv2
          by simp
        have "extractFrame (!P) = \<langle>\<epsilon>, \<one>\<rangle>" by simp
	moreover obtain A\<^isub>Q \<Psi>\<^isub>Q where FrQ: "extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>" and "A\<^isub>Q \<sharp>* \<Psi>" and "a \<sharp> A\<^isub>Q" and "A\<^isub>Q \<sharp>* M\<^isub>a" and "A\<^isub>Q \<sharp>* \<Psi>\<^isub>c"
	  by(rule_tac C="(\<Psi>, a, M\<^isub>a, \<Psi>\<^isub>c)" in freshFrame) auto
        ultimately have "extractFrame (Q \<parallel> !P) = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q \<otimes> \<one>\<rangle>" by simp
        have "extractFrame(\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<langle>[a], \<one> \<otimes> \<Psi>\<^isub>c\<rangle>" by simp

        have "\<langle>A\<^isub>Q, \<Psi>\<^isub>Q \<otimes> \<one>\<rangle> \<simeq>\<^sub>F \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>" by(metis FrameStatEqTrans frameIntIdentity)
        with `extractFrame (Q \<parallel> !P) = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q \<otimes> \<one>\<rangle>`
        have "extractFrame (Q \<parallel> !P) \<simeq>\<^sub>F \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>" by simp

        have "\<langle>[a], \<one> \<otimes> \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle>" by(metis FrameStatEqTrans frameIntCommutativity frameIntIdentity)
        with `extractFrame(\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<langle>[a], \<one> \<otimes> \<Psi>\<^isub>c\<rangle>`
        have "extractFrame(\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) \<simeq>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle>" by simp

        have "a \<sharp> \<Psi>\<^isub>Q"
          proof -
            have "a \<sharp> extractFrame Q" using `a \<sharp> Q`
              by(rule extractFrameFresh)
            hence "a \<sharp> \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>"
              by(subst `extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>`[symmetric])
            moreover have "a \<notin> set A\<^isub>Q" using `a \<sharp> A\<^isub>Q`
              by(simp add: fresh_def)
            ultimately show ?thesis
              by(simp add: frameResChainFresh)
          qed
        from `a \<sharp> A\<^isub>Q` have "A\<^isub>Q \<sharp>* [a]" by simp
        from `a \<sharp> \<Psi>\<^isub>Q` have "[a] \<sharp>* \<Psi>\<^isub>Q" by simp
        from `A\<^isub>Q \<sharp>* [a]` `A\<^isub>Q \<sharp>* \<Psi>\<^isub>c` `[a] \<sharp>* \<Psi>\<^isub>Q` have fk1: "(\<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>) \<otimes>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle> = \<langle>(A\<^isub>Q@[a]), \<Psi>\<^isub>Q \<otimes> \<Psi>\<^isub>c\<rangle>" by (rule mergeFrames)
        have fk2: "\<langle>(A\<^isub>Q@[a]), \<Psi>\<^isub>Q \<otimes> \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F \<lparr>\<nu>*A\<^isub>Q\<rparr>((\<langle>\<epsilon>, \<Psi>\<^isub>Q\<rangle>) \<otimes>\<^sub>F \<langle>[a],\<Psi>\<^isub>c\<rangle>)"
          apply(subst frameChainAppend)
          apply(rule frameResChainPres)
          apply(rule FrameStatEqTrans[OF frameIntCommutativity])
          apply(subst mergeFrameSimps[symmetric, OF `[a] \<sharp>* \<Psi>\<^isub>Q`])
          by(rule frameCommutativity)
        have "(\<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>) \<otimes>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>"
          apply(subst fk1)
          apply(rule FrameStatEqTrans[OF fk2])
          apply(rule frameResChainPres)
          by(rule Equiv1)
        then have frameCoding: "extractFrame(\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<langle>[a], \<one> \<otimes> \<Psi>\<^isub>c\<rangle>"
          by simp
        have frameCoding': "extractFrame(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) \<simeq>\<^sub>F \<langle>A\<^isub>Q, \<Psi>\<^isub>Q \<otimes> \<one>\<rangle>"
          apply(subst extractFrameSimps)
          apply(subst frameCoding)
          apply(subst `extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>`)
          apply(subst mergeFrames[OF `A\<^isub>Q \<sharp>* [a]`])
          apply(rule freshCompChain)
          apply(rule freshBottoChain)
          apply(rule `A\<^isub>Q \<sharp>* \<Psi>\<^isub>c`)
          apply(rule `[a] \<sharp>* \<Psi>\<^isub>Q`)
          apply(subst frameChainAppend)
          apply(rule frameResChainPres)
          apply(rule FrameStatEqTrans)
          apply(rule FrameStatEqSym[OF frameIntAssociativity])
          apply(rule FrameStatEqTrans[OF frameIntCommutativity])
          apply(subst mergeFrameSimps[symmetric])         
          apply(rule freshCompChain)
          apply(rule `[a] \<sharp>* \<Psi>\<^isub>Q`)
          apply(rule freshBottoChain)
          apply(rule FrameStatEqTrans[OF frameCommutativity])
          by(rule Equiv1)
        have frameCoding'': "extractFrame((\<lbrace>\<Psi>\<rbrace> \<parallel> Q) \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) \<simeq>\<^sub>F \<langle>A\<^isub>Q, \<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<one>\<rangle>"
          apply(subst extractFrameSimps)
          apply(subst frameCoding)
          apply(subst extractFrameSimps)
          apply(subst `extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>`)
          apply(subst extractFrameSimps)
          apply(subst mergeFrameSimps'[OF `A\<^isub>Q \<sharp>* \<Psi>`])
          apply(subst mergeFrames[OF `A\<^isub>Q \<sharp>* [a]`])
          apply(rule freshCompChain)
          apply(rule freshBottoChain)
          apply(rule `A\<^isub>Q \<sharp>* \<Psi>\<^isub>c`)
          apply(rule freshCompChain)
          apply(simp add: `a \<sharp> \<Psi>`)
          apply(rule `[a] \<sharp>* \<Psi>\<^isub>Q`)
          apply(subst frameChainAppend)
          apply(rule frameResChainPres)
          apply(rule FrameStatEqTrans)
          apply(rule FrameStatEqSym[OF frameIntAssociativity])
          apply(rule FrameStatEqTrans[OF frameIntCommutativity])
          apply(subst mergeFrameSimps[symmetric])         
          apply(rule freshCompChain)
          apply(rule freshCompChain)
          apply(simp add: `a \<sharp> \<Psi>`)
          apply(rule `[a] \<sharp>* \<Psi>\<^isub>Q`)
          apply(rule freshBottoChain)
          apply(rule FrameStatEqTrans[OF frameCommutativity])
          apply(rule FrameStatEqSym)
          apply(rule FrameStatEqTrans)
          apply(subst frameNilStatEq)
          apply(rule AssertionStatEqSym[OF Associativity])
          apply(rule FrameStatEqSym)
          by(rule Equiv1)
        have frameResMv: "\<lparr>\<nu>a\<rparr> \<langle>\<epsilon>, \<one> \<otimes> \<Psi>\<^isub>c\<rangle> = \<langle>[a], \<one> \<otimes> \<Psi>\<^isub>c\<rangle>"
          by simp
        have frameCodingStatEq: "extractFrame((\<lbrace>\<Psi>\<rbrace> \<parallel> Q) \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) \<simeq>\<^sub>F insertAssertion (extractFrame (Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))) \<Psi> "
          unfolding extractFrameSimps mergeFrame.simps `extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>`
          apply(subst insertAssertionResChain[OF `A\<^isub>Q \<sharp>* \<Psi>`])
          unfolding insertAssertion.simps
          apply(subst (1 2) frameResMv)
          apply(subst mergeFrames[OF `A\<^isub>Q \<sharp>* [a]`])
          apply(rule freshCompChain)
          apply(rule freshBottoChain)
          apply(rule `A\<^isub>Q \<sharp>* \<Psi>\<^isub>c`)
          apply(rule freshCompChain)
          apply(simp add: `a \<sharp> \<Psi>`)
          apply(rule `[a] \<sharp>* \<Psi>\<^isub>Q`)
          apply(subst mergeFrames[OF `A\<^isub>Q \<sharp>* [a]`])
          apply(rule freshCompChain)
          apply(rule freshBottoChain)
          apply(rule `A\<^isub>Q \<sharp>* \<Psi>\<^isub>c`)
          apply(rule `[a] \<sharp>* \<Psi>\<^isub>Q`)
          apply(subst insertAssertionResChain)
          apply(simp add: `A\<^isub>Q \<sharp>* \<Psi>` `a \<sharp> \<Psi>`)
          unfolding insertAssertion.simps
          apply(rule frameResChainPres)
          apply(subst frameNilStatEq)
          apply(rule AssertionStatEqTrans[OF Associativity])
          by(rule AssertionStatEqRefl)
        have "extractFrame (\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) \<simeq>\<^sub>F extractFrame(!P)"
          apply(subst frameCoding)
          unfolding extractFrameSimps
          apply(rule FrameStatEqTrans[OF frameIntCommutativity])
          apply(subst mergeFrameSimps[symmetric])
          apply(rule freshBottoChain)
          apply(rule FrameStatEqTrans[OF frameCommutativity])
          by(rule Equiv1)
        hence "extractFrame (Q \<parallel> (\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))) \<simeq>\<^sub>F extractFrame(Q \<parallel> !P)" using `a \<sharp> A\<^isub>Q`
          apply -
          apply(rule FrameStatEqTrans[OF frameCoding'])
          unfolding extractFrameSimps `extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>`
          by simp
	show ?case
          unfolding  `S = \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))` `T = \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> !P)`
          apply(rule FrameStatEqTrans[OF insertAssertionExtractFrameFresh, OF `xvec \<sharp>* \<Psi>`])
          apply(rule FrameStatEqSym)
          apply(rule FrameStatEqTrans[OF insertAssertionExtractFrameFresh, OF `xvec \<sharp>* \<Psi>`])
          apply(rule FrameStatEqSym[OF frameResChainPres])
          apply(rule FrameStatEqTrans)
          apply(rule FrameStatEqSym[OF frameCodingStatEq])
          apply(rule FrameStatEqTrans[OF frameCoding''])
          unfolding extractFrameSimps `extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>`
          apply(subst frameBottomExpand)
          apply(subst mergeFrames)
          apply simp
          apply(rule freshBottoChain)
          apply simp
          apply(subst append_Nil2)
          apply(subst insertAssertionResChain)
          apply fact
          unfolding insertAssertion.simps
          by(rule FrameStatEqRefl)
      next
        assume  "(\<Psi>, S, T) \<notin> ?X1"
        hence "(\<Psi>, S, T) \<in> ?X2" using cStatEq
          by blast
	then obtain a P Q M\<^isub>a \<Psi>\<^isub>c xvec where "T = \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))" and "S = \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> !P)" and "a \<sharp> P" and "a \<sharp> Q" and "a \<sharp> \<Psi>" and "\<not> a \<sharp> M\<^isub>a" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> R \<longrightarrow> (R = (P \<parallel> \<langle>M\<^isub>a\<rangle>)))" and "guarded P" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<longrightarrow> (supp M\<^isub>a) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P'))" and "((supp P)::name set) \<subseteq> supp M\<^isub>a" and "((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a" and "xvec \<sharp>* \<Psi>" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
          by auto
        have Equiv2: "\<And>\<Psi> R. \<lbrakk>a \<sharp> \<Psi>\<rbrakk> \<Longrightarrow> (\<lbrakk>\<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> R\<rbrakk> \<Longrightarrow> R = (P \<parallel> \<langle>M\<^isub>a\<rangle>))" using Equiv2
          by simp
        have "extractFrame (!P) = \<langle>\<epsilon>, \<one>\<rangle>" by simp
	moreover obtain A\<^isub>Q \<Psi>\<^isub>Q where FrQ: "extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>" and "A\<^isub>Q \<sharp>* \<Psi>" and "a \<sharp> A\<^isub>Q" and "A\<^isub>Q \<sharp>* M\<^isub>a" and "A\<^isub>Q \<sharp>* \<Psi>\<^isub>c"
	  by(rule_tac C="(\<Psi>, a, M\<^isub>a, \<Psi>\<^isub>c)" in freshFrame) auto
        ultimately have "extractFrame (Q \<parallel> !P) = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q \<otimes> \<one>\<rangle>" by simp
        have "extractFrame(\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<langle>[a], \<one> \<otimes> \<Psi>\<^isub>c\<rangle>" by simp
        have "\<langle>A\<^isub>Q, \<Psi>\<^isub>Q \<otimes> \<one>\<rangle> \<simeq>\<^sub>F \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>" by(metis FrameStatEqTrans frameIntIdentity)
        with `extractFrame (Q \<parallel> !P) = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q \<otimes> \<one>\<rangle>`
        have "extractFrame (Q \<parallel> !P) \<simeq>\<^sub>F \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>" by simp

        have "\<langle>[a], \<one> \<otimes> \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle>" by(metis FrameStatEqTrans frameIntCommutativity frameIntIdentity)
        with `extractFrame(\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<langle>[a], \<one> \<otimes> \<Psi>\<^isub>c\<rangle>`
        have "extractFrame(\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) \<simeq>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle>" by simp
        have "a \<sharp> \<Psi>\<^isub>Q"
          proof -
            have "a \<sharp> extractFrame Q" using `a \<sharp> Q`
              by(rule extractFrameFresh)
            hence "a \<sharp> \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>"
              by(subst `extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>`[symmetric])
            moreover have "a \<notin> set A\<^isub>Q" using `a \<sharp> A\<^isub>Q`
              by(simp add: fresh_def)
            ultimately show ?thesis
              by(simp add: frameResChainFresh)
          qed
        from `a \<sharp> A\<^isub>Q` have "A\<^isub>Q \<sharp>* [a]" by simp
        from `a \<sharp> \<Psi>\<^isub>Q` have "[a] \<sharp>* \<Psi>\<^isub>Q" by simp
        from `A\<^isub>Q \<sharp>* [a]` `A\<^isub>Q \<sharp>* \<Psi>\<^isub>c` `[a] \<sharp>* \<Psi>\<^isub>Q` have fk1: "(\<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>) \<otimes>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle> = \<langle>(A\<^isub>Q@[a]), \<Psi>\<^isub>Q \<otimes> \<Psi>\<^isub>c\<rangle>" by (rule mergeFrames)
        have fk2: "\<langle>(A\<^isub>Q@[a]), \<Psi>\<^isub>Q \<otimes> \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F \<lparr>\<nu>*A\<^isub>Q\<rparr>((\<langle>\<epsilon>, \<Psi>\<^isub>Q\<rangle>) \<otimes>\<^sub>F \<langle>[a],\<Psi>\<^isub>c\<rangle>)"
          apply(subst frameChainAppend)
          apply(rule frameResChainPres)
          apply(rule FrameStatEqTrans[OF frameIntCommutativity])
          apply(subst mergeFrameSimps[symmetric, OF `[a] \<sharp>* \<Psi>\<^isub>Q`])
          by(rule frameCommutativity)
        have "(\<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>) \<otimes>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>"
          apply(subst fk1)
          apply(rule FrameStatEqTrans[OF fk2])
          apply(rule frameResChainPres)
          by(rule Equiv1)
        then have frameCoding: "extractFrame(\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<langle>[a], \<one> \<otimes> \<Psi>\<^isub>c\<rangle>"
          by simp
        have frameCoding': "extractFrame(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) \<simeq>\<^sub>F \<langle>A\<^isub>Q, \<Psi>\<^isub>Q \<otimes> \<one>\<rangle>"
          apply(subst extractFrameSimps)
          apply(subst frameCoding)
          apply(subst `extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>`)
          apply(subst mergeFrames[OF `A\<^isub>Q \<sharp>* [a]`])
          apply(rule freshCompChain)
          apply(rule freshBottoChain)
          apply(rule `A\<^isub>Q \<sharp>* \<Psi>\<^isub>c`)
          apply(rule `[a] \<sharp>* \<Psi>\<^isub>Q`)
          apply(subst frameChainAppend)
          apply(rule frameResChainPres)
          apply(rule FrameStatEqTrans)
          apply(rule FrameStatEqSym[OF frameIntAssociativity])
          apply(rule FrameStatEqTrans[OF frameIntCommutativity])
          apply(subst mergeFrameSimps[symmetric])         
          apply(rule freshCompChain)
          apply(rule `[a] \<sharp>* \<Psi>\<^isub>Q`)
          apply(rule freshBottoChain)
          apply(rule FrameStatEqTrans[OF frameCommutativity])
          by(rule Equiv1)
        have frameCoding'': "extractFrame((\<lbrace>\<Psi>\<rbrace> \<parallel> Q) \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) \<simeq>\<^sub>F \<langle>A\<^isub>Q, \<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<one>\<rangle>"
          apply(subst extractFrameSimps)
          apply(subst frameCoding)
          apply(subst extractFrameSimps)
          apply(subst `extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>`)
          apply(subst extractFrameSimps)
          apply(subst mergeFrameSimps'[OF `A\<^isub>Q \<sharp>* \<Psi>`])
          apply(subst mergeFrames[OF `A\<^isub>Q \<sharp>* [a]`])
          apply(rule freshCompChain)
          apply(rule freshBottoChain)
          apply(rule `A\<^isub>Q \<sharp>* \<Psi>\<^isub>c`)
          apply(rule freshCompChain)
          apply(simp add: `a \<sharp> \<Psi>`)
          apply(rule `[a] \<sharp>* \<Psi>\<^isub>Q`)
          apply(subst frameChainAppend)
          apply(rule frameResChainPres)
          apply(rule FrameStatEqTrans)
          apply(rule FrameStatEqSym[OF frameIntAssociativity])
          apply(rule FrameStatEqTrans[OF frameIntCommutativity])
          apply(subst mergeFrameSimps[symmetric])         
          apply(rule freshCompChain)
          apply(rule freshCompChain)
          apply(simp add: `a \<sharp> \<Psi>`)
          apply(rule `[a] \<sharp>* \<Psi>\<^isub>Q`)
          apply(rule freshBottoChain)
          apply(rule FrameStatEqTrans[OF frameCommutativity])
          apply(rule FrameStatEqSym)
          apply(rule FrameStatEqTrans)
          apply(subst frameNilStatEq)
          apply(rule AssertionStatEqSym[OF Associativity])
          apply(rule FrameStatEqSym)
          by(rule Equiv1)
        have frameResMv: "\<lparr>\<nu>a\<rparr> \<langle>\<epsilon>, \<one> \<otimes> \<Psi>\<^isub>c\<rangle> = \<langle>[a], \<one> \<otimes> \<Psi>\<^isub>c\<rangle>"
          by simp
        have frameCodingStatEq: "extractFrame((\<lbrace>\<Psi>\<rbrace> \<parallel> Q) \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) \<simeq>\<^sub>F insertAssertion (extractFrame (Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))) \<Psi> "
          unfolding extractFrameSimps mergeFrame.simps `extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>`
          apply(subst insertAssertionResChain[OF `A\<^isub>Q \<sharp>* \<Psi>`])
          unfolding insertAssertion.simps
          apply(subst (1 2) frameResMv)
          apply(subst mergeFrames[OF `A\<^isub>Q \<sharp>* [a]`])
          apply(rule freshCompChain)
          apply(rule freshBottoChain)
          apply(rule `A\<^isub>Q \<sharp>* \<Psi>\<^isub>c`)
          apply(rule freshCompChain)
          apply(simp add: `a \<sharp> \<Psi>`)
          apply(rule `[a] \<sharp>* \<Psi>\<^isub>Q`)
          apply(subst mergeFrames[OF `A\<^isub>Q \<sharp>* [a]`])
          apply(rule freshCompChain)
          apply(rule freshBottoChain)
          apply(rule `A\<^isub>Q \<sharp>* \<Psi>\<^isub>c`)
          apply(rule `[a] \<sharp>* \<Psi>\<^isub>Q`)
          apply(subst insertAssertionResChain)
          apply(simp add: `A\<^isub>Q \<sharp>* \<Psi>` `a \<sharp> \<Psi>`)
          unfolding insertAssertion.simps
          apply(rule frameResChainPres)
          apply(subst frameNilStatEq)
          apply(rule AssertionStatEqTrans[OF Associativity])
          by(rule AssertionStatEqRefl)
        have "extractFrame (\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) \<simeq>\<^sub>F extractFrame(!P)"
          apply(subst frameCoding)
          unfolding extractFrameSimps
          apply(rule FrameStatEqTrans[OF frameIntCommutativity])
          apply(subst mergeFrameSimps[symmetric])
          apply(rule freshBottoChain)
          apply(rule FrameStatEqTrans[OF frameCommutativity])
          by(rule Equiv1)
        hence "extractFrame (Q \<parallel> (\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))) \<simeq>\<^sub>F extractFrame(Q \<parallel> !P)" using `a \<sharp> A\<^isub>Q`
          apply -
          apply(rule FrameStatEqTrans[OF frameCoding'])
          unfolding extractFrameSimps `extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>`
          by simp
	show ?case
          unfolding  `T = \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))` `S = \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> !P)`
          apply(rule FrameStatEqTrans[OF insertAssertionExtractFrameFresh, OF `xvec \<sharp>* \<Psi>`])
          apply(rule FrameStatEqSym)
          apply(rule FrameStatEqTrans[OF insertAssertionExtractFrameFresh, OF `xvec \<sharp>* \<Psi>`])
          apply(rule frameResChainPres)
          apply(rule FrameStatEqTrans)
          apply(rule FrameStatEqSym[OF frameCodingStatEq])
          apply(rule FrameStatEqTrans[OF frameCoding''])
          unfolding extractFrameSimps `extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>`
          apply(subst frameBottomExpand)
          apply(subst mergeFrames)
          apply simp
          apply(rule freshBottoChain)
          apply simp
          apply(subst append_Nil2)
          apply(subst insertAssertionResChain)
          apply fact
          unfolding insertAssertion.simps
          by(rule FrameStatEqRefl)
      qed
    next
      case (cSym \<Psi> R S)
      thus ?case
        by blast
    next
      case (cExt \<Psi> S T \<Psi>')
      thus ?case
        by(rule_tac disjI1) (rule XswapPsi)
    next
      case(cSim \<Psi> S T)
      let ?Y = "{(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim> P' \<and> ((\<Psi>, P', Q') \<in> ?X \<or> \<Psi> \<rhd> P' \<sim> Q') \<and> \<Psi> \<rhd> Q' \<sim> Q}"
      let ?Y1 = "{(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim> P' \<and> ((\<Psi>, P', Q') \<in> ?X1 \<or> \<Psi> \<rhd> P' \<sim> Q') \<and> \<Psi> \<rhd> Q' \<sim> Q}"
      let ?Y2 = "{(\<Psi>, P, Q) | \<Psi> P P' Q' Q. \<Psi> \<rhd> P \<sim> P' \<and> ((\<Psi>, P', Q') \<in> ?X2 \<or> \<Psi> \<rhd> P' \<sim> Q') \<and> \<Psi> \<rhd> Q' \<sim> Q}"
      let ?YA1 = "{(\<Psi>, P, Q, A::name list) | \<Psi> P P' Q' Q A. \<Psi> \<rhd> P \<sim> P' \<and> ((\<Psi>, P', Q') \<in> ?X1 \<or> \<Psi> \<rhd> P' \<sim> Q') \<and> \<Psi> \<rhd> Q' \<sim> Q \<and> A \<sharp>* (P,P',Q',Q)}"
      let ?YA2 = "{(\<Psi>, P, Q, A::name list) | \<Psi> P P' Q' Q A. \<Psi> \<rhd> P \<sim> P' \<and> ((\<Psi>, P', Q') \<in> ?X2 \<or> \<Psi> \<rhd> P' \<sim> Q') \<and> \<Psi> \<rhd> Q' \<sim> Q \<and> A \<sharp>* (P,P',Q',Q)}"
      let ?YA = "{(\<Psi>, P, Q, A::name list) | \<Psi> P P' Q' Q A. \<Psi> \<rhd> P \<sim> P' \<and> ((\<Psi>, P', Q') \<in> ?X \<or> \<Psi> \<rhd> P' \<sim> Q') \<and> \<Psi> \<rhd> Q' \<sim> Q \<and> A \<sharp>* (P,P',Q',Q)}"
      have "eqvt(?Y)" using `eqvt ?X` by blast
      have "eqvt(?Y1)" using `eqvt ?X1` by blast
      have "eqvt(?Y2)" using `eqvt ?X2` by blast
      have "eqvt(?YA1)" using `eqvt ?X1` by blast
      have "eqvt(?YA2)" using `eqvt ?X2` by blast
      have "eqvt(?YA)" using `eqvt ?X` by blast
      {
        fix \<Psi> P Q
        assume "(\<Psi>,P,Q) \<in> ?Y1"
        then obtain P' Q' where "\<Psi> \<rhd> P \<sim> P'" and "(\<Psi>,P',Q') \<in> ?X1 \<or> \<Psi> \<rhd> P' \<sim> Q'" and "\<Psi> \<rhd> Q' \<sim> Q"
          by auto blast
        from `(\<Psi>,P',Q') \<in> ?X1 \<or> \<Psi> \<rhd> P' \<sim> Q'` have "(\<Psi>,P,Q) \<in> ?Y"
        proof(rule disjE)
          assume "(\<Psi>,P',Q') \<in> ?X1"
          hence  "(\<Psi>,P',Q') \<in> ?X"
            by(rule UnI1)
          show ?thesis
            apply(rule lol)
            by(rule refl|fact|rule disjI1)+
        next
          assume "\<Psi> \<rhd> P' \<sim> Q'"
          show ?thesis
            apply(rule lol)
            by(rule refl|fact|rule disjI2)+
        qed
      }
      note Y1inY=this
      {
        fix \<Psi> P Q
        assume "(\<Psi>,P,Q) \<in> ?Y2"
        then obtain P' Q' where "\<Psi> \<rhd> P \<sim> P'" and "(\<Psi>,P',Q') \<in> ?X2 \<or> \<Psi> \<rhd> P' \<sim> Q'" and "\<Psi> \<rhd> Q' \<sim> Q"
          by auto blast
        from `(\<Psi>,P',Q') \<in> ?X2 \<or> \<Psi> \<rhd> P' \<sim> Q'` have "(\<Psi>,P,Q) \<in> ?Y"
        proof(rule disjE)
          assume "(\<Psi>,P',Q') \<in> ?X2"
          hence  "(\<Psi>,P',Q') \<in> ?X"
            by(rule UnI2)
          show ?thesis
            apply(rule lol)
            by(rule refl|fact|rule disjI1)+
        next
          assume "\<Psi> \<rhd> P' \<sim> Q'"
          show ?thesis
            apply(rule lol)
            by(rule refl|fact|rule disjI2)+
        qed
      }
      note Y2inY=this
      {
        fix \<Psi> P Q
          and x::name
        assume "(\<Psi>,P,Q) \<in> ?X1"
          and    "x \<sharp> \<Psi>"
        hence  "(\<Psi>,\<lparr>\<nu>x\<rparr>P,\<lparr>\<nu>x\<rparr>Q) \<in> ?X1"
          apply auto
          apply(rule_tac x=a in exI)
          apply(rule_tac x=P in exI)
          apply(rule_tac x=Q in exI)
          apply(rule_tac x=M\<^isub>a in exI)
          apply(rule_tac x=\<Psi>\<^isub>c in exI)
          apply(rule_tac x="x#xvec" in exI)
          apply auto
          done
      }
      note X1resClose = this
      {
        fix \<Psi> P Q
          and x::name
        assume "(\<Psi>,P,Q) \<in> ?X2"
          and    "x \<sharp> \<Psi>"
        hence  "(\<Psi>,\<lparr>\<nu>x\<rparr>P,\<lparr>\<nu>x\<rparr>Q) \<in> ?X2"
          apply auto
          apply(rule_tac x=a in exI)
          apply(rule_tac x=P in exI)
          apply(rule_tac x=Q in exI)
          apply(rule_tac x=M\<^isub>a in exI)
          apply(rule_tac x=\<Psi>\<^isub>c in exI)
          apply(rule_tac x="x#xvec" in exI)
          apply auto
          done
      }
      note X2resClose = this
      {
        fix \<Psi> P Q
          and x::name
        assume "(\<Psi>,P,Q) \<in> ?X"
          and    "x \<sharp> \<Psi>"
        hence "(\<Psi>,P,Q) \<in> ?X1 \<or> (\<Psi>,P,Q) \<in> ?X2"
          by blast
        hence  "(\<Psi>,\<lparr>\<nu>x\<rparr>P,\<lparr>\<nu>x\<rparr>Q) \<in> ?X" using `(\<Psi>,P,Q) \<in> ?X1 \<or> (\<Psi>,P,Q) \<in> ?X2`
        proof(erule_tac disjE)
          assume "(\<Psi>,P,Q) \<in> ?X1"
          hence "(\<Psi>,\<lparr>\<nu>x\<rparr>P,\<lparr>\<nu>x\<rparr>Q) \<in> ?X1" using `x \<sharp> \<Psi>`
            by(rule X1resClose)
          thus ?thesis
            by(rule UnI1)
        next
          assume "(\<Psi>,P,Q) \<in> ?X2"
          hence "(\<Psi>,\<lparr>\<nu>x\<rparr>P,\<lparr>\<nu>x\<rparr>Q) \<in> ?X2" using `x \<sharp> \<Psi>`
            by(rule X2resClose)
          thus ?thesis
            by(rule UnI2)
        qed
      }
      note XresClose = this
      {
        fix   \<Psi> P Q
          and x::name
        assume "(\<Psi>,P,Q) \<in> ?Y1"
          and    "x \<sharp> \<Psi>"
        from `(\<Psi>,P,Q) \<in> ?Y1` obtain P' Q' where "\<Psi> \<rhd> P \<sim> P'" and "(\<Psi>,P',Q') \<in> ?X1 \<or> \<Psi> \<rhd> P' \<sim> Q'" and "\<Psi> \<rhd> Q' \<sim> Q"
          by auto blast
        have "(\<Psi>,\<lparr>\<nu>x\<rparr>P,\<lparr>\<nu>x\<rparr>Q) \<in> ?Y1" using `(\<Psi>,P',Q') \<in> ?X1 \<or> \<Psi> \<rhd> P' \<sim> Q'`
        proof(erule_tac disjE)
          assume "(\<Psi>,P',Q') \<in> ?X1"
          hence "(\<Psi>,\<lparr>\<nu>x\<rparr>P',\<lparr>\<nu>x\<rparr>Q') \<in> ?X1" using `x \<sharp> \<Psi>`
            by(rule X1resClose)
          moreover have "\<Psi> \<rhd> \<lparr>\<nu>x\<rparr>P \<sim> \<lparr>\<nu>x\<rparr>P'" using `\<Psi> \<rhd> P \<sim> P'` `x \<sharp> \<Psi>`
            by(rule bisimResPres)
          moreover have "\<Psi> \<rhd> \<lparr>\<nu>x\<rparr>Q' \<sim> \<lparr>\<nu>x\<rparr>Q" using `\<Psi> \<rhd> Q' \<sim> Q` `x \<sharp> \<Psi>`
            by(rule bisimResPres)
          ultimately show ?thesis
            apply(rule_tac lol)
            by(rule refl | assumption | rule disjI1)+
        next
          assume "\<Psi> \<rhd> P' \<sim> Q'"
          hence "\<Psi> \<rhd> \<lparr>\<nu>x\<rparr>P' \<sim> \<lparr>\<nu>x\<rparr>Q'" using `x \<sharp> \<Psi>`
            by(rule bisimResPres)
          moreover have "\<Psi> \<rhd> \<lparr>\<nu>x\<rparr>P \<sim> \<lparr>\<nu>x\<rparr>P'" using `\<Psi> \<rhd> P \<sim> P'` `x \<sharp> \<Psi>`
            by(rule bisimResPres)
          moreover have "\<Psi> \<rhd> \<lparr>\<nu>x\<rparr>Q' \<sim> \<lparr>\<nu>x\<rparr>Q" using `\<Psi> \<rhd> Q' \<sim> Q` `x \<sharp> \<Psi>`
            by(rule bisimResPres)
          ultimately show ?thesis
            apply(rule_tac lol)
            by(rule refl | assumption | rule disjI2)+
        qed
      }
      note Y1resClose = this
      {
        fix   \<Psi> P Q
          and x::name
        assume "(\<Psi>,P,Q) \<in> ?Y2"
          and    "x \<sharp> \<Psi>"
        from `(\<Psi>,P,Q) \<in> ?Y2` obtain P' Q' where "\<Psi> \<rhd> P \<sim> P'" and "(\<Psi>,P',Q') \<in> ?X2 \<or> \<Psi> \<rhd> P' \<sim> Q'" and "\<Psi> \<rhd> Q' \<sim> Q"
          by auto blast
        have "(\<Psi>,\<lparr>\<nu>x\<rparr>P,\<lparr>\<nu>x\<rparr>Q) \<in> ?Y2" using `(\<Psi>,P',Q') \<in> ?X2 \<or> \<Psi> \<rhd> P' \<sim> Q'`
        proof(erule_tac disjE)
          assume "(\<Psi>,P',Q') \<in> ?X2"
          hence "(\<Psi>,\<lparr>\<nu>x\<rparr>P',\<lparr>\<nu>x\<rparr>Q') \<in> ?X2" using `x \<sharp> \<Psi>`
            by(rule X2resClose)
          moreover have "\<Psi> \<rhd> \<lparr>\<nu>x\<rparr>P \<sim> \<lparr>\<nu>x\<rparr>P'" using `\<Psi> \<rhd> P \<sim> P'` `x \<sharp> \<Psi>`
            by(rule bisimResPres)
          moreover have "\<Psi> \<rhd> \<lparr>\<nu>x\<rparr>Q' \<sim> \<lparr>\<nu>x\<rparr>Q" using `\<Psi> \<rhd> Q' \<sim> Q` `x \<sharp> \<Psi>`
            by(rule bisimResPres)
          ultimately show ?thesis
            apply(rule_tac lol)
            by(rule refl | assumption | rule disjI1)+
        next
          assume "\<Psi> \<rhd> P' \<sim> Q'"
          hence "\<Psi> \<rhd> \<lparr>\<nu>x\<rparr>P' \<sim> \<lparr>\<nu>x\<rparr>Q'" using `x \<sharp> \<Psi>`
            by(rule bisimResPres)
          moreover have "\<Psi> \<rhd> \<lparr>\<nu>x\<rparr>P \<sim> \<lparr>\<nu>x\<rparr>P'" using `\<Psi> \<rhd> P \<sim> P'` `x \<sharp> \<Psi>`
            by(rule bisimResPres)
          moreover have "\<Psi> \<rhd> \<lparr>\<nu>x\<rparr>Q' \<sim> \<lparr>\<nu>x\<rparr>Q" using `\<Psi> \<rhd> Q' \<sim> Q` `x \<sharp> \<Psi>`
            by(rule bisimResPres)
          ultimately show ?thesis
            apply(rule_tac lol)
            by(rule refl | assumption | rule disjI2)+
        qed
      }
      note Y2resClose = this
      {
        fix   \<Psi> P Q
          and x::name
        assume "(\<Psi>,P,Q) \<in> ?Y"
          and    "x \<sharp> \<Psi>"
        from `(\<Psi>,P,Q) \<in> ?Y` obtain P' Q' where "\<Psi> \<rhd> P \<sim> P'" and "(\<Psi>,P',Q') \<in> ?X \<or> \<Psi> \<rhd> P' \<sim> Q'" and "\<Psi> \<rhd> Q' \<sim> Q"
          by auto (blast+)
        have "(\<Psi>,\<lparr>\<nu>x\<rparr>P,\<lparr>\<nu>x\<rparr>Q) \<in> ?Y" using `(\<Psi>,P',Q') \<in> ?X \<or> \<Psi> \<rhd> P' \<sim> Q'`
        proof(erule_tac disjE)
          assume "(\<Psi>,P',Q') \<in> ?X"
          hence "(\<Psi>,\<lparr>\<nu>x\<rparr>P',\<lparr>\<nu>x\<rparr>Q') \<in> ?X" using `x \<sharp> \<Psi>`
            by(rule XresClose)
          moreover have "\<Psi> \<rhd> \<lparr>\<nu>x\<rparr>P \<sim> \<lparr>\<nu>x\<rparr>P'" using `\<Psi> \<rhd> P \<sim> P'` `x \<sharp> \<Psi>`
            by(rule bisimResPres)
          moreover have "\<Psi> \<rhd> \<lparr>\<nu>x\<rparr>Q' \<sim> \<lparr>\<nu>x\<rparr>Q" using `\<Psi> \<rhd> Q' \<sim> Q` `x \<sharp> \<Psi>`
            by(rule bisimResPres)
          ultimately show ?thesis
            apply(rule_tac lol)
            by(rule refl | assumption | rule disjI1)+
        next
          assume "\<Psi> \<rhd> P' \<sim> Q'"
          hence "\<Psi> \<rhd> \<lparr>\<nu>x\<rparr>P' \<sim> \<lparr>\<nu>x\<rparr>Q'" using `x \<sharp> \<Psi>`
            by(rule bisimResPres)
          moreover have "\<Psi> \<rhd> \<lparr>\<nu>x\<rparr>P \<sim> \<lparr>\<nu>x\<rparr>P'" using `\<Psi> \<rhd> P \<sim> P'` `x \<sharp> \<Psi>`
            by(rule bisimResPres)
          moreover have "\<Psi> \<rhd> \<lparr>\<nu>x\<rparr>Q' \<sim> \<lparr>\<nu>x\<rparr>Q" using `\<Psi> \<rhd> Q' \<sim> Q` `x \<sharp> \<Psi>`
            by(rule bisimResPres)
          ultimately show ?thesis
            apply(rule_tac lol)
            by(rule refl | assumption | rule disjI2)+
        qed
      }
      note YresClose = this
      {
        fix   \<Psi> P Q
          and xvec::"name list"
        assume iny: "(\<Psi>,P,Q) \<in> ?Y"
          and  f: "xvec \<sharp>* \<Psi>"
        from f have "(\<Psi>,\<lparr>\<nu>*xvec\<rparr>P,\<lparr>\<nu>*xvec\<rparr>Q) \<in> ?Y"
        proof(induct xvec)
          case Nil show ?case unfolding resChain.simps by(rule iny)
        next
          case(Cons x xvec)
          hence "xvec \<sharp>* \<Psi>" and "x \<sharp> \<Psi>"
            by simp+
          from `xvec \<sharp>* \<Psi>` have "(\<Psi>,\<lparr>\<nu>*xvec\<rparr>P,\<lparr>\<nu>*xvec\<rparr>Q) \<in> ?Y"
            by(rule Cons)
          hence "(\<Psi>,\<lparr>\<nu>x\<rparr>(\<lparr>\<nu>*xvec\<rparr>P),\<lparr>\<nu>x\<rparr>(\<lparr>\<nu>*xvec\<rparr>Q)) \<in> ?Y" using `x \<sharp> \<Psi>`
            by(rule_tac YresClose)
          thus ?case by simp
        qed
      }
      note YresChainClose' = this
      show ?case
      proof(cases "(\<Psi>,S,T) \<in> ?X1")
        assume "(\<Psi>,S,T) \<in> ?X1"
        then obtain a P Q M\<^isub>a \<Psi>\<^isub>c xvec where "S = \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))" and "T = \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> !P)" and "a \<sharp> P" and "a \<sharp> Q" and "a \<sharp> \<Psi>" and "\<not> a \<sharp> M\<^isub>a" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> R \<longrightarrow> (R = (P \<parallel> \<langle>M\<^isub>a\<rangle>)))" and "guarded P" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<longrightarrow> (supp M\<^isub>a) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P'))" and "((supp P)::name set) \<subseteq> supp M\<^isub>a" and "((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a" and "xvec \<sharp>* \<Psi>" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
          by auto
        have Equiv2: "\<And>\<Psi> R. \<lbrakk>a \<sharp> \<Psi>\<rbrakk> \<Longrightarrow> (\<lbrakk>\<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> R\<rbrakk> \<Longrightarrow> R = (P \<parallel> \<langle>M\<^isub>a\<rangle>))" using Equiv2
          by simp
        have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) \<leadsto>[?Y] \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> !P)" using `xvec \<sharp>* \<Psi>`
        proof(induct xvec)
          case(Cons x xvec)
          from `(x # xvec) \<sharp>* \<Psi>` have "x \<sharp> \<Psi>" and "xvec \<sharp>* \<Psi>" by simp+
          have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) \<leadsto>[?Y] \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> !P)" using `xvec \<sharp>* \<Psi>`
            by(rule Cons)
          show ?case
            unfolding resChain.simps
            by(rule resPres[where Rel="?Y" and Rel'="?Y"]) (fact | rule subset_refl)+
        next
          case Nil
          show ?case unfolding resChain.simps
          proof(rule simI[where C= "a", OF `eqvt(?Y)`])
            fix \<alpha> Q'
            assume "\<Psi> \<rhd> Q \<parallel> !P \<longmapsto> \<alpha> \<prec> Q'"
              and  "bn \<alpha> \<sharp>* \<Psi>"
              and  "bn \<alpha> \<sharp>* (Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))"
              and  "bn \<alpha> \<sharp>* (Q \<parallel> !P)"
              and  "bn \<alpha> \<sharp>* subject \<alpha>"
              and  "distinct(bn \<alpha>)"
              and  "bn \<alpha> \<sharp>* a"
            hence "bn \<alpha> \<sharp>* M\<^isub>a" and "bn \<alpha> \<sharp>* \<Psi>\<^isub>c" using `bn \<alpha> \<sharp>* a`
              by simp+
            have "bn \<alpha> \<sharp>* Q" and  "bn \<alpha> \<sharp>* P" and  "bn \<alpha> \<sharp>* !P" using `bn \<alpha> \<sharp>* (Q \<parallel> !P)` by simp+          
            show "\<exists> P'. \<Psi> \<rhd> Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> P' \<and> (\<Psi>, P', Q') \<in> (?Y)" using `\<Psi> \<rhd> Q \<parallel> !P \<longmapsto> \<alpha> \<prec> Q'` `bn \<alpha> \<sharp>* \<Psi>` `bn \<alpha> \<sharp>* Q` `bn \<alpha> \<sharp>* !P` `bn \<alpha> \<sharp>* subject \<alpha>`
            proof(induct rule: parCases[where C="(a,M\<^isub>a,\<Psi>\<^isub>c)"])
              case(cPar1 Q' A\<^isub>Q \<Psi>\<^isub>Q)
              obtain b::name where "b \<sharp> P" and "b \<sharp> Q" and "b \<sharp> \<Psi>" and "b \<sharp> M\<^isub>a" "b \<sharp> \<Psi>\<^isub>c" and "b \<sharp> \<alpha>" and "b \<sharp> Q'"
                apply(rule at_exists_fresh[where x="(P,Q,\<Psi>,M\<^isub>a,\<Psi>\<^isub>c,\<alpha>,Q')", OF at_name_inst])
                apply(finite_guess)
                by force
              have sl_def: "\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) = \<lparr>\<nu>b\<rparr>([(a,b)] \<bullet> (\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))"
                apply(subst alphaRes[where y=b])
                apply(simp add: `b \<sharp> M\<^isub>a` `b \<sharp> \<Psi>\<^isub>c`)
                by(rule refl)
              have "\<Psi>\<^isub>Q = \<one>" using `extractFrame (!P) = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>`
                by simp
              have "\<Psi> \<rhd> Q\<parallel>!P \<longmapsto> \<alpha> \<prec> Q'\<parallel>!P"
                apply(rule Par1[where A\<^isub>Q = A\<^isub>Q and \<Psi>\<^isub>Q = \<Psi>\<^isub>Q])              
                by fact+
              have "extractFrame (\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = (\<langle>[a], \<one> \<otimes> \<Psi>\<^isub>c\<rangle>)"
                by simp
              have unb: "\<langle>[b], [(a, b)] \<bullet> \<Psi>\<^isub>c\<rangle> = \<langle>[a], \<Psi>\<^isub>c\<rangle>"
                apply(subst pt3[OF pt_name_inst, OF at_ds5, OF at_name_inst])
                apply(subst frameChainAlpha[where xvec = "[b]" and p="[(b,a)]"])
                apply(simp add: eqvts pt_fresh_bij1[OF pt_name_inst, OF at_name_inst] `b \<sharp> \<Psi>\<^isub>c`)
                apply(simp add: at2[OF at_name_inst] at3[OF at_name_inst])
                by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
              have "[b] \<sharp>* \<alpha>" using `b \<sharp> \<alpha>`
                by simp
              have "bn \<alpha> \<sharp>* (\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))" using `bn \<alpha> \<sharp>* (Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))` `bn \<alpha> \<sharp>* a`
                by simp
              {
                fix F::"'b frame"
                obtain A\<^isub>F \<Psi>\<^isub>F where "F=\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>" and "distinct A\<^isub>F" and "A\<^isub>F \<sharp>* \<Psi>" and "A\<^isub>F \<sharp>* \<Psi>\<^isub>Q" and "A\<^isub>F \<sharp>* a" and "A\<^isub>F \<sharp>* b" and "A\<^isub>F \<sharp>* \<Psi>\<^isub>c"
                  apply(rule freshFrame[where C="(\<Psi>,\<Psi>\<^isub>Q,a,b,\<Psi>\<^isub>c)"])
                  by auto
                have lhs: "(\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>\<epsilon>, \<Psi> \<otimes> \<Psi>\<^isub>Q\<rangle>) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>\<rangle>)"
                  apply(subst mergeFrameSimps)
                  apply(rule freshCompChain)
                  apply fact+
                  apply(subst `\<Psi>\<^isub>Q = \<one>`)
                  apply(rule frameResChainPres)
                  apply(subst frameNilStatEq)
                  apply(rule AssertionStatEqTrans[OF AssertionStatEqSym[OF Associativity]])
                  by(rule Identity)
                obtain c::name where "c \<sharp> A\<^isub>F" and "c \<sharp> \<Psi>\<^isub>F" and "c \<noteq> b" and "c \<noteq> a" and "c \<sharp> \<Psi>"  and "c \<sharp> \<Psi>\<^isub>c"
                  apply(generate_fresh "name")
                  by force
                have af: "([(a,c)]\<bullet>A\<^isub>F) = A\<^isub>F" using `A\<^isub>F \<sharp>* a` `c \<sharp> A\<^isub>F`
                  by simp
                have "((\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[c], \<Psi> \<otimes> \<one> \<otimes> ([(a, c)] \<bullet> \<Psi>\<^isub>c)\<rangle>)) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>\<rangle>) \<otimes>\<^sub>F (\<langle>[c], ([(a,c)]\<bullet>\<Psi>\<^isub>c)\<rangle>)"
                  apply(subst (1 2) mergeFrames)
                  apply(simp add: `c \<sharp> A\<^isub>F`)
                  apply(subst af[symmetric])
                  apply(subst fresh_star_bij)
                  apply fact
                  apply(rule freshCompChain)
                  apply(simp add: `c \<sharp> \<Psi>\<^isub>F`)
                  apply(simp add: `c \<sharp> \<Psi>`)
                  apply(rule freshCompChain)
                  apply fact
                  apply(rule freshCompChain)
                  apply simp
                  apply(subst af[symmetric])
                  apply(subst fresh_star_bij)
                  apply fact
                  apply(simp add: `c \<sharp> \<Psi>\<^isub>F`)
                  apply(rule frameResChainPres)
                  apply(subst frameNilStatEq)
                  apply(rule AssertionStatEqSym)
                  apply(rule AssertionStatEqTrans[OF Associativity])
                  apply(rule compositionSym)
                  apply(rule compositionSym)
                  apply(rule AssertionStatEqSym)
                  apply(rule AssertionStatEqTrans[OF Commutativity])
                  by(rule Identity)
                hence "((\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[a], \<Psi> \<otimes> \<one> \<otimes> \<Psi>\<^isub>c\<rangle>)) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>\<rangle>) \<otimes>\<^sub>F (\<langle>[a], \<Psi>\<^isub>c\<rangle>)"
                  apply(subst (1 2) frameChainAlpha'[where A\<^isub>P="[a]" and p="[(a,c)]"])
                  apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `c \<sharp> \<Psi>\<^isub>c`)
                  apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                  apply(rule freshCompChain)
                  apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `a \<sharp> \<Psi>` `c \<sharp> \<Psi>`)
                  apply(rule freshCompChain)
                  apply simp
                  apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `c \<sharp> \<Psi>\<^isub>c`)
                  apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] eqvts `a \<sharp> \<Psi>` `c \<sharp> \<Psi>`)
                  done
                hence rhs: "((\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[b], \<Psi> \<otimes> \<one> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rangle>)) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>\<rangle>) \<otimes>\<^sub>F (\<langle>[b], ([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rangle>)"
                  apply(subst unb)
                  apply(subst frameChainAlpha'[where A\<^isub>P="[b]" and p="[(b,a)]"])
                  apply(rule freshCompChain)
                  apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `a \<sharp> \<Psi>` `b \<sharp> \<Psi>`)
                  apply(rule freshCompChain)
                  apply simp
                  apply(subst pt3[OF pt_name_inst, OF at_ds5, OF at_name_inst])
                  apply(subst pt_fresh_star_bij[OF pt_name_inst, OF at_name_inst])
                  apply(simp add: `b \<sharp> \<Psi>\<^isub>c`)
                  apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                  apply(subst (1 2) pt3[OF pt_name_inst, OF at_ds5, OF at_name_inst])
                  by(force simp add: eqvts `a \<sharp> \<Psi>` `b \<sharp> \<Psi>` at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                have "(F \<otimes>\<^sub>F (\<langle>\<epsilon>, \<Psi> \<otimes> \<Psi>\<^isub>Q\<rangle>)) \<simeq>\<^sub>F (F \<otimes>\<^sub>F (\<langle>[b], \<Psi> \<otimes> \<one> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rangle>))"
                  unfolding `F = \<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>`
                  apply(rule FrameStatEqTrans[OF lhs])
                  apply(rule FrameStatEqSym)
                  apply(rule FrameStatEqTrans[OF rhs])
                  apply(subst unb)
                  by(rule Equiv1)
              }
              have "\<Psi> \<rhd> Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> Q' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)"
                unfolding sl_def
                apply(rule Par1[where A\<^isub>Q = "[b]" and \<Psi>\<^isub>Q = "\<one> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)"])
                apply(rule frameStatEqTransition[where \<Psi> = "\<Psi> \<otimes> \<Psi>\<^isub>Q" and A = "[b]"])
                apply fact
                apply fact
                apply fact
                apply fact
                apply(simp del: actionFresh actionFreshChain add: `b \<sharp> \<alpha>`)
                apply(simp add: `b \<sharp> Q`)
                apply simp
                apply(subst sl_def[symmetric])
                apply fact
                apply(simp add: `b \<sharp> \<Psi>`)
                apply(simp add: `b \<sharp> Q`)
                apply(simp del: actionFresh actionFreshChain add: `b \<sharp> \<alpha>`)
                done
              moreover have "(\<Psi>, Q' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>), Q' \<parallel> !P) \<in> ?X1"
              proof(subst sl_def, simp only: eqvts)
                note `b \<sharp> \<Psi>` `b \<sharp> P` `b \<sharp> Q'` `guarded P`
                moreover have "\<not> b \<sharp> ([(a,b)]\<bullet>M\<^isub>a)" using `\<not> a \<sharp> (M\<^isub>a)`
                  apply(subst perm_bool[where pi="[(a,b)]",symmetric])
                  by(force simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                moreover have "\<forall> \<Psi>. \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>"
                  using EntCl
                  apply(subst perm_bool[where pi="[(a,b)]", symmetric])
                  by(simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
                moreover have "\<forall>F. (F \<otimes>\<^sub>F (\<langle>[b],[(a,b)]\<bullet>\<Psi>\<^isub>c\<rangle>) \<simeq>\<^sub>F F)"
                  using Equiv1
                  apply(subst perm_bool[where pi="[(a,b)]", symmetric])
                  by(force simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                moreover have "(\<forall>\<Psi> R. b \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a) \<Leftarrow> R \<longrightarrow> (R = (P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)))" using Equiv2
                  apply(subst perm_bool[where pi="[(a,b)]", symmetric])
                  by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst] `a \<sharp> P` `b \<sharp> P`)
                moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<longrightarrow> (supp ([(a,b)]\<bullet>M\<^isub>a)) \<subseteq> ((supp \<Psi>)::name set))" using EntSupp
                  apply(subst perm_bool[where pi="[(a,b)]", symmetric])
                  by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                moreover have "\<forall> P P'. ((([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P = ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P') = (P = P'))" using Inj
                  apply(subst perm_bool[where pi="[(a,b)]", symmetric])
                  by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                moreover have "((supp P)::name set) \<subseteq> supp ([(a,b)]\<bullet>M\<^isub>a)" using `((supp P)::name set) \<subseteq> supp M\<^isub>a`
                  apply(subst perm_bool[where pi="[(a,b)]", symmetric])
                  by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst] `a \<sharp> P` `b \<sharp> P`)
                moreover have "((supp ([(a,b)]\<bullet>\<Psi>\<^isub>c))::name set) = supp ([(a,b)]\<bullet>M\<^isub>a)" using `((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a`
                  apply(subst perm_bool[where pi="[(a,b)]", symmetric])
                  by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                moreover have "([]::name list) \<sharp>* \<Psi>"
                  by simp
                moreover from EntEnt have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                  by(subst perm_bool[where pi="[(a,b)]",symmetric]) (simp add: eqvts swap_simps `a \<sharp> P` `b \<sharp> P`)
                ultimately have "(\<Psi>, \<lparr>\<nu>*[]\<rparr>(Q' \<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)), \<lparr>\<nu>*[]\<rparr>(Q' \<parallel> !P)) \<in> ?X1"
                  by blast
                then show "(\<Psi>, Q' \<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>), Q' \<parallel> !P) \<in> ?X1"
                  unfolding resChain.simps
                  by -
              qed
              moreover have "\<Psi> \<rhd> (Q' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) \<sim> (Q' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))"
                by(rule bisimReflexive)
              moreover have "\<Psi> \<rhd> (Q' \<parallel> !P) \<sim> (Q' \<parallel> !P)"
                by(rule bisimReflexive)
              ultimately show ?case
                apply -
                apply(rule_tac x = "Q' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)" in exI)
                apply(rule conjI, assumption)
                apply(rule lol)
                apply(rule refl)
                apply assumption
                apply(rule disjI1)
                apply blast
                apply assumption
                done
            next
              case(cPar2 P' A\<^isub>Q \<Psi>\<^isub>Q)
              have "bn \<alpha> \<sharp>* \<Psi>\<^isub>Q"
              proof -
                have "bn \<alpha> \<sharp>* extractFrame Q" using `bn \<alpha> \<sharp>* Q`
                  by(rule extractFrameFreshChain)
                hence "bn \<alpha> \<sharp>* (\<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>)" unfolding `extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>`
                  by assumption
                thus ?thesis using `A\<^isub>Q \<sharp>* \<alpha>`
                  by simp
              qed
              with `bn \<alpha> \<sharp>* \<Psi>` have "bn \<alpha> \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)"
                by(rule freshCompChain)
              have "bn \<alpha> \<sharp>* P" using `bn \<alpha> \<sharp>* (Q \<parallel> !P)`
                by simp
              have "bn \<alpha> \<sharp>* M\<^isub>a" using `bn \<alpha> \<sharp>* a` `bn \<alpha> \<sharp>* (Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))`
                by simp
              have "a \<sharp> \<Psi>\<^isub>Q" using `a \<sharp> Q` `extractFrame Q = \<langle>A\<^isub>Q,\<Psi>\<^isub>Q\<rangle>` `A\<^isub>Q \<sharp>* (a, M\<^isub>a,\<Psi>\<^isub>c)`
              proof -
                have "a \<sharp> \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>" using `a \<sharp> Q` 
                  apply(subst `extractFrame Q = \<langle>A\<^isub>Q,\<Psi>\<^isub>Q\<rangle>`[symmetric])
                  by(rule extractFrameFresh)
                hence "a \<sharp> \<langle>\<epsilon>, \<Psi>\<^isub>Q\<rangle>" using `A\<^isub>Q \<sharp>* (a, M\<^isub>a,\<Psi>\<^isub>c)`
                  apply(subst (asm) frameResChainFresh)
                  by(simp add: fresh_def frame.supp name_list_supp)
                thus ?thesis
                  by simp
              qed
              from `A\<^isub>Q \<sharp>* !P` have "A\<^isub>Q \<sharp>* P" by simp
              from `A\<^isub>Q \<sharp>* (a,M\<^isub>a,\<Psi>\<^isub>c)` have "A\<^isub>Q \<sharp>* a" and "A\<^isub>Q \<sharp>* M\<^isub>a" and "A\<^isub>Q \<sharp>* \<Psi>\<^isub>c"
                by simp+
              have "A\<^isub>Q \<sharp>* bn \<alpha>" using `A\<^isub>Q \<sharp>* \<alpha>`
                by simp
              have "(\<exists> P''. ((\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> P'') \<and> (\<Psi> \<otimes> \<Psi>\<^isub>Q, P'', P',A\<^isub>Q) \<in> ?YA1))" using `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P'` `bn \<alpha> \<sharp>* subject \<alpha>` (*`extractFrame Q = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>` `A\<^isub>Q \<sharp>* \<alpha>`*) `bn \<alpha> \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)` `bn \<alpha> \<sharp>* P` `bn \<alpha> \<sharp>* M\<^isub>a` `(supp \<Psi>\<^isub>c) = supp M\<^isub>a` `((supp P)::name set) \<subseteq> supp M\<^isub>a` `a \<sharp> P` EntSupp EntCl Equiv1 `a \<sharp> \<Psi>` `a \<sharp> \<Psi>\<^isub>Q` Equiv2 Inj `\<not> a \<sharp> M\<^isub>a` `A\<^isub>Q \<sharp>* bn \<alpha>` `A\<^isub>Q \<sharp>* P` `A\<^isub>Q \<sharp>* a` `A\<^isub>Q \<sharp>* M\<^isub>a` `A\<^isub>Q \<sharp>* \<Psi>\<^isub>c` `A\<^isub>Q \<sharp>* P'` EntEnt
              proof(nominal_induct avoiding: a M\<^isub>a \<Psi>\<^isub>c A\<^isub>Q rule: bangInduct')
                case(cAlpha \<alpha> P' p)
                from `A\<^isub>Q \<sharp>* bn(p \<bullet> \<alpha>)` have "bn(p\<bullet>\<alpha>) \<sharp>* A\<^isub>Q"
                  by simp
                from `bn(p\<bullet>\<alpha>) \<sharp>* A\<^isub>Q` `bn(\<alpha>) \<sharp>* A\<^isub>Q` `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` have "(p\<bullet>A\<^isub>Q) = A\<^isub>Q"
                  by simp
                from `A\<^isub>Q \<sharp>* bn(p \<bullet> \<alpha>)` have "A\<^isub>Q \<sharp>* bn \<alpha>"
                  apply(subst (asm) `(p\<bullet>A\<^isub>Q) = A\<^isub>Q`[symmetric])
                  apply(subst (asm) bnEqvt[symmetric])
                  by(subst (asm) fresh_star_bij)
                from `A\<^isub>Q \<sharp>* (p \<bullet> P')` have "A\<^isub>Q \<sharp>* P'"
                  by(subst (asm) `(p\<bullet>A\<^isub>Q) = A\<^isub>Q`[symmetric]) (subst (asm) fresh_star_bij)
                have "\<exists>P''. (\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> P'' \<and> (\<Psi> \<otimes> \<Psi>\<^isub>Q, P'', P',A\<^isub>Q) \<in> ?YA1)"
                  by(rule cAlpha) fact+
                then obtain P'' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> P'' \<and> (\<Psi> \<otimes> \<Psi>\<^isub>Q, P'', P',A\<^isub>Q) \<in> ?YA1"
                  by(drule HOL.exE)
                hence "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> P''" and "(\<Psi> \<otimes> \<Psi>\<^isub>Q, P'', P',A\<^isub>Q) \<in> ?YA1"
                  by simp+
                have ppsi: "(p\<bullet>(\<Psi> \<otimes> \<Psi>\<^isub>Q)) = (\<Psi> \<otimes> \<Psi>\<^isub>Q)" using `bn(p\<bullet>\<alpha>) \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)` `bn \<alpha> \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)` `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))`
                  by simp
                have lhs: "(\<Psi> \<otimes> \<Psi>\<^isub>Q, p\<bullet>P'', p\<bullet>P',A\<^isub>Q) \<in> ?YA1"
                proof -
                  have pout: "(p \<bullet> \<Psi> \<otimes> \<Psi>\<^isub>Q, p \<bullet> P'', p \<bullet> P',p\<bullet>A\<^isub>Q) = p\<bullet>(\<Psi> \<otimes> \<Psi>\<^isub>Q, P'', P',A\<^isub>Q)"
                    by(simp only: eqvts)
                  show ?thesis
                    apply(subst ppsi[symmetric])
                    apply(subst `(p\<bullet>A\<^isub>Q) = A\<^isub>Q`[symmetric])
                    apply(subst pout)
                    apply(rule eqvtI)
                    by fact+
                qed
                have "distinct(bn \<alpha>)" using `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> P''`
                  by(rule boundOutputDistinct)
                from `bn (p \<bullet> \<alpha>) \<sharp>* \<alpha>` have "bn (p \<bullet> \<alpha>) \<sharp>* object \<alpha>" and "bn (p \<bullet> \<alpha>) \<sharp>* subject \<alpha>" by simp+
                have "bn (p \<bullet> \<alpha>) \<sharp>* \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)"
                proof -
                  have "bn(p\<bullet>\<alpha>) \<sharp>* (\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)" using `bn(p\<bullet>\<alpha>) \<sharp>* M\<^isub>a` `(supp \<Psi>\<^isub>c) = supp M\<^isub>a`
                    unfolding fresh_def fresh_star_def
                    by(auto simp add: psi.supp)
                  thus ?thesis
                    apply(subst psiFreshVec)
                    unfolding fresh_star_def fresh_def
                    apply(subst abs_fun_supp[OF pt_name_inst, OF at_name_inst])
                    apply(finite_guess)
                    by blast
                qed
                have "bn (p \<bullet> \<alpha>) \<sharp>* P''"
                  apply(rule freeFreshChainDerivative[OF `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> P''`])
                  by fact+
                have rhs: "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> (p \<bullet> \<alpha>) \<prec> (p\<bullet>P'')"
                  apply(subst residualAlpha[symmetric])
                  by fact+
                from lhs rhs show ?case
                  apply -
                  apply(rule_tac x="p\<bullet>P''" in exI)
                  apply(rule conjI)
                  by assumption+
              next
                case(cPar1 \<alpha> P')
                obtain b::name where "b \<sharp> \<Psi>" and "b \<sharp> \<Psi>\<^isub>Q" and "b \<sharp> P" and "b \<sharp> \<alpha>" and "b \<noteq> a" and "b \<sharp> bn \<alpha>" and "b \<sharp> M\<^isub>a" and "b \<sharp> \<Psi>\<^isub>c" and "b \<sharp> A\<^isub>Q" and "b \<sharp> P'"
                  apply(generate_fresh "name")
                  by force
                have unb: "\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) = \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                  apply(subst alphaRes[where x=a and y=b])
                  apply(simp add: `b \<sharp> M\<^isub>a` `b \<sharp> \<Psi>\<^isub>c`)
                  unfolding eqvts
                  by(rule refl)
                have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>" using `\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>`
                  apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                  by(simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
                have "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one> \<rhd> P \<longmapsto> \<alpha> \<prec> P'"
                proof -
                  {
                    fix F::"'b frame"
                    obtain A\<^isub>F \<Psi>\<^isub>F where "F=\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>" and "distinct A\<^isub>F" and "A\<^isub>F \<sharp>* \<Psi>" and "A\<^isub>F \<sharp>* \<Psi>\<^isub>Q" and "A\<^isub>F \<sharp>* a" and "A\<^isub>F \<sharp>* b" and "A\<^isub>F \<sharp>* \<Psi>\<^isub>c"
                      apply(rule freshFrame[where C="(\<Psi>,\<Psi>\<^isub>Q,a,b,\<Psi>\<^isub>c)"])
                      by auto
                    have lhs: "(\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>\<epsilon>, \<Psi> \<otimes> \<Psi>\<^isub>Q\<rangle>) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi> \<otimes> \<Psi>\<^isub>Q\<rangle>)"
                      apply(subst mergeFrameSimps)
                      apply(rule freshCompChain)
                      apply fact+
                      apply(rule FrameStatEqRefl)
                      done
                    obtain c::name where "c \<sharp> A\<^isub>F" and "c \<sharp> \<Psi>\<^isub>F" and "c \<noteq> b" and "c \<noteq> a" and "c \<sharp> \<Psi>"  and "c \<sharp> \<Psi>\<^isub>c" and "c \<sharp> \<Psi>\<^isub>Q"
                      apply(generate_fresh "name")
                      by force
                    have af: "([(a,c)]\<bullet>A\<^isub>F) = A\<^isub>F" using `A\<^isub>F \<sharp>* a` `c \<sharp> A\<^isub>F`
                      by simp
                    have "((\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[c], ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, c)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one>\<rangle>)) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi> \<otimes> \<Psi>\<^isub>Q\<rangle>) \<otimes>\<^sub>F (\<langle>[c], ([(a,c)]\<bullet>\<Psi>\<^isub>c)\<rangle>)"
                      apply(subst (1 2) mergeFrames)
                      apply(simp add: `c \<sharp> A\<^isub>F`)
                      apply(subst af[symmetric])
                      apply(subst fresh_star_bij)
                      apply fact
                      apply(rule freshCompChain)
                      apply(simp add: `c \<sharp> \<Psi>\<^isub>F`)
                      apply(rule freshCompChain)
                      apply(simp add: `c \<sharp> \<Psi>`)
                      apply(simp add: `c \<sharp> \<Psi>\<^isub>Q`)
                      apply(rule freshCompChain)
                      apply(rule freshCompChain)
                      apply(rule freshCompChain)
                      apply fact
                      apply fact
                      apply(subst af[symmetric])
                      apply(subst fresh_star_bij)
                      apply fact
                      apply simp
                      apply(simp add: `c \<sharp> \<Psi>\<^isub>F`)
                      apply(rule frameResChainPres)
                      apply(subst frameNilStatEq)
                      apply(rule AssertionStatEqSym)
                      apply(rule AssertionStatEqTrans[OF Associativity])
                      apply(rule compositionSym)
                      by(rule AssertionStatEqSym[OF Identity])
                    hence "((\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[a], ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c) \<otimes> \<one>\<rangle>)) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi> \<otimes> \<Psi>\<^isub>Q\<rangle>) \<otimes>\<^sub>F (\<langle>[a], \<Psi>\<^isub>c\<rangle>)"
                      apply(subst (1 2) frameChainAlpha'[where A\<^isub>P="[a]" and p="[(a,c)]"])
                      apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `c \<sharp> \<Psi>\<^isub>c`)
                      apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                      apply(rule freshCompChain)
                      apply(rule freshCompChain)
                      apply(rule freshCompChain)
                      apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `a \<sharp> \<Psi>` `c \<sharp> \<Psi>`)
                      apply(simp add: `a \<sharp> \<Psi>\<^isub>Q`)
                      apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `c \<sharp> \<Psi>\<^isub>c` `c \<sharp> \<Psi>\<^isub>Q`)
                      apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] eqvts `c \<sharp> \<Psi>\<^isub>c`)
                      apply simp
                      apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] eqvts `a \<sharp> \<Psi>` `c \<sharp> \<Psi>`)
                      by(simp add: `a \<sharp> \<Psi>\<^isub>Q` `c \<sharp> \<Psi>\<^isub>Q`)
                    hence rhs: "((\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[b], ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one>\<rangle>)) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi> \<otimes> \<Psi>\<^isub>Q\<rangle>) \<otimes>\<^sub>F (\<langle>[b], ([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rangle>)"
                      apply(subst frameChainAlpha'[where A\<^isub>P="[b]" and p="[(b,a)]"])
                      apply(rule freshCompChain)
                      apply(rule freshCompChain)
                      apply(rule freshCompChain)
                      apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `a \<sharp> \<Psi>` `b \<sharp> \<Psi>`)
                      apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `a \<sharp> \<Psi>\<^isub>Q` `b \<sharp> \<Psi>\<^isub>Q`)
                      apply(subst pt3[OF pt_name_inst, OF at_ds5, OF at_name_inst])
                      apply(subst pt_fresh_star_bij[OF pt_name_inst, OF at_name_inst])
                      apply(simp add: `b \<sharp> \<Psi>\<^isub>c`)
                      apply simp
                      apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                      apply(subst frameChainAlpha'[where A\<^isub>P="[b]" and p="[(b,a)]"])
                      apply(subst pt3[OF pt_name_inst, OF at_ds5, OF at_name_inst])
                      apply(subst pt_fresh_star_bij[OF pt_name_inst, OF at_name_inst])
                      apply(simp add: `b \<sharp> \<Psi>\<^isub>c`)
                      apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                      apply(simp add: eqvts at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `a \<sharp> \<Psi>\<^isub>Q` `a \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q` `b \<sharp> \<Psi>` pt3[OF pt_name_inst, OF at_ds5[where a=b and b=a], OF at_name_inst])
                      done
                    have unb: "\<langle>[a],\<Psi>\<^isub>c\<rangle> = (\<langle>[b],([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rangle>)"
                      apply(subst frameChainAlpha'[where A\<^isub>P="[a]" and p="[(a,b)]"])
                      apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `b \<sharp> \<Psi>\<^isub>c`)
                      apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                      by(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                    have "(F \<otimes>\<^sub>F (\<langle>\<epsilon>, \<Psi> \<otimes> \<Psi>\<^isub>Q\<rangle>)) \<simeq>\<^sub>F (F \<otimes>\<^sub>F (\<langle>[b], ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one>\<rangle>))"
                      unfolding `F = \<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>`
                      apply(rule FrameStatEqTrans[OF lhs])
                      apply(rule FrameStatEqSym)
                      apply(rule FrameStatEqTrans[OF rhs])
                      apply(subst unb[symmetric])
                      by fact
                  }
                  show ?thesis
                    apply(rule frameStatEqTransition[where A="[b]" and \<Psi>="(\<Psi> \<otimes> \<Psi>\<^isub>Q)"])
                    by(fact | simp del: actionFresh actionFreshChain add: `b \<sharp> \<alpha>` `b \<sharp> P`)+
                qed
                have "bn \<alpha> \<sharp>* \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>" and "bn \<alpha> \<sharp>* \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>"
                proof -
                  have "[(a,b)] \<bullet> (bn \<alpha>) = bn \<alpha>" using `bn \<alpha> \<sharp>* a` `b \<sharp> bn \<alpha>`
                    by simp
                  show "bn \<alpha> \<sharp>* \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>"
                    apply(subst `[(a,b)] \<bullet> (bn \<alpha>) = bn \<alpha>`[symmetric])
                    apply(subst perm_bool[symmetric, where pi="[(a,b)]"])
                    by(simp add: eqvts `bn \<alpha> \<sharp>* M\<^isub>a`)
                  show "bn \<alpha> \<sharp>* \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>"
                    apply(subst `[(a,b)] \<bullet> (bn \<alpha>) = bn \<alpha>`[symmetric])
                    apply(subst perm_bool[symmetric, where pi="[(a,b)]"])
                    by(simp add: eqvts `bn \<alpha> \<sharp>* \<Psi>\<^isub>c`)
                qed
                have "supp (P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>) \<subseteq> (supp ([(a, b)] \<bullet> M\<^isub>a)::name set)" using `((supp P)::name set) \<subseteq> supp M\<^isub>a`
                  apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                  by(simp add: eqvts psi.supp `a \<sharp> P` `b \<sharp> P`)
                have "guarded (P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>)" using `guarded P`
                  by simp
                have "b \<sharp> \<Psi> \<otimes> \<Psi>\<^isub>Q"
                  by(rule freshComp) fact+
                have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> \<lparr>\<nu>b\<rparr>((P' \<parallel> (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)) \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                  unfolding unb
                  apply(rule Scope)
                  apply(rule Par1[where \<Psi>\<^isub>Q="([(a, b)] \<bullet> \<Psi>\<^isub>c)" and A\<^isub>Q="[]"])
                  apply(rule Inv[where P="P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>"])
                  apply fact
                  apply(rule Par1[where \<Psi>\<^isub>Q="\<one>" and A\<^isub>Q="[]"])
                  by(fact | simp)+
                moreover have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>((P' \<parallel> (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)) \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> P' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)"
                proof -
                  have lhs: "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>((P' \<parallel> (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)) \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>b\<rparr>(P' \<parallel> ((\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>) \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                    apply(rule bisimResPres)
                    apply(rule bisimParAssoc)
                    by fact
                  show ?thesis
                    unfolding unb
                    apply(rule bisimTransitive[OF lhs])
                    apply(rule bisimScopeExt)
                    by fact
                qed
                moreover have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> P' \<parallel> !P \<sim> P' \<parallel> !P"
                  by(rule bisimReflexive)
                moreover have "(\<Psi> \<otimes> \<Psi>\<^isub>Q, P' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>), P' \<parallel> !P) \<in> ?X1 \<or> \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> P' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<sim> P' \<parallel> !P"
                proof -
                  have "b \<sharp> \<Psi> \<otimes> \<Psi>\<^isub>Q"
                    by(rule freshComp) fact+
                  moreover note `b \<sharp> P` `b \<sharp> P'`
                  moreover have "\<not> b \<sharp> [(a, b)] \<bullet> M\<^isub>a" using `\<not> a \<sharp> M\<^isub>a`
                    apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                    by(force simp add: eqvts at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                  moreover note `guarded P`
                  moreover have "supp P \<subseteq> (supp ([(a, b)] \<bullet> M\<^isub>a)::name set)" using `supp P \<subseteq> supp M\<^isub>a`
                    apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                    by(simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
                  moreover have "supp ([(a, b)] \<bullet> \<Psi>\<^isub>c) = (supp ([(a, b)] \<bullet> M\<^isub>a)::name set)" using `supp (\<Psi>\<^isub>c) = supp (M\<^isub>a)`
                    apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                    by(simp add: eqvts)
                  moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<longrightarrow> (supp ([(a,b)]\<bullet>M\<^isub>a)) \<subseteq> ((supp \<Psi>)::name set))" using `\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<longrightarrow> (supp M\<^isub>a) \<subseteq> ((supp \<Psi>)::name set))`
                    apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                    by(simp add: eqvts)
                  moreover have "\<forall>\<Psi>. \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>" using `\<And>\<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>`
                    apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                    by(simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
                  moreover have "\<forall> F. F \<otimes>\<^sub>F (\<langle>[b],([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rangle>) \<simeq>\<^sub>F F" using `\<And> F. F \<otimes>\<^sub>F (\<langle>[a],\<Psi>\<^isub>c\<rangle>) \<simeq>\<^sub>F F`
                    apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                    by(force simp add: eqvts at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                  moreover have "\<forall> P P'. ((([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P = ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P') = (P = P'))" using `\<forall>P P'. (M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P')`
                    apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                    by(simp add: eqvts)
                  moreover have "\<forall> \<Psi> R. b \<sharp> \<Psi> \<longrightarrow> (\<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a) \<Leftarrow> R \<longrightarrow> R = (P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>))" using `\<And> \<Psi> R. \<lbrakk>a \<sharp> \<Psi>; \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>R\<rbrakk> \<Longrightarrow> R = P \<parallel> \<langle>M\<^isub>a\<rangle>`
                    apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                    by(simp add: eqvts `a \<sharp> P` `b \<sharp> P` at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                  moreover have "([]::name list) \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)"
                    by simp
                  moreover from `\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> \<phi> \<Longrightarrow> \<phi> = M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> \<phi>` have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                    by(subst perm_bool[where pi="[(a,b)]",symmetric]) (simp add: eqvts swap_simps `a \<sharp> P` `b \<sharp> P`)
                  ultimately have "(\<Psi> \<otimes> \<Psi>\<^isub>Q, \<lparr>\<nu>*[]\<rparr>P' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>), \<lparr>\<nu>*[]\<rparr>P' \<parallel> !P) \<in> ?X1"
                    apply(subst unb)
                    by blast
                  hence "(\<Psi> \<otimes> \<Psi>\<^isub>Q, P' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>), P' \<parallel> !P) \<in> ?X1"
                    unfolding unb resChain.simps
                    by -
                  thus ?thesis
                    by blast
                qed
                moreover have "A\<^isub>Q \<sharp>* (\<lparr>\<nu>b\<rparr>((P' \<parallel> (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)) \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>), P' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>), P' \<parallel> !P, P' \<parallel> !P)"
                proof -
                  have "a \<sharp> A\<^isub>Q" using `A\<^isub>Q \<sharp>* a`
                    by simp
                  thus ?thesis using `b \<sharp> A\<^isub>Q` `A\<^isub>Q \<sharp>* (P' \<parallel> !P)` `A\<^isub>Q \<sharp>* M\<^isub>a` `A\<^isub>Q \<sharp>* \<Psi>\<^isub>c`
                    by(simp add: fresh_abs_fun_iff[OF pt_name_inst, OF at_name_inst, OF fin_supp])
                qed
                ultimately have "(\<Psi> \<otimes> \<Psi>\<^isub>Q, \<lparr>\<nu>b\<rparr>((P' \<parallel> (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)) \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>), P' \<parallel> !P,A\<^isub>Q) \<in> ?YA1"
                  apply -
                  apply(rule lol')
                  apply(rule refl)
                  by assumption+
                with `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> \<lparr>\<nu>b\<rparr>((P' \<parallel> (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)) \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)` show ?case
                  apply(rule_tac exI[where x="\<lparr>\<nu>b\<rparr>(P' \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)"])
                  apply(rule conjI)
                  by assumption+
              next
                case(cPar2 \<alpha> P')
                from `A\<^isub>Q \<sharp>* (P \<parallel> P')` have "A\<^isub>Q \<sharp>* P'" by simp
                have "\<exists> P''. \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> P'' \<and> (\<Psi> \<otimes> \<Psi>\<^isub>Q, P'', P',A\<^isub>Q) \<in> ?YA1"
                  apply(rule cPar2)
                  apply(rule cPar2)
                  by fact+
                then obtain P'' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> P'' \<and> (\<Psi> \<otimes> \<Psi>\<^isub>Q, P'', P',A\<^isub>Q) \<in> ?YA1"
                  by(rule exE)
                hence "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> P''" and "(\<Psi> \<otimes> \<Psi>\<^isub>Q, P'', P', A\<^isub>Q) \<in> ?YA1"
                  by simp+
                from `(\<Psi> \<otimes> \<Psi>\<^isub>Q, P'', P', A\<^isub>Q) \<in> ?YA1`
                obtain R R' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> P'' \<sim> R" and "(\<Psi> \<otimes> \<Psi>\<^isub>Q,R,R') \<in> ?X1 \<or> \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> R \<sim> R'" and "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> R' \<sim> P'" and "A\<^isub>Q \<sharp>* R" and "A\<^isub>Q \<sharp>* R'" and "A\<^isub>Q \<sharp>* P''" and "A\<^isub>Q \<sharp>* P'"
                  by auto blast+
                have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> R' \<sim> P \<parallel> P'"
                proof -
                  obtain R'' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> P' \<sim> R'' \<parallel> !P"
                    apply(rule bangDerivative[where \<Psi>="\<Psi>\<otimes>\<Psi>\<^isub>Q" and P=P and Q=P and \<alpha>=\<alpha> and P'=P'])
                    apply(rule cPar2)
                    apply(rule bisimReflexive)
                    apply(rule cPar2)
                    apply(rule cPar2)
                    apply(rule cPar2)
                    apply(rule cPar2)
                    apply(rule `guarded P`)
                    by simp
                  have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> R'' \<parallel> !P \<sim> R'' \<parallel> (P \<parallel> !P)"
                    apply(rule bisimTransitive[OF bisimParComm])
                    apply(rule bisimSymmetric)
                    apply(rule bisimTransitive[OF bisimParComm])
                    apply(rule bisimParPres)
                    apply(rule bisimSymmetric)
                    apply(rule bangExt[OF `guarded P`])
                    done
                  show ?thesis
                    apply(rule bisimTransitive[OF `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> R' \<sim> P'`])
                    apply(rule bisimTransitive[OF `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> P' \<sim> R'' \<parallel> !P`])
                    apply(rule bisimSymmetric)
                    apply(rule bisimTransitive[OF bisimParComm])
                    apply(rule bisimTransitive[OF bisimParPres[OF `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> P' \<sim> R'' \<parallel> !P`]])
                    apply(rule bisimTransitive[OF bisimParAssoc])
                    apply(rule bisimTransitive[OF bisimParComm])
                    apply(rule bisimSymmetric)
                    apply(rule bisimTransitive[OF bisimParComm])
                    apply(rule bisimParPres)
                    apply(rule bisimSymmetric)
                    apply(rule bisimTransitive[OF bisimParComm])
                    apply(rule bisimSymmetric[OF bangExt, OF `guarded P`])
                    done
                qed      
                show ?case
                  apply(rule exI[where x=P''])
                  apply(rule conjI)
                  apply fact
                  apply(rule lol')
                  apply(rule refl)
                  apply(rule `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> P'' \<sim> R`)
                  apply(rule `(\<Psi> \<otimes> \<Psi>\<^isub>Q,R,R') \<in> ?X1 \<or> \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> R \<sim> R'`)
                  apply(rule `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> R' \<sim> P \<parallel> P'`)
                  apply(simp add: `A\<^isub>Q \<sharp>* P''` `A\<^isub>Q \<sharp>* P'` `A\<^isub>Q \<sharp>* R` `A\<^isub>Q \<sharp>* R'` `A\<^isub>Q \<sharp>* P`)
                  done
              next
                case(cBang \<alpha> P')
                show ?case
                  apply(rule cBang)
                  apply(rule cBang)
                  by fact+
              next
                case(cComm1 M N P' K xvec P'')
                from `xvec \<sharp>* A\<^isub>Q` have "A\<^isub>Q \<sharp>* xvec" by simp
                with `A\<^isub>Q \<sharp>* (\<lparr>\<nu>*xvec\<rparr>P'\<parallel>P'')` have "A\<^isub>Q \<sharp>* P'" and "A\<^isub>Q \<sharp>* P''"
                  by simp+
                have "\<exists> Q''. \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q'' \<and> (\<Psi> \<otimes> \<Psi>\<^isub>Q, Q'', P'',A\<^isub>Q) \<in> ?YA1"
                  by(rule cComm1) (fact | simp add: `xvec \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)` `xvec \<sharp>* P` `xvec \<sharp>* M\<^isub>a`)+
                then obtain Q'' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q'' \<and> (\<Psi> \<otimes> \<Psi>\<^isub>Q, Q'', P'',A\<^isub>Q) \<in> ?YA1"
                  by(rule exE)
                hence toq'': "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q''" and "(\<Psi> \<otimes> \<Psi>\<^isub>Q, Q'', P'',A\<^isub>Q) \<in> ?YA1"
                  by simp+
                from `(\<Psi> \<otimes> \<Psi>\<^isub>Q, Q'', P'',A\<^isub>Q) \<in> ?YA1` obtain Q'''' P''' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> Q'' \<sim> Q''''" and "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> P''' \<sim> P''" and "(\<Psi>\<otimes>\<Psi>\<^isub>Q,Q'''',P''') \<in> ?X1 \<or> \<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> Q'''' \<sim> P'''" and "A\<^isub>Q \<sharp>* Q''" and "A\<^isub>Q \<sharp>* Q''''" and "A\<^isub>Q \<sharp>* P''" and "A\<^isub>Q \<sharp>* P'''"
                  by auto blast
                obtain b::name where "b \<sharp> \<Psi>\<^isub>Q" and "b \<sharp> \<Psi>" and "b \<sharp> \<Psi> \<otimes> \<Psi>\<^isub>Q" and "b \<sharp> \<Psi>\<^isub>c" and "b \<sharp> Q''" and "b \<sharp> P'" and "b \<sharp> P''" and "b \<sharp> M" and "b \<sharp> N" and "b \<sharp> K" and "b \<sharp> xvec" and "b \<sharp> M\<^isub>a" and "b\<noteq>a" and "b \<sharp> P" and "b \<sharp> A\<^isub>Q"
                  by(generate_fresh "name") force
                have vadef: "\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) = \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                  apply(subst alphaRes[where x=a and y=b])
                  apply(simp add: `b \<sharp> M\<^isub>a` `b \<sharp> \<Psi>\<^isub>c`)
                  unfolding eqvts
                  by(rule refl)
(*                have fst: "finite(supp(([(a, b)] \<bullet> Q''') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)::name set)"
                  apply(rule fin_supp)
                proof -
                  have "finite(supp(Q''' \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)::name set)"
                    by finite_guess
                  hence "finite(supp([(a,b)]\<bullet>(Q''' \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))::name set)"
                    by(rule pt_supp_finite_pi[OF pt_name_inst, OF at_name_inst])
                  thus ?thesis
                    by(simp only: eqvts)
                qed*)
                have "a \<sharp> xvec" using `xvec \<sharp>* a` by simp
                have "b \<sharp> (\<Psi> \<otimes> \<Psi>\<^isub>Q)" and "b \<sharp> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>" and "b \<sharp> \<lparr>\<nu>b\<rparr>(([(a,b)]\<bullet>Q''') \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)" and "bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* (\<Psi>\<otimes>\<Psi>\<^isub>Q)" and "bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* (\<Psi>\<otimes>\<Psi>\<^isub>Q)" and "bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)" and "bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* subject(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>)"
                  by(simp add: `b \<sharp> \<Psi> \<otimes> \<Psi>\<^isub>Q` `b \<sharp> xvec` `a \<sharp> xvec` `b \<sharp> K` `b \<sharp> N` fresh_abs_funI2[OF pt_name_inst, OF at_name_inst, OF fin_supp] `xvec \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)` `xvec \<sharp>* M\<^isub>a` `xvec \<sharp>* \<Psi>\<^isub>c` `xvec \<sharp>* K`)+
                from toq'' obtain Q''' where q''def: "Q'' = \<lparr>\<nu>b\<rparr>(Q''' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                proof -
                  assume asm: "(\<And>Q'''. Q'' = \<lparr>\<nu>b\<rparr>(Q''' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<Longrightarrow> thesis)"
                  {
                    fix yvec K' N' P''
                    assume "yvec=xvec"
                      and    "K' = K"
                      and    "N' = N"
                    hence "K'\<lparr>\<nu>*yvec\<rparr>\<langle>N'\<rangle> = K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>" by simp
                    have "b \<sharp> yvec" unfolding `yvec=xvec` by fact
                    have "b \<sharp> K'" unfolding `K'=K` by fact
                    have "b \<sharp> N'" unfolding `N'=N` by fact
                    from toq'' have "\<exists> Q'''. Q''= \<lparr>\<nu>b\<rparr>(Q''' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)" using `b \<sharp> (\<Psi> \<otimes> \<Psi>\<^isub>Q)` `b \<sharp> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>` `b \<sharp> Q''` `bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* (\<Psi>\<otimes>\<Psi>\<^isub>Q)` `bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)` `bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* subject(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>)` `K'\<lparr>\<nu>*yvec\<rparr>\<langle>N'\<rangle>=K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>` `b \<sharp> K'` `b \<sharp> yvec` `b \<sharp> N'` unfolding vadef
                    proof(induct rule: resCases)
                      case(cRes P')
                      show ?case
                        apply(rule parOutputCases[OF `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace> \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'`])
                        apply(rule `xvec \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)`)
                        by(simp add: `a \<sharp> xvec` `b \<sharp> xvec` `xvec \<sharp>* M\<^isub>a``xvec \<sharp>* \<Psi>\<^isub>c` `xvec \<sharp>* K` | blast)+
                    next
                      case(cOpen M xvec zvec y N P')
                      have "b \<in> (supp([(b,y)]\<bullet>N)::name set)" using `y \<in> supp N`
                        apply(subst (asm) perm_bool[symmetric,where pi="[(b,y)]"])
                        by(simp only: eqvts swap_simps)
                      note `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace> \<longmapsto> M\<lparr>\<nu>*(xvec @ zvec)\<rparr>\<langle>([(b, y)] \<bullet> N)\<rangle> \<prec> [(b, y)] \<bullet> P'`
                      moreover have "(xvec @ zvec) \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)" using `zvec \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)` `xvec \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)`
                        by simp
                      moreover have "(xvec @ zvec) \<sharp>* \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>" and "(xvec @ zvec) \<sharp>* \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>" using `xvec \<sharp>* (\<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)` `zvec \<sharp>* (\<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`
                        by simp+
                      moreover have "(xvec @ zvec) \<sharp>* M" using `xvec \<sharp>* M` `zvec \<sharp>* M`
                        by simp
                      ultimately have False
                      proof(induct rule: parOutputCases[of _ _ _ _ _ _ _ "()"])
                        case(cPar1 P' A\<^isub>R \<Psi>\<^isub>R)
                        have psirdef: "\<Psi>\<^isub>R = ([(a,b)]\<bullet>\<Psi>\<^isub>c)" using cPar1
                          by simp
                        have "b \<sharp> (xvec @ zvec)" using `b \<sharp> yvec` `K'\<lparr>\<nu>*yvec\<rparr>\<langle>N'\<rangle> = M\<lparr>\<nu>*(xvec @ y # zvec)\<rparr>\<langle>N\<rangle>`
                          by(simp add: action.inject)
                        hence inbnl: "b \<in> bnlNames (\<lparr>\<nu>*(xvec @ zvec)\<rparr>([(b, y)] \<bullet> N) \<prec>' P')"
                          by(simp add: boundOutputFresh bnlNamesJudge `b \<in> supp ([(b, y)] \<bullet> N)`)
                        note `(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>R \<rhd> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<longmapsto> M\<lparr>\<nu>*(xvec @ zvec)\<rparr>\<langle>([(b, y)] \<bullet> N)\<rangle> \<prec> P'` `guarded P`
                        moreover have "\<forall> \<Psi>'. ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>R) \<otimes> \<Psi>' \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>"
                        proof -
                          {
                            fix \<Psi>'
                            have "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>') \<otimes> \<Psi>\<^isub>R \<simeq> ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>R) \<otimes> \<Psi>'"
                              apply(rule AssertionStatEqTrans[OF Commutativity])
                              apply(rule AssertionStatEqTrans)
                              apply(rule AssertionStatEqSym[OF Associativity])
                              apply(rule Composition)
                              apply(rule AssertionStatEqTrans[OF Commutativity])
                              by(rule AssertionStatEqRefl)
                            have psirdef: "\<Psi>\<^isub>R = ([(a,b)]\<bullet>\<Psi>\<^isub>c)" using cPar1
                              by simp
                            have "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>') \<otimes> \<Psi>\<^isub>R \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>" unfolding psirdef using `\<And>\<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>`
                              apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                              by(simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
                            hence "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>R) \<otimes> \<Psi>' \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>" using `((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>') \<otimes> \<Psi>\<^isub>R \<simeq> ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>R) \<otimes> \<Psi>'`
                              by(rule_tac statEqEnt)
                          }
                          thus ?thesis by simp
                        qed
                        moreover have "\<forall>\<Psi>' Q. b \<sharp> \<Psi>' \<and> ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>R) \<otimes> \<Psi>' \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>Q \<longrightarrow> Q = P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>"
                        proof -
                          {
                            fix \<Psi>' Q
                            assume "b \<sharp> \<Psi>'"
                              and    "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>R) \<otimes> \<Psi>' \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>Q"
                            have "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>R) \<otimes> \<Psi>' \<simeq> ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>') \<otimes> \<Psi>\<^isub>R"
                              apply(rule AssertionStatEqSym)
                              apply(rule AssertionStatEqTrans[OF Commutativity])
                              apply(rule AssertionStatEqTrans)
                              apply(rule AssertionStatEqSym[OF Associativity])
                              apply(rule Composition)
                              apply(rule AssertionStatEqTrans[OF Commutativity])
                              by(rule AssertionStatEqRefl)
                            have psirdef: "\<Psi>\<^isub>R = ([(a,b)]\<bullet>\<Psi>\<^isub>c)" using cPar1
                              by simp
                            have np: "\<forall> \<Psi> Q. (a \<sharp> \<Psi> \<and> \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> Q) \<longrightarrow> Q = P \<parallel> \<langle>M\<^isub>a\<rangle>" using `\<And> \<Psi> Q. \<lbrakk>a \<sharp> \<Psi>; \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>Q\<rbrakk> \<Longrightarrow> Q = P \<parallel> \<langle>M\<^isub>a\<rangle>`
                              by auto
                            moreover have "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>') \<otimes> \<Psi>\<^isub>R \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>Q" using `((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>R) \<otimes> \<Psi>' \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>Q` `((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>R) \<otimes> \<Psi>' \<simeq> ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>') \<otimes> \<Psi>\<^isub>R`
                              by(rule_tac statEqEnt)
                            moreover have "b \<sharp> ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>')"
                              using `b \<sharp> (\<Psi> \<otimes> \<Psi>\<^isub>Q)` `b \<sharp> \<Psi>'`
                              by(simp add: freshComp)
                            ultimately have "Q = P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>" unfolding psirdef
                              apply(subst (asm) perm_bool[symmetric, where pi="[(a,b)]"])
                              by(auto simp add: eqvts swap_simps `a \<sharp> P` `b \<sharp> P`)
                          }
                          thus ?thesis by auto
                        qed
                        ultimately show False using inbnl
                        proof(nominal_induct rule: replOutputInduct)
                          case(cPar1 xvec N P')
                          have "b \<in> supp([(a, b)] \<bullet> M\<^isub>a)" using `\<not> a \<sharp> M\<^isub>a`
                            unfolding fresh_def
                            by(subst (asm) perm_bool[symmetric,where pi="[(a,b)]"]) (simp add: eqvts swap_simps)
                          hence "b \<sharp> xvec" using `xvec \<sharp>* ([(a, b)] \<bullet> M\<^isub>a)` unfolding fresh_star_def fresh_def
                            by(auto simp add: name_list_supp)
                          have "b \<sharp> N"
                            by(rule outputFreshDerivative) fact+
                          moreover have "b \<in> supp N" using `b \<sharp> xvec` `b \<in> bnlNames (\<lparr>\<nu>*xvec\<rparr>N \<prec>' (P' \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>))`
                            unfolding bnlNamesJudge fresh_def name_list_supp
                            by simp
                          ultimately show ?case unfolding fresh_def
                            by simp                           
                        next
                          case(cPar2 B)
                          thus ?case by(simp add: bnlNamesJudge)
                        next
                          case(cInv B)
                          thus ?case by simp
                        qed
                      next
                        case cPar2  thus ?case by auto
                      qed
                      thus ?case by(rule FalseE)
                    qed
                  }
                  note Ex=this
                  from Ex have "\<exists>Q'''. Q'' = \<lparr>\<nu>b\<rparr>(Q''' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)" by simp                
                  hence "(\<And>Q'''. Q'' = \<lparr>\<nu>b\<rparr>(Q''' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<Longrightarrow> thesis) \<Longrightarrow> thesis"
                    by force
                  thus ?thesis using asm
                    by force
                qed
                have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<rhd> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q'''" using `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q''`
                  unfolding vadef using `b \<sharp> (\<Psi> \<otimes> \<Psi>\<^isub>Q)` `b \<sharp> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>` `b \<sharp> Q''` `bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* (\<Psi>\<otimes>\<Psi>\<^isub>Q)` `bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)` `bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* subject(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>)` q''def 
                proof(induct rule: resCases)
                  case(cRes P')
                  have P'eq: "P' = Q''' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>" using `\<lparr>\<nu>b\<rparr>P' = \<lparr>\<nu>b\<rparr>(Q''' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`
                    by(simp add: psi.inject abs_fun_eq1)
                  note `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace> \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'` `xvec \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)`
                  moreover have "xvec \<sharp>* (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)" using `xvec \<sharp>* M\<^isub>a` `xvec \<sharp>* a` `b \<sharp> xvec`
                    by simp
                  moreover have "xvec \<sharp>* \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>" using `xvec \<sharp>* \<Psi>\<^isub>c` `xvec \<sharp>* a` `b \<sharp> xvec`
                    by simp
                  moreover note `xvec \<sharp>* K`
                  ultimately show ?case using P'eq
                  proof(induct rule: parOutputCases[of _ _ _ _ _ _ _ "()"])
                    case cPar2 thus ?case by(auto dest: nilTrans)
                  next
                    case(cPar1 P' A\<^isub>R \<Psi>\<^isub>R) thus ?case by(simp add: psi.inject)
                  qed
                next
                  case(cOpen M xvec yvec y N P')
                  have bypeq: "([(b,y)]\<bullet>P') = ([(b,y)]\<bullet> \<lparr>\<nu>b\<rparr>(Q''' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>))" using `P' = \<lparr>\<nu>b\<rparr>(Q''' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`
                    by simp
                  note `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace> \<longmapsto> M\<lparr>\<nu>*(xvec @ yvec)\<rparr>\<langle>([(b,y)]\<bullet>N)\<rangle> \<prec> ([(b,y)]\<bullet>P')`
                  moreover have "(xvec @ yvec) \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)" using `yvec \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)` `xvec \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)` by simp
                  moreover have "(xvec @ yvec) \<sharp>* (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)" using `xvec \<sharp>* (\<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)` `yvec \<sharp>* (\<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`
                    by simp
                  moreover have "(xvec @ yvec) \<sharp>* \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>" using `xvec \<sharp>* (\<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)` `yvec \<sharp>* (\<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`
                    by simp
                  moreover have "(xvec@yvec) \<sharp>* M" using `xvec \<sharp>* M` `yvec \<sharp>* M`
                    by simp
                  ultimately show ?case using bypeq
                  proof(induct rule: parOutputCases[of _ _ _ _ _ _ _ "()"])
                    case cPar1 thus ?case by(simp add: psi.inject)
                  next
                    case cPar2 thus ?case by(simp add: psi.inject)
                  qed
                qed
                hence "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one> \<rhd> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q'''"
                  by(rule transferFrame[where A\<^isub>F="[]" and A\<^isub>G="[]" and A\<^isub>P="[]" and \<Psi>\<^isub>P="\<one>"]) (auto simp add: assertionStatEqImp[OF AssertionStatEqSym[OF Identity]])
                
                {
                  fix F:: "'b frame"
                  obtain A\<^isub>F \<Psi>\<^isub>F where "F=\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>" and "A\<^isub>F \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)" and "A\<^isub>F \<sharp>* b" and "A\<^isub>F \<sharp>* (((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one>)" and "A\<^isub>F \<sharp>* ([(a, b)] \<bullet> \<Psi>\<^isub>c)"
                    apply(rule freshFrame[where C="((\<Psi> \<otimes> \<Psi>\<^isub>Q),((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one>,b,([(a, b)] \<bullet> \<Psi>\<^isub>c))"])
                    by auto
                  hence "b \<sharp> A\<^isub>F" by simp
                  obtain c::name where "c \<sharp> A\<^isub>F" and "c \<sharp> \<Psi>\<^isub>F" and "c \<noteq> b" and "c \<noteq> a" and "c \<sharp> \<Psi>"  and "c \<sharp> \<Psi>\<^isub>c" and "c \<sharp> \<Psi>\<^isub>Q"
                    apply(generate_fresh "name")
                    by force
                  hence "[(a,b)] \<bullet> c = c" by simp
                  have "c \<sharp> ([(a,b)]\<bullet>\<Psi>\<^isub>c)"
                    apply(subst `[(a,b)] \<bullet> c = c`[symmetric])
                    apply(subst fresh_bij)
                    by fact
                  have lhs: "(\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>\<epsilon>,(\<Psi> \<otimes> \<Psi>\<^isub>Q)\<rangle>) \<simeq>\<^sub>F \<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> (\<Psi> \<otimes> \<Psi>\<^isub>Q)\<rangle>"
                    apply(subst mergeFrameSimps)
                    by(fact,rule FrameStatEqRefl)
                  have "(\<langle>A\<^isub>F, \<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[c], ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(b, c)] \<bullet> [(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one>\<rangle>) \<simeq>\<^sub>F (\<langle>A\<^isub>F, \<Psi>\<^isub>F \<otimes> (\<Psi> \<otimes> \<Psi>\<^isub>Q)\<rangle>) \<otimes>\<^sub>F (\<langle>[c], [(b, c)] \<bullet> [(a, b)] \<bullet> \<Psi>\<^isub>c\<rangle>)"
                    apply(subst (1 2) mergeFrames)
                    apply(simp add: `c \<sharp> A\<^isub>F`)
                    apply(subst fresh_star_bij[symmetric, where pi="[(b,c)]"])
                    apply(simp add: `b \<sharp> A\<^isub>F` `c \<sharp> A\<^isub>F` `A\<^isub>F \<sharp>* ([(a, b)] \<bullet> \<Psi>\<^isub>c)`)
                    apply(simp add: freshComp `c \<sharp> \<Psi>` `c \<sharp> \<Psi>\<^isub>Q` `c \<sharp> \<Psi>\<^isub>F`)
                    apply(subst fresh_star_bij[symmetric, where pi="[(b,c)]"])
                    apply(simp add: `b \<sharp> A\<^isub>F` `c \<sharp> A\<^isub>F` `A\<^isub>F \<sharp>* ([(a, b)] \<bullet> \<Psi>\<^isub>c)` eqvts `b \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q` `c \<sharp> \<Psi>` `c \<sharp> \<Psi>\<^isub>Q` `A\<^isub>F \<sharp>* (((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one>)`)
                    apply(simp add: `c \<sharp> \<Psi>\<^isub>F`)
                    apply(rule frameResChainPres)
                    apply(subst frameNilStatEq)                    
                    apply(rule AssertionStatEqSym)
                    apply(rule AssertionStatEqTrans[OF Associativity])
                    apply(rule compositionSym)
                    apply(rule AssertionStatEqSym[OF Identity])
                    done
                  hence rhs: "(\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[b],(((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one>)\<rangle>) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> (\<Psi> \<otimes> \<Psi>\<^isub>Q)\<rangle>) \<otimes>\<^sub>F (\<langle>[b], ([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rangle>)"
                    apply(subst (1 2) frameChainAlpha[where p="[(b,c)]" and xvec="[b]"])
                    apply(simp add: `c \<sharp> ([(a,b)]\<bullet>\<Psi>\<^isub>c)` at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] freshComp `c \<sharp> \<Psi>\<^isub>Q` `c \<sharp> \<Psi>` `c \<sharp> ([(a,b)]\<bullet>\<Psi>\<^isub>c)`)
                    apply(simp add: eqvts at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `b \<sharp> \<Psi>` `c \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q` `c \<sharp> \<Psi>\<^isub>Q` del: frameResChain.simps)
                    done
                  have bfeq: "(\<langle>[b], [(a, b)] \<bullet> \<Psi>\<^isub>c\<rangle>) = (\<langle>[a], \<Psi>\<^isub>c\<rangle>)"
                    by(subst frameChainAlpha[where p="[(a,b)]" and xvec="[a]"]) (simp add: `b \<sharp> \<Psi>\<^isub>c` at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])+
                  have "F \<otimes>\<^sub>F (\<langle>\<epsilon>,(\<Psi> \<otimes> \<Psi>\<^isub>Q)\<rangle>) \<simeq>\<^sub>F F \<otimes>\<^sub>F (\<langle>[b],(((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one>)\<rangle>)"
                    unfolding `F=\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>`
                    apply(rule FrameStatEqTrans[OF lhs])
                    apply(rule FrameStatEqSym)
                    apply(rule FrameStatEqTrans[OF rhs])
                    apply(subst bfeq)
                    by fact
                }
                have "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)) \<otimes> \<one> \<rhd> P \<longmapsto> M\<lparr>N\<rparr> \<prec> P'"
                  by(rule frameStatEqTransition[where A="[b]"]) (fact | simp add:  `b \<sharp> M` `b \<sharp> N` `b \<sharp> P`)+
                obtain A\<^isub>P \<Psi>\<^isub>P where "extractFrame P = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>" and "A\<^isub>P \<sharp>* \<Psi>" and "A\<^isub>P \<sharp>* \<Psi>\<^isub>Q" and "A\<^isub>P \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)" and "A\<^isub>P \<sharp>* a" and "A\<^isub>P \<sharp>* b" and "A\<^isub>P \<sharp>* P" and "A\<^isub>P \<sharp>* xvec" and "A\<^isub>P \<sharp>* M" and "A\<^isub>P \<sharp>* N" and "A\<^isub>P \<sharp>* K" and "distinct A\<^isub>P" and "A\<^isub>P \<sharp>* \<Psi>\<^isub>c" and "A\<^isub>P \<sharp>* M\<^isub>a"
                  by(rule freshFrame[where F="extractFrame P" and C="(\<Psi>,\<Psi>\<^isub>Q,\<Psi> \<otimes> \<Psi>\<^isub>Q,a,b,P,xvec,M,N,K,\<Psi>\<^isub>c,M\<^isub>a)"]) force
                have "\<Psi>\<^isub>P \<simeq> \<one>" and "supp \<Psi>\<^isub>P = ({}::name set)" using `guarded P` `extractFrame P = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>`
                  by(simp add: guardedStatEq)+
                have "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)) \<otimes> \<Psi>\<^isub>P \<rhd> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q'''" using `((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one> \<rhd> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q'''`
                  apply(rule transferFrame[where A\<^isub>P="[]" and \<Psi>\<^isub>P=\<one> and A\<^isub>F="[]" and A\<^isub>G="[]"])
                  apply simp
                  apply simp
                  apply(subst (1 2) frameResChain.simps)
                  apply(subst frameImpNilStatEq)
                  apply(rule assertionStatEqImp)
                  apply(rule Composition)
                  apply(rule compositionSym)
                  apply(rule AssertionStatEqSym[OF `\<Psi>\<^isub>P \<simeq> \<one>`])
                  by simp+
                have "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)) \<otimes> \<Psi>\<^isub>P \<otimes> \<one> \<turnstile> M \<leftrightarrow> K"
                proof -
                  have "(\<langle>\<epsilon>,(\<Psi> \<otimes> \<Psi>\<^isub>Q)\<rangle>) \<turnstile>\<^sub>F M \<leftrightarrow> K" using `\<Psi> \<otimes> \<Psi>\<^isub>Q \<turnstile> M \<leftrightarrow> K`
                    by simp
                  have rhs: "(\<langle>[a],((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> (\<Psi>\<^isub>c)) \<otimes> \<Psi>\<^isub>P \<otimes> \<one>\<rangle>) \<simeq>\<^sub>F (\<langle>\<epsilon>,\<Psi> \<otimes> \<Psi>\<^isub>Q\<rangle>) \<otimes>\<^sub>F (\<langle>[a],\<Psi>\<^isub>c\<rangle>)"
                    apply(subst mergeFrameSimps')
                    apply(simp add: freshComp `a \<sharp> \<Psi>` `a \<sharp> \<Psi>\<^isub>Q`)
                    apply(rule frameResChainPres)
                    apply(subst frameNilStatEq)
                    apply(rule AssertionStatEqTrans[OF AssertionStatEqSym[OF Associativity]])
                    apply(rule AssertionStatEqTrans[OF Identity])
                    apply(rule AssertionStatEqTrans[OF compositionSym[OF `\<Psi>\<^isub>P \<simeq> \<one>`]])
                    by(rule Identity)
                  have "b \<sharp> \<Psi>\<^isub>P" and "a \<sharp> \<Psi>\<^isub>P" unfolding fresh_def using `supp \<Psi>\<^isub>P = {}`
                    by simp+
                  have abeq: "(\<langle>[a],((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>\<^isub>P \<otimes> \<one>\<rangle>) = (\<langle>[b],((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)) \<otimes> \<Psi>\<^isub>P \<otimes> \<one>\<rangle>)"
                    by(subst frameChainAlpha[where p="[(a,b)]" and xvec="[a]"]) (simp add: swap_simps freshComp `b \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q` `b \<sharp> \<Psi>\<^isub>c` `b \<sharp> \<Psi>\<^isub>P` `a \<sharp> \<Psi>` `a \<sharp> \<Psi>\<^isub>Q` `a \<sharp> \<Psi>\<^isub>P` eqvts)+
                  have "(\<langle>\<epsilon>,(\<Psi> \<otimes> \<Psi>\<^isub>Q)\<rangle>) \<simeq>\<^sub>F (\<langle>[a],((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>\<^isub>P \<otimes> \<one>\<rangle>)"
                    apply(rule FrameStatEqSym)
                    apply(rule FrameStatEqTrans[OF rhs])
                    by fact
                  hence "(\<langle>[a],((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>\<^isub>P \<otimes> \<one>\<rangle>) \<turnstile>\<^sub>F M \<leftrightarrow> K" using `(\<langle>\<epsilon>,(\<Psi> \<otimes> \<Psi>\<^isub>Q)\<rangle>) \<turnstile>\<^sub>F M \<leftrightarrow> K`
                    by(rule frameStatEqEnt)
                  moreover have "[b] \<sharp>* (M \<leftrightarrow> K)" using `b \<sharp> M` `b \<sharp> K`
                    by auto
                  ultimately show ?thesis
                    unfolding abeq
                    by(rule frameImpE)
                qed
                have "a \<sharp> xvec" using `xvec \<sharp>* a`
                  by simp
                have abeq: "\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) = \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"(* and abeq': "\<lparr>\<nu>a\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>Q'''))\<parallel>\<lbrace>\<Psi>\<^isub>c\<rbrace>) = \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(([(a,b)]\<bullet>P')\<parallel>([(a,b)]\<bullet>Q''')))\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"*)
                  by(subst alphaRes[where x=a and y=b],(simp add: `b \<sharp> M\<^isub>a` `b \<sharp> \<Psi>\<^isub>c` `b \<sharp> P'` `a \<sharp> xvec` `b \<sharp> xvec` resChainFresh eqvts)+)+
                have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>" using `(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>`
                  by(subst perm_bool[where pi="[(a,b)]", symmetric]) (simp add: eqvts `a \<sharp> \<Psi>` `a \<sharp> \<Psi>\<^isub>Q` `a \<sharp> P` `b \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q` `b \<sharp> P`)
                have "A\<^isub>P \<sharp>* ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c))"
                proof -
                  have "A\<^isub>P \<sharp>* ([(a, b)] \<bullet> \<Psi>\<^isub>c)" using `A\<^isub>P \<sharp>* a` `A\<^isub>P \<sharp>* b` `A\<^isub>P \<sharp>* \<Psi>\<^isub>c`
                    by simp
                  thus ?thesis
                    by(simp add: freshCompChain `A\<^isub>P \<sharp>* \<Psi>` `A\<^isub>P \<sharp>* \<Psi>\<^isub>Q`)
                qed
                have "A\<^isub>P \<sharp>* \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>" using `A\<^isub>P \<sharp>* a` `A\<^isub>P \<sharp>* b` `A\<^isub>P \<sharp>* M\<^isub>a`
                  by simp
                have "(supp (P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>)::name set) \<subseteq> supp ([(a, b)] \<bullet> M\<^isub>a)"
                proof -
                  have "(supp P::name set) \<subseteq> supp([(a,b)]\<bullet>M\<^isub>a)" using `supp P \<subseteq> supp M\<^isub>a`
                    by(subst (asm) perm_bool[symmetric,where pi="[(a,b)]"]) (simp add: `a \<sharp> P` `b \<sharp> P` eqvts)
                  thus ?thesis
                    by(simp add: psi.supp)
                qed
                have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<tau> \<prec> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(Q''')))\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                  unfolding abeq
                  apply(rule Scope)
                  apply(rule Par1[where \<Psi>\<^isub>Q="[(a,b)]\<bullet>\<Psi>\<^isub>c" and A\<^isub>Q="[]"])
                  apply(rule Inv[where P="P\<parallel>\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>"])
                  apply fact
                  apply(rule Comm1[where \<Psi>\<^isub>Q=\<one> and \<Psi>\<^isub>P=\<Psi>\<^isub>P and A\<^isub>Q="[]" and A\<^isub>P=A\<^isub>P and M=M and N=N and K=K])
                  by(fact | simp add: `guarded P`)+
                show ?case
                proof(cases "\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> Q'''' \<sim> P'''")
                  assume "\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> Q'''' \<sim> P'''"
                  hence "\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> Q'' \<sim> P''" using `\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> Q'' \<sim> Q''''` `\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> P''' \<sim> P''`
                    by(auto dest: bisimTransitive)
                  have "\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(Q''')))\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>*xvec\<rparr>(P'\<parallel>\<lparr>\<nu>b\<rparr>(Q'''\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                  proof -
                    have "\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(Q''')))\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>b\<rparr>((P'\<parallel>Q''')\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                      apply(rule bisimTransitive[OF bisimResPres])
                      apply(rule bisimTransitive[OF bisimParComm])
                      apply(rule bisimSymmetric[OF bisimScopeExtChain])
                      apply fact
                      apply(simp add: `a \<sharp> xvec` `b \<sharp> xvec` `xvec \<sharp>* \<Psi>\<^isub>c`)
                      apply fact
                      apply(rule bisimTransitive[OF bisimResComm'])
                      apply fact+
                      apply(rule bisimResChainPres)
                      apply(rule bisimResPres)
                      apply(rule bisimParComm)
                      by fact+
                    moreover have "\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>b\<rparr>((P'\<parallel>Q''')\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> \<lparr>\<nu>b\<rparr>(Q'''\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                      apply(rule bisimResChainPres)
                      apply(rule bisimTransitive[OF bisimResPres])
                      apply(rule bisimParAssoc)
                      apply fact
                      apply(rule bisimScopeExt)
                      by fact+
                    ultimately show ?thesis
                      by(auto dest: bisimTransitive)
                  qed
                  hence bis: "\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(Q''')))\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q''" unfolding q''def by -
                  have bis': "\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q'' \<sim> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> P''" using `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> Q'' \<sim> P''` `xvec \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)`
                    by(metis bisimResChainPres bisimTransitive bisimParComm bisimSymmetric bisimParPres)
                  have fresh: "A\<^isub>Q \<sharp>* (\<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q''') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>), \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q'', \<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'', \<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'')"
                  proof -
                    have "A\<^isub>Q \<sharp>* Q'''" using `A\<^isub>Q \<sharp>* Q''` `b \<sharp> A\<^isub>Q`
                      unfolding q''def
                      by simp
                    thus ?thesis using `A\<^isub>Q \<sharp>* xvec` `b \<sharp> A\<^isub>Q` `A\<^isub>Q \<sharp>* a` `A\<^isub>Q \<sharp>* \<Psi>\<^isub>c` `A\<^isub>Q \<sharp>* P'` `A\<^isub>Q \<sharp>* P''` `A\<^isub>Q \<sharp>* Q''`
                      by simp
                  qed
                  have "(\<Psi>\<otimes>\<Psi>\<^isub>Q,\<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(Q''')))\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>),\<lparr>\<nu>*xvec\<rparr>P'\<parallel>P'',A\<^isub>Q) \<in> ?YA1"
                    apply(rule lol')
                    apply(rule refl)
                    apply(rule bis)
                    apply(rule disjI2[OF bis'])
                    apply(rule bisimReflexive)
                    by(rule fresh)
                  with `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<tau> \<prec> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(Q''')))\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)` show ?thesis
                    apply(rule_tac exI)
                    by(rule conjI)
                next
                  assume "\<not> \<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> Q'''' \<sim> P'''"
                  hence "(\<Psi>\<otimes>\<Psi>\<^isub>Q,Q'''',P''') \<in> ?X1" using `(\<Psi>\<otimes>\<Psi>\<^isub>Q,Q'''',P''') \<in> ?X1 \<or> \<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> Q'''' \<sim> P'''`
                    by simp
                  then obtain yvec Q\<^isub>i a\<^isub>i M\<^isub>i \<Psi>\<^isub>i P\<^isub>i where q''''def: "Q'''' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)" and p'''def: "P''' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i" and "a\<^isub>i \<sharp> (\<Psi> \<otimes> \<Psi>\<^isub>Q)" and "a\<^isub>i \<sharp> P\<^isub>i" and "a\<^isub>i \<sharp> Q\<^isub>i" and "\<not> a\<^isub>i \<sharp> M\<^isub>i" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i\<Leftarrow>P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a\<^isub>i], \<Psi>\<^isub>i\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i \<Leftarrow> R \<longrightarrow> (R = (P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>)))" and "guarded P\<^isub>i" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>i\<Leftarrow>P \<longrightarrow> (supp M\<^isub>i) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>i\<Leftarrow>P = M\<^isub>i\<Leftarrow>P') = (P = P'))" and "((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i" and "((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i" and "yvec \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>i \<Leftarrow> P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
                    by auto
                  have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>Q'''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))))"
                  proof -
                    have "xvec \<sharp>* ([(a,b)] \<bullet> \<Psi>\<^isub>c)" using `a \<sharp> xvec` `b \<sharp> xvec` `xvec \<sharp>* \<Psi>\<^isub>c`
                      by simp
                    have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>Q'''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>b\<rparr>(\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>Q'''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                      apply(rule bisimResPres)
                      apply(rule bisimTransitive[OF bisimParComm])
                      apply(rule bisimSymmetric)
                      apply(rule bisimTransitive[OF bisimResChainPres])
                      apply(rule bisimParComm)
                      apply fact
                      apply(rule bisimScopeExtChain)
                      by(fact|simp add: `xvec \<sharp>* ([(a,b)] \<bullet> \<Psi>\<^isub>c)`)+
                    moreover have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>(\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>Q'''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>b\<rparr>(P'\<parallel>Q'''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                      by(rule bisimResComm') fact+
                    moreover have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>b\<rparr>(P'\<parallel>Q'''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>(P'\<parallel>Q'')"
                      unfolding q''def
                      apply(rule bisimResChainPres)
                      apply(rule bisimTransitive[OF bisimResPres[OF bisimParAssoc]])
                      apply fact
                      apply(rule bisimScopeExt)
                      by fact+
                    moreover have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>*xvec\<rparr>(P'\<parallel>Q'') \<sim> (\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))))"
                      apply(rule bisimResChainPres)
                      apply(rule bisimTransitive[OF bisimParComm])
                      apply(rule bisimSymmetric[OF bisimTransitive[OF bisimParComm]])
                      apply(rule bisimParPres)
                      apply(rule bisimSymmetric)
                      apply(rule bisimTransitive[OF `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> Q'' \<sim> Q''''`])
                      apply(subst q''''def)
                      apply(rule bisimReflexive)
                      by fact
                    ultimately show ?thesis
                      by(auto dest: bisimTransitive)
                  qed
                  obtain p::"name prm" where "set p \<subseteq> set yvec \<times> set(p\<bullet>yvec)" and "(p\<bullet>yvec) \<sharp>* a\<^isub>i" and "distinctPerm p" and "(p\<bullet>yvec) \<sharp>* M\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>\<^isub>i" and "(p\<bullet>yvec) \<sharp>* P\<^isub>i" and "(p\<bullet>yvec) \<sharp>* Q\<^isub>i" and "(p\<bullet>yvec) \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)" and "(p\<bullet>yvec) \<sharp>* P'" and "(p\<bullet>yvec) \<sharp>* xvec" and "(p\<bullet>yvec) \<sharp>* A\<^isub>Q"
                    by(rule chain.name_list_avoiding[where xvec=yvec and c="(M\<^isub>i,\<Psi>\<^isub>i,a\<^isub>i,P\<^isub>i,Q\<^isub>i,\<Psi>\<otimes>\<Psi>\<^isub>Q,P',xvec,A\<^isub>Q)"]) auto
                  hence yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>(p\<bullet>a\<^isub>i)\<rparr>(\<langle>(p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" and yvrdef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)"
                    by(subst resChainAlpha[where p=p],simp,simp,simp)+
                  obtain b\<^isub>i::name where "b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)" and "b\<^isub>i \<sharp> \<Psi>" and "b\<^isub>i \<sharp> \<Psi>\<^isub>Q" and "b\<^isub>i \<sharp> P'" and "b\<^isub>i \<sharp> A\<^isub>Q"
                    apply(rule at_exists_fresh[OF at_name_inst, where x="(p\<bullet>M\<^isub>i,p\<bullet>\<Psi>\<^isub>i,p\<bullet>P\<^isub>i,p\<bullet>Q\<^isub>i,\<Psi>,\<Psi>\<^isub>Q,P',A\<^isub>Q)"])
                    unfolding supp_prod pt_perm_supp[OF pt_name_inst,OF at_name_inst, symmetric]
                    apply(rule finite_UnI, (finite_guess | (subst pt_set_finite_ineq[OF pt_name_inst,OF at_name_inst], finite_guess)))+
                    apply(finite_guess)
                    by force
                  from `b\<^isub>i \<sharp> \<Psi>` `b\<^isub>i \<sharp> \<Psi>\<^isub>Q` have "b\<^isub>i \<sharp> (\<Psi>\<otimes>\<Psi>\<^isub>Q)"
                    by(rule freshComp)
                  have yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" unfolding yvldef
                    by(subst alphaRes[where x="(p\<bullet>a\<^isub>i)" and y="b\<^isub>i"]) (simp add: `b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)`)+
                  have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>Q'''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (\<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                  proof -
                    note `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>Q'''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))`
                    moreover have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                      unfolding yvldef
                      by(rule bisimReflexive)
                    moreover have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)) \<sim> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>P'\<parallel>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)"
                      apply(subst resChainAppend)
                      apply(rule bisimResChainPres)
                      apply(rule bisimSymmetric)
                      apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                      apply fact
                      apply(rule bisimScopeExtChain)
                      by fact+
                    ultimately show ?thesis
                      by(auto dest: bisimTransitive)
                  qed
                  moreover have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> P''"
                  proof -
                    have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i))" unfolding resChainAppend
                      by(rule bisimReflexive)
                    moreover have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i)) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)))"
                      apply(rule bisimResChainPres)
                      apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                      apply fact
                      apply(rule bisimScopeExtChain)
                      by fact+
                    moreover have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i))) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> P'')" using `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> P''' \<sim> P''` `xvec \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)`
                      unfolding yvrdef[symmetric] p'''def[symmetric]
                      by(metis bisimResChainPres bisimTransitive bisimParComm bisimSymmetric bisimParPres)
                    ultimately show ?thesis
                      by(auto dest: bisimTransitive)
                  qed
                  moreover have "(\<Psi> \<otimes> \<Psi>\<^isub>Q, (\<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)), \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i)) \<in> ?X1"
                  proof -
                    have "(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)" using `a\<^isub>i \<sharp> P\<^isub>i` by(subst fresh_bij)
                    note `b\<^isub>i \<sharp> (\<Psi> \<otimes> \<Psi>\<^isub>Q)`
                    moreover note `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`
                    moreover have "b\<^isub>i \<sharp> (P' \<parallel> (p\<bullet>Q\<^isub>i))" using `b\<^isub>i \<sharp> P'` `b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)`
                      by simp
                    moreover have "guarded(p\<bullet>P\<^isub>i)" using `guarded P\<^isub>i` by simp
                    moreover have "\<not> b\<^isub>i \<sharp> ([(p\<bullet>a\<^isub>i,b\<^isub>i)] \<bullet> (p\<bullet>M\<^isub>i))" using `\<not> a\<^isub>i \<sharp> M\<^isub>i`
                      apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                      apply(subst (asm) perm_bool[where pi="p",symmetric])
                      by(simp add: eqvts swap_simps)
                    moreover have "\<And> \<Psi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>" using EntCl
                      apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                      apply(subst perm_bool[where pi="rev p",symmetric])
                      by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`)
                    moreover have "\<forall>F. (F \<otimes>\<^sub>F (\<langle>[b\<^isub>i],[(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)\<rangle>) \<simeq>\<^sub>F F)"
                      using Equiv1
                      apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                      apply(subst perm_bool[where pi="rev p",symmetric])
                      by(simp add: eqvts swap_simps)
                    moreover have "(\<forall>\<Psi> R. b\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i)) \<Leftarrow> R \<longrightarrow> (R = ((p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>)))" using Equiv2
                      apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                      apply(subst (asm) perm_bool[where pi="p",symmetric])
                      by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i` swap_simps)
                    moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P \<longrightarrow> (supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))) \<subseteq> ((supp \<Psi>)::name set))" using EntSupp
                      apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                      apply(subst (asm) perm_bool[where pi="p",symmetric])
                      by(simp add: eqvts swap_simps)
                    moreover have "\<forall> P P'. ((([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P') = (P = P'))" using Inj
                      apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                      apply(subst (asm) perm_bool[where pi="p",symmetric])
                      by(simp add: eqvts swap_simps)
                    moreover have "((supp(p\<bullet>P\<^isub>i))::name set) \<subseteq> supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i`
                      apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                      apply(subst (asm) perm_bool[where pi="p",symmetric])
                      by(simp add: eqvts swap_simps `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                    moreover have "((supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)))::name set) = supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i`
                      apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                      apply(subst (asm) perm_bool[where pi="p",symmetric])
                      by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                    moreover have "(xvec@(p\<bullet>yvec)) \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)" using `xvec \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)` `(p\<bullet>yvec) \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)`
                      by simp
                    moreover from EntEnt have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i)\<parallel>\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                      apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                      apply(subst perm_bool[where pi="p",symmetric])
                      apply(simp add: eqvts swap_simps `distinctPerm p` `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                      done
                    ultimately show ?thesis
                      by blast
                  qed
                  moreover have "A\<^isub>Q \<sharp>* (\<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q''') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>), \<lparr>\<nu>*(xvec @p \<bullet> yvec)\<rparr>P' \<parallel> (p \<bullet> Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p \<bullet> a\<^isub>i, b\<^isub>i)] \<bullet> p \<bullet> M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p \<bullet> a\<^isub>i, b\<^isub>i)] \<bullet> p \<bullet> \<Psi>\<^isub>i)\<rbrace>), \<lparr>\<nu>*(xvec @ p \<bullet> yvec)\<rparr>P' \<parallel> (p \<bullet> Q\<^isub>i) \<parallel> !(p \<bullet> P\<^isub>i), \<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'')"
                  proof -
                    from `A\<^isub>Q \<sharp>* Q''` `b \<sharp> A\<^isub>Q` have "A\<^isub>Q \<sharp>* Q'''" unfolding q''def
                      by simp
                    moreover from `A\<^isub>Q \<sharp>* Q''''` have "A\<^isub>Q \<sharp>* (p \<bullet> Q\<^isub>i)" and "A\<^isub>Q \<sharp>* ([(p \<bullet> a\<^isub>i, b\<^isub>i)] \<bullet> p \<bullet> M\<^isub>i)" and "A\<^isub>Q \<sharp>* ([(p \<bullet> a\<^isub>i, b\<^isub>i)] \<bullet> p \<bullet> \<Psi>\<^isub>i)" unfolding q''''def yvldef using `(p \<bullet> yvec) \<sharp>* A\<^isub>Q` `b\<^isub>i \<sharp> A\<^isub>Q`
                      by(simp+)
                    moreover from `A\<^isub>Q \<sharp>* P'''` have "A\<^isub>Q \<sharp>* (p\<bullet>P\<^isub>i)" unfolding p'''def yvrdef using `(p \<bullet> yvec) \<sharp>* A\<^isub>Q`
                      by simp
                    ultimately show ?thesis using `b \<sharp> A\<^isub>Q` `A\<^isub>Q \<sharp>* xvec` `A\<^isub>Q \<sharp>* P'` `A\<^isub>Q \<sharp>* Q'''` `A\<^isub>Q \<sharp>* \<Psi>\<^isub>c` `A\<^isub>Q \<sharp>* a` `(p\<bullet>yvec) \<sharp>* A\<^isub>Q` `b\<^isub>i \<sharp> A\<^isub>Q` `A\<^isub>Q \<sharp>* P''`
                      by(simp add: resChainAppend)
                  qed
                  ultimately have "(\<Psi>\<otimes>\<Psi>\<^isub>Q,\<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(Q''')))\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>),\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'',A\<^isub>Q) \<in> ?YA1"
                    apply -
                    apply(rule lol')
                    by(rule refl|assumption|rule disjI1)+
                  moreover note `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<tau> \<prec> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(Q''')))\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)`
                  ultimately show ?thesis
                    by(rule_tac exI) (rule conjI)
                qed
              next
                case(cComm2 M xvec N P' K P'')
                from `xvec \<sharp>* A\<^isub>Q` have "A\<^isub>Q \<sharp>* xvec" by simp
                with `A\<^isub>Q \<sharp>* (\<lparr>\<nu>*xvec\<rparr>P'\<parallel>P'')` have "A\<^isub>Q \<sharp>* P'" and "A\<^isub>Q \<sharp>* P''"
                  by simp+
                have "\<exists> Q''. \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>N\<rparr> \<prec> Q'' \<and> (\<Psi> \<otimes> \<Psi>\<^isub>Q, Q'', P'',A\<^isub>Q) \<in> ?YA1"
                  by(rule cComm2) (fact | simp add: `xvec \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)` `xvec \<sharp>* P` `xvec \<sharp>* M\<^isub>a`)+
                then obtain Q'' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>N\<rparr> \<prec> Q'' \<and> (\<Psi> \<otimes> \<Psi>\<^isub>Q, Q'', P'',A\<^isub>Q) \<in> ?YA1"
                  by(rule exE)
                hence toq'': "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>N\<rparr> \<prec> Q''" and "(\<Psi> \<otimes> \<Psi>\<^isub>Q, Q'', P'',A\<^isub>Q) \<in> ?YA1"
                  by simp+
                from `(\<Psi> \<otimes> \<Psi>\<^isub>Q, Q'', P'',A\<^isub>Q) \<in> ?YA1` obtain Q'''' P''' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> Q'' \<sim> Q''''" and "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> P''' \<sim> P''" and "(\<Psi>\<otimes>\<Psi>\<^isub>Q,Q'''',P''') \<in> ?X1 \<or> \<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> Q'''' \<sim> P'''" and "A\<^isub>Q \<sharp>* Q''" and "A\<^isub>Q \<sharp>* Q''''" and "A\<^isub>Q \<sharp>* P''" and "A\<^isub>Q \<sharp>* P'''"
                  by auto blast
                obtain b::name where "b \<sharp> \<Psi>\<^isub>Q" and "b \<sharp> \<Psi>" and "b \<sharp> \<Psi> \<otimes> \<Psi>\<^isub>Q" and "b \<sharp> \<Psi>\<^isub>c" and "b \<sharp> Q''" and "b \<sharp> P'" and "b \<sharp> P''" and "b \<sharp> M" and "b \<sharp> N" and "b \<sharp> K" and "b \<sharp> xvec" and "b \<sharp> M\<^isub>a" and "b\<noteq>a" and "b \<sharp> P" and "b \<sharp> A\<^isub>Q"
                  by(generate_fresh "name") force
                have vadef: "\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) = \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                  apply(subst alphaRes[where x=a and y=b])
                  apply(simp add: `b \<sharp> M\<^isub>a` `b \<sharp> \<Psi>\<^isub>c`)
                  unfolding eqvts
                  by(rule refl)
                have "a \<sharp> xvec" using `xvec \<sharp>* a` by simp
                have "b \<sharp> (\<Psi> \<otimes> \<Psi>\<^isub>Q)" and "b \<sharp> K\<lparr>N\<rparr>" and "b \<sharp> \<lparr>\<nu>b\<rparr>(([(a,b)]\<bullet>Q''') \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)" and "bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* (\<Psi>\<otimes>\<Psi>\<^isub>Q)" and "bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* (\<Psi>\<otimes>\<Psi>\<^isub>Q)" and "bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)" and "bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* subject(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>)"
                  by(simp add: `b \<sharp> \<Psi> \<otimes> \<Psi>\<^isub>Q` `b \<sharp> xvec` `a \<sharp> xvec` `b \<sharp> K` `b \<sharp> N` fresh_abs_funI2[OF pt_name_inst, OF at_name_inst, OF fin_supp] `xvec \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)` `xvec \<sharp>* M\<^isub>a` `xvec \<sharp>* \<Psi>\<^isub>c` `xvec \<sharp>* K`)+
                from toq'' obtain Q''' where q''def: "Q'' = \<lparr>\<nu>b\<rparr>(Q''' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                proof -
                  assume asm: "(\<And>Q'''. Q'' = \<lparr>\<nu>b\<rparr>(Q''' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<Longrightarrow> thesis)"
                  from toq'' have "\<exists> Q'''. Q''= \<lparr>\<nu>b\<rparr>(Q''' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)" using `b \<sharp> (\<Psi> \<otimes> \<Psi>\<^isub>Q)` `b \<sharp> K\<lparr>N\<rparr>` `b \<sharp> Q''` unfolding vadef
                  proof(induct rule: resInputCases')
                    case(cRes P')
                    show ?case by(rule parInputCases[OF `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace> \<longmapsto> K\<lparr>N\<rparr> \<prec> P'`]) blast+
                  qed
                  hence "(\<And>Q'''. Q'' = \<lparr>\<nu>b\<rparr>(Q''' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<Longrightarrow> thesis) \<Longrightarrow> thesis"
                    by force
                  thus ?thesis using asm
                    by force
                qed
                have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<rhd> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<longmapsto> K\<lparr>N\<rparr> \<prec> Q'''" using `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>N\<rparr> \<prec> Q''`
                  unfolding vadef using `b \<sharp> (\<Psi> \<otimes> \<Psi>\<^isub>Q)` `b \<sharp> K\<lparr>N\<rparr>` `b \<sharp> Q''` q''def 
                proof(induct rule: resInputCases')
                  case(cRes P')
                  have P'eq: "P' = Q''' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>" using `\<lparr>\<nu>b\<rparr>P' = \<lparr>\<nu>b\<rparr>(Q''' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`
                    by(simp add: psi.inject abs_fun_eq1)
                  note `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace> \<longmapsto> K\<lparr>N\<rparr> \<prec> P'`
                  then show ?case using P'eq
                  proof(induct rule: parInputCases[of _ _ _ _ _ _ "()"])
                    case cPar2 thus ?case by(auto dest: nilTrans)
                  next
                    case(cPar1 P' A\<^isub>R \<Psi>\<^isub>R) thus ?case by(simp add: psi.inject)
                  qed
                qed
                hence "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one> \<rhd> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<longmapsto> K\<lparr>N\<rparr> \<prec> Q'''"
                  by(rule transferFrame[where A\<^isub>F="[]" and A\<^isub>G="[]" and A\<^isub>P="[]" and \<Psi>\<^isub>P="\<one>"]) (auto simp add: assertionStatEqImp[OF AssertionStatEqSym[OF Identity]])
                
                {
                  fix F:: "'b frame"
                  obtain A\<^isub>F \<Psi>\<^isub>F where "F=\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>" and "A\<^isub>F \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)" and "A\<^isub>F \<sharp>* b" and "A\<^isub>F \<sharp>* (((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one>)" and "A\<^isub>F \<sharp>* ([(a, b)] \<bullet> \<Psi>\<^isub>c)"
                    apply(rule freshFrame[where C="((\<Psi> \<otimes> \<Psi>\<^isub>Q),((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one>,b,([(a, b)] \<bullet> \<Psi>\<^isub>c))"])
                    by auto
                  hence "b \<sharp> A\<^isub>F" by simp
                  obtain c::name where "c \<sharp> A\<^isub>F" and "c \<sharp> \<Psi>\<^isub>F" and "c \<noteq> b" and "c \<noteq> a" and "c \<sharp> \<Psi>"  and "c \<sharp> \<Psi>\<^isub>c" and "c \<sharp> \<Psi>\<^isub>Q"
                    apply(generate_fresh "name")
                    by force
                  hence "[(a,b)] \<bullet> c = c" by simp
                  have "c \<sharp> ([(a,b)]\<bullet>\<Psi>\<^isub>c)"
                    apply(subst `[(a,b)] \<bullet> c = c`[symmetric])
                    apply(subst fresh_bij)
                    by fact
                  have lhs: "(\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>\<epsilon>,(\<Psi> \<otimes> \<Psi>\<^isub>Q)\<rangle>) \<simeq>\<^sub>F \<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> (\<Psi> \<otimes> \<Psi>\<^isub>Q)\<rangle>"
                    apply(subst mergeFrameSimps)
                    by(fact,rule FrameStatEqRefl)
                  have "(\<langle>A\<^isub>F, \<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[c], ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(b, c)] \<bullet> [(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one>\<rangle>) \<simeq>\<^sub>F (\<langle>A\<^isub>F, \<Psi>\<^isub>F \<otimes> (\<Psi> \<otimes> \<Psi>\<^isub>Q)\<rangle>) \<otimes>\<^sub>F (\<langle>[c], [(b, c)] \<bullet> [(a, b)] \<bullet> \<Psi>\<^isub>c\<rangle>)"
                    apply(subst (1 2) mergeFrames)
                    apply(simp add: `c \<sharp> A\<^isub>F`)
                    apply(subst fresh_star_bij[symmetric, where pi="[(b,c)]"])
                    apply(simp add: `b \<sharp> A\<^isub>F` `c \<sharp> A\<^isub>F` `A\<^isub>F \<sharp>* ([(a, b)] \<bullet> \<Psi>\<^isub>c)`)
                    apply(simp add: freshComp `c \<sharp> \<Psi>` `c \<sharp> \<Psi>\<^isub>Q` `c \<sharp> \<Psi>\<^isub>F`)
                    apply(subst fresh_star_bij[symmetric, where pi="[(b,c)]"])
                    apply(simp add: `b \<sharp> A\<^isub>F` `c \<sharp> A\<^isub>F` `A\<^isub>F \<sharp>* ([(a, b)] \<bullet> \<Psi>\<^isub>c)` eqvts `b \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q` `c \<sharp> \<Psi>` `c \<sharp> \<Psi>\<^isub>Q` `A\<^isub>F \<sharp>* (((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one>)`)
                    apply(simp add: `c \<sharp> \<Psi>\<^isub>F`)
                    apply(rule frameResChainPres)
                    apply(subst frameNilStatEq)                    
                    apply(rule AssertionStatEqSym)
                    apply(rule AssertionStatEqTrans[OF Associativity])
                    apply(rule compositionSym)
                    apply(rule AssertionStatEqSym[OF Identity])
                    done
                  hence rhs: "(\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[b],(((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one>)\<rangle>) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> (\<Psi> \<otimes> \<Psi>\<^isub>Q)\<rangle>) \<otimes>\<^sub>F (\<langle>[b], ([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rangle>)"
                    apply(subst (1 2) frameChainAlpha[where p="[(b,c)]" and xvec="[b]"])
                    apply(simp add: `c \<sharp> ([(a,b)]\<bullet>\<Psi>\<^isub>c)` at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] freshComp `c \<sharp> \<Psi>\<^isub>Q` `c \<sharp> \<Psi>` `c \<sharp> ([(a,b)]\<bullet>\<Psi>\<^isub>c)`)
                    apply(simp add: eqvts at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `b \<sharp> \<Psi>` `c \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q` `c \<sharp> \<Psi>\<^isub>Q` del: frameResChain.simps)
                    done
                  have bfeq: "(\<langle>[b], [(a, b)] \<bullet> \<Psi>\<^isub>c\<rangle>) = (\<langle>[a], \<Psi>\<^isub>c\<rangle>)"
                    by(subst frameChainAlpha[where p="[(a,b)]" and xvec="[a]"]) (simp add: `b \<sharp> \<Psi>\<^isub>c` at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])+
                  have "F \<otimes>\<^sub>F (\<langle>\<epsilon>,(\<Psi> \<otimes> \<Psi>\<^isub>Q)\<rangle>) \<simeq>\<^sub>F F \<otimes>\<^sub>F (\<langle>[b],(((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one>)\<rangle>)"
                    unfolding `F=\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>`
                    apply(rule FrameStatEqTrans[OF lhs])
                    apply(rule FrameStatEqSym)
                    apply(rule FrameStatEqTrans[OF rhs])
                    apply(subst bfeq)
                    by fact
                }
                have "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)) \<otimes> \<one> \<rhd> P \<longmapsto> M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'"
                  by(rule frameStatEqTransition[where A="[b]"]) (fact | simp add:  `b \<sharp> M` `b \<sharp> N` `b \<sharp> P` `b \<sharp> xvec`)+
                obtain A\<^isub>P \<Psi>\<^isub>P where "extractFrame P = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>" and "A\<^isub>P \<sharp>* \<Psi>" and "A\<^isub>P \<sharp>* \<Psi>\<^isub>Q" and "A\<^isub>P \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)" and "A\<^isub>P \<sharp>* a" and "A\<^isub>P \<sharp>* b" and "A\<^isub>P \<sharp>* P" and "A\<^isub>P \<sharp>* xvec" and "A\<^isub>P \<sharp>* M" and "A\<^isub>P \<sharp>* N" and "A\<^isub>P \<sharp>* K" and "distinct A\<^isub>P" and "A\<^isub>P \<sharp>* \<Psi>\<^isub>c" and "A\<^isub>P \<sharp>* M\<^isub>a"
                  by(rule freshFrame[where F="extractFrame P" and C="(\<Psi>,\<Psi>\<^isub>Q,\<Psi> \<otimes> \<Psi>\<^isub>Q,a,b,P,xvec,M,N,K,\<Psi>\<^isub>c,M\<^isub>a)"]) force
                have "\<Psi>\<^isub>P \<simeq> \<one>" and "supp \<Psi>\<^isub>P = ({}::name set)" using `guarded P` `extractFrame P = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>`
                  by(simp add: guardedStatEq)+
                have "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)) \<otimes> \<Psi>\<^isub>P \<rhd> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<longmapsto> K\<lparr>N\<rparr> \<prec> Q'''" using `((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one> \<rhd> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<longmapsto> K\<lparr>N\<rparr> \<prec> Q'''`
                  apply(rule transferFrame[where A\<^isub>P="[]" and \<Psi>\<^isub>P=\<one> and A\<^isub>F="[]" and A\<^isub>G="[]"])
                  apply simp
                  apply simp
                  apply(subst (1 2) frameResChain.simps)
                  apply(subst frameImpNilStatEq)
                  apply(rule assertionStatEqImp)
                  apply(rule Composition)
                  apply(rule compositionSym)
                  apply(rule AssertionStatEqSym[OF `\<Psi>\<^isub>P \<simeq> \<one>`])
                  by simp+
                have "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)) \<otimes> \<Psi>\<^isub>P \<otimes> \<one> \<turnstile> M \<leftrightarrow> K"
                proof -
                  have "(\<langle>\<epsilon>,(\<Psi> \<otimes> \<Psi>\<^isub>Q)\<rangle>) \<turnstile>\<^sub>F M \<leftrightarrow> K" using `\<Psi> \<otimes> \<Psi>\<^isub>Q \<turnstile> M \<leftrightarrow> K`
                    by simp
                  have rhs: "(\<langle>[a],((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> (\<Psi>\<^isub>c)) \<otimes> \<Psi>\<^isub>P \<otimes> \<one>\<rangle>) \<simeq>\<^sub>F (\<langle>\<epsilon>,\<Psi> \<otimes> \<Psi>\<^isub>Q\<rangle>) \<otimes>\<^sub>F (\<langle>[a],\<Psi>\<^isub>c\<rangle>)"
                    apply(subst mergeFrameSimps')
                    apply(simp add: freshComp `a \<sharp> \<Psi>` `a \<sharp> \<Psi>\<^isub>Q`)
                    apply(rule frameResChainPres)
                    apply(subst frameNilStatEq)
                    apply(rule AssertionStatEqTrans[OF AssertionStatEqSym[OF Associativity]])
                    apply(rule AssertionStatEqTrans[OF Identity])
                    apply(rule AssertionStatEqTrans[OF compositionSym[OF `\<Psi>\<^isub>P \<simeq> \<one>`]])
                    by(rule Identity)
                  have "b \<sharp> \<Psi>\<^isub>P" and "a \<sharp> \<Psi>\<^isub>P" unfolding fresh_def using `supp \<Psi>\<^isub>P = {}`
                    by simp+
                  have abeq: "(\<langle>[a],((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>\<^isub>P \<otimes> \<one>\<rangle>) = (\<langle>[b],((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)) \<otimes> \<Psi>\<^isub>P \<otimes> \<one>\<rangle>)"
                    by(subst frameChainAlpha[where p="[(a,b)]" and xvec="[a]"]) (simp add: swap_simps freshComp `b \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q` `b \<sharp> \<Psi>\<^isub>c` `b \<sharp> \<Psi>\<^isub>P` `a \<sharp> \<Psi>` `a \<sharp> \<Psi>\<^isub>Q` `a \<sharp> \<Psi>\<^isub>P` eqvts)+
                  have "(\<langle>\<epsilon>,(\<Psi> \<otimes> \<Psi>\<^isub>Q)\<rangle>) \<simeq>\<^sub>F (\<langle>[a],((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>\<^isub>P \<otimes> \<one>\<rangle>)"
                    apply(rule FrameStatEqSym)
                    apply(rule FrameStatEqTrans[OF rhs])
                    by fact
                  hence "(\<langle>[a],((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>\<^isub>P \<otimes> \<one>\<rangle>) \<turnstile>\<^sub>F M \<leftrightarrow> K" using `(\<langle>\<epsilon>,(\<Psi> \<otimes> \<Psi>\<^isub>Q)\<rangle>) \<turnstile>\<^sub>F M \<leftrightarrow> K`
                    by(rule frameStatEqEnt)
                  moreover have "[b] \<sharp>* (M \<leftrightarrow> K)" using `b \<sharp> M` `b \<sharp> K`
                    by auto
                  ultimately show ?thesis
                    unfolding abeq
                    by(rule frameImpE)
                qed
                have "a \<sharp> xvec" using `xvec \<sharp>* a`
                  by simp
                have abeq: "\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) = \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"(* and abeq': "\<lparr>\<nu>a\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>Q'''))\<parallel>\<lbrace>\<Psi>\<^isub>c\<rbrace>) = \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(([(a,b)]\<bullet>P')\<parallel>([(a,b)]\<bullet>Q''')))\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"*)
                  by(subst alphaRes[where x=a and y=b],(simp add: `b \<sharp> M\<^isub>a` `b \<sharp> \<Psi>\<^isub>c` `b \<sharp> P'` `a \<sharp> xvec` `b \<sharp> xvec` resChainFresh eqvts)+)+
                have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>" using `(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>`
                  by(subst perm_bool[where pi="[(a,b)]", symmetric]) (simp add: eqvts `a \<sharp> \<Psi>` `a \<sharp> \<Psi>\<^isub>Q` `a \<sharp> P` `b \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q` `b \<sharp> P`)
                have "A\<^isub>P \<sharp>* ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c))"
                proof -
                  have "A\<^isub>P \<sharp>* ([(a, b)] \<bullet> \<Psi>\<^isub>c)" using `A\<^isub>P \<sharp>* a` `A\<^isub>P \<sharp>* b` `A\<^isub>P \<sharp>* \<Psi>\<^isub>c`
                    by simp
                  thus ?thesis
                    by(simp add: freshCompChain `A\<^isub>P \<sharp>* \<Psi>` `A\<^isub>P \<sharp>* \<Psi>\<^isub>Q`)
                qed
                have "A\<^isub>P \<sharp>* \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>" using `A\<^isub>P \<sharp>* a` `A\<^isub>P \<sharp>* b` `A\<^isub>P \<sharp>* M\<^isub>a`
                  by simp
                have "(supp (P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>)::name set) \<subseteq> supp ([(a, b)] \<bullet> M\<^isub>a)"
                proof -
                  have "(supp P::name set) \<subseteq> supp([(a,b)]\<bullet>M\<^isub>a)" using `supp P \<subseteq> supp M\<^isub>a`
                    by(subst (asm) perm_bool[symmetric,where pi="[(a,b)]"]) (simp add: `a \<sharp> P` `b \<sharp> P` eqvts)
                  thus ?thesis
                    by(simp add: psi.supp)
                qed
                have "xvec \<sharp>* \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>" using `a \<sharp> xvec` `b \<sharp> xvec` `xvec \<sharp>* M\<^isub>a`
                  by simp
                have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<tau> \<prec> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(Q''')))\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                  unfolding abeq
                  apply(rule Scope)
                  apply(rule Par1[where \<Psi>\<^isub>Q="[(a,b)]\<bullet>\<Psi>\<^isub>c" and A\<^isub>Q="[]"])
                  apply(rule Inv[where P="P\<parallel>\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>"])
                  apply fact
                  apply(rule Comm2[where \<Psi>\<^isub>Q=\<one> and \<Psi>\<^isub>P=\<Psi>\<^isub>P and A\<^isub>Q="[]" and A\<^isub>P=A\<^isub>P and M=M and N=N and K=K])
                  by(fact | simp add: `guarded P `)+
                show ?case
                proof(cases "\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> Q'''' \<sim> P'''")
                  assume "\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> Q'''' \<sim> P'''"
                  hence "\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> Q'' \<sim> P''" using `\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> Q'' \<sim> Q''''` `\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> P''' \<sim> P''`
                    by(auto dest: bisimTransitive)
                  have "\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(Q''')))\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>*xvec\<rparr>(P'\<parallel>\<lparr>\<nu>b\<rparr>(Q'''\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                  proof -
                    have "\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(Q''')))\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>b\<rparr>((P'\<parallel>Q''')\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                      apply(rule bisimTransitive[OF bisimResPres])
                      apply(rule bisimTransitive[OF bisimParComm])
                      apply(rule bisimSymmetric[OF bisimScopeExtChain])
                      apply fact
                      apply(simp add: `a \<sharp> xvec` `b \<sharp> xvec` `xvec \<sharp>* \<Psi>\<^isub>c`)
                      apply fact
                      apply(rule bisimTransitive[OF bisimResComm'])
                      apply fact+
                      apply(rule bisimResChainPres)
                      apply(rule bisimResPres)
                      apply(rule bisimParComm)
                      by fact+
                    moreover have "\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>b\<rparr>((P'\<parallel>Q''')\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> \<lparr>\<nu>b\<rparr>(Q'''\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                      apply(rule bisimResChainPres)
                      apply(rule bisimTransitive[OF bisimResPres])
                      apply(rule bisimParAssoc)
                      apply fact
                      apply(rule bisimScopeExt)
                      by fact+
                    ultimately show ?thesis
                      by(auto dest: bisimTransitive)
                  qed
                  hence bis: "\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(Q''')))\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q''" unfolding q''def by -
                  have bis': "\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q'' \<sim> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> P''" using `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> Q'' \<sim> P''` `xvec \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)`
                    by(metis bisimResChainPres bisimTransitive bisimParComm bisimSymmetric bisimParPres)
                  have fresh: "A\<^isub>Q \<sharp>* (\<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q''') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>), \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q'', \<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'', \<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'')"
                  proof -
                    have "A\<^isub>Q \<sharp>* Q'''" using `A\<^isub>Q \<sharp>* Q''` `b \<sharp> A\<^isub>Q`
                      unfolding q''def
                      by simp
                    thus ?thesis using `A\<^isub>Q \<sharp>* xvec` `b \<sharp> A\<^isub>Q` `A\<^isub>Q \<sharp>* a` `A\<^isub>Q \<sharp>* \<Psi>\<^isub>c` `A\<^isub>Q \<sharp>* P'` `A\<^isub>Q \<sharp>* P''` `A\<^isub>Q \<sharp>* Q''`
                      by simp
                  qed
                  have "(\<Psi>\<otimes>\<Psi>\<^isub>Q,\<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(Q''')))\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>),\<lparr>\<nu>*xvec\<rparr>P'\<parallel>P'',A\<^isub>Q) \<in> ?YA1"
                    apply(rule lol')
                    apply(rule refl)
                    apply(rule bis)
                    apply(rule disjI2[OF bis'])
                    apply(rule bisimReflexive)
                    by(rule fresh)
                  with `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<tau> \<prec> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(Q''')))\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)` show ?thesis
                    apply(rule_tac exI)
                    by(rule conjI)
                next
                  assume "\<not> \<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> Q'''' \<sim> P'''"
                  hence "(\<Psi>\<otimes>\<Psi>\<^isub>Q,Q'''',P''') \<in> ?X1" using `(\<Psi>\<otimes>\<Psi>\<^isub>Q,Q'''',P''') \<in> ?X1 \<or> \<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> Q'''' \<sim> P'''`
                    by simp
                  then obtain yvec Q\<^isub>i a\<^isub>i M\<^isub>i \<Psi>\<^isub>i P\<^isub>i where q''''def: "Q'''' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)" and p'''def: "P''' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i" and "a\<^isub>i \<sharp> (\<Psi> \<otimes> \<Psi>\<^isub>Q)" and "a\<^isub>i \<sharp> P\<^isub>i" and "a\<^isub>i \<sharp> Q\<^isub>i" and "\<not> a\<^isub>i \<sharp> M\<^isub>i" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i\<Leftarrow>P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a\<^isub>i], \<Psi>\<^isub>i\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i \<Leftarrow> R \<longrightarrow> (R = (P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>)))" and "guarded P\<^isub>i" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>i\<Leftarrow>P \<longrightarrow> (supp M\<^isub>i) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>i\<Leftarrow>P = M\<^isub>i\<Leftarrow>P') = (P = P'))" and "((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i" and "((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i" and "yvec \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>i \<Leftarrow> P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
                    by auto
                  have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>Q'''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))))"
                  proof -
                    have "xvec \<sharp>* ([(a,b)] \<bullet> \<Psi>\<^isub>c)" using `a \<sharp> xvec` `b \<sharp> xvec` `xvec \<sharp>* \<Psi>\<^isub>c`
                      by simp
                    have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>Q'''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>b\<rparr>(\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>Q'''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                      apply(rule bisimResPres)
                      apply(rule bisimTransitive[OF bisimParComm])
                      apply(rule bisimSymmetric)
                      apply(rule bisimTransitive[OF bisimResChainPres])
                      apply(rule bisimParComm)
                      apply fact
                      apply(rule bisimScopeExtChain)
                      by(fact|simp add: `xvec \<sharp>* ([(a,b)] \<bullet> \<Psi>\<^isub>c)`)+
                    moreover have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>(\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>Q'''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>b\<rparr>(P'\<parallel>Q'''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                      by(rule bisimResComm') fact+
                    moreover have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>b\<rparr>(P'\<parallel>Q'''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>(P'\<parallel>Q'')"
                      unfolding q''def
                      apply(rule bisimResChainPres)
                      apply(rule bisimTransitive[OF bisimResPres[OF bisimParAssoc]])
                      apply fact
                      apply(rule bisimScopeExt)
                      by fact+
                    moreover have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>*xvec\<rparr>(P'\<parallel>Q'') \<sim> (\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))))"
                      apply(rule bisimResChainPres)
                      apply(rule bisimTransitive[OF bisimParComm])
                      apply(rule bisimSymmetric[OF bisimTransitive[OF bisimParComm]])
                      apply(rule bisimParPres)
                      apply(rule bisimSymmetric)
                      apply(rule bisimTransitive[OF `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> Q'' \<sim> Q''''`])
                      apply(subst q''''def)
                      apply(rule bisimReflexive)
                      by fact
                    ultimately show ?thesis
                      by(auto dest: bisimTransitive)
                  qed
                  obtain p::"name prm" where "set p \<subseteq> set yvec \<times> set(p\<bullet>yvec)" and "(p\<bullet>yvec) \<sharp>* a\<^isub>i" and "distinctPerm p" and "(p\<bullet>yvec) \<sharp>* M\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>\<^isub>i" and "(p\<bullet>yvec) \<sharp>* P\<^isub>i" and "(p\<bullet>yvec) \<sharp>* Q\<^isub>i" and "(p\<bullet>yvec) \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)" and "(p\<bullet>yvec) \<sharp>* P'" and "(p\<bullet>yvec) \<sharp>* xvec" and "(p\<bullet>yvec) \<sharp>* A\<^isub>Q"
                    by(rule chain.name_list_avoiding[where xvec=yvec and c="(M\<^isub>i,\<Psi>\<^isub>i,a\<^isub>i,P\<^isub>i,Q\<^isub>i,\<Psi>\<otimes>\<Psi>\<^isub>Q,P',xvec,A\<^isub>Q)"]) auto
                  hence yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>(p\<bullet>a\<^isub>i)\<rparr>(\<langle>(p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" and yvrdef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)"
                    by(subst resChainAlpha[where p=p],simp,simp,simp)+
                  obtain b\<^isub>i::name where "b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)" and "b\<^isub>i \<sharp> \<Psi>" and "b\<^isub>i \<sharp> \<Psi>\<^isub>Q" and "b\<^isub>i \<sharp> P'" and "b\<^isub>i \<sharp> A\<^isub>Q"
                    apply(rule at_exists_fresh[OF at_name_inst, where x="(p\<bullet>M\<^isub>i,p\<bullet>\<Psi>\<^isub>i,p\<bullet>P\<^isub>i,p\<bullet>Q\<^isub>i,\<Psi>,\<Psi>\<^isub>Q,P',A\<^isub>Q)"])
                    unfolding supp_prod pt_perm_supp[OF pt_name_inst,OF at_name_inst, symmetric]
                    apply(rule finite_UnI, (finite_guess | (subst pt_set_finite_ineq[OF pt_name_inst,OF at_name_inst], finite_guess)))+
                    apply(finite_guess)
                    by force
                  from `b\<^isub>i \<sharp> \<Psi>` `b\<^isub>i \<sharp> \<Psi>\<^isub>Q` have "b\<^isub>i \<sharp> (\<Psi>\<otimes>\<Psi>\<^isub>Q)"
                    by(rule freshComp)
                  have yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" unfolding yvldef
                    by(subst alphaRes[where x="(p\<bullet>a\<^isub>i)" and y="b\<^isub>i"]) (simp add: `b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)`)+
                  have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>Q'''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (\<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                  proof -
                    note `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>Q'''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))`
                    moreover have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                      unfolding yvldef
                      by(rule bisimReflexive)
                    moreover have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)) \<sim> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>P'\<parallel>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)"
                      apply(subst resChainAppend)
                      apply(rule bisimResChainPres)
                      apply(rule bisimSymmetric)
                      apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                      apply fact
                      apply(rule bisimScopeExtChain)
                      by fact+
                    ultimately show ?thesis
                      by(auto dest: bisimTransitive)
                  qed
                  moreover have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> P''"
                  proof -
                    have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i))" unfolding resChainAppend
                      by(rule bisimReflexive)
                    moreover have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i)) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)))"
                      apply(rule bisimResChainPres)
                      apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                      apply fact
                      apply(rule bisimScopeExtChain)
                      by fact+
                    moreover have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i))) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> P'')" using `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> P''' \<sim> P''` `xvec \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)`
                      unfolding yvrdef[symmetric] p'''def[symmetric]
                      by(metis bisimResChainPres bisimTransitive bisimParComm bisimSymmetric bisimParPres)
                    ultimately show ?thesis
                      by(auto dest: bisimTransitive)
                  qed
                  moreover have "(\<Psi> \<otimes> \<Psi>\<^isub>Q, (\<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)), \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i)) \<in> ?X1"
                  proof -
                    have "(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)" using `a\<^isub>i \<sharp> P\<^isub>i` by(subst fresh_bij)
                    note `b\<^isub>i \<sharp> (\<Psi> \<otimes> \<Psi>\<^isub>Q)`
                    moreover note `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`
                    moreover have "b\<^isub>i \<sharp> (P' \<parallel> (p\<bullet>Q\<^isub>i))" using `b\<^isub>i \<sharp> P'` `b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)`
                      by simp
                    moreover have "guarded(p\<bullet>P\<^isub>i)" using `guarded P\<^isub>i` by simp
                    moreover have "\<not> b\<^isub>i \<sharp> ([(p\<bullet>a\<^isub>i,b\<^isub>i)] \<bullet> (p\<bullet>M\<^isub>i))" using `\<not> a\<^isub>i \<sharp> M\<^isub>i`
                      apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                      apply(subst (asm) perm_bool[where pi="p",symmetric])
                      by(simp add: eqvts swap_simps)
                    moreover have "\<And> \<Psi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>" using EntCl
                      apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                      apply(subst perm_bool[where pi="rev p",symmetric])
                      by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`)
                    moreover have "\<forall>F. (F \<otimes>\<^sub>F (\<langle>[b\<^isub>i],[(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)\<rangle>) \<simeq>\<^sub>F F)"
                      using Equiv1
                      apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                      apply(subst perm_bool[where pi="rev p",symmetric])
                      by(simp add: eqvts swap_simps)
                    moreover have "(\<forall>\<Psi> R. b\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i)) \<Leftarrow> R \<longrightarrow> (R = ((p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>)))" using Equiv2
                      apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                      apply(subst (asm) perm_bool[where pi="p",symmetric])
                      by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i` swap_simps)
                    moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P \<longrightarrow> (supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))) \<subseteq> ((supp \<Psi>)::name set))" using EntSupp
                      apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                      apply(subst (asm) perm_bool[where pi="p",symmetric])
                      by(simp add: eqvts swap_simps)
                    moreover have "\<forall> P P'. ((([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P') = (P = P'))" using Inj
                      apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                      apply(subst (asm) perm_bool[where pi="p",symmetric])
                      by(simp add: eqvts swap_simps)
                    moreover have "((supp(p\<bullet>P\<^isub>i))::name set) \<subseteq> supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i`
                      apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                      apply(subst (asm) perm_bool[where pi="p",symmetric])
                      by(simp add: eqvts swap_simps `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                    moreover have "((supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)))::name set) = supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i`
                      apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                      apply(subst (asm) perm_bool[where pi="p",symmetric])
                      by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                    moreover have "(xvec@(p\<bullet>yvec)) \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)" using `xvec \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)` `(p\<bullet>yvec) \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)`
                      by simp
                    moreover from EntEnt have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i)\<parallel>\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>  \<or> \<Psi> \<turnstile> \<phi>"
                      apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                      apply(subst perm_bool[where pi="p",symmetric])
                      apply(simp add: eqvts swap_simps `distinctPerm p` `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                      done
                    ultimately show ?thesis
                      by blast
                  qed
                  moreover have "A\<^isub>Q \<sharp>* (\<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q''') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>), \<lparr>\<nu>*(xvec @p \<bullet> yvec)\<rparr>P' \<parallel> (p \<bullet> Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p \<bullet> a\<^isub>i, b\<^isub>i)] \<bullet> p \<bullet> M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p \<bullet> a\<^isub>i, b\<^isub>i)] \<bullet> p \<bullet> \<Psi>\<^isub>i)\<rbrace>), \<lparr>\<nu>*(xvec @ p \<bullet> yvec)\<rparr>P' \<parallel> (p \<bullet> Q\<^isub>i) \<parallel> !(p \<bullet> P\<^isub>i), \<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'')"
                  proof -
                    from `A\<^isub>Q \<sharp>* Q''` `b \<sharp> A\<^isub>Q` have "A\<^isub>Q \<sharp>* Q'''" unfolding q''def
                      by simp
                    moreover from `A\<^isub>Q \<sharp>* Q''''` have "A\<^isub>Q \<sharp>* (p \<bullet> Q\<^isub>i)" and "A\<^isub>Q \<sharp>* ([(p \<bullet> a\<^isub>i, b\<^isub>i)] \<bullet> p \<bullet> M\<^isub>i)" and "A\<^isub>Q \<sharp>* ([(p \<bullet> a\<^isub>i, b\<^isub>i)] \<bullet> p \<bullet> \<Psi>\<^isub>i)" unfolding q''''def yvldef using `(p \<bullet> yvec) \<sharp>* A\<^isub>Q` `b\<^isub>i \<sharp> A\<^isub>Q`
                      by(simp+)
                    moreover from `A\<^isub>Q \<sharp>* P'''` have "A\<^isub>Q \<sharp>* (p\<bullet>P\<^isub>i)" unfolding p'''def yvrdef using `(p \<bullet> yvec) \<sharp>* A\<^isub>Q`
                      by simp
                    ultimately show ?thesis using `b \<sharp> A\<^isub>Q` `A\<^isub>Q \<sharp>* xvec` `A\<^isub>Q \<sharp>* P'` `A\<^isub>Q \<sharp>* Q'''` `A\<^isub>Q \<sharp>* \<Psi>\<^isub>c` `A\<^isub>Q \<sharp>* a` `(p\<bullet>yvec) \<sharp>* A\<^isub>Q` `b\<^isub>i \<sharp> A\<^isub>Q` `A\<^isub>Q \<sharp>* P''`
                      by(simp add: resChainAppend)
                  qed
                  ultimately have "(\<Psi>\<otimes>\<Psi>\<^isub>Q,\<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(Q''')))\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>),\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'',A\<^isub>Q) \<in> ?YA1"
                    apply -
                    apply(rule lol')
                    by(rule refl|assumption|rule disjI1)+
                  moreover note `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<tau> \<prec> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(Q''')))\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)`
                  ultimately show ?thesis
                    by(rule_tac exI) (rule conjI)
                qed
              qed
              then obtain Q' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> Q' \<and> (\<Psi>\<otimes>\<Psi>\<^isub>Q,Q',P',A\<^isub>Q) \<in> ?YA1"
                by(drule exE)
              hence "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> Q'" and "(\<Psi>\<otimes>\<Psi>\<^isub>Q,Q',P',A\<^isub>Q) \<in> ?YA1"
                by simp+
              have "A\<^isub>Q \<sharp>* \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)" using `A\<^isub>Q \<sharp>* (a,M\<^isub>a,\<Psi>\<^isub>c)`
                by simp
              from `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> Q'` `bn \<alpha> \<sharp>* subject \<alpha>` `distinct(bn \<alpha>)` `A\<^isub>Q \<sharp>* \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)` `A\<^isub>Q \<sharp>* \<alpha>` have "A\<^isub>Q \<sharp>* Q'"
                by(rule freeFreshChainDerivative)
              from `(\<Psi>\<otimes>\<Psi>\<^isub>Q,Q',P',A\<^isub>Q) \<in> ?YA1` obtain Q'' P'' where "\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> Q' \<sim> Q''" and "(\<Psi>\<otimes>\<Psi>\<^isub>Q,Q'',P'')\<in>?X1 \<or> \<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> Q'' \<sim> P''" and "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> P'' \<sim> P'" and "A\<^isub>Q \<sharp>* P'" and "A\<^isub>Q \<sharp>* Q'" and "A\<^isub>Q \<sharp>* P''" and "A\<^isub>Q \<sharp>* Q''"
                by auto blast
              have Trans: "\<Psi> \<rhd> Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> Q \<parallel> Q'"
                by(rule Par2) fact+
              have "(\<Psi>,Q\<parallel>Q',Q\<parallel>P') \<in> ?Y1" using `(\<Psi>\<otimes>\<Psi>\<^isub>Q,Q'',P'')\<in>?X1 \<or> \<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> Q'' \<sim> P''`
              proof(rule disjE)
                assume "\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> Q'' \<sim> P''"
                hence "\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> Q' \<sim> P'" using `\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> Q' \<sim> Q''` `\<Psi>\<otimes>\<Psi>\<^isub>Q \<rhd> P'' \<sim> P'`
                  by(auto dest: bisimTransitive)
                have "\<Psi> \<rhd> Q\<parallel>Q' \<sim> Q\<parallel>P'"
                  apply(rule bisimSymmetric[OF bisimTransitive[OF bisimParComm]])
                  apply(rule bisimSymmetric[OF bisimTransitive[OF bisimParComm]])
                  apply(rule bisimParPresAux)
                  by fact+
                show ?thesis
                  apply(rule lol)
                  by(rule refl|rule `\<Psi> \<rhd> Q\<parallel>Q' \<sim> Q\<parallel>P'`|rule disjI2|rule bisimReflexive)+
              next
                assume "(\<Psi> \<otimes> \<Psi>\<^isub>Q, Q'', P'') \<in> ?X1"
                hence "(\<Psi>, Q'', P'') \<in> ?X1" by(rule X1swapPsi)
                then obtain a P Q\<^isub>i M\<^isub>a \<Psi>\<^isub>c xvec where q''_def: "Q'' = \<lparr>\<nu>*xvec\<rparr>(Q\<^isub>i \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))" and p''_def: "P'' = \<lparr>\<nu>*xvec\<rparr>(Q\<^isub>i \<parallel> !P)" and "a \<sharp> P" and "a \<sharp> Q\<^isub>i" and "a \<sharp> \<Psi>" and "\<not> a \<sharp> M\<^isub>a" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> R \<longrightarrow> (R = (P \<parallel> \<langle>M\<^isub>a\<rangle>)))" and "guarded P" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<longrightarrow> (supp M\<^isub>a) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P'))" and "((supp P)::name set) \<subseteq> supp M\<^isub>a" and "((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a" and "xvec \<sharp>* \<Psi>" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle>\<or> \<Psi> \<turnstile> \<phi>)"
                  by auto
                obtain p::"name prm" where "distinctPerm p" and "set p \<subseteq> set xvec \<times> set(p\<bullet>xvec)" "(p\<bullet>xvec) \<sharp>* Q" and "(p\<bullet>xvec) \<sharp>* \<Psi>" and "(p\<bullet>xvec) \<sharp>* (Q\<^isub>i \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))" and "(p\<bullet>xvec) \<sharp>* (Q\<^isub>i \<parallel> !P)"
                  by(rule chain.name_list_avoiding[where xvec=xvec and c="(Q,\<Psi>,Q\<^isub>i \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>),Q\<^isub>i \<parallel> !P)"]) auto
                obtain b::name where "b \<sharp> P" and "b \<sharp> Q" and "b \<sharp> \<Psi>" and "b \<sharp> M\<^isub>a" "b \<sharp> \<Psi>\<^isub>c" and "b \<sharp> (p\<bullet>a)" and "b \<sharp> (p\<bullet>M\<^isub>a)" and "b \<sharp> (p\<bullet>\<Psi>\<^isub>c)" and "b \<sharp> (p\<bullet>P)" and "b \<sharp> Q" and "b \<sharp> (p\<bullet>Q\<^isub>i)"
                  apply(rule at_exists_fresh[OF at_name_inst, where x="(P,Q,\<Psi>,M\<^isub>a,\<Psi>\<^isub>c,p\<bullet>a,p\<bullet>M\<^isub>a,p\<bullet>\<Psi>\<^isub>c,p\<bullet>P,p\<bullet>Q\<^isub>i,Q)"])
                  unfolding supp_prod pt_perm_supp[OF pt_name_inst,OF at_name_inst, symmetric]
                  apply(rule finite_UnI, (finite_guess | (subst pt_set_finite_ineq[OF pt_name_inst,OF at_name_inst], finite_guess)))+
                  apply(finite_guess)
                  by force
                have "\<Psi> \<rhd> Q\<parallel>Q' \<sim> \<lparr>\<nu>*(p\<bullet>xvec)\<rparr>((Q \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>p\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                proof -
                  have "\<Psi> \<rhd> Q\<parallel>Q' \<sim> Q\<parallel>Q''"
                    apply(rule bisimSymmetric[OF bisimTransitive[OF bisimParComm]])
                    apply(rule bisimSymmetric[OF bisimTransitive[OF bisimParComm]])
                    apply(rule bisimParPresAux)
                    by fact+
                  moreover have "\<Psi> \<rhd> Q\<parallel>Q'' \<sim> Q\<parallel>(\<lparr>\<nu>*xvec\<rparr>Q\<^isub>i \<parallel>\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))"
                    unfolding q''_def
                    by(rule bisimReflexive)
                  moreover have "\<Psi> \<rhd> Q\<parallel>(\<lparr>\<nu>*xvec\<rparr>Q\<^isub>i \<parallel>\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) \<sim> Q\<parallel>(\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>((p\<bullet>Q\<^isub>i)\<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>p\<bullet>\<Psi>\<^isub>c)\<rbrace>)))"
                    apply(subst resChainAlpha[where p=p])
                    apply fact
                    apply fact
                    unfolding eqvts
                    apply(subst alphaRes[where x="p\<bullet>a" and y=b])
                    apply(simp add: `b \<sharp> (p\<bullet>M\<^isub>a)` `b \<sharp> (p\<bullet>\<Psi>\<^isub>c)`)
                    unfolding eqvts
                    by(rule bisimReflexive)
                  moreover have "\<Psi> \<rhd> Q\<parallel>(\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>((p\<bullet>Q\<^isub>i)\<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>p\<bullet>\<Psi>\<^isub>c)\<rbrace>))) \<sim> (\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(Q \<parallel>((p\<bullet>Q\<^isub>i)\<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>p\<bullet>\<Psi>\<^isub>c)\<rbrace>))))"
                    by(rule bisimSymmetric[OF bisimScopeExtChain]) fact+
                  moreover have "\<Psi> \<rhd> (\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(Q \<parallel>((p\<bullet>Q\<^isub>i)\<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>p\<bullet>\<Psi>\<^isub>c)\<rbrace>)))) \<sim> \<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(Q \<parallel>(p\<bullet>Q\<^isub>i))\<parallel>(\<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>p\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                    apply(rule bisimResChainPres)
                    apply(rule bisimSymmetric[OF bisimParAssoc])
                    by fact
                  ultimately show ?thesis
                    by(auto dest: bisimTransitive)
                qed
                moreover have "\<Psi> \<rhd> \<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(Q\<parallel>(p\<bullet>Q\<^isub>i))\<parallel>!(p\<bullet>P) \<sim> Q \<parallel> P'"
                proof -
                  have "\<Psi> \<rhd> Q \<parallel> P'' \<sim> Q \<parallel> P'"
                    apply(rule bisimSymmetric[OF bisimTransitive[OF bisimParComm]])
                    apply(rule bisimSymmetric[OF bisimTransitive[OF bisimParComm]])
                    apply(rule bisimParPresAux)
                    by fact+
                  moreover have "\<Psi> \<rhd> Q\<parallel>(\<lparr>\<nu>*xvec\<rparr>Q\<^isub>i \<parallel>!P) \<sim> Q\<parallel>P''"
                    unfolding p''_def
                    by(rule bisimReflexive)
                  moreover have "\<Psi> \<rhd> Q\<parallel>(\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>((p\<bullet>Q\<^isub>i)\<parallel> !(p\<bullet>P))) \<sim> Q\<parallel>(\<lparr>\<nu>*xvec\<rparr>Q\<^isub>i \<parallel>!P)"
                    apply(subst resChainAlpha[where p=p and xvec=xvec])
                    apply fact
                    apply fact
                    unfolding eqvts
                    by(rule bisimReflexive)
                  moreover have "\<Psi> \<rhd> \<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(Q \<parallel>((p\<bullet>Q\<^isub>i)\<parallel> (!(p\<bullet>P)))) \<sim> Q\<parallel>(\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>((p\<bullet>Q\<^isub>i)\<parallel> !(p\<bullet>P)))"                    
                    by(rule bisimScopeExtChain) fact+
                  moreover have "\<Psi> \<rhd> \<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(Q \<parallel>(p\<bullet>Q\<^isub>i))\<parallel>(!(p\<bullet>P)) \<sim> \<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(Q \<parallel>((p\<bullet>Q\<^isub>i)\<parallel> (!(p\<bullet>P))))"
                    apply(rule bisimResChainPres)
                    by(rule bisimParAssoc) fact
                  ultimately show ?thesis
                    by(metis bisimTransitive)
                qed
                moreover have "(\<Psi>,\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>((Q \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>p\<bullet>\<Psi>\<^isub>c)\<rbrace>)),\<lparr>\<nu>*p \<bullet> xvec\<rparr>Q \<parallel> (p \<bullet> Q\<^isub>i) \<parallel> !(p \<bullet> P)) \<in> ?X1"
                proof -
                  have "(p\<bullet>a) \<sharp> (p\<bullet>P)" using `a \<sharp> P`
                    by(subst pt_fresh_bij[OF pt_name_inst, OF at_name_inst])
                  note `b \<sharp> \<Psi>`
                  moreover note `b \<sharp> (p\<bullet>P)`
                  moreover have "b \<sharp> Q \<parallel> (p\<bullet>Q\<^isub>i)" using `b \<sharp> Q` `b \<sharp> (p\<bullet>Q\<^isub>i)`
                    by simp
                  moreover have "guarded(p\<bullet>P)" using `guarded P` by simp
                  moreover have "\<not> b \<sharp> ([(p\<bullet>a,b)] \<bullet> (p\<bullet>M\<^isub>a))" using `\<not> a \<sharp> M\<^isub>a`        
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps)
                  moreover have "\<And> \<Psi>. \<Psi> \<otimes> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)) \<turnstile> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>(p\<bullet>P) \<parallel> \<langle>([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<rangle>" using EntCl
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst perm_bool[where pi="rev p",symmetric])
                    by(simp add: eqvts `(p\<bullet>a) \<sharp> (p\<bullet>P)` `b \<sharp> (p\<bullet>P)`)
                  moreover have "\<forall>F. (F \<otimes>\<^sub>F (\<langle>[b],[(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)\<rangle>) \<simeq>\<^sub>F F)"
                    using Equiv1
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst perm_bool[where pi="rev p",symmetric])
                    by(simp add: eqvts swap_simps)
                  moreover have "(\<forall>\<Psi> R. b \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)) \<turnstile> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a)) \<Leftarrow> R \<longrightarrow> (R = ((p\<bullet>P) \<parallel> \<langle>([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<rangle>)))" using Equiv2
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts `(p\<bullet>a) \<sharp> p\<bullet>P` `b \<sharp> p\<bullet>P` swap_simps)
                  moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>P \<longrightarrow> (supp ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))) \<subseteq> ((supp \<Psi>)::name set))" using EntSupp
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps)
                  moreover have "\<forall> P P'. ((([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>P = ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>P') = (P = P'))" using Inj
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps)
                  moreover have "((supp(p\<bullet>P))::name set) \<subseteq> supp ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))" using `((supp P)::name set) \<subseteq> supp M\<^isub>a`
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps `(p\<bullet>a) \<sharp> p\<bullet>P` `b \<sharp> p\<bullet>P`)
                  moreover have "((supp ([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)))::name set) = supp ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))" using `((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a`
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                  moreover note `(p\<bullet>xvec)\<sharp>*\<Psi>`
                  moreover from EntEnt have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>(p\<bullet>P)\<parallel>\<langle>([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]",symmetric])
                    apply(subst perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps `distinctPerm p` `(p\<bullet>a) \<sharp> p\<bullet>P` `b \<sharp> p\<bullet>P`)
                  ultimately show ?thesis by blast
                qed
                ultimately show ?thesis
                  apply(rule_tac lol)
                  by(rule refl|assumption|rule disjI1)+
              qed
              hence "(\<Psi>,Q\<parallel>Q',Q\<parallel>P') \<in> ?Y"
                by(rule Y1inY)
              with Trans show ?case
                by(rule_tac exI) (rule conjI)
            next
              case(cComm1 \<Psi>\<^isub>P M N Q' A\<^isub>Q \<Psi>\<^isub>Q K xvec P' A\<^isub>P)
              hence "bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* M\<^isub>a" and "A\<^isub>Q \<sharp>* bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>)" and "bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)"
                by auto
              have "a \<sharp> \<Psi>\<^isub>Q" using `a \<sharp> Q` `extractFrame Q = \<langle>A\<^isub>Q,\<Psi>\<^isub>Q\<rangle>` `A\<^isub>Q \<sharp>* (a, M\<^isub>a,\<Psi>\<^isub>c)`
              proof -
                have "a \<sharp> \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>" using `a \<sharp> Q` 
                  apply(subst `extractFrame Q = \<langle>A\<^isub>Q,\<Psi>\<^isub>Q\<rangle>`[symmetric])
                  by(rule extractFrameFresh)
                hence "a \<sharp> \<langle>\<epsilon>, \<Psi>\<^isub>Q\<rangle>" using `A\<^isub>Q \<sharp>* (a, M\<^isub>a,\<Psi>\<^isub>c)`
                  apply(subst (asm) frameResChainFresh)
                  by(simp add: fresh_def frame.supp name_list_supp)
                thus ?thesis
                  by simp
              qed
              obtain \<Psi>' where psi'def: "\<Psi>'=\<Psi>\<otimes>\<Psi>\<^isub>Q" by simp
              obtain A\<^isub>Q' \<Psi>\<^isub>Q' where "extractFrame Q = \<langle>A\<^isub>Q',\<Psi>\<^isub>Q'\<rangle>" and "distinct A\<^isub>Q'" and "A\<^isub>Q' \<sharp>* \<Psi>" and "A\<^isub>Q' \<sharp>* P" and "A\<^isub>Q' \<sharp>* a" and "A\<^isub>Q' \<sharp>* M\<^isub>a" and "A\<^isub>Q' \<sharp>* \<Psi>\<^isub>c" and "A\<^isub>Q' \<sharp>* P'" and "A\<^isub>Q' \<sharp>* bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>)" and "A\<^isub>Q' \<sharp>* \<Psi>'"
                by(rule freshFrame[where F="extractFrame Q" and C="(\<Psi>,P,a,M\<^isub>a,\<Psi>\<^isub>c,P',xvec,\<Psi>')"]) auto
              from `A\<^isub>Q \<sharp>* (a, M\<^isub>a,\<Psi>\<^isub>c)` have "a \<sharp> A\<^isub>Q" and "A\<^isub>Q \<sharp>* a" and "A\<^isub>Q \<sharp>* M\<^isub>a" and "A\<^isub>Q \<sharp>* \<Psi>\<^isub>c"
                by simp+
              from `A\<^isub>Q \<sharp>* !P` have "A\<^isub>Q \<sharp>* P"
                by simp
              from `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'` have "\<Psi>' \<rhd> !P \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'"
                unfolding psi'def by-
              from `bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)` have "bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* \<Psi>'"
                unfolding psi'def by-
              from `a \<sharp> \<Psi>` `a \<sharp> \<Psi>\<^isub>Q` have "a \<sharp> \<Psi>'"
                unfolding psi'def by force
              from `xvec \<sharp>* \<Psi>` have "bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* \<Psi>" by simp
              from `\<Psi>' \<rhd> !P \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'` `xvec \<sharp>* K`
              have "\<exists> Q'. \<Psi>' \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q' \<and> (\<Psi>,Q',P',A\<^isub>Q) \<in> ?YA1" using `bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* \<Psi>'` `bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* \<Psi>` `bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* M\<^isub>a` `(supp \<Psi>\<^isub>c) = supp M\<^isub>a` `((supp P)::name set) \<subseteq> supp M\<^isub>a` `a \<sharp> P` EntSupp EntCl Equiv1 `a \<sharp> \<Psi>` `a \<sharp> \<Psi>'` Equiv2 Inj `\<not> a \<sharp> M\<^isub>a` `A\<^isub>Q \<sharp>* bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>)` `A\<^isub>Q \<sharp>* P` `A\<^isub>Q \<sharp>* a` `A\<^isub>Q \<sharp>* M\<^isub>a` `A\<^isub>Q \<sharp>* \<Psi>\<^isub>c` `A\<^isub>Q \<sharp>* P'` EntEnt
              proof(nominal_induct avoiding: a \<Psi> \<Psi>\<^isub>c A\<^isub>Q M\<^isub>a rule: bangOutputInduct')
                case(cAlpha xvec N P' p)
                from `A\<^isub>Q \<sharp>* bn (K\<lparr>\<nu>*(p \<bullet> xvec)\<rparr>\<langle>(p \<bullet> N)\<rangle>)` have "A\<^isub>Q \<sharp>* (p\<bullet>xvec)"
                  by simp
                from `A\<^isub>Q \<sharp>* (p \<bullet> xvec)` have "(p\<bullet>xvec) \<sharp>* A\<^isub>Q"
                  by simp
                from `(p\<bullet>xvec) \<sharp>* A\<^isub>Q` `xvec \<sharp>* A\<^isub>Q` `set p \<subseteq> set xvec \<times> set (p \<bullet> xvec)` have "(p\<bullet>A\<^isub>Q) = A\<^isub>Q"
                  by simp
                from `A\<^isub>Q \<sharp>* (p \<bullet> xvec)` have "A\<^isub>Q \<sharp>* xvec"
                  apply(subst (asm) `(p\<bullet>A\<^isub>Q) = A\<^isub>Q`[symmetric])
                  by(subst (asm) fresh_star_bij)
                from `A\<^isub>Q \<sharp>* (p \<bullet> P')` have "A\<^isub>Q \<sharp>* P'"
                  by(subst (asm) `(p\<bullet>A\<^isub>Q) = A\<^isub>Q`[symmetric]) (subst (asm) fresh_star_bij)
                have "\<exists>P''. (\<Psi>' \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'' \<and> (\<Psi>, P'', P',A\<^isub>Q) \<in> ?YA1)"
                  by(rule cAlpha) (fact|simp add: `xvec \<sharp>* \<Psi>'` `xvec \<sharp>* M\<^isub>a` `A\<^isub>Q \<sharp>* xvec`)+
                then obtain P'' where "\<Psi>' \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'' \<and> (\<Psi>, P'', P',A\<^isub>Q) \<in> ?YA1"
                  by(drule HOL.exE)
                hence "\<Psi>' \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P''" and "(\<Psi>, P'', P',A\<^isub>Q) \<in> ?YA1"
                  by simp+
                have ppsi: "(p\<bullet>\<Psi>) = \<Psi>" using `bn(K\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>\<langle>(p\<bullet>N)\<rangle>) \<sharp>* \<Psi>` `xvec \<sharp>* \<Psi>` `set p \<subseteq> set xvec \<times> set (p \<bullet> xvec)`
                  by simp
                have lhs: "(\<Psi>, p\<bullet>P'', p\<bullet>P',A\<^isub>Q) \<in> ?YA1"
                proof -
                  have pout: "(p \<bullet> \<Psi>, p \<bullet> P'', p \<bullet> P',p\<bullet>A\<^isub>Q) = p\<bullet>(\<Psi>, P'', P',A\<^isub>Q)"
                    by(simp only: eqvts)
                  show ?thesis
                    apply(subst ppsi[symmetric])
                    apply(subst `(p\<bullet>A\<^isub>Q) = A\<^isub>Q`[symmetric])
                    apply(subst pout)
                    apply(rule eqvtI)
                    by fact+
                qed
                have "distinct(bn (K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>))" using `\<Psi>' \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P''`
                  by(rule boundOutputDistinct)
                hence "distinct xvec" by simp
                have "(p \<bullet> xvec) \<sharp>* \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)"
                proof -
                  have "(p\<bullet>xvec) \<sharp>* (\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)" using `bn (K\<lparr>\<nu>*(p \<bullet> xvec)\<rparr>\<langle>(p \<bullet> N)\<rangle>) \<sharp>* M\<^isub>a` `(supp \<Psi>\<^isub>c) = supp M\<^isub>a`
                    unfolding fresh_def fresh_star_def
                    by(auto simp add: psi.supp)
                  thus ?thesis
                    apply(subst psiFreshVec)
                    unfolding fresh_star_def fresh_def
                    apply(subst abs_fun_supp[OF pt_name_inst, OF at_name_inst])
                    apply(finite_guess)
                    by blast
                qed
                have "(p \<bullet> xvec) \<sharp>* P''"
                  by(rule outputFreshChainDerivative[OF `\<Psi>' \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P''`]) fact+
                have rhs: "\<Psi>' \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>\<langle>(p\<bullet>N)\<rangle> \<prec> (p\<bullet>P'')"
                  by(subst alphaOutputResidual[symmetric]) (fact|simp)+
                from lhs rhs show ?case
                  apply(rule_tac x="p\<bullet>P''" in exI)
                  by(rule conjI)
              next
                case cBang
                thus ?case by auto
              next
                case(cPar1 xvec N P')
                obtain b::name where "b \<sharp> \<Psi>" and "b \<sharp> \<Psi>'" and "b \<sharp> P" and "b \<sharp> K" and "b \<sharp> xvec" and "b \<sharp> N" and "b \<noteq> a" and "b \<sharp> M\<^isub>a" and "b \<sharp> \<Psi>\<^isub>c" and "b \<sharp> A\<^isub>Q" and "b \<sharp> P'"
                  apply(generate_fresh "name")
                  by force
                have unb: "\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) = \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                  apply(subst alphaRes[where x=a and y=b])
                  apply(simp add: `b \<sharp> M\<^isub>a` `b \<sharp> \<Psi>\<^isub>c`)
                  unfolding eqvts
                  by(rule refl)
                have "\<Psi>' \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>" using `\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>`
                  apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                  by(simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
                have "(\<Psi>' \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one> \<rhd> P \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'"
                proof -
                  {
                    fix F::"'b frame"
                    obtain A\<^isub>F \<Psi>\<^isub>F where "F=\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>" and "distinct A\<^isub>F" and "A\<^isub>F \<sharp>* \<Psi>" and "A\<^isub>F \<sharp>* \<Psi>'" and "A\<^isub>F \<sharp>* a" and "A\<^isub>F \<sharp>* b" and "A\<^isub>F \<sharp>* \<Psi>\<^isub>c"
                      apply(rule freshFrame[where C="(\<Psi>,\<Psi>',a,b,\<Psi>\<^isub>c)"])
                      by auto
                    have lhs: "(\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>\<epsilon>, \<Psi>'\<rangle>) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>'\<rangle>)"
                      apply(subst mergeFrameSimps)
                      apply fact
                      apply(rule FrameStatEqRefl)
                      done
                    obtain c::name where "c \<sharp> A\<^isub>F" and "c \<sharp> \<Psi>\<^isub>F" and "c \<noteq> b" and "c \<noteq> a" and "c \<sharp> \<Psi>"  and "c \<sharp> \<Psi>\<^isub>c" and "c \<sharp> \<Psi>'"
                      apply(generate_fresh "name")
                      by force
                    have af: "([(a,c)]\<bullet>A\<^isub>F) = A\<^isub>F" using `A\<^isub>F \<sharp>* a` `c \<sharp> A\<^isub>F`
                      by simp
                    have "((\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[c], (\<Psi>' \<otimes> ([(a, c)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one>\<rangle>)) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>'\<rangle>) \<otimes>\<^sub>F (\<langle>[c], ([(a,c)]\<bullet>\<Psi>\<^isub>c)\<rangle>)"
                      apply(subst (1 2) mergeFrames)
                      apply(simp add: `c \<sharp> A\<^isub>F`)
                      apply(subst af[symmetric])
                      apply(subst fresh_star_bij)
                      apply fact
                      apply(rule freshCompChain)
                      apply(simp add: `c \<sharp> \<Psi>\<^isub>F`)
                      apply(simp add: `c \<sharp> \<Psi>'`)
                      apply(rule freshCompChain)
                      apply(rule freshCompChain)
                      apply fact
                      apply(subst af[symmetric])
                      apply(subst fresh_star_bij)
                      apply fact
                      apply simp
                      apply(simp add: `c \<sharp> \<Psi>\<^isub>F`)
                      apply(rule frameResChainPres)
                      apply(subst frameNilStatEq)
                      apply(rule AssertionStatEqSym)
                      apply(rule AssertionStatEqTrans[OF Associativity])
                      apply(rule compositionSym)
                      by(rule AssertionStatEqSym[OF Identity])
                    hence "((\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[a], (\<Psi>' \<otimes> \<Psi>\<^isub>c) \<otimes> \<one>\<rangle>)) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>'\<rangle>) \<otimes>\<^sub>F (\<langle>[a], \<Psi>\<^isub>c\<rangle>)"
                      apply(subst (1 2) frameChainAlpha'[where A\<^isub>P="[a]" and p="[(a,c)]"])
                      apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `c \<sharp> \<Psi>\<^isub>c`)
                      apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                      apply(rule freshCompChain)
                      apply(rule freshCompChain)
                      apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `a \<sharp> \<Psi>'` `c \<sharp> \<Psi>'`)
                      apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `c \<sharp> \<Psi>\<^isub>c`)
                      apply simp
                      by(simp add: `a \<sharp> \<Psi>'` `c \<sharp> \<Psi>'` swap_simps eqvts)
                    hence rhs: "((\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[b], (\<Psi>' \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one>\<rangle>)) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>'\<rangle>) \<otimes>\<^sub>F (\<langle>[b], ([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rangle>)"
                      apply(subst frameChainAlpha'[where A\<^isub>P="[b]" and p="[(b,a)]"])
                      apply(rule freshCompChain)
                      apply(rule freshCompChain)
                      apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `a \<sharp> \<Psi>'` `b \<sharp> \<Psi>'`)
                      apply(subst pt3[OF pt_name_inst, OF at_ds5, OF at_name_inst])
                      apply(subst pt_fresh_star_bij[OF pt_name_inst, OF at_name_inst])
                      apply(simp add: `b \<sharp> \<Psi>\<^isub>c`)
                      apply simp
                      apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                      apply(subst frameChainAlpha'[where A\<^isub>P="[b]" and p="[(b,a)]"])
                      apply(subst pt3[OF pt_name_inst, OF at_ds5, OF at_name_inst])
                      apply(subst pt_fresh_star_bij[OF pt_name_inst, OF at_name_inst])
                      apply(simp add: `b \<sharp> \<Psi>\<^isub>c`)
                      apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                      apply(simp add: eqvts at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `a \<sharp> \<Psi>'` `b \<sharp> \<Psi>'` pt3[OF pt_name_inst, OF at_ds5[where a=b and b=a], OF at_name_inst])
                      done
                    have unb: "\<langle>[a],\<Psi>\<^isub>c\<rangle> = (\<langle>[b],([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rangle>)"
                      apply(subst frameChainAlpha'[where A\<^isub>P="[a]" and p="[(a,b)]"])
                      apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `b \<sharp> \<Psi>\<^isub>c`)
                      apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                      by(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                    have "(F \<otimes>\<^sub>F (\<langle>\<epsilon>, \<Psi>'\<rangle>)) \<simeq>\<^sub>F (F \<otimes>\<^sub>F (\<langle>[b], (\<Psi>' \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one>\<rangle>))"
                      unfolding `F = \<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>`
                      apply(rule FrameStatEqTrans[OF lhs])
                      apply(rule FrameStatEqSym)
                      apply(rule FrameStatEqTrans[OF rhs])
                      apply(subst unb[symmetric])
                      by fact
                  }
                  show ?thesis
                    apply(rule frameStatEqTransition[where A="[b]" and \<Psi>="\<Psi>'"])
                    by(fact|simp add: `xvec \<sharp>* K` `b \<sharp> K` `b \<sharp> xvec` `b \<sharp> N`)+
                qed
                have "xvec \<sharp>* ([(a, b)] \<bullet> M\<^isub>a)" and "xvec \<sharp>* ([(a, b)] \<bullet> \<Psi>\<^isub>c)"
                proof -
                  have "[(a,b)] \<bullet> xvec = xvec" using `xvec \<sharp>* a` `b \<sharp> xvec`
                    by simp
                  show "xvec \<sharp>* ([(a, b)] \<bullet> M\<^isub>a)"
                    apply(subst `[(a,b)] \<bullet> xvec = xvec`[symmetric])
                    apply(subst perm_bool[symmetric, where pi="[(a,b)]"])
                    by(simp add: eqvts `xvec \<sharp>* M\<^isub>a`)
                  show "xvec \<sharp>* ([(a, b)] \<bullet> \<Psi>\<^isub>c)"
                    apply(subst `[(a,b)] \<bullet> xvec = xvec`[symmetric])
                    apply(subst perm_bool[symmetric, where pi="[(a,b)]"])
                    by(simp add: eqvts `xvec \<sharp>* \<Psi>\<^isub>c`)
                qed
                have "supp (P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>) \<subseteq> (supp ([(a, b)] \<bullet> M\<^isub>a)::name set)" using `((supp P)::name set) \<subseteq> supp M\<^isub>a`
                  apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                  by(simp add: eqvts psi.supp `a \<sharp> P` `b \<sharp> P`)
                have "guarded (P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>)" using `guarded P`
                  by simp
                have "\<Psi>' \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> \<lparr>\<nu>b\<rparr>((P' \<parallel> (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)) \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                  unfolding unb
                  apply(rule Scope)
                  apply(rule Par1[where \<Psi>\<^isub>Q="([(a, b)] \<bullet> \<Psi>\<^isub>c)" and A\<^isub>Q="[]"])
                  apply(rule Inv[where P="P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>"])
                  apply fact
                  apply(rule Par1[where \<Psi>\<^isub>Q="\<one>" and A\<^isub>Q="[]"])
                  by(fact|simp|rule conjI)+
                moreover have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((P' \<parallel> (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)) \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> P' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)"
                proof -
                  have lhs: "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((P' \<parallel> (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)) \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>b\<rparr>(P' \<parallel> ((\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>) \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                    apply(rule bisimResPres)
                    apply(rule bisimParAssoc)
                    by fact
                  show ?thesis
                    unfolding unb
                    apply(rule bisimTransitive[OF lhs])
                    apply(rule bisimScopeExt)
                    by fact
                qed
                moreover have "\<Psi> \<rhd> P' \<parallel> !P \<sim> P' \<parallel> !P"
                  by(rule bisimReflexive)
                moreover have "(\<Psi>, P' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>), P' \<parallel> !P) \<in> ?X1 \<or> \<Psi> \<rhd> P' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<sim> P' \<parallel> !P"
                proof -
                  note `b \<sharp> \<Psi>`
                  moreover note `b \<sharp> P` `b \<sharp> P'`
                  moreover have "\<not> b \<sharp> [(a, b)] \<bullet> M\<^isub>a" using `\<not> a \<sharp> M\<^isub>a`
                    apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                    by(force simp add: eqvts at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                  moreover note `guarded P`
                  moreover have "supp P \<subseteq> (supp ([(a, b)] \<bullet> M\<^isub>a)::name set)" using `supp P \<subseteq> supp M\<^isub>a`
                    apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                    by(simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
                  moreover have "supp ([(a, b)] \<bullet> \<Psi>\<^isub>c) = (supp ([(a, b)] \<bullet> M\<^isub>a)::name set)" using `supp (\<Psi>\<^isub>c) = supp (M\<^isub>a)`
                    apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                    by(simp add: eqvts)
                  moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<longrightarrow> (supp ([(a,b)]\<bullet>M\<^isub>a)) \<subseteq> ((supp \<Psi>)::name set))" using `\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<longrightarrow> (supp M\<^isub>a) \<subseteq> ((supp \<Psi>)::name set))`
                    apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                    by(simp add: eqvts)
                  moreover have "\<forall>\<Psi>. \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>" using `\<And>\<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>`
                    apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                    by(simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
                  moreover have "\<forall> F. F \<otimes>\<^sub>F (\<langle>[b],([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rangle>) \<simeq>\<^sub>F F" using `\<And> F. F \<otimes>\<^sub>F (\<langle>[a],\<Psi>\<^isub>c\<rangle>) \<simeq>\<^sub>F F`
                    apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                    by(force simp add: eqvts at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                  moreover have "\<forall> P P'. ((([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P = ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P') = (P = P'))" using `\<forall>P P'. (M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P')`
                    apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                    by(simp add: eqvts)
                  moreover have "\<forall> \<Psi> R. b \<sharp> \<Psi> \<longrightarrow> (\<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a) \<Leftarrow> R \<longrightarrow> R = (P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>))" using `\<And> \<Psi> R. \<lbrakk>a \<sharp> \<Psi>; \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>R\<rbrakk> \<Longrightarrow> R = P \<parallel> \<langle>M\<^isub>a\<rangle>`
                    apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                    by(simp add: eqvts `a \<sharp> P` `b \<sharp> P` at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                  moreover have "([]::name list) \<sharp>* \<Psi>"
                    by simp
                  moreover have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<or> \<Psi> \<turnstile> \<phi>" using `\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> \<phi> \<Longrightarrow> \<phi> = M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> \<phi>`
                    by(subst perm_bool[where pi="[(a,b)]",symmetric]) (simp add: eqvts swap_simps `a \<sharp> P` `b \<sharp> P`)
                  ultimately have "(\<Psi>, \<lparr>\<nu>*[]\<rparr>P' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>), \<lparr>\<nu>*[]\<rparr>P' \<parallel> !P) \<in> ?X1"
                    apply(subst unb)
                    by blast
                  hence "(\<Psi>, P' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>), P' \<parallel> !P) \<in> ?X1"
                    unfolding unb resChain.simps
                    by -
                  thus ?thesis
                    by blast
                qed
                moreover have "A\<^isub>Q \<sharp>* (\<lparr>\<nu>b\<rparr>((P' \<parallel> (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)) \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>), P' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>), P' \<parallel> !P, P' \<parallel> !P)"
                proof -
                  have "a \<sharp> A\<^isub>Q" using `A\<^isub>Q \<sharp>* a`
                    by simp
                  thus ?thesis using `b \<sharp> A\<^isub>Q` `A\<^isub>Q \<sharp>* (P' \<parallel> !P)` `A\<^isub>Q \<sharp>* M\<^isub>a` `A\<^isub>Q \<sharp>* \<Psi>\<^isub>c`
                    by(simp add: fresh_abs_fun_iff[OF pt_name_inst, OF at_name_inst, OF fin_supp])
                qed
                ultimately have "(\<Psi>, \<lparr>\<nu>b\<rparr>((P' \<parallel> (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)) \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>), P' \<parallel> !P,A\<^isub>Q) \<in> ?YA1"
                  apply -
                  apply(rule lol')
                  apply(rule refl)
                  by assumption+
                with `\<Psi>' \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> \<lparr>\<nu>b\<rparr>((P' \<parallel> (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)) \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)` show ?case
                  by(rule_tac exI, rule conjI)
              next
                case(cPar2 xvec N P')
                from `A\<^isub>Q \<sharp>* (P \<parallel> P')` have "A\<^isub>Q \<sharp>* P'" by simp
                have "\<exists> P''. \<Psi>' \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'' \<and> (\<Psi>, P'', P',A\<^isub>Q) \<in> ?YA1"
                  apply(rule cPar2)
                  apply(rule cPar2)
                  by fact+
                then obtain P'' where "\<Psi>' \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'' \<and> (\<Psi>, P'', P',A\<^isub>Q) \<in> ?YA1"
                  by(rule exE)
                hence "\<Psi>' \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P''" and "(\<Psi>, P'', P', A\<^isub>Q) \<in> ?YA1"
                  by simp+
                from `(\<Psi>, P'', P', A\<^isub>Q) \<in> ?YA1`
                obtain R R' where "\<Psi> \<rhd> P'' \<sim> R" and "(\<Psi>,R,R') \<in> ?X1 \<or> \<Psi> \<rhd> R \<sim> R'" and "\<Psi> \<rhd> R' \<sim> P'" and "A\<^isub>Q \<sharp>* R" and "A\<^isub>Q \<sharp>* R'" and "A\<^isub>Q \<sharp>* P''" and "A\<^isub>Q \<sharp>* P'"
                  by auto blast+
                have "\<Psi> \<rhd> R' \<sim> P \<parallel> P'"
                proof -
                  obtain R'' where "\<Psi> \<rhd> P' \<sim> R'' \<parallel> !P"
                    apply(rule bangDerivative'[where \<Psi>="\<Psi>'" and \<Psi>'=\<Psi> and P=P and \<alpha>="K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>" and P'=P'])
                    apply(rule cPar2)
                    apply(rule cPar2)
                    apply(rule cPar2)
                    apply(simp add: `xvec \<sharp>* P`)
                    apply(simp add: `xvec \<sharp>* K`)
                    by simp
                  have "\<Psi> \<rhd> R'' \<parallel> !P \<sim> R'' \<parallel> (P \<parallel> !P)"
                    apply(rule bisimTransitive[OF bisimParComm])
                    apply(rule bisimSymmetric)
                    apply(rule bisimTransitive[OF bisimParComm])
                    apply(rule bisimParPres)
                    apply(rule bisimSymmetric)
                    apply(rule bangExt[OF `guarded P`])
                    done
                  show ?thesis
                    apply(rule bisimTransitive[OF `\<Psi> \<rhd> R' \<sim> P'`])
                    apply(rule bisimTransitive[OF `\<Psi> \<rhd> P' \<sim> R'' \<parallel> !P`])
                    apply(rule bisimSymmetric)
                    apply(rule bisimTransitive[OF bisimParComm])
                    apply(rule bisimTransitive[OF bisimParPres[OF `\<Psi> \<rhd> P' \<sim> R'' \<parallel> !P`]])
                    apply(rule bisimTransitive[OF bisimParAssoc])
                    apply(rule bisimTransitive[OF bisimParComm])
                    apply(rule bisimSymmetric)
                    apply(rule bisimTransitive[OF bisimParComm])
                    apply(rule bisimParPres)
                    apply(rule bisimSymmetric)
                    apply(rule bisimTransitive[OF bisimParComm])
                    apply(rule bisimSymmetric[OF bangExt, OF `guarded P`])
                    done
                qed      
                show ?case
                  apply(rule exI[where x=P''])
                  apply(rule conjI)
                  apply fact
                  apply(rule lol')
                  apply(rule refl)
                  apply(rule `\<Psi> \<rhd> P'' \<sim> R`)
                  apply(rule `(\<Psi>,R,R') \<in> ?X1 \<or> \<Psi> \<rhd> R \<sim> R'`)
                  apply(rule `\<Psi> \<rhd> R' \<sim> P \<parallel> P'`)
                  apply(simp add: `A\<^isub>Q \<sharp>* P''` `A\<^isub>Q \<sharp>* P'` `A\<^isub>Q \<sharp>* R` `A\<^isub>Q \<sharp>* R'` `A\<^isub>Q \<sharp>* P`)
                  done
              qed
              then obtain Q'' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q'' \<and> (\<Psi>,Q'',P',A\<^isub>Q) \<in> ?YA1" unfolding psi'def
                by(drule exE)
              hence "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q''" and "(\<Psi>,Q'',P',A\<^isub>Q) \<in> ?YA1"
                by simp+
              have "A\<^isub>Q \<sharp>* \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)" using `A\<^isub>Q \<sharp>* (a,M\<^isub>a,\<Psi>\<^isub>c)`
                by simp
              from `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q''` `xvec \<sharp>* K` `distinct(xvec)` `A\<^isub>Q \<sharp>* \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)` `A\<^isub>Q \<sharp>* xvec` have "A\<^isub>Q \<sharp>* Q''"
                by(rule outputFreshChainDerivative)
              from `(\<Psi>,Q'',P',A\<^isub>Q) \<in> ?YA1` obtain Q''' P'' where "\<Psi> \<rhd> Q'' \<sim> Q'''" and "(\<Psi>,Q''',P'')\<in>?X1 \<or> \<Psi> \<rhd> Q''' \<sim> P''" and "\<Psi> \<rhd> P'' \<sim> P'" and "A\<^isub>Q \<sharp>* P'" and "A\<^isub>Q \<sharp>* Q''" and "A\<^isub>Q \<sharp>* P''" and "A\<^isub>Q \<sharp>* Q'''"
                by auto blast
              obtain b::name where "b \<sharp> P" and "b \<sharp> Q" and "b \<sharp> \<Psi>" and "b \<sharp> M\<^isub>a" "b \<sharp> \<Psi>\<^isub>c" and "b \<sharp> \<Psi>\<^isub>Q" and "b \<sharp> Q'" and "b \<sharp> Q''" and "b \<sharp> Q'''" and "b \<sharp> P'" and "b \<sharp> A\<^isub>Q" and "b \<sharp> A\<^isub>P" and "b \<sharp> \<Psi>\<^isub>P" and "b \<sharp> \<Psi>\<^isub>Q" and "b \<sharp> M" and "b \<sharp> K" and "b \<sharp> xvec" and "b \<sharp> N"
                by(generate_fresh "name") force
              from `extractFrame (!P) = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>` have "\<Psi>\<^isub>P = \<one>"
                by simp
              have unb: "\<langle>[b], [(a, b)] \<bullet> \<Psi>\<^isub>c\<rangle> = \<langle>[a], \<Psi>\<^isub>c\<rangle>"
                by(subst frameChainAlpha[where xvec="[a]" and p="[(a,b)]"]) (simp add: swap_simps `b \<sharp> \<Psi>\<^isub>c`)+
              have "\<Psi> \<otimes> \<one> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<rhd> Q \<longmapsto> M\<lparr>N\<rparr> \<prec> Q'"
              proof -
                {
                  fix F::"'b frame"
                  obtain A\<^isub>F \<Psi>\<^isub>F where "F=\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>" and "distinct A\<^isub>F" and "A\<^isub>F \<sharp>* \<Psi>" and "A\<^isub>F \<sharp>* \<Psi>\<^isub>P" and "A\<^isub>F \<sharp>* a" and "A\<^isub>F \<sharp>* b" and "A\<^isub>F \<sharp>* \<Psi>\<^isub>c"
                    apply(rule freshFrame[where C="(\<Psi>,\<Psi>\<^isub>P,a,b,\<Psi>\<^isub>c)"])
                    by auto
                  have lhs: "(\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>\<epsilon>, \<Psi> \<otimes> \<Psi>\<^isub>P\<rangle>) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>\<rangle>)"
                    apply(subst mergeFrameSimps)
                    apply(rule freshCompChain)
                    apply fact+
                    apply(subst `\<Psi>\<^isub>P=\<one>`)
                    apply(rule frameResChainPres)
                    apply(subst frameNilStatEq)
                    apply(rule AssertionStatEqTrans[OF AssertionStatEqSym[OF Associativity]])
                    by(rule Identity)
                  obtain c::name where "c \<sharp> A\<^isub>F" and "c \<sharp> \<Psi>\<^isub>F" and "c \<noteq> b" and "c \<noteq> a" and "c \<sharp> \<Psi>"  and "c \<sharp> \<Psi>\<^isub>c"
                    apply(generate_fresh "name")
                    by force
                  have af: "([(a,c)]\<bullet>A\<^isub>F) = A\<^isub>F" using `A\<^isub>F \<sharp>* a` `c \<sharp> A\<^isub>F`
                    by simp
                  have "((\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[c], \<Psi> \<otimes> \<one> \<otimes> ([(a, c)] \<bullet> \<Psi>\<^isub>c)\<rangle>)) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>\<rangle>) \<otimes>\<^sub>F (\<langle>[c], ([(a,c)]\<bullet>\<Psi>\<^isub>c)\<rangle>)"
                    apply(subst (1 2) mergeFrames)
                    apply(simp add: `c \<sharp> A\<^isub>F`)
                    apply(subst af[symmetric])
                    apply(subst fresh_star_bij)
                    apply fact
                    apply(rule freshCompChain)
                    apply(simp add: `c \<sharp> \<Psi>\<^isub>F`)
                    apply(simp add: `c \<sharp> \<Psi>`)
                    apply(rule freshCompChain)
                    apply fact
                    apply(rule freshCompChain)
                    apply simp
                    apply(subst af[symmetric])
                    apply(subst fresh_star_bij)
                    apply fact
                    apply(simp add: `c \<sharp> \<Psi>\<^isub>F`)
                    apply(rule frameResChainPres)
                    apply(subst frameNilStatEq)
                    apply(rule AssertionStatEqSym)
                    apply(rule AssertionStatEqTrans[OF Associativity])
                    apply(rule compositionSym)
                    apply(rule compositionSym)
                    apply(rule AssertionStatEqSym)
                    apply(rule AssertionStatEqTrans[OF Commutativity])
                    by(rule Identity)
                  hence "((\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[a], \<Psi> \<otimes> \<one> \<otimes> \<Psi>\<^isub>c\<rangle>)) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>\<rangle>) \<otimes>\<^sub>F (\<langle>[a], \<Psi>\<^isub>c\<rangle>)"
                    apply(subst (1 2) frameChainAlpha'[where A\<^isub>P="[a]" and p="[(a,c)]"])
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `c \<sharp> \<Psi>\<^isub>c`)
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                    apply(rule freshCompChain)
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `a \<sharp> \<Psi>` `c \<sharp> \<Psi>`)
                    apply(rule freshCompChain)
                    apply simp
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `c \<sharp> \<Psi>\<^isub>c`)
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] eqvts `a \<sharp> \<Psi>` `c \<sharp> \<Psi>`)
                    done
                  hence rhs: "((\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[b], \<Psi> \<otimes> \<one> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rangle>)) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>\<rangle>) \<otimes>\<^sub>F (\<langle>[b], ([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rangle>)"
                    apply(subst unb)
                    apply(subst frameChainAlpha'[where A\<^isub>P="[b]" and p="[(b,a)]"])
                    apply(rule freshCompChain)
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `a \<sharp> \<Psi>` `b \<sharp> \<Psi>`)
                    apply(rule freshCompChain)
                    apply simp
                    apply(subst pt3[OF pt_name_inst, OF at_ds5, OF at_name_inst])
                    apply(subst pt_fresh_star_bij[OF pt_name_inst, OF at_name_inst])
                    apply(simp add: `b \<sharp> \<Psi>\<^isub>c`)
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                    apply(subst (1 2) pt3[OF pt_name_inst, OF at_ds5, OF at_name_inst])
                    by(force simp add: eqvts `a \<sharp> \<Psi>` `b \<sharp> \<Psi>` at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                  have "(F \<otimes>\<^sub>F (\<langle>\<epsilon>, \<Psi> \<otimes> \<Psi>\<^isub>P\<rangle>)) \<simeq>\<^sub>F (F \<otimes>\<^sub>F (\<langle>[b], \<Psi> \<otimes> \<one> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rangle>))"
                    unfolding `F = \<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>`
                    apply(rule FrameStatEqTrans[OF lhs])
                    apply(rule FrameStatEqSym)
                    apply(rule FrameStatEqTrans[OF rhs])
                    apply(subst unb)
                    by(rule Equiv1)
                }
                show ?thesis
                  by(rule frameStatEqTransition[where A="[b]"]) (fact|simp add: `b \<sharp> M` `b \<sharp> N` `b \<sharp> Q`)+
              qed
              have "extractFrame(\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<langle>[b],\<one> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rangle>"
                by(subst alphaRes[where y=b]) (simp add: `b \<sharp> M\<^isub>a` `b \<sharp> \<Psi>\<^isub>c`)+
              have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<one> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> M \<leftrightarrow> K"
              proof -
                have "(\<langle>[], \<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<Psi>\<^isub>P\<rangle>) \<simeq>\<^sub>F (\<langle>[b], \<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<one> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rangle>)"
                proof -
                  have lhs: "(\<langle>[], \<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<Psi>\<^isub>P\<rangle>) \<simeq>\<^sub>F (\<langle>[], \<Psi> \<otimes> \<Psi>\<^isub>Q\<rangle>)" unfolding `\<Psi>\<^isub>P = \<one>`
                    by (metis AssertionStatEqSym FrameStatEqSym Identity compositionSym frameNilStatEq frameResChainbase)
                  have rhs: "(\<langle>[b], \<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<one> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rangle>) \<simeq>\<^sub>F (\<langle>[], \<Psi> \<otimes> \<Psi>\<^isub>Q\<rangle>) \<otimes>\<^sub>F (\<langle>[a], \<Psi>\<^isub>c\<rangle>)" unfolding unb[symmetric] frameResChainbase
                    apply(subst mergeFrameSimps')
                    apply(force simp add: `b \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q`)
                    apply(rule frameResChainPres)
                    apply(subst frameNilStatEq)
                    apply(rule AssertionStatEqTrans[OF AssertionStatEqSym[OF Associativity]])
                    apply(rule compositionSym)
                    apply(rule AssertionStatEqTrans[OF Commutativity])
                    by(rule Identity)
                  show ?thesis
                    apply(rule FrameStatEqTrans[OF lhs])
                    apply(rule FrameStatEqSym)
                    apply(rule FrameStatEqTrans[OF rhs])
                    by fact
                qed
                with `\<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<Psi>\<^isub>P \<turnstile> M \<leftrightarrow> K` `b \<sharp> M` `b \<sharp> K` show ?thesis 
                  by(rule_tac frameImpE[where A\<^isub>F="[b]"]) (auto dest: frameStatEqEnt)
              qed
              have Trans: "\<Psi> \<rhd> Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<tau> \<prec> \<lparr>\<nu>*xvec\<rparr>(Q' \<parallel> Q'')"
                by(rule Comm1[where A\<^isub>Q="[b]" and \<Psi>\<^isub>Q="\<one> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)"]) (fact | simp add: `b \<sharp> A\<^isub>Q` `b \<sharp> \<Psi>` `b \<sharp> Q` `b \<sharp> K` `xvec \<sharp>* Q` `b \<sharp> M\<^isub>a` `b \<sharp> \<Psi>\<^isub>c` fresh_abs_fun_iff[OF pt_name_inst,OF at_name_inst, OF fin_supp])+
              have "(\<Psi>,Q'\<parallel>Q'',Q'\<parallel>P') \<in> ?Y1" using `(\<Psi>,Q''',P'')\<in>?X1 \<or> \<Psi> \<rhd> Q''' \<sim> P''`
              proof(rule disjE)
                assume "\<Psi> \<rhd> Q''' \<sim> P''"
                hence "\<Psi> \<rhd> Q'' \<sim> P'" using `\<Psi> \<rhd> Q'' \<sim> Q'''` `\<Psi> \<rhd> P'' \<sim> P'`
                  by(auto dest: bisimTransitive)
                have "\<Psi> \<rhd> Q'\<parallel>Q'' \<sim> Q'\<parallel>P'"
                  by(rule bisimParPresSym) fact
                show ?thesis
                  apply(rule lol)
                  by(rule refl|rule `\<Psi> \<rhd> Q'\<parallel>Q'' \<sim> Q'\<parallel>P'`|rule disjI2|rule bisimReflexive)+
              next
                assume "(\<Psi>, Q''', P'') \<in> ?X1"
                then obtain a P Q\<^isub>i M\<^isub>a \<Psi>\<^isub>c xvec where q''_def: "Q''' = \<lparr>\<nu>*xvec\<rparr>(Q\<^isub>i \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))" and p''_def: "P'' = \<lparr>\<nu>*xvec\<rparr>(Q\<^isub>i \<parallel> !P)" and "a \<sharp> P" and "a \<sharp> Q\<^isub>i" and "a \<sharp> \<Psi>" and "\<not> a \<sharp> M\<^isub>a" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> R \<longrightarrow> (R = (P \<parallel> \<langle>M\<^isub>a\<rangle>)))" and "guarded P" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<longrightarrow> (supp M\<^isub>a) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P'))" and "((supp P)::name set) \<subseteq> supp M\<^isub>a" and "((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a" and "xvec \<sharp>* \<Psi>" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
                  by auto
                obtain p::"name prm" where "distinctPerm p" and "set p \<subseteq> set xvec \<times> set(p\<bullet>xvec)" "(p\<bullet>xvec) \<sharp>* Q'" and "(p\<bullet>xvec) \<sharp>* \<Psi>" and "(p\<bullet>xvec) \<sharp>* (Q\<^isub>i \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))" and "(p\<bullet>xvec) \<sharp>* (Q\<^isub>i \<parallel> !P)"
                  by(rule chain.name_list_avoiding[where xvec=xvec and c="(Q',\<Psi>,Q\<^isub>i \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>),Q\<^isub>i \<parallel> !P)"]) auto
                obtain b::name where "b \<sharp> P" and "b \<sharp> Q'" and "b \<sharp> \<Psi>" and "b \<sharp> M\<^isub>a" "b \<sharp> \<Psi>\<^isub>c" and "b \<sharp> (p\<bullet>a)" and "b \<sharp> (p\<bullet>M\<^isub>a)" and "b \<sharp> (p\<bullet>\<Psi>\<^isub>c)" and "b \<sharp> (p\<bullet>P)" and "b \<sharp> Q" and "b \<sharp> (p\<bullet>Q\<^isub>i)"
                  apply(rule at_exists_fresh[OF at_name_inst, where x="(P,Q',\<Psi>,M\<^isub>a,\<Psi>\<^isub>c,p\<bullet>a,p\<bullet>M\<^isub>a,p\<bullet>\<Psi>\<^isub>c,p\<bullet>P,p\<bullet>Q\<^isub>i,Q)"])
                  unfolding supp_prod pt_perm_supp[OF pt_name_inst,OF at_name_inst, symmetric]
                  apply(rule finite_UnI, (finite_guess | (subst pt_set_finite_ineq[OF pt_name_inst,OF at_name_inst], finite_guess)))+
                  apply(finite_guess)
                  by force
                have "\<Psi> \<rhd> Q'\<parallel>Q'' \<sim> \<lparr>\<nu>*(p\<bullet>xvec)\<rparr>((Q' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>p\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                proof -
                  have "\<Psi> \<rhd> Q'\<parallel>Q'' \<sim> Q'\<parallel>Q'''"
                    apply(rule bisimParPresSym)
                    by fact
                  moreover have "\<Psi> \<rhd> Q'\<parallel>Q''' \<sim> Q'\<parallel>(\<lparr>\<nu>*xvec\<rparr>Q\<^isub>i \<parallel>\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))"
                    unfolding q''_def
                    by(rule bisimReflexive)
                  moreover have "\<Psi> \<rhd> Q'\<parallel>(\<lparr>\<nu>*xvec\<rparr>Q\<^isub>i \<parallel>\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) \<sim> Q'\<parallel>(\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>((p\<bullet>Q\<^isub>i)\<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>p\<bullet>\<Psi>\<^isub>c)\<rbrace>)))"
                    apply(subst resChainAlpha[where p=p])
                    apply fact
                    apply fact
                    unfolding eqvts
                    apply(subst alphaRes[where x="p\<bullet>a" and y=b])
                    apply(simp add: `b \<sharp> (p\<bullet>M\<^isub>a)` `b \<sharp> (p\<bullet>\<Psi>\<^isub>c)`)
                    unfolding eqvts
                    by(rule bisimReflexive)
                  moreover have "\<Psi> \<rhd> Q'\<parallel>(\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>((p\<bullet>Q\<^isub>i)\<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>p\<bullet>\<Psi>\<^isub>c)\<rbrace>))) \<sim> (\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(Q' \<parallel>((p\<bullet>Q\<^isub>i)\<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>p\<bullet>\<Psi>\<^isub>c)\<rbrace>))))"
                    by(rule bisimSymmetric[OF bisimScopeExtChain]) fact+
                  moreover have "\<Psi> \<rhd> (\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(Q' \<parallel>((p\<bullet>Q\<^isub>i)\<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>p\<bullet>\<Psi>\<^isub>c)\<rbrace>)))) \<sim> \<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(Q' \<parallel>(p\<bullet>Q\<^isub>i))\<parallel>(\<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>p\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                    apply(rule bisimResChainPres)
                    apply(rule bisimSymmetric[OF bisimParAssoc])
                    by fact
                  ultimately show ?thesis
                    by(auto dest: bisimTransitive)
                qed
                moreover have "\<Psi> \<rhd> \<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(Q'\<parallel>(p\<bullet>Q\<^isub>i))\<parallel>!(p\<bullet>P) \<sim> Q' \<parallel> P'"
                proof -
                  have "\<Psi> \<rhd> Q' \<parallel> P'' \<sim> Q' \<parallel> P'"
                    apply(rule bisimParPresSym)
                    by fact+
                  moreover have "\<Psi> \<rhd> Q'\<parallel>(\<lparr>\<nu>*xvec\<rparr>Q\<^isub>i \<parallel>!P) \<sim> Q'\<parallel>P''"
                    unfolding p''_def
                    by(rule bisimReflexive)
                  moreover have "\<Psi> \<rhd> Q'\<parallel>(\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>((p\<bullet>Q\<^isub>i)\<parallel> !(p\<bullet>P))) \<sim> Q'\<parallel>(\<lparr>\<nu>*xvec\<rparr>Q\<^isub>i \<parallel>!P)"
                    apply(subst resChainAlpha[where p=p and xvec=xvec])
                    apply fact
                    apply fact
                    unfolding eqvts
                    by(rule bisimReflexive)
                  moreover have "\<Psi> \<rhd> \<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(Q' \<parallel>((p\<bullet>Q\<^isub>i)\<parallel> (!(p\<bullet>P)))) \<sim> Q'\<parallel>(\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>((p\<bullet>Q\<^isub>i)\<parallel> !(p\<bullet>P)))"                    
                    by(rule bisimScopeExtChain) fact+
                  moreover have "\<Psi> \<rhd> \<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(Q' \<parallel>(p\<bullet>Q\<^isub>i))\<parallel>(!(p\<bullet>P)) \<sim> \<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(Q' \<parallel>((p\<bullet>Q\<^isub>i)\<parallel> (!(p\<bullet>P))))"
                    apply(rule bisimResChainPres)
                    by(rule bisimParAssoc) fact
                  ultimately show ?thesis
                    by(metis bisimTransitive)
                qed
                moreover have "(\<Psi>,\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>((Q' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>p\<bullet>\<Psi>\<^isub>c)\<rbrace>)),\<lparr>\<nu>*p \<bullet> xvec\<rparr>Q' \<parallel> (p \<bullet> Q\<^isub>i) \<parallel> !(p \<bullet> P)) \<in> ?X1"
                proof -
                  have "(p\<bullet>a) \<sharp> (p\<bullet>P)" using `a \<sharp> P` by(subst fresh_bij) 
                  note `b \<sharp> \<Psi>`
                  moreover note `b \<sharp> (p\<bullet>P)`
                  moreover have "b \<sharp> (Q'\<parallel>(p\<bullet>Q\<^isub>i))" using `b \<sharp> Q'` `b \<sharp>(p\<bullet>Q\<^isub>i)`
                    by simp
                  moreover have "guarded(p\<bullet>P)" using `guarded P` by simp
                  moreover have "\<not> b \<sharp> ([(p\<bullet>a,b)] \<bullet> (p\<bullet>M\<^isub>a))" using `\<not> a \<sharp> M\<^isub>a`
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps)
                  moreover have "\<And> \<Psi>. \<Psi> \<otimes> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)) \<turnstile> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>(p\<bullet>P) \<parallel> \<langle>([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<rangle>" using EntCl
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst perm_bool[where pi="rev p",symmetric])
                    by(simp add: eqvts `(p\<bullet>a) \<sharp> (p\<bullet>P)` `b \<sharp> (p\<bullet>P)`)
                  moreover have "\<forall>F. (F \<otimes>\<^sub>F (\<langle>[b],[(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)\<rangle>) \<simeq>\<^sub>F F)"
                    using Equiv1
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst perm_bool[where pi="rev p",symmetric])
                    by(simp add: eqvts swap_simps)
                  moreover have "(\<forall>\<Psi> R. b \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)) \<turnstile> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a)) \<Leftarrow> R \<longrightarrow> (R = ((p\<bullet>P) \<parallel> \<langle>([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<rangle>)))" using Equiv2
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts `(p\<bullet>a) \<sharp> p\<bullet>P` `b \<sharp> p\<bullet>P` swap_simps)
                  moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>P \<longrightarrow> (supp ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))) \<subseteq> ((supp \<Psi>)::name set))" using EntSupp
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps)
                  moreover have "\<forall> P P'. ((([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>P = ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>P') = (P = P'))" using Inj
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps)
                  moreover have "((supp(p\<bullet>P))::name set) \<subseteq> supp ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))" using `((supp P)::name set) \<subseteq> supp M\<^isub>a`
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps `(p\<bullet>a) \<sharp> p\<bullet>P` `b \<sharp> p\<bullet>P`)
                  moreover have "((supp ([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)))::name set) = supp ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))" using `((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a`
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                  moreover note `(p\<bullet>xvec)\<sharp>*\<Psi>`
                  moreover from EntEnt have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>(p\<bullet>P)\<parallel>\<langle>([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps `distinctPerm p` `(p\<bullet>a) \<sharp> p\<bullet>P` `b \<sharp> p\<bullet>P`)
                  ultimately show ?thesis
                    by blast
                qed
                ultimately show ?thesis
                  apply(rule_tac lol)
                  by(rule refl|assumption|rule disjI1)+
              qed
              hence "(\<Psi>,Q'\<parallel>Q'',Q'\<parallel>P') \<in> ?Y"
                by(rule Y1inY)
              hence "(\<Psi>,\<lparr>\<nu>*xvec\<rparr>Q'\<parallel>Q'',\<lparr>\<nu>*xvec\<rparr>Q'\<parallel>P') \<in> ?Y" using `xvec \<sharp>* \<Psi>`
                by(rule YresChainClose')
              with Trans show ?case
                by(rule_tac exI) (rule conjI)
          next
            case(cComm2 \<Psi>\<^isub>P M xvec N Q' A\<^isub>Q \<Psi>\<^isub>Q K P' A\<^isub>P)
            hence "bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* M\<^isub>a" and "A\<^isub>Q \<sharp>* bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>)" and "bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)"
              by auto
            have "a \<sharp> \<Psi>\<^isub>Q" using `a \<sharp> Q` `extractFrame Q = \<langle>A\<^isub>Q,\<Psi>\<^isub>Q\<rangle>` `A\<^isub>Q \<sharp>* (a, M\<^isub>a,\<Psi>\<^isub>c)`
            proof -
              have "a \<sharp> \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>" using `a \<sharp> Q` 
                apply(subst `extractFrame Q = \<langle>A\<^isub>Q,\<Psi>\<^isub>Q\<rangle>`[symmetric])
                by(rule extractFrameFresh)
              hence "a \<sharp> \<langle>\<epsilon>, \<Psi>\<^isub>Q\<rangle>" using `A\<^isub>Q \<sharp>* (a, M\<^isub>a,\<Psi>\<^isub>c)`
                apply(subst (asm) frameResChainFresh)
                by(simp add: fresh_def frame.supp name_list_supp)
              thus ?thesis
                by simp
            qed
            obtain \<Psi>' where psi'def: "\<Psi>'=\<Psi>\<otimes>\<Psi>\<^isub>Q" by simp
            from `A\<^isub>Q \<sharp>* (a, M\<^isub>a,\<Psi>\<^isub>c)` have "a \<sharp> A\<^isub>Q" and "A\<^isub>Q \<sharp>* a" and "A\<^isub>Q \<sharp>* M\<^isub>a" and "A\<^isub>Q \<sharp>* \<Psi>\<^isub>c"
              by simp+
            from `A\<^isub>Q \<sharp>* !P` have "A\<^isub>Q \<sharp>* P"
              by simp
            thm cComm2
            from `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K\<lparr>N\<rparr> \<prec> P'` have "\<Psi>' \<rhd> !P \<longmapsto> K\<lparr>N\<rparr> \<prec> P'"
              unfolding psi'def by-
            from `bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)` have "bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* \<Psi>'"
              unfolding psi'def by-
            from `a \<sharp> \<Psi>` `a \<sharp> \<Psi>\<^isub>Q` have "a \<sharp> \<Psi>'"
              unfolding psi'def by force
            from `xvec \<sharp>* \<Psi>` have "bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* \<Psi>" by simp
            from `\<Psi>' \<rhd> !P \<longmapsto> K\<lparr>N\<rparr> \<prec> P'`
            have "\<exists> Q'. \<Psi>' \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>N\<rparr> \<prec> Q' \<and> (\<Psi>,Q',P',A\<^isub>Q) \<in> ?YA1" using `A\<^isub>Q \<sharp>* P'`
            proof(induct rule: bangInputInduct)
              case cBang
              thus ?case by auto
            next
              case(cPar1 P')
              obtain b::name where "b \<sharp> \<Psi>" and "b \<sharp> \<Psi>'" and "b \<sharp> P" and "b \<sharp> K" and "b \<sharp> xvec" and "b \<sharp> N" and "b \<noteq> a" and "b \<sharp> M\<^isub>a" and "b \<sharp> \<Psi>\<^isub>c" and "b \<sharp> A\<^isub>Q" and "b \<sharp> P'"
                apply(generate_fresh "name")
                by force
              have unb: "\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) = \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                apply(subst alphaRes[where x=a and y=b])
                apply(simp add: `b \<sharp> M\<^isub>a` `b \<sharp> \<Psi>\<^isub>c`)
                unfolding eqvts
                by(rule refl)
              have "\<Psi>' \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>" using `\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>`
                apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                by(simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
              have "(\<Psi>' \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one> \<rhd> P \<longmapsto> K\<lparr>N\<rparr> \<prec> P'"
              proof -
                {
                  fix F::"'b frame"
                  obtain A\<^isub>F \<Psi>\<^isub>F where "F=\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>" and "distinct A\<^isub>F" and "A\<^isub>F \<sharp>* \<Psi>" and "A\<^isub>F \<sharp>* \<Psi>'" and "A\<^isub>F \<sharp>* a" and "A\<^isub>F \<sharp>* b" and "A\<^isub>F \<sharp>* \<Psi>\<^isub>c"
                    apply(rule freshFrame[where C="(\<Psi>,\<Psi>',a,b,\<Psi>\<^isub>c)"])
                    by auto
                  have lhs: "(\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>\<epsilon>, \<Psi>'\<rangle>) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>'\<rangle>)"
                    apply(subst mergeFrameSimps)
                    apply fact
                    apply(rule FrameStatEqRefl)
                    done
                  obtain c::name where "c \<sharp> A\<^isub>F" and "c \<sharp> \<Psi>\<^isub>F" and "c \<noteq> b" and "c \<noteq> a" and "c \<sharp> \<Psi>"  and "c \<sharp> \<Psi>\<^isub>c" and "c \<sharp> \<Psi>'"
                    apply(generate_fresh "name")
                    by force
                  have af: "([(a,c)]\<bullet>A\<^isub>F) = A\<^isub>F" using `A\<^isub>F \<sharp>* a` `c \<sharp> A\<^isub>F`
                    by simp
                  have "((\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[c], (\<Psi>' \<otimes> ([(a, c)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one>\<rangle>)) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>'\<rangle>) \<otimes>\<^sub>F (\<langle>[c], ([(a,c)]\<bullet>\<Psi>\<^isub>c)\<rangle>)"
                    apply(subst (1 2) mergeFrames)
                    apply(simp add: `c \<sharp> A\<^isub>F`)
                    apply(subst af[symmetric])
                    apply(subst fresh_star_bij)
                    apply fact
                    apply(rule freshCompChain)
                    apply(simp add: `c \<sharp> \<Psi>\<^isub>F`)
                    apply(simp add: `c \<sharp> \<Psi>'`)
                    apply(rule freshCompChain)
                    apply(rule freshCompChain)
                    apply fact
                    apply(subst af[symmetric])
                    apply(subst fresh_star_bij)
                    apply fact
                    apply simp
                    apply(simp add: `c \<sharp> \<Psi>\<^isub>F`)
                    apply(rule frameResChainPres)
                    apply(subst frameNilStatEq)
                    apply(rule AssertionStatEqSym)
                    apply(rule AssertionStatEqTrans[OF Associativity])
                    apply(rule compositionSym)
                    by(rule AssertionStatEqSym[OF Identity])
                  hence "((\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[a], (\<Psi>' \<otimes> \<Psi>\<^isub>c) \<otimes> \<one>\<rangle>)) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>'\<rangle>) \<otimes>\<^sub>F (\<langle>[a], \<Psi>\<^isub>c\<rangle>)"
                    apply(subst (1 2) frameChainAlpha'[where A\<^isub>P="[a]" and p="[(a,c)]"])
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `c \<sharp> \<Psi>\<^isub>c`)
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                    apply(rule freshCompChain)
                    apply(rule freshCompChain)
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `a \<sharp> \<Psi>'` `c \<sharp> \<Psi>'`)
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `c \<sharp> \<Psi>\<^isub>c`)
                    apply simp
                    by(simp add: `a \<sharp> \<Psi>'` `c \<sharp> \<Psi>'` swap_simps eqvts)
                  hence rhs: "((\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[b], (\<Psi>' \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one>\<rangle>)) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>'\<rangle>) \<otimes>\<^sub>F (\<langle>[b], ([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rangle>)"
                    apply(subst frameChainAlpha'[where A\<^isub>P="[b]" and p="[(b,a)]"])
                    apply(rule freshCompChain)
                    apply(rule freshCompChain)
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `a \<sharp> \<Psi>'` `b \<sharp> \<Psi>'`)
                    apply(subst pt3[OF pt_name_inst, OF at_ds5, OF at_name_inst])
                    apply(subst pt_fresh_star_bij[OF pt_name_inst, OF at_name_inst])
                    apply(simp add: `b \<sharp> \<Psi>\<^isub>c`)
                    apply simp
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                    apply(subst frameChainAlpha'[where A\<^isub>P="[b]" and p="[(b,a)]"])
                    apply(subst pt3[OF pt_name_inst, OF at_ds5, OF at_name_inst])
                    apply(subst pt_fresh_star_bij[OF pt_name_inst, OF at_name_inst])
                    apply(simp add: `b \<sharp> \<Psi>\<^isub>c`)
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                    apply(simp add: eqvts at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `a \<sharp> \<Psi>'` `b \<sharp> \<Psi>'` pt3[OF pt_name_inst, OF at_ds5[where a=b and b=a], OF at_name_inst])
                    done
                  have unb: "\<langle>[a],\<Psi>\<^isub>c\<rangle> = (\<langle>[b],([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rangle>)"
                    apply(subst frameChainAlpha'[where A\<^isub>P="[a]" and p="[(a,b)]"])
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `b \<sharp> \<Psi>\<^isub>c`)
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                    by(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                  have "(F \<otimes>\<^sub>F (\<langle>\<epsilon>, \<Psi>'\<rangle>)) \<simeq>\<^sub>F (F \<otimes>\<^sub>F (\<langle>[b], (\<Psi>' \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<one>\<rangle>))"
                    unfolding `F = \<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>`
                    apply(rule FrameStatEqTrans[OF lhs])
                    apply(rule FrameStatEqSym)
                    apply(rule FrameStatEqTrans[OF rhs])
                    apply(subst unb[symmetric])
                    by fact
                }
                show ?thesis
                  apply(rule frameStatEqTransition[where A="[b]" and \<Psi>="\<Psi>'"])
                  by(fact|simp add: `xvec \<sharp>* K` `b \<sharp> K` `b \<sharp> xvec` `b \<sharp> N`)+
              qed
              have "supp (P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>) \<subseteq> (supp ([(a, b)] \<bullet> M\<^isub>a)::name set)" using `((supp P)::name set) \<subseteq> supp M\<^isub>a`
                apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                by(simp add: eqvts psi.supp `a \<sharp> P` `b \<sharp> P`)
              have "guarded (P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>)" using `guarded P`
                by simp
              have "\<Psi>' \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>N\<rparr> \<prec> \<lparr>\<nu>b\<rparr>((P' \<parallel> (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)) \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                unfolding unb
                apply(rule Scope)
                apply(rule Par1[where \<Psi>\<^isub>Q="([(a, b)] \<bullet> \<Psi>\<^isub>c)" and A\<^isub>Q="[]"])
                apply(rule Inv[where P="P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>"])
                apply fact
                apply(rule Par1[where \<Psi>\<^isub>Q="\<one>" and A\<^isub>Q="[]"])
                by(fact|simp|rule conjI)+
              moreover have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((P' \<parallel> (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)) \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> P' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)"
              proof -
                have lhs: "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((P' \<parallel> (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)) \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>b\<rparr>(P' \<parallel> ((\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>) \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                  apply(rule bisimResPres)
                  apply(rule bisimParAssoc)
                  by fact
                show ?thesis
                  unfolding unb
                  apply(rule bisimTransitive[OF lhs])
                  apply(rule bisimScopeExt)
                  by fact
              qed
              moreover have "\<Psi> \<rhd> P' \<parallel> !P \<sim> P' \<parallel> !P"
                by(rule bisimReflexive)
              moreover have "(\<Psi>, P' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>), P' \<parallel> !P) \<in> ?X1 \<or> \<Psi> \<rhd> P' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<sim> P' \<parallel> !P"
              proof -
                note `b \<sharp> \<Psi>`
                moreover note `b \<sharp> P` `b \<sharp> P'`
                moreover have "\<not> b \<sharp> [(a, b)] \<bullet> M\<^isub>a" using `\<not> a \<sharp> M\<^isub>a`
                  apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                  by(force simp add: eqvts at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                moreover note `guarded P`
                moreover have "supp P \<subseteq> (supp ([(a, b)] \<bullet> M\<^isub>a)::name set)" using `supp P \<subseteq> supp M\<^isub>a`
                  apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                  by(simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
                moreover have "supp ([(a, b)] \<bullet> \<Psi>\<^isub>c) = (supp ([(a, b)] \<bullet> M\<^isub>a)::name set)" using `supp (\<Psi>\<^isub>c) = supp (M\<^isub>a)`
                  apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                  by(simp add: eqvts)
                moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<longrightarrow> (supp ([(a,b)]\<bullet>M\<^isub>a)) \<subseteq> ((supp \<Psi>)::name set))" using `\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<longrightarrow> (supp M\<^isub>a) \<subseteq> ((supp \<Psi>)::name set))`
                  apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                  by(simp add: eqvts)
                moreover have "\<forall>\<Psi>. \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>" using `\<And>\<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>`
                  apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                  by(simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
                moreover have "\<forall> F. F \<otimes>\<^sub>F (\<langle>[b],([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rangle>) \<simeq>\<^sub>F F" using `\<And> F. F \<otimes>\<^sub>F (\<langle>[a],\<Psi>\<^isub>c\<rangle>) \<simeq>\<^sub>F F`
                  apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                  by(force simp add: eqvts at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                moreover have "\<forall> P P'. ((([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P = ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P') = (P = P'))" using `\<forall>P P'. (M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P')`
                  apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                  by(simp add: eqvts)
                moreover have "\<forall> \<Psi> R. b \<sharp> \<Psi> \<longrightarrow> (\<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a) \<Leftarrow> R \<longrightarrow> R = (P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>))" using `\<And> \<Psi> R. \<lbrakk>a \<sharp> \<Psi>; \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>R\<rbrakk> \<Longrightarrow> R = P \<parallel> \<langle>M\<^isub>a\<rangle>`
                  apply(subst perm_bool[symmetric,where pi="[(a,b)]"])
                  by(simp add: eqvts `a \<sharp> P` `b \<sharp> P` at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                moreover have "([]::name list) \<sharp>* \<Psi>"
                  by simp
                moreover from EntEnt have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                  by(subst perm_bool[where pi="[(a,b)]",symmetric]) (simp add: eqvts swap_simps `a \<sharp> P` `b \<sharp> P`)
                ultimately have "(\<Psi>, \<lparr>\<nu>*[]\<rparr>P' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>), \<lparr>\<nu>*[]\<rparr>P' \<parallel> !P) \<in> ?X1"
                  apply(subst unb)
                  by blast
                hence "(\<Psi>, P' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>), P' \<parallel> !P) \<in> ?X1"
                  unfolding unb resChain.simps
                  by -
                thus ?thesis
                  by blast
              qed
              moreover have "A\<^isub>Q \<sharp>* (\<lparr>\<nu>b\<rparr>((P' \<parallel> (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)) \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>), P' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>), P' \<parallel> !P, P' \<parallel> !P)"
              proof -
                  have "a \<sharp> A\<^isub>Q" using `A\<^isub>Q \<sharp>* a`
                    by simp
                  thus ?thesis using `b \<sharp> A\<^isub>Q` `A\<^isub>Q \<sharp>* (P' \<parallel> !P)` `A\<^isub>Q \<sharp>* M\<^isub>a` `A\<^isub>Q \<sharp>* \<Psi>\<^isub>c`
                    by(simp add: fresh_abs_fun_iff[OF pt_name_inst, OF at_name_inst, OF fin_supp])
                qed
                ultimately have "(\<Psi>, \<lparr>\<nu>b\<rparr>((P' \<parallel> (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)) \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>), P' \<parallel> !P,A\<^isub>Q) \<in> ?YA1"
                  apply -
                  apply(rule lol')
                  apply(rule refl)
                  by assumption+
                with `\<Psi>' \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>N\<rparr> \<prec> \<lparr>\<nu>b\<rparr>((P' \<parallel> (\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)) \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)` show ?case
                  by(rule_tac exI, rule conjI)
              next
                case(cPar2 P')
                from `A\<^isub>Q \<sharp>* (P \<parallel> P')` have "A\<^isub>Q \<sharp>* P'" by simp
                have "\<exists> P''. \<Psi>' \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>N\<rparr> \<prec> P'' \<and> (\<Psi>, P'', P',A\<^isub>Q) \<in> ?YA1"
                  apply(rule cPar2)
                  by fact
                then obtain P'' where "\<Psi>' \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>N\<rparr> \<prec> P'' \<and> (\<Psi>, P'', P',A\<^isub>Q) \<in> ?YA1"
                  by(rule exE)
                hence "\<Psi>' \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>N\<rparr> \<prec> P''" and "(\<Psi>, P'', P', A\<^isub>Q) \<in> ?YA1"
                  by simp+
                from `(\<Psi>, P'', P', A\<^isub>Q) \<in> ?YA1`
                obtain R R' where "\<Psi> \<rhd> P'' \<sim> R" and "(\<Psi>,R,R') \<in> ?X1 \<or> \<Psi> \<rhd> R \<sim> R'" and "\<Psi> \<rhd> R' \<sim> P'" and "A\<^isub>Q \<sharp>* R" and "A\<^isub>Q \<sharp>* R'" and "A\<^isub>Q \<sharp>* P''" and "A\<^isub>Q \<sharp>* P'"
                  by auto blast+
                have "\<Psi> \<rhd> R' \<sim> P \<parallel> P'"
                proof -
                  obtain R'' where "\<Psi> \<rhd> P' \<sim> R'' \<parallel> !P"
                    apply(rule bangDerivative'[where \<Psi>="\<Psi>'" and \<Psi>'=\<Psi> and P=P and \<alpha>="K\<lparr>N\<rparr>" and P'=P'])
                    apply(rule cPar2)
                    by simp+
                  have "\<Psi> \<rhd> R'' \<parallel> !P \<sim> R'' \<parallel> (P \<parallel> !P)"
                    apply(rule bisimTransitive[OF bisimParComm])
                    apply(rule bisimSymmetric)
                    apply(rule bisimTransitive[OF bisimParComm])
                    apply(rule bisimParPres)
                    apply(rule bisimSymmetric)
                    apply(rule bangExt[OF `guarded P`])
                    done
                  show ?thesis
                    apply(rule bisimTransitive[OF `\<Psi> \<rhd> R' \<sim> P'`])
                    apply(rule bisimTransitive[OF `\<Psi> \<rhd> P' \<sim> R'' \<parallel> !P`])
                    apply(rule bisimSymmetric)
                    apply(rule bisimTransitive[OF bisimParComm])
                    apply(rule bisimTransitive[OF bisimParPres[OF `\<Psi> \<rhd> P' \<sim> R'' \<parallel> !P`]])
                    apply(rule bisimTransitive[OF bisimParAssoc])
                    apply(rule bisimTransitive[OF bisimParComm])
                    apply(rule bisimSymmetric)
                    apply(rule bisimTransitive[OF bisimParComm])
                    apply(rule bisimParPres)
                    apply(rule bisimSymmetric)
                    apply(rule bisimTransitive[OF bisimParComm])
                    apply(rule bisimSymmetric[OF bangExt, OF `guarded P`])
                    done
                qed      
                show ?case
                  apply(rule exI[where x=P''])
                  apply(rule conjI)
                  apply fact
                  apply(rule lol')
                  apply(rule refl)
                  apply(rule `\<Psi> \<rhd> P'' \<sim> R`)
                  apply(rule `(\<Psi>,R,R') \<in> ?X1 \<or> \<Psi> \<rhd> R \<sim> R'`)
                  apply(rule `\<Psi> \<rhd> R' \<sim> P \<parallel> P'`)
                  apply(simp add: `A\<^isub>Q \<sharp>* P''` `A\<^isub>Q \<sharp>* P'` `A\<^isub>Q \<sharp>* R` `A\<^isub>Q \<sharp>* R'` `A\<^isub>Q \<sharp>* P`)
                  done
              qed
              then obtain Q'' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>N\<rparr> \<prec> Q'' \<and> (\<Psi>,Q'',P',A\<^isub>Q) \<in> ?YA1" unfolding psi'def
                by(drule exE)
              hence "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>N\<rparr> \<prec> Q''" and "(\<Psi>,Q'',P',A\<^isub>Q) \<in> ?YA1"
                by simp+
              have "A\<^isub>Q \<sharp>* \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)" using `A\<^isub>Q \<sharp>* (a,M\<^isub>a,\<Psi>\<^isub>c)`
                by simp
              from `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>N\<rparr> \<prec> Q''` `A\<^isub>Q \<sharp>* \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)` `A\<^isub>Q \<sharp>* N` have "A\<^isub>Q \<sharp>* Q''"
                by(rule inputFreshChainDerivative)
              from `(\<Psi>,Q'',P',A\<^isub>Q) \<in> ?YA1` obtain Q''' P'' where "\<Psi> \<rhd> Q'' \<sim> Q'''" and "(\<Psi>,Q''',P'')\<in>?X1 \<or> \<Psi> \<rhd> Q''' \<sim> P''" and "\<Psi> \<rhd> P'' \<sim> P'" and "A\<^isub>Q \<sharp>* P'" and "A\<^isub>Q \<sharp>* Q''" and "A\<^isub>Q \<sharp>* P''" and "A\<^isub>Q \<sharp>* Q'''"
                by auto blast
              obtain b::name where "b \<sharp> P" and "b \<sharp> Q" and "b \<sharp> \<Psi>" and "b \<sharp> M\<^isub>a" "b \<sharp> \<Psi>\<^isub>c" and "b \<sharp> \<Psi>\<^isub>Q" and "b \<sharp> Q'" and "b \<sharp> Q''" and "b \<sharp> Q'''" and "b \<sharp> P'" and "b \<sharp> A\<^isub>Q" and "b \<sharp> A\<^isub>P" and "b \<sharp> \<Psi>\<^isub>P" and "b \<sharp> \<Psi>\<^isub>Q" and "b \<sharp> M" and "b \<sharp> K" and "b \<sharp> xvec" and "b \<sharp> N"
                by(generate_fresh "name") force
              have unb: "(\<langle>[b], [(a, b)] \<bullet> \<Psi>\<^isub>c\<rangle>) = (\<langle>[a], \<Psi>\<^isub>c\<rangle>)"
                by(subst frameChainAlpha[where xvec="[a]" and p="[(a,b)]"]) (auto simp add: swap_simps `b \<sharp> \<Psi>\<^isub>c`)
              have "\<Psi>\<^isub>P=\<one>" using `extractFrame(!P) = \<langle>A\<^isub>P,\<Psi>\<^isub>P\<rangle>` by simp
              have "\<Psi> \<otimes> \<one> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<rhd> Q \<longmapsto> M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q'"
              proof -
                {
                  fix F::"'b frame"
                  obtain A\<^isub>F \<Psi>\<^isub>F where "F=\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>" and "distinct A\<^isub>F" and "A\<^isub>F \<sharp>* \<Psi>" and "A\<^isub>F \<sharp>* \<Psi>\<^isub>P" and "A\<^isub>F \<sharp>* a" and "A\<^isub>F \<sharp>* b" and "A\<^isub>F \<sharp>* \<Psi>\<^isub>c"
                    apply(rule freshFrame[where C="(\<Psi>,\<Psi>\<^isub>P,a,b,\<Psi>\<^isub>c)"])
                    by auto
                  have lhs: "(\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>\<epsilon>, \<Psi> \<otimes> \<Psi>\<^isub>P\<rangle>) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>\<rangle>)"
                    apply(subst mergeFrameSimps)
                    apply(rule freshCompChain)
                    apply fact+
                    apply(subst `\<Psi>\<^isub>P=\<one>`)
                    apply(rule frameResChainPres)
                    apply(subst frameNilStatEq)
                    apply(rule AssertionStatEqTrans[OF AssertionStatEqSym[OF Associativity]])
                    by(rule Identity)
                  obtain c::name where "c \<sharp> A\<^isub>F" and "c \<sharp> \<Psi>\<^isub>F" and "c \<noteq> b" and "c \<noteq> a" and "c \<sharp> \<Psi>"  and "c \<sharp> \<Psi>\<^isub>c"
                    apply(generate_fresh "name")
                    by force
                  have af: "([(a,c)]\<bullet>A\<^isub>F) = A\<^isub>F" using `A\<^isub>F \<sharp>* a` `c \<sharp> A\<^isub>F`
                    by simp
                  have "((\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[c], \<Psi> \<otimes> \<one> \<otimes> ([(a, c)] \<bullet> \<Psi>\<^isub>c)\<rangle>)) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>\<rangle>) \<otimes>\<^sub>F (\<langle>[c], ([(a,c)]\<bullet>\<Psi>\<^isub>c)\<rangle>)"
                    apply(subst (1 2) mergeFrames)
                    apply(simp add: `c \<sharp> A\<^isub>F`)
                    apply(subst af[symmetric])
                    apply(subst fresh_star_bij)
                    apply fact
                    apply(rule freshCompChain)
                    apply(simp add: `c \<sharp> \<Psi>\<^isub>F`)
                    apply(simp add: `c \<sharp> \<Psi>`)
                    apply(rule freshCompChain)
                    apply fact
                    apply(rule freshCompChain)
                    apply simp
                    apply(subst af[symmetric])
                    apply(subst fresh_star_bij)
                    apply fact
                    apply(simp add: `c \<sharp> \<Psi>\<^isub>F`)
                    apply(rule frameResChainPres)
                    apply(subst frameNilStatEq)
                    apply(rule AssertionStatEqSym)
                    apply(rule AssertionStatEqTrans[OF Associativity])
                    apply(rule compositionSym)
                    apply(rule compositionSym)
                    apply(rule AssertionStatEqSym)
                    apply(rule AssertionStatEqTrans[OF Commutativity])
                    by(rule Identity)
                  hence "((\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[a], \<Psi> \<otimes> \<one> \<otimes> \<Psi>\<^isub>c\<rangle>)) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>\<rangle>) \<otimes>\<^sub>F (\<langle>[a], \<Psi>\<^isub>c\<rangle>)"
                    apply(subst (1 2) frameChainAlpha'[where A\<^isub>P="[a]" and p="[(a,c)]"])
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `c \<sharp> \<Psi>\<^isub>c`)
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                    apply(rule freshCompChain)
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `a \<sharp> \<Psi>` `c \<sharp> \<Psi>`)
                    apply(rule freshCompChain)
                    apply simp
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `c \<sharp> \<Psi>\<^isub>c`)
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] eqvts `a \<sharp> \<Psi>` `c \<sharp> \<Psi>`)
                    done
                  hence rhs: "((\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[b], \<Psi> \<otimes> \<one> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rangle>)) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>\<rangle>) \<otimes>\<^sub>F (\<langle>[b], ([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rangle>)"
                    apply(subst unb)
                    apply(subst frameChainAlpha'[where A\<^isub>P="[b]" and p="[(b,a)]"])
                    apply(rule freshCompChain)
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `a \<sharp> \<Psi>` `b \<sharp> \<Psi>`)
                    apply(rule freshCompChain)
                    apply simp
                    apply(subst pt3[OF pt_name_inst, OF at_ds5, OF at_name_inst])
                    apply(subst pt_fresh_star_bij[OF pt_name_inst, OF at_name_inst])
                    apply(simp add: `b \<sharp> \<Psi>\<^isub>c`)
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                    apply(subst (1 2) pt3[OF pt_name_inst, OF at_ds5, OF at_name_inst])
                    by(force simp add: eqvts `a \<sharp> \<Psi>` `b \<sharp> \<Psi>` at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                  have "(F \<otimes>\<^sub>F (\<langle>\<epsilon>, \<Psi> \<otimes> \<Psi>\<^isub>P\<rangle>)) \<simeq>\<^sub>F (F \<otimes>\<^sub>F (\<langle>[b], \<Psi> \<otimes> \<one> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rangle>))"
                    unfolding `F = \<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>`
                    apply(rule FrameStatEqTrans[OF lhs])
                    apply(rule FrameStatEqSym)
                    apply(rule FrameStatEqTrans[OF rhs])
                    apply(subst unb)
                    by(rule Equiv1)
                }
                show ?thesis
                  by(rule frameStatEqTransition[where A="[b]"]) (fact|simp add: `b \<sharp> M` `b \<sharp> N` `b \<sharp> xvec` `b \<sharp> Q`)+
              qed
              have "extractFrame(\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<langle>[b],\<one> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rangle>"
                by(subst alphaRes[where y=b]) (simp add: `b \<sharp> M\<^isub>a` `b \<sharp> \<Psi>\<^isub>c`)+
              have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<one> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> M \<leftrightarrow> K"
              proof -
                have "(\<langle>[], \<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<Psi>\<^isub>P\<rangle>) \<simeq>\<^sub>F (\<langle>[b], \<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<one> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rangle>)"
                proof -
                  have lhs: "(\<langle>[], \<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<Psi>\<^isub>P\<rangle>) \<simeq>\<^sub>F (\<langle>[], \<Psi> \<otimes> \<Psi>\<^isub>Q\<rangle>)" unfolding `\<Psi>\<^isub>P = \<one>`
                    by (metis AssertionStatEqSym FrameStatEqSym Identity compositionSym frameNilStatEq frameResChainbase)
                  have rhs: "(\<langle>[b], \<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<one> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rangle>) \<simeq>\<^sub>F (\<langle>[], \<Psi> \<otimes> \<Psi>\<^isub>Q\<rangle>) \<otimes>\<^sub>F (\<langle>[a], \<Psi>\<^isub>c\<rangle>)" unfolding unb[symmetric] frameResChainbase
                    apply(subst mergeFrameSimps')
                    apply(force simp add: `b \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q`)
                    apply(rule frameResChainPres)
                    apply(subst frameNilStatEq)
                    apply(rule AssertionStatEqTrans[OF AssertionStatEqSym[OF Associativity]])
                    apply(rule compositionSym)
                    apply(rule AssertionStatEqTrans[OF Commutativity])
                    by(rule Identity)
                  show ?thesis
                    apply(rule FrameStatEqTrans[OF lhs])
                    apply(rule FrameStatEqSym)
                    apply(rule FrameStatEqTrans[OF rhs])
                    by fact
                qed
                with `\<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<Psi>\<^isub>P \<turnstile> M \<leftrightarrow> K` `b \<sharp> M` `b \<sharp> K` show ?thesis
                  by(rule_tac frameImpE[where A\<^isub>F="[b]"]) (auto dest: frameStatEqEnt)
              qed
              from `xvec \<sharp>* (a, M\<^isub>a, \<Psi>\<^isub>c)` have "a \<sharp> xvec" and "xvec \<sharp>* M\<^isub>a" and "xvec \<sharp>* \<Psi>\<^isub>c"
                by simp+
              have Trans: "\<Psi> \<rhd> Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<tau> \<prec> \<lparr>\<nu>*xvec\<rparr>(Q' \<parallel> Q'')"
                by(rule Comm2[where A\<^isub>Q="[b]" and \<Psi>\<^isub>Q="\<one> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)"]) (fact | simp add: `b \<sharp> A\<^isub>Q` `b \<sharp> \<Psi>` `b \<sharp> Q` fresh_abs_fun_iff[OF pt_name_inst,OF at_name_inst, OF fin_supp] `b \<sharp> M\<^isub>a` `b \<sharp> \<Psi>\<^isub>c` `b \<sharp> K` `a \<sharp> xvec` `xvec \<sharp>* M\<^isub>a` `xvec \<sharp>* \<Psi>\<^isub>c`)+
              have "(\<Psi>,Q'\<parallel>Q'',Q'\<parallel>P') \<in> ?Y1" using `(\<Psi>,Q''',P'')\<in>?X1 \<or> \<Psi> \<rhd> Q''' \<sim> P''`
              proof(rule disjE)
                assume "\<Psi> \<rhd> Q''' \<sim> P''"
                hence "\<Psi> \<rhd> Q'' \<sim> P'" using `\<Psi> \<rhd> Q'' \<sim> Q'''` `\<Psi> \<rhd> P'' \<sim> P'`
                  by(auto dest: bisimTransitive)
                have "\<Psi> \<rhd> Q'\<parallel>Q'' \<sim> Q'\<parallel>P'"
                  by(rule bisimParPresSym) fact
                show ?thesis
                  apply(rule lol)
                  by(rule refl|rule `\<Psi> \<rhd> Q'\<parallel>Q'' \<sim> Q'\<parallel>P'`|rule disjI2|rule bisimReflexive)+
              next
                assume "(\<Psi>, Q''', P'') \<in> ?X1"
                then obtain a P Q\<^isub>i M\<^isub>a \<Psi>\<^isub>c xvec where q''_def: "Q''' = \<lparr>\<nu>*xvec\<rparr>(Q\<^isub>i \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))" and p''_def: "P'' = \<lparr>\<nu>*xvec\<rparr>(Q\<^isub>i \<parallel> !P)" and "a \<sharp> P" and "a \<sharp> Q\<^isub>i" and "a \<sharp> \<Psi>" and "\<not> a \<sharp> M\<^isub>a" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> R \<longrightarrow> (R = (P \<parallel> \<langle>M\<^isub>a\<rangle>)))" and "guarded P" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<longrightarrow> (supp M\<^isub>a) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P'))" and "((supp P)::name set) \<subseteq> supp M\<^isub>a" and "((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a" and "xvec \<sharp>* \<Psi>" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>a \<Leftarrow> P\<parallel>\<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
                  by auto
                obtain p::"name prm" where "distinctPerm p" and "set p \<subseteq> set xvec \<times> set(p\<bullet>xvec)" "(p\<bullet>xvec) \<sharp>* Q'" and "(p\<bullet>xvec) \<sharp>* \<Psi>" and "(p\<bullet>xvec) \<sharp>* (Q\<^isub>i \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))" and "(p\<bullet>xvec) \<sharp>* (Q\<^isub>i \<parallel> !P)"
                  by(rule chain.name_list_avoiding[where xvec=xvec and c="(Q',\<Psi>,Q\<^isub>i \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>),Q\<^isub>i \<parallel> !P)"]) auto
                obtain b::name where "b \<sharp> P" and "b \<sharp> Q'" and "b \<sharp> \<Psi>" and "b \<sharp> M\<^isub>a" "b \<sharp> \<Psi>\<^isub>c" and "b \<sharp> (p\<bullet>a)" and "b \<sharp> (p\<bullet>M\<^isub>a)" and "b \<sharp> (p\<bullet>\<Psi>\<^isub>c)" and "b \<sharp> (p\<bullet>P)" and "b \<sharp> Q" and "b \<sharp> (p\<bullet>Q\<^isub>i)"
                  apply(rule at_exists_fresh[OF at_name_inst, where x="(P,Q',\<Psi>,M\<^isub>a,\<Psi>\<^isub>c,p\<bullet>a,p\<bullet>M\<^isub>a,p\<bullet>\<Psi>\<^isub>c,p\<bullet>P,p\<bullet>Q\<^isub>i,Q)"])
                  unfolding supp_prod pt_perm_supp[OF pt_name_inst,OF at_name_inst, symmetric]
                  apply(rule finite_UnI, (finite_guess | (subst pt_set_finite_ineq[OF pt_name_inst,OF at_name_inst], finite_guess)))+
                  apply(finite_guess)
                  by force
                have "\<Psi> \<rhd> Q'\<parallel>Q'' \<sim> \<lparr>\<nu>*(p\<bullet>xvec)\<rparr>((Q' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>p\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                proof -
                  have "\<Psi> \<rhd> Q'\<parallel>Q'' \<sim> Q'\<parallel>Q'''"
                    apply(rule bisimParPresSym)
                    by fact
                  moreover have "\<Psi> \<rhd> Q'\<parallel>Q''' \<sim> Q'\<parallel>(\<lparr>\<nu>*xvec\<rparr>Q\<^isub>i \<parallel>\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))"
                    unfolding q''_def
                    by(rule bisimReflexive)
                  moreover have "\<Psi> \<rhd> Q'\<parallel>(\<lparr>\<nu>*xvec\<rparr>Q\<^isub>i \<parallel>\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) \<sim> Q'\<parallel>(\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>((p\<bullet>Q\<^isub>i)\<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>p\<bullet>\<Psi>\<^isub>c)\<rbrace>)))"
                    apply(subst resChainAlpha[where p=p])
                    apply fact
                    apply fact
                    unfolding eqvts
                    apply(subst alphaRes[where x="p\<bullet>a" and y=b])
                    apply(simp add: `b \<sharp> (p\<bullet>M\<^isub>a)` `b \<sharp> (p\<bullet>\<Psi>\<^isub>c)`)
                    unfolding eqvts
                    by(rule bisimReflexive)
                  moreover have "\<Psi> \<rhd> Q'\<parallel>(\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>((p\<bullet>Q\<^isub>i)\<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>p\<bullet>\<Psi>\<^isub>c)\<rbrace>))) \<sim> (\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(Q' \<parallel>((p\<bullet>Q\<^isub>i)\<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>p\<bullet>\<Psi>\<^isub>c)\<rbrace>))))"
                    by(rule bisimSymmetric[OF bisimScopeExtChain]) fact+
                  moreover have "\<Psi> \<rhd> (\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(Q' \<parallel>((p\<bullet>Q\<^isub>i)\<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>p\<bullet>\<Psi>\<^isub>c)\<rbrace>)))) \<sim> \<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(Q' \<parallel>(p\<bullet>Q\<^isub>i))\<parallel>(\<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>p\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                    apply(rule bisimResChainPres)
                    apply(rule bisimSymmetric[OF bisimParAssoc])
                    by fact
                  ultimately show ?thesis
                    by(auto dest: bisimTransitive)
                qed
                moreover have "\<Psi> \<rhd> \<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(Q'\<parallel>(p\<bullet>Q\<^isub>i))\<parallel>!(p\<bullet>P) \<sim> Q' \<parallel> P'"
                proof -
                  have "\<Psi> \<rhd> Q' \<parallel> P'' \<sim> Q' \<parallel> P'"
                    apply(rule bisimParPresSym)
                    by fact+
                  moreover have "\<Psi> \<rhd> Q'\<parallel>(\<lparr>\<nu>*xvec\<rparr>Q\<^isub>i \<parallel>!P) \<sim> Q'\<parallel>P''"
                    unfolding p''_def
                    by(rule bisimReflexive)
                  moreover have "\<Psi> \<rhd> Q'\<parallel>(\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>((p\<bullet>Q\<^isub>i)\<parallel> !(p\<bullet>P))) \<sim> Q'\<parallel>(\<lparr>\<nu>*xvec\<rparr>Q\<^isub>i \<parallel>!P)"
                    apply(subst resChainAlpha[where p=p and xvec=xvec])
                    apply fact
                    apply fact
                    unfolding eqvts
                    by(rule bisimReflexive)
                  moreover have "\<Psi> \<rhd> \<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(Q' \<parallel>((p\<bullet>Q\<^isub>i)\<parallel> (!(p\<bullet>P)))) \<sim> Q'\<parallel>(\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>((p\<bullet>Q\<^isub>i)\<parallel> !(p\<bullet>P)))"                    
                    by(rule bisimScopeExtChain) fact+
                  moreover have "\<Psi> \<rhd> \<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(Q' \<parallel>(p\<bullet>Q\<^isub>i))\<parallel>(!(p\<bullet>P)) \<sim> \<lparr>\<nu>*(p\<bullet>xvec)\<rparr>(Q' \<parallel>((p\<bullet>Q\<^isub>i)\<parallel> (!(p\<bullet>P))))"
                    apply(rule bisimResChainPres)
                    by(rule bisimParAssoc) fact
                  ultimately show ?thesis
                    by(metis bisimTransitive)
                qed
                moreover have "(\<Psi>,\<lparr>\<nu>*(p\<bullet>xvec)\<rparr>((Q' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(p\<bullet>a,b)]\<bullet>p\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a,b)]\<bullet>p\<bullet>\<Psi>\<^isub>c)\<rbrace>)),\<lparr>\<nu>*p \<bullet> xvec\<rparr>Q' \<parallel> (p \<bullet> Q\<^isub>i) \<parallel> !(p \<bullet> P)) \<in> ?X1"
                proof -
                  have "(p\<bullet>a) \<sharp> (p\<bullet>P)" using `a \<sharp> P` by(subst fresh_bij) 
                  note `b \<sharp> \<Psi>`
                  moreover note `b \<sharp> (p\<bullet>P)`
                  moreover have "b \<sharp> (Q'\<parallel>(p\<bullet>Q\<^isub>i))" using `b \<sharp> Q'` `b \<sharp>(p\<bullet>Q\<^isub>i)`
                    by simp
                  moreover have "guarded(p\<bullet>P)" using `guarded P` by simp
                  moreover have "\<not> b \<sharp> ([(p\<bullet>a,b)] \<bullet> (p\<bullet>M\<^isub>a))" using `\<not> a \<sharp> M\<^isub>a`
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps)
                  moreover have "\<And> \<Psi>. \<Psi> \<otimes> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)) \<turnstile> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>(p\<bullet>P) \<parallel> \<langle>([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<rangle>" using EntCl
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst perm_bool[where pi="rev p",symmetric])
                    by(simp add: eqvts `(p\<bullet>a) \<sharp> (p\<bullet>P)` `b \<sharp> (p\<bullet>P)`)
                  moreover have "\<forall>F. (F \<otimes>\<^sub>F (\<langle>[b],[(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)\<rangle>) \<simeq>\<^sub>F F)"
                    using Equiv1
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst perm_bool[where pi="rev p",symmetric])
                    by(simp add: eqvts swap_simps)
                  moreover have "(\<forall>\<Psi> R. b \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)) \<turnstile> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a)) \<Leftarrow> R \<longrightarrow> (R = ((p\<bullet>P) \<parallel> \<langle>([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<rangle>)))" using Equiv2
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts `(p\<bullet>a) \<sharp> p\<bullet>P` `b \<sharp> p\<bullet>P` swap_simps)
                  moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>P \<longrightarrow> (supp ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))) \<subseteq> ((supp \<Psi>)::name set))" using EntSupp
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps)
                  moreover have "\<forall> P P'. ((([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>P = ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>P') = (P = P'))" using Inj
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps)
                  moreover have "((supp(p\<bullet>P))::name set) \<subseteq> supp ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))" using `((supp P)::name set) \<subseteq> supp M\<^isub>a`
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps `(p\<bullet>a) \<sharp> p\<bullet>P` `b \<sharp> p\<bullet>P`)
                  moreover have "((supp ([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)))::name set) = supp ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))" using `((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a`
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                  moreover note `(p\<bullet>xvec)\<sharp>*\<Psi>`
                  moreover from EntEnt have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(p\<bullet>a,b)]\<bullet>(p\<bullet>\<Psi>\<^isub>c)) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<Leftarrow>(p\<bullet>P)\<parallel>\<langle>([(p\<bullet>a,b)]\<bullet>(p\<bullet>M\<^isub>a))\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                    apply(subst perm_bool[where pi="[(p\<bullet>a,b)]", symmetric])
                    apply(subst perm_bool[where pi="p",symmetric])
                    apply(simp add: eqvts swap_simps `distinctPerm p` `(p\<bullet>a) \<sharp> p\<bullet>P` `b \<sharp> p\<bullet>P`)
                    done
                  ultimately show ?thesis
                    by blast
                qed
                ultimately show ?thesis
                  apply(rule_tac lol)
                  by(rule refl|assumption|rule disjI1)+
              qed
              hence "(\<Psi>,Q'\<parallel>Q'',Q'\<parallel>P') \<in> ?Y"
                by(rule Y1inY)
              hence "(\<Psi>,\<lparr>\<nu>*xvec\<rparr>Q'\<parallel>Q'',\<lparr>\<nu>*xvec\<rparr>Q'\<parallel>P') \<in> ?Y" using `xvec \<sharp>* \<Psi>`
                by(rule YresChainClose')
              with Trans show ?case
                by(rule_tac exI) (rule conjI)
            qed
          qed
        qed
        thus ?thesis unfolding `S = \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))` `T = \<lparr>\<nu>*xvec\<rparr>Q \<parallel> !P`
          by -
      next
        assume "(\<Psi>,S,T) \<notin> ?X1"
        hence "(\<Psi>,S,T) \<in> ?X2" using `(\<Psi>,S,T) \<in> ?X`
          by blast
        then obtain a P Q M\<^isub>a \<Psi>\<^isub>c xvec where "T = \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))" and "S = \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> !P)" and "a \<sharp> P" and "a \<sharp> Q" and "a \<sharp> \<Psi>" and "\<not> a \<sharp> M\<^isub>a" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a], \<Psi>\<^isub>c\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> R \<longrightarrow> (R = (P \<parallel> \<langle>M\<^isub>a\<rangle>)))" and "guarded P" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<longrightarrow> (supp M\<^isub>a) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P'))" and "((supp P)::name set) \<subseteq> supp M\<^isub>a" and "((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a" and "xvec \<sharp>* \<Psi>" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
          by auto
        have Equiv2: "\<And>\<Psi> R. \<lbrakk>a \<sharp> \<Psi>\<rbrakk> \<Longrightarrow> (\<lbrakk>\<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> R\<rbrakk> \<Longrightarrow> R = (P \<parallel> \<langle>M\<^isub>a\<rangle>))" using Equiv2
          by simp
        have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> !P) \<leadsto>[?Y] \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))" using `xvec \<sharp>* \<Psi>`
        proof(induct xvec)
          case(Cons x xvec)
          from `(x # xvec) \<sharp>* \<Psi>` have "x \<sharp> \<Psi>" and "xvec \<sharp>* \<Psi>" by simp+
          have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> !P) \<leadsto>[?Y] \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))" using `xvec \<sharp>* \<Psi>`
            by(rule Cons)
          show ?case
            unfolding resChain.simps
            by(rule resPres[where Rel="?Y" and Rel'="?Y"]) (fact | rule subset_refl)+
        next
          case Nil
          show ?case unfolding resChain.simps
          proof(rule simI[where C= "a", OF `eqvt(?Y)`])
            fix \<alpha> Q'
            assume "\<Psi> \<rhd> Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> Q'"
              and  "bn \<alpha> \<sharp>* (Q \<parallel> !P)"
              and  "bn \<alpha> \<sharp>* (Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))"
              and  "bn \<alpha> \<sharp>* \<Psi>"
              and  "distinct (bn \<alpha>)"
              and  "bn \<alpha> \<sharp>* subject \<alpha>"
              and  "bn \<alpha> \<sharp>* a"
            from `bn \<alpha> \<sharp>* (Q \<parallel> !P)` have "bn \<alpha> \<sharp>* P"
              by simp
            from `bn \<alpha> \<sharp>* (Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))` `bn \<alpha> \<sharp>* a` have "bn \<alpha> \<sharp>* M\<^isub>a" and "bn \<alpha> \<sharp>* \<Psi>\<^isub>c"
              by auto
            note `\<Psi> \<rhd> Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> Q'` `bn \<alpha> \<sharp>* \<Psi>`
            moreover from `bn \<alpha> \<sharp>* (Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))` have "bn \<alpha> \<sharp>* Q" and "bn \<alpha> \<sharp>* (\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)::('a,'b,'c) psi)"
              by auto
            moreover note `bn \<alpha> \<sharp>* subject \<alpha>`
            ultimately show "\<exists> P'. \<Psi> \<rhd> Q \<parallel> !P \<longmapsto> \<alpha> \<prec> P' \<and> (\<Psi>,P',Q') \<in> ?Y"
            proof(induct rule: parCases[where C="(a,M\<^isub>a,\<Psi>\<^isub>c,P)"])
              case(cPar1 Q' A\<^isub>Q \<Psi>\<^isub>Q)
              obtain b where "A\<^isub>Q=[b]"
              proof -
                assume "\<And>b. A\<^isub>Q = [b] \<Longrightarrow> thesis"
                moreover from `extractFrame (\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>`
                have "\<exists> b. A\<^isub>Q=[b]"
                proof(induct A\<^isub>Q)
                  case Nil thus ?case by(simp)
                next
                  case(Cons a\<^isub>Q A\<^isub>Q)
                  have "[(a, a\<^isub>Q)] \<bullet> A\<^isub>Q = [] \<Longrightarrow> A\<^isub>Q = []"
                    by(induct A\<^isub>Q) (auto simp add: eqvts)
                  with Cons(2) have "A\<^isub>Q=[]"
                    by(auto simp add: abs_fun_eq[OF pt_name_inst, OF at_name_inst] frame.inject eqvts)
                  thus ?case by simp
                qed
                ultimately show ?thesis
                  by auto
              qed
              with cPar1 have "b \<sharp> \<Psi>" and "b \<sharp> Q" and "b \<sharp> a" and "b \<sharp> M\<^isub>a" and "b \<sharp> \<alpha>" and "b \<sharp> \<Psi>\<^isub>c" and "b \<sharp> Q'"
                by auto
              have "\<Psi>\<^isub>Q = \<one> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)" using `extractFrame (\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>` unfolding `A\<^isub>Q = [b]`
                by(auto simp add: frame.inject abs_fun_eq'[OF pt_name_inst, OF at_name_inst] eqvts pt3[OF pt_name_inst, OF at_ds5, OF at_name_inst])
              have unb: "(\<langle>[b],([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rangle>) = (\<langle>[a],\<Psi>\<^isub>c\<rangle>)"
                by(subst frameChainAlpha[where xvec="[a]" and p="[(a,b)]"]) (auto simp add: swap_simps `b \<sharp> \<Psi>\<^isub>c`)
              have "\<Psi> \<otimes> \<one> \<rhd> Q \<longmapsto> \<alpha> \<prec> Q'"
              proof -
                {
                  fix F::"'b frame"
                  obtain A\<^isub>F \<Psi>\<^isub>F where "F=\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>" and "distinct A\<^isub>F" and "A\<^isub>F \<sharp>* \<Psi>" and "A\<^isub>F \<sharp>* \<Psi>\<^isub>Q" and "A\<^isub>F \<sharp>* a" and "A\<^isub>F \<sharp>* b" and "A\<^isub>F \<sharp>* \<Psi>\<^isub>c"
                    apply(rule freshFrame[where C="(\<Psi>,\<Psi>\<^isub>Q,a,b,\<Psi>\<^isub>c)"])
                    by auto
                  have lhs: "(\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>\<epsilon>, \<Psi> \<otimes> \<one>\<rangle>) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>\<rangle>)"
                    apply(subst mergeFrameSimps)
                    apply(rule freshCompChain)
                    apply fact
                    apply simp
                    apply(rule frameResChainPres)
                    apply(subst frameNilStatEq)
                    apply(rule AssertionStatEqTrans[OF AssertionStatEqSym[OF Associativity]])
                    by(rule Identity)
                  obtain c::name where "c \<sharp> A\<^isub>F" and "c \<sharp> \<Psi>\<^isub>F" and "c \<noteq> b" and "c \<noteq> a" and "c \<sharp> \<Psi>"  and "c \<sharp> \<Psi>\<^isub>c"
                    apply(generate_fresh "name")
                    by force
                  have af: "([(a,c)]\<bullet>A\<^isub>F) = A\<^isub>F" using `A\<^isub>F \<sharp>* a` `c \<sharp> A\<^isub>F`
                    by simp
                  have "((\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[c], \<Psi> \<otimes> \<one> \<otimes> ([(a, c)] \<bullet> \<Psi>\<^isub>c)\<rangle>)) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>\<rangle>) \<otimes>\<^sub>F (\<langle>[c], ([(a,c)]\<bullet>\<Psi>\<^isub>c)\<rangle>)"
                    apply(subst (1 2) mergeFrames)
                    apply(simp add: `c \<sharp> A\<^isub>F`)
                    apply(subst af[symmetric])
                    apply(subst fresh_star_bij)
                    apply fact
                    apply(rule freshCompChain)
                    apply(simp add: `c \<sharp> \<Psi>\<^isub>F`)
                    apply(simp add: `c \<sharp> \<Psi>`)
                    apply(rule freshCompChain)
                    apply fact
                    apply(rule freshCompChain)
                    apply simp
                    apply(subst af[symmetric])
                    apply(subst fresh_star_bij)
                    apply fact
                    apply(simp add: `c \<sharp> \<Psi>\<^isub>F`)
                    apply(rule frameResChainPres)
                    apply(subst frameNilStatEq)
                    apply(rule AssertionStatEqSym)
                    apply(rule AssertionStatEqTrans[OF Associativity])
                    apply(rule compositionSym)
                    apply(rule compositionSym)
                    apply(rule AssertionStatEqSym)
                    apply(rule AssertionStatEqTrans[OF Commutativity])
                    by(rule Identity)
                  hence "((\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[a], \<Psi> \<otimes> \<one> \<otimes> \<Psi>\<^isub>c\<rangle>)) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>\<rangle>) \<otimes>\<^sub>F (\<langle>[a], \<Psi>\<^isub>c\<rangle>)"
                    apply(subst (1 2) frameChainAlpha'[where A\<^isub>P="[a]" and p="[(a,c)]"])
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `c \<sharp> \<Psi>\<^isub>c`)
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                    apply(rule freshCompChain)
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `a \<sharp> \<Psi>` `c \<sharp> \<Psi>`)
                    apply(rule freshCompChain)
                    apply simp
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `c \<sharp> \<Psi>\<^isub>c`)
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] eqvts `a \<sharp> \<Psi>` `c \<sharp> \<Psi>`)
                    done
                  hence rhs: "((\<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>) \<otimes>\<^sub>F (\<langle>[b], \<Psi> \<otimes> \<one> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rangle>)) \<simeq>\<^sub>F (\<langle>A\<^isub>F,\<Psi>\<^isub>F \<otimes> \<Psi>\<rangle>) \<otimes>\<^sub>F (\<langle>[b], ([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rangle>)"
                    apply(subst unb)
                    apply(subst frameChainAlpha'[where A\<^isub>P="[b]" and p="[(b,a)]"])
                    apply(rule freshCompChain)
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst] `a \<sharp> \<Psi>` `b \<sharp> \<Psi>`)
                    apply(rule freshCompChain)
                    apply simp
                    apply(subst pt3[OF pt_name_inst, OF at_ds5, OF at_name_inst])
                    apply(subst pt_fresh_star_bij[OF pt_name_inst, OF at_name_inst])
                    apply(simp add: `b \<sharp> \<Psi>\<^isub>c`)
                    apply(simp add: at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                    apply(subst (1 2) pt3[OF pt_name_inst, OF at_ds5, OF at_name_inst])
                    by(force simp add: eqvts `a \<sharp> \<Psi>` `b \<sharp> \<Psi>` at1[OF at_name_inst] at2[OF at_name_inst] at3[OF at_name_inst])
                  have "F \<otimes>\<^sub>F (\<langle>\<epsilon>, \<Psi> \<otimes> \<one>\<rangle>) \<simeq>\<^sub>F F \<otimes>\<^sub>F (\<langle>A\<^isub>Q, \<Psi> \<otimes> \<Psi>\<^isub>Q\<rangle>)" unfolding `\<Psi>\<^isub>Q = \<one> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)` `A\<^isub>Q = [b]` `F = \<langle>A\<^isub>F,\<Psi>\<^isub>F\<rangle>`
                    apply(rule FrameStatEqTrans[OF lhs])
                    apply(rule FrameStatEqSym[OF FrameStatEqTrans[OF rhs]])
                    apply(subst unb)
                    by fact
                }
                show ?thesis
                  by(rule frameStatEqTransition'[where A="A\<^isub>Q"]) fact+
              qed
              hence "\<Psi> \<rhd> Q \<parallel> !P \<longmapsto> \<alpha> \<prec> Q' \<parallel> !P"
                by(rule Par1) (simp add: `bn \<alpha> \<sharp>* P`)+
              moreover have "(\<Psi>, Q' \<parallel> !P, Q' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) \<in> ?Y"
              proof -
                have abeq: "\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) = \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                  by(subst alphaRes[where y=b and x=a]) (auto simp add: eqvts `b \<sharp> M\<^isub>a` `b \<sharp> \<Psi>\<^isub>c`)
                have "(\<Psi>, \<lparr>\<nu>*[]\<rparr>Q' \<parallel> !P, \<lparr>\<nu>*[]\<rparr>Q' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) \<in> ?X2"
                proof -
                  note `b \<sharp> \<Psi>`
                  moreover have "b \<sharp> P" using `b \<sharp> M\<^isub>a` `supp P \<subseteq> supp M\<^isub>a`
                    unfolding fresh_def by auto
                  moreover note `b \<sharp> Q'`
                  moreover have "\<not> b \<sharp> ([(a,b)]\<bullet>M\<^isub>a)" using `\<not> a \<sharp> M\<^isub>a`
                    by(simp add: fresh_left swap_simps)
                  moreover have "(\<forall>\<Psi>. \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)" using EntCl
                    by(subst perm_bool[symmetric, where pi="[(a,b)]"]) (simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
                  moreover have "\<forall>F. F \<otimes>\<^sub>F (\<langle>[b],([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rangle>)\<simeq>\<^sub>F F" using Equiv1
                    unfolding unb by simp
                  moreover have "(\<forall>\<Psi> R. b \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>R \<longrightarrow> R = P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)" using Equiv2
                    by(subst perm_bool[symmetric, where pi="[(a,b)]"]) (simp add: eqvts `a \<sharp> P` `b \<sharp> P` swap_simps)
                  moreover note `guarded P`
                  moreover have "(supp P::name set) \<subseteq> supp([(a,b)]\<bullet>M\<^isub>a)" using `supp P \<subseteq> supp M\<^isub>a`
                    by(subst perm_bool[symmetric, where pi="[(a,b)]"]) (simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
                  moreover have "(supp([(a,b)]\<bullet>\<Psi>\<^isub>c)::name set) = supp([(a,b)]\<bullet>M\<^isub>a)" using `supp \<Psi>\<^isub>c = supp M\<^isub>a`
                    by(subst perm_bool[symmetric, where pi="[(a,b)]"]) (simp add: eqvts)
                  moreover have "\<forall> P P'. (([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P = ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P') = (P = P')" using Inj
                    by(subst perm_bool[symmetric, where pi="[(a,b)]"]) (simp add: eqvts)
                  moreover have "\<forall>\<Psi> P. \<Psi> \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<longrightarrow> supp([(a,b)]\<bullet>M\<^isub>a) \<subseteq> (supp \<Psi>::name set)" using EntSupp
                    by(subst perm_bool[symmetric, where pi="[(a,b)]"]) (simp add: eqvts)
                  moreover have "([]::name list) \<sharp>* \<Psi>"
                    by simp
                  moreover from EntEnt have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                    by(subst perm_bool[where pi="[(a,b)]",symmetric]) (simp add: eqvts swap_simps `a \<sharp> P` `b \<sharp> P`)
                  ultimately show ?thesis unfolding abeq
                    by blast
                qed
                hence "(\<Psi>, Q' \<parallel> !P, Q' \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) \<in> ?X" unfolding resChain.simps
                  by blast
                show ?thesis
                  by(rule lol|rule refl|rule bisimReflexive|rule disjI1|fact)+
              qed
              ultimately show ?case
                by(rule_tac exI) (rule conjI)
            next
              case(cPar2 Q' A\<^isub>Q \<Psi>\<^isub>Q)
              from `A\<^isub>Q \<sharp>* (a, M\<^isub>a, \<Psi>\<^isub>c, P)` have "a \<sharp> A\<^isub>Q" and "A\<^isub>Q \<sharp>* M\<^isub>a" and "A\<^isub>Q \<sharp>* \<Psi>\<^isub>c" and "A\<^isub>Q \<sharp>* P"
                by auto
              obtain b::name where "b \<sharp> A\<^isub>Q" and "b \<sharp> \<Psi>" and "b \<sharp> Q" and "b \<sharp> a" and "b \<sharp> M\<^isub>a" and "b \<sharp> \<Psi>\<^isub>c" and "b \<sharp> \<alpha>" and "b \<sharp> Q'" and "b \<sharp> \<Psi>\<^isub>Q" and "b \<sharp> P"
                by(generate_fresh "name") force
              have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<longmapsto> \<alpha> \<prec> Q'" using `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> \<alpha> \<prec> Q'`  `b \<sharp> M\<^isub>a` `b \<sharp> \<Psi>\<^isub>c`
                by(subst (asm) alphaRes[where y=b]) auto
              moreover have "b \<sharp> (\<Psi>\<otimes>\<Psi>\<^isub>Q)" by(rule freshComp) fact+
              moreover note `b \<sharp> \<alpha>` `b \<sharp> Q'`
              moreover have "bn \<alpha> \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)" using `bn \<alpha> \<sharp>* \<Psi>` `bn \<alpha> \<sharp>* Q` `A\<^isub>Q \<sharp>* \<alpha>` `extractFrame Q=\<langle>A\<^isub>Q, \<Psi>\<^isub>Q\<rangle>`
                by(drule_tac extractFrameFreshChain) auto
              moreover have "bn \<alpha> \<sharp>* ([(a,b)]\<bullet>M\<^isub>a)" using `bn \<alpha> \<sharp>* M\<^isub>a` `bn \<alpha> \<sharp>* a` `b \<sharp> \<alpha>`
                by simp
              moreover have "bn \<alpha> \<sharp>* ([(a,b)]\<bullet>\<Psi>\<^isub>c)" using `bn \<alpha> \<sharp>* \<Psi>\<^isub>c` `bn \<alpha> \<sharp>* a` `b \<sharp> \<alpha>`
                by simp
              moreover note `bn \<alpha> \<sharp>* subject \<alpha>` `b \<sharp> P` `guarded P`
              moreover have "\<not> b \<sharp> ([(a,b)]\<bullet>M\<^isub>a)" using `\<not> a \<sharp> M\<^isub>a`
                by(auto simp add: fresh_left swap_simps)
              moreover have "\<forall>\<Psi>'. \<Psi>' \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>" using EntCl
                apply(subst perm_bool[where pi="[(a,b)]",symmetric])
                by(simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
              moreover have "\<forall>\<Psi>' Q. b \<sharp> \<Psi>' \<and> \<Psi>' \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>Q \<longrightarrow> Q = P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>" using Equiv2
                apply(subst perm_bool[where pi="[(a,b)]",symmetric])
                by(auto simp add: eqvts `a \<sharp> P` `b \<sharp> P` swap_simps)
              ultimately show ?case
              proof(induct rule: resReplCases)
                case(cRes P')
                note `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace> \<longmapsto> \<alpha> \<prec> P'`
                moreover note `bn \<alpha> \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)`
                moreover have "bn \<alpha> \<sharp>* \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>" using `bn \<alpha> \<sharp>* ([(a,b)]\<bullet>M\<^isub>a)`
                  by simp
                moreover have "bn \<alpha> \<sharp>* \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>" using `bn \<alpha> \<sharp>* ([(a,b)]\<bullet>\<Psi>\<^isub>c)`
                  by simp
                moreover note `bn \<alpha> \<sharp>* subject \<alpha>`
                ultimately show ?case
                proof(induct rule: parCases[where C="()"])
                  case cPar2 thus ?case by(drule_tac nilTrans) (rule FalseE)
                next
                  case cComm1 thus ?case by(drule_tac nilTrans) (rule FalseE)
                next
                  case cComm2 thus ?case by(drule_tac nilTrans) (rule FalseE)
                next
                  case(cPar1 Q' A\<^isub>R \<Psi>\<^isub>R)
                  from `extractFrame (\<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>) = \<langle>A\<^isub>R, \<Psi>\<^isub>R\<rangle>` have "A\<^isub>R = []" and psirdef: "\<Psi>\<^isub>R = ([(a,b)]\<bullet>\<Psi>\<^isub>c)"
                    by auto
                  obtain \<Psi>' where psi'def: "\<Psi>'=(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)"
                    by simp
                  obtain M\<^isub>b \<Psi>\<^isub>b where mbdef: "M\<^isub>b = [(a,b)]\<bullet>M\<^isub>a" and psibdef: "\<Psi>\<^isub>b = [(a,b)]\<bullet>\<Psi>\<^isub>c"
                    by auto
                  have "b \<sharp> bn \<alpha>" and "b \<sharp> object \<alpha>" using `b \<sharp> \<alpha>` by auto
                  have "a \<sharp> \<Psi>\<^isub>Q" using `a \<sharp> Q` `extractFrame Q = \<langle>A\<^isub>Q,\<Psi>\<^isub>Q\<rangle>` `a \<sharp> A\<^isub>Q`
                    by(drule_tac extractFrameFresh) simp
                  have "bn \<alpha> \<sharp>* \<Psi>\<^isub>Q" using `bn \<alpha> \<sharp>* Q` `extractFrame Q = \<langle>A\<^isub>Q,\<Psi>\<^isub>Q\<rangle>` `A\<^isub>Q \<sharp>* \<alpha>`
                    by(drule_tac extractFrameFreshChain) simp
                  have Equiv1': "\<And>F. F \<otimes>\<^sub>F (\<langle>[b], [(a, b)] \<bullet> \<Psi>\<^isub>c\<rangle>) \<simeq>\<^sub>F F" using `\<And>F. F \<otimes>\<^sub>F (\<langle>[a], \<Psi>\<^isub>c\<rangle>) \<simeq>\<^sub>F F`
                    by(subst (asm) frameChainAlpha [where p="[(a,b)]"]) (simp add: swap_simps `b \<sharp> \<Psi>\<^isub>c`)+
                  have Equiv2': "\<And>\<Psi> R. \<lbrakk>b \<sharp> \<Psi>; \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a) \<Leftarrow> R\<rbrakk> \<Longrightarrow> R = P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>"
                  proof -
                    fix \<Psi> R
                    assume "b \<sharp> \<Psi>" and "\<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a) \<Leftarrow> R"
                    moreover have "\<forall> \<Psi> R. (b \<sharp> \<Psi> \<and> \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a) \<Leftarrow> R) \<longrightarrow> R = P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>" using Equiv2
                      by(subst perm_bool[where pi="[(a,b)]",symmetric]) (force simp add: eqvts swap_simps `a \<sharp> P` `b \<sharp> P`)
                    ultimately show "R = P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>"
                      by auto
                  qed
                  have "((supp P)::name set) \<subseteq> supp([(a,b)]\<bullet>M\<^isub>a)" using `((supp P)::name set) \<subseteq> supp M\<^isub>a`
                    apply(subst perm_bool[where pi="[(a,b)]",symmetric])
                    by(simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
                  from `((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a` have "((supp([(a,b)]\<bullet>\<Psi>\<^isub>c))::name set) = supp([(a,b)]\<bullet>M\<^isub>a)"
                    apply(subst (asm) perm_bool[where pi="[(a,b)]",symmetric])
                    by(simp only: eqvts)
                  have EntCl': "\<And>\<Psi>. \<Psi> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>" using `\<And>\<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>`
                    by(subst perm_bool[where pi="[(a,b)]",symmetric]) (simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
                  have EntEnt': "\<And>\<Psi> \<phi>. \<Psi> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> \<phi> \<Longrightarrow>
                    \<phi> = ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                  proof -
                    fix \<Psi> \<phi>
                    assume "\<Psi> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> \<phi>"
                    moreover have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<or> \<Psi> \<turnstile> \<phi>" using `\<And>\<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> \<phi> \<Longrightarrow> \<phi> = M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> \<phi>`
                      apply(subst perm_bool[where pi="[(a,b)]",symmetric])
                      by(simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
                    ultimately show "\<phi> = ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                      by simp
                  qed
                  {
                    fix \<Psi> P
                    assume "\<Psi> \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P"
                    moreover have "\<forall> \<Psi> P. \<Psi> \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<longrightarrow> supp ([(a, b)] \<bullet> M\<^isub>a) \<subseteq> (supp \<Psi>::name set)" using `\<forall>\<Psi> P. \<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<longrightarrow> supp M\<^isub>a \<subseteq> supp \<Psi>`
                      apply(subst perm_bool[where pi="[(a,b)]",symmetric])
                      by(simp add: eqvts)
                    ultimately have "supp ([(a, b)] \<bullet> M\<^isub>a) \<subseteq> (supp \<Psi>::name set)"
                      by auto
                  }
                  note EntSupp' = this
                  have Inj': "\<And>P P'. (([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P = ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P') = (P = P')" using `\<forall>P P'. (M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P')`
                    by(subst (asm) perm_bool[where pi="[(a,b)]",symmetric]) (simp add: eqvts)
                  have "\<not> b \<sharp> [(a, b)] \<bullet> M\<^isub>a" using `\<not> a \<sharp> M\<^isub>a`
                    by(subst (asm) perm_bool [where pi="[(a,b)]",symmetric]) (simp add: eqvts swap_simps)
                  have "\<Psi>' \<rhd> \<langle>M\<^isub>b\<rangle> \<longmapsto> \<alpha> \<prec> Q'" unfolding psi'def mbdef psirdef[symmetric] by fact
                  moreover note `bn \<alpha> \<sharp>* subject \<alpha>`
                  moreover note `guarded P`
                  moreover have "\<forall> \<Psi>''. \<Psi>' \<otimes> \<Psi>'' \<turnstile> M\<^isub>b \<Leftarrow> P \<parallel> \<langle>M\<^isub>b\<rangle>"
                  proof -
                    {
                      fix \<Psi>''
                      have "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>'' \<turnstile> M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle>" using EntCl[where \<Psi>="\<Psi>\<otimes>\<Psi>\<^isub>Q\<otimes>\<Psi>''"]
                        by(rule_tac statEqEnt[where \<Psi>="(\<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<Psi>'') \<otimes> \<Psi>\<^isub>c"]) (metis AssertionStatEqTrans Composition AssertionStatEqSym Associativity Commutativity)
                    }
                    thus ?thesis unfolding psi'def psirdef mbdef
                      by(subst perm_bool[where pi="[(a,b)]",symmetric]) (simp add: eqvts `a \<sharp> \<Psi>` `a \<sharp> \<Psi>\<^isub>Q` `b \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q` `a \<sharp> P` `b \<sharp> P`)
                  qed
                  moreover have "\<forall> \<Psi>'' Q. b \<sharp> \<Psi>'' \<and> \<Psi>' \<otimes> \<Psi>'' \<turnstile> M\<^isub>b \<Leftarrow> Q \<longrightarrow> Q = P \<parallel> \<langle>M\<^isub>b\<rangle>"
                  proof -
                    {
                      fix \<Psi>'' Q
                      assume "a \<sharp> \<Psi>''"
                      and    "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>'' \<turnstile> M\<^isub>a \<Leftarrow> Q"
                      hence "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>'') \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> Q"
                        by(rule_tac statEqEnt[where \<Psi>="((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>''"]) (metis AssertionStatEqTrans Composition AssertionStatEqSym Associativity Commutativity)
                      moreover have "a \<sharp> (\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>''" using `a \<sharp> \<Psi>` `a \<sharp> \<Psi>\<^isub>Q` `a \<sharp> \<Psi>''`
                        by auto
                      ultimately have "Q = P \<parallel> \<langle>M\<^isub>a\<rangle>"
                        by(rule_tac Equiv2)
                    }
                    thus ?thesis unfolding psi'def psirdef mbdef
                      by(subst perm_bool[where pi="[(a,b)]",symmetric]) (auto simp add: eqvts `a \<sharp> \<Psi>` `a \<sharp> \<Psi>\<^isub>Q` `b \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q` `a \<sharp> P` `b \<sharp> P` swap_simps)
                  qed
                  ultimately have "\<exists> P'. \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P' \<and> (\<Psi>, P', \<lparr>\<nu>b\<rparr>(Q'\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2" using mbdef psi'def `bn \<alpha> \<sharp>* a` `b \<sharp> bn \<alpha>` `b \<sharp> object \<alpha>` `bn \<alpha> \<sharp>* M\<^isub>a` `bn \<alpha> \<sharp>* P` `bn \<alpha> \<sharp>* \<Psi>\<^isub>c` `bn \<alpha> \<sharp>* \<Psi>` `bn \<alpha> \<sharp>* \<Psi>\<^isub>Q` `b \<sharp> P` Equiv1' EntCl' EntSupp' Equiv2' Inj' EntEnt' `a \<sharp> P` `\<not>(b \<sharp> ([(a,b)]\<bullet>M\<^isub>a))` `((supp P)::name set) \<subseteq> supp([(a,b)]\<bullet>M\<^isub>a)` `((supp([(a,b)]\<bullet>\<Psi>\<^isub>c))::name set) = supp([(a,b)]\<bullet>M\<^isub>a)` `b \<sharp> M\<^isub>a` `b \<sharp> \<Psi>\<^isub>c` `b \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q`
                  proof(nominal_induct avoiding: a b M\<^isub>a \<Psi> \<Psi>\<^isub>Q \<Psi>\<^isub>c rule: replInduct')
                    case(cAlpha \<alpha> P' p)
                    from `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` `b \<sharp> bn(p \<bullet> \<alpha>)` `bn \<alpha> \<sharp>* b` have "p \<bullet> b = b"
                      by simp
                    from `bn \<alpha> \<sharp>* b` have "b \<sharp> bn \<alpha>" by simp
                    from `b \<sharp> object(p \<bullet> \<alpha>)` have "b \<sharp> object \<alpha>"
                      by(subst perm_bool[where pi=p,symmetric]) (simp only: eqvts `p\<bullet>b=b`)
                    have "\<exists> P''. \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P'' \<and> (\<Psi>, P'', \<lparr>\<nu>b\<rparr>(P' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                      by(rule cAlpha) fact+
                    then obtain P'' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P'' \<and> (\<Psi>, P'', \<lparr>\<nu>b\<rparr>(P' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                      by(rule exE)
                    hence "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P''" and "(\<Psi>, P'', \<lparr>\<nu>b\<rparr>(P' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                      by simp+
                    from `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` `bn (p \<bullet> \<alpha>) \<sharp>* a` `bn \<alpha> \<sharp>* a` have "p \<bullet> a = a"
                      by simp
                    from `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` `bn (p \<bullet> \<alpha>) \<sharp>* M\<^isub>a` `bn \<alpha> \<sharp>* M\<^isub>a` have "p \<bullet> M\<^isub>a = M\<^isub>a"
                      by simp
                    from `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` `bn (p \<bullet> \<alpha>) \<sharp>* \<Psi>\<^isub>c` `bn \<alpha> \<sharp>* \<Psi>\<^isub>c` have "p \<bullet> \<Psi>\<^isub>c = \<Psi>\<^isub>c"
                      by simp
                    from `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` `bn (p \<bullet> \<alpha>) \<sharp>* P` `bn \<alpha> \<sharp>* P` have "p \<bullet> P = P"
                      by simp
                    from `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` `bn (p \<bullet> \<alpha>) \<sharp>* \<Psi>\<^isub>Q` `bn \<alpha> \<sharp>* \<Psi>\<^isub>Q` have "p \<bullet> \<Psi>\<^isub>Q = \<Psi>\<^isub>Q"
                      by simp
                    from `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` `bn (p \<bullet> \<alpha>) \<sharp>* \<Psi>` `bn \<alpha> \<sharp>* \<Psi>` have "p \<bullet> \<Psi> = \<Psi>"
                      by simp
                    from `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P''` have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> (p\<bullet>\<alpha>) \<prec> (p\<bullet>P'')"
                      by(subst (asm) perm_bool[where pi=p,symmetric])(simp only: eqvts `p\<bullet>\<Psi>=\<Psi>` `p\<bullet>\<Psi>\<^isub>Q=\<Psi>\<^isub>Q` `p\<bullet>P=P`)
                    moreover have "(\<Psi>, p\<bullet>P'', \<lparr>\<nu>b\<rparr>((p \<bullet> P') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    proof -
                      from `(\<Psi>, P'', \<lparr>\<nu>b\<rparr>(P' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2` `eqvt ?Y2` have "(p\<bullet>(\<Psi>, P'', \<lparr>\<nu>b\<rparr>(P' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>))) \<in> ?Y2"
                        by(rule_tac eqvtI)
                      moreover have "(p\<bullet>(\<Psi>, P'', \<lparr>\<nu>b\<rparr>(P' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>))) = (\<Psi>, p\<bullet>P'', \<lparr>\<nu>b\<rparr>((p \<bullet> P') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>))"
                        by(simp add: eqvts  `p\<bullet>\<Psi>=\<Psi>` `p\<bullet>\<Psi>\<^isub>Q=\<Psi>\<^isub>Q` `p\<bullet>P=P` `p\<bullet>b=b` `p\<bullet>a=a` `p\<bullet>\<Psi>\<^isub>c=\<Psi>\<^isub>c` cp1[OF cp_pt_inst, OF pt_name_inst, OF at_name_inst, of p "[(a,b)]" \<Psi>\<^isub>c])
                      ultimately show ?thesis
                        by simp
                    qed
                    ultimately show ?case
                      by(rule_tac exI) (rule conjI)
                  next
                    case(cPar1 \<alpha> P')
                    have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> P \<longmapsto> \<alpha> \<prec> P'"
                    proof -
                      have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<rhd> P \<longmapsto> \<alpha> \<prec> P'"
                        apply(rule statEqTransition)
                        apply fact
                        apply(subst cPar1)
                        apply(rule AssertionStatEqRefl)
                        done
                      thus ?thesis
                        apply(rule replTrans[where a="b" and M\<^isub>a="[(a,b)]\<bullet>M\<^isub>a"])
                        by(fact | rule freshComp)+
                    qed
                    hence "b \<sharp> P'"
                      by(rule freeFreshDerivative') fact+
                    have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P' \<parallel> !P"
                      apply(rule_tac Bang)
                      apply(rule Par1[where A\<^isub>Q="[]" and \<Psi>\<^isub>Q=\<one>])
                      apply(rule statEqTransition[OF `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> P \<longmapsto> \<alpha> \<prec> P'`])
                      apply(rule AssertionStatEqSym[OF Identity])
                      by(fact|simp add: `bn \<alpha> \<sharp>* P`)+
                    moreover have "(\<Psi>,P'\<parallel>!P,\<lparr>\<nu>b\<rparr>(P' \<parallel> \<langle>M\<^isub>b\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    proof -
                      have "\<Psi> \<rhd> P'\<parallel>!P \<sim> P' \<parallel> !P" by(rule bisimReflexive)
                      moreover have "\<Psi> \<rhd> P' \<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>b\<rparr>(P' \<parallel> \<langle>M\<^isub>b\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)"
                        apply(subst `M\<^isub>b = [(a,b)]\<bullet>M\<^isub>a`)
                        apply(rule bisimSymmetric)
                        apply(rule bisimTransitive[OF bisimResPres, OF bisimParAssoc, OF `b \<sharp> \<Psi>`])
                        apply(rule bisimScopeExt[OF `b \<sharp> P'`])
                        done
                      moreover have "(\<Psi>,P' \<parallel> !P, P' \<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?X2"
                      proof -
                        note `b \<sharp> \<Psi>` `b \<sharp> P` `b \<sharp> P'` `\<not> b \<sharp> ([(a,b)]\<bullet>M\<^isub>a)`
                        moreover have "(\<forall>\<Psi>. \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)" using cPar1
                          by simp
                        moreover have "\<forall> F. F \<otimes>\<^sub>F (\<langle>[b],([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rangle>) \<simeq>\<^sub>F F" using cPar1 by simp
                        moreover have "(\<forall>\<Psi> R. b \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>R \<longrightarrow> R = P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)" using cPar1
                          by simp
                        moreover note `guarded P` `supp P \<subseteq> supp([(a,b)]\<bullet>M\<^isub>a)` `supp([(a,b)]\<bullet>\<Psi>\<^isub>c) = (supp([(a,b)]\<bullet>M\<^isub>a)::name set)`
                        moreover have "(\<forall>P P'. (([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P = ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P') = (P = P'))" using cPar1
                          by simp
                        moreover have "(\<forall>\<Psi> P. \<Psi> \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<longrightarrow> supp([(a,b)]\<bullet>M\<^isub>a) \<subseteq> (supp \<Psi>::name set))"
                          using cPar1 by auto
                        moreover have "([]::name list) \<sharp>* \<Psi>" by simp
                        moreover have "(\<forall>\<Psi> \<phi>. \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<or> \<Psi> \<turnstile> \<phi>)" using cPar1 by simp
                        ultimately have "(\<Psi>,\<lparr>\<nu>*[]\<rparr>P' \<parallel> !P, \<lparr>\<nu>*[]\<rparr>(P' \<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>))) \<in> ?X2"
                          by blast
                        thus ?thesis by simp
                      qed                        
                      ultimately show ?thesis
                        apply(rule_tac lol)
                        apply(rule refl)
                        apply(assumption)
                        apply(rule disjI1)
                        by -
                    qed
                    ultimately show ?case
                      by(rule_tac exI) (rule conjI)
                  next
                    case(cComm1 M N P' K xvec P'')
                    have "b \<sharp> N"
                    proof -
                      note `\<Psi>' \<rhd> \<langle>M\<^isub>b\<rangle> \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P''`
                      moreover note `guarded P`
                      moreover have "\<forall> \<Psi>''. \<Psi>' \<otimes> \<Psi>'' \<turnstile> M\<^isub>b \<Leftarrow> P \<parallel> \<langle>M\<^isub>b\<rangle>"
                      proof -
                        {
                          fix \<Psi>''
                          have seq: "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<Psi>'' \<simeq>  (\<Psi>'' \<otimes> ((\<Psi> \<otimes> \<Psi>\<^isub>Q))) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)"
                            by(metis AssertionStatEqTrans Commutativity AssertionStatEqSym Associativity AssertionStatEqRefl)
                          have "\<Psi>' \<otimes> \<Psi>'' \<turnstile> M\<^isub>b \<Leftarrow> P \<parallel> \<langle>M\<^isub>b\<rangle>"
                            unfolding cComm1
                            apply(rule statEqEnt[OF AssertionStatEqSym, OF seq])
                            by fact
                        }
                        thus ?thesis by simp
                      qed
                      moreover have "\<forall> \<Psi>'' Q. b \<sharp> \<Psi>'' \<and> \<Psi>' \<otimes> \<Psi>'' \<turnstile> M\<^isub>b \<Leftarrow> Q \<longrightarrow> Q = P \<parallel> \<langle>M\<^isub>b\<rangle>"
                      proof -
                        {
                          fix \<Psi>'' Q
                          assume "b \<sharp> \<Psi>''" and "\<Psi>' \<otimes> \<Psi>'' \<turnstile> M\<^isub>b \<Leftarrow> Q"
                          have seq: "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<Psi>'' \<simeq>  (\<Psi>'' \<otimes> ((\<Psi> \<otimes> \<Psi>\<^isub>Q))) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)"
                            by(metis AssertionStatEqTrans Commutativity AssertionStatEqSym Associativity AssertionStatEqRefl)
                          have "(\<Psi>'' \<otimes> ((\<Psi> \<otimes> \<Psi>\<^isub>Q))) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> M\<^isub>b \<Leftarrow> Q" using `\<Psi>' \<otimes> \<Psi>'' \<turnstile> M\<^isub>b \<Leftarrow> Q` unfolding cComm1
                            by(rule statEqEnt[OF seq])
                          moreover have "b \<sharp> (\<Psi>'' \<otimes> ((\<Psi> \<otimes> \<Psi>\<^isub>Q)))" using `b \<sharp> \<Psi>` `b \<sharp> \<Psi>''` `b \<sharp> \<Psi>\<^isub>Q` by force
                          ultimately have "Q = P \<parallel> \<langle>M\<^isub>b\<rangle>"
                            unfolding cComm1
                            by(rule_tac cComm1)
                        }
                        thus ?thesis by auto
                      qed
                      ultimately have "b \<sharp> bnlNames(\<lparr>\<nu>*xvec\<rparr>N \<prec>' P'')" using `b \<sharp> P`
                      proof(nominal_induct avoiding: b rule: replOutputInduct)
                        case cInv thus ?case by -
                      next
                        case(cPar1 xvec N P')
                        from `\<Psi>' \<rhd> P \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'`  `xvec \<sharp>* b` `b \<sharp> P` `distinct xvec` `xvec \<sharp>* K`
                        have "b \<sharp> N"
                          by(rule_tac outputFreshDerivative) auto
                        thus ?case unfolding bnlNamesJudge fresh_def
                          by(auto simp add: name_set_supp[OF finite_Diff, OF fin_supp])
                      next
                        case(cPar2 xvec N P')
                        have "b \<sharp> bnlNames (\<lparr>\<nu>*xvec\<rparr>N \<prec>' P')" by(rule cPar2) fact
                        thus ?case unfolding bnlNamesJudge by -
                      qed
                      moreover have "b \<sharp> xvec" using `xvec \<sharp>* b` by simp
                      ultimately show "b \<sharp> N" unfolding bnlNamesJudge
                        by(auto simp add: name_set_supp[OF finite_Diff, OF fin_supp] fresh_def)
                    qed
                    hence "b \<sharp> P'" using `\<Psi>' \<rhd> P \<longmapsto> M\<lparr>N\<rparr> \<prec> P'` `b \<sharp> P`
                      by(rule_tac inputFreshDerivative)
                    have "a \<sharp> xvec" and "b \<sharp> xvec" using `xvec \<sharp>* a` `xvec \<sharp>* b`
                      by simp+
                    have "\<exists> P'. \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K \<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P' \<and> (\<Psi>, P', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                      by(rule_tac cComm1) (fact|simp add: `b \<sharp> N` `b \<sharp> xvec` `a \<sharp> xvec`)+
                    then obtain Q' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K \<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q' \<and> (\<Psi>, Q', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                      by(rule exE)
                    hence "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K \<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q'" and "(\<Psi>, Q', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                      by simp+
                    have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<rhd> P \<longmapsto> M\<lparr>N\<rparr> \<prec> P'" using `\<Psi>' \<rhd> P \<longmapsto> M\<lparr>N\<rparr> \<prec> P'` by(subst (asm) cComm1)
                    have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<one> \<rhd> P \<longmapsto> M\<lparr>N\<rparr> \<prec> P'"
                      apply(rule_tac statEqTransition[where \<Psi>="\<Psi>\<otimes>\<Psi>\<^isub>Q"])
                      apply(rule replTrans)
                      by(fact|simp|rule freshComp|rule AssertionStatEqSym[OF Identity])+
                    have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<one> \<rhd> !P \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q'"
                      apply(rule statEqTransition)
                      by(fact) (rule AssertionStatEqSym[OF Identity])
                    obtain A\<^isub>P \<Psi>\<^isub>P where "extractFrame P = \<langle>A\<^isub>P,\<Psi>\<^isub>P\<rangle>" and "A\<^isub>P \<sharp>* \<Psi>" and "A\<^isub>P \<sharp>* \<Psi>\<^isub>Q" and "A\<^isub>P \<sharp>* !P" and "A\<^isub>P \<sharp>* P" and "A\<^isub>P \<sharp>* M" and "distinct A\<^isub>P"
                      by(rule freshFrame[where F="extractFrame P" and C="(\<Psi>,\<Psi>\<^isub>Q,P,M)"]) force
                    have "\<Psi>\<^isub>P \<simeq> \<one>" and "supp \<Psi>\<^isub>P = ({}::name set)" using `guarded P` `extractFrame P = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>`
                      by(auto dest: guardedStatEq)
                    have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>P \<rhd> !P \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q'"
                      apply(rule statEqTransition)
                      apply fact
                      apply(rule compositionSym)
                      apply(rule AssertionStatEqSym)
                      by fact
                    have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>P \<otimes> \<one> \<turnstile> M \<leftrightarrow> K"
                    proof -
                      from `\<Psi>' \<turnstile> M \<leftrightarrow> K` have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<turnstile> M \<leftrightarrow> K" using DisjointCCl[where M=M and N=K]
                        unfolding cComm1
                        apply(drule_tac cComm1)
                        apply(erule disjE)
                        by blast
                      thus ?thesis
                        apply(rule_tac statEqEnt)
                        apply(rule AssertionStatEqSym[OF AssertionStatEqTrans[OF compositionSym[OF Identity]]])
                        apply(rule AssertionStatEqTrans[OF compositionSym[OF `\<Psi>\<^isub>P \<simeq> \<one>`]])
                        by(rule Identity)
                    qed
                    have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<tau> \<prec> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q'"
                      apply(rule Bang)
                      apply(rule Comm1[where M=M and K=K and N=N and A\<^isub>P="A\<^isub>P" and \<Psi>\<^isub>P="\<Psi>\<^isub>P" and A\<^isub>Q="[]" and \<Psi>\<^isub>Q="\<one>"])
                      by(fact|rule freshCompChain|simp)+
                    moreover have "(\<Psi>, \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q', \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    proof -
                      from `(\<Psi>, Q', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2` obtain Q'' P''' where "\<Psi> \<rhd> Q' \<sim> Q''" and Disj: "(\<Psi>,Q'',P''') \<in> ?X2 \<or> \<Psi> \<rhd> Q'' \<sim> P'''" and "\<Psi> \<rhd> P''' \<sim> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                        by auto blast
                      have "xvec \<sharp>* \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>" using `xvec \<sharp>* a` `xvec \<sharp>* b` `xvec \<sharp>* \<Psi>\<^isub>c`
                        by simp
                      have rhs: "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>))"
                        apply(rule bisimTransitive[OF bisimResPres])
                        apply(rule bisimTransitive[OF bisimParComm])
                        apply(rule bisimSymmetric[OF bisimScopeExtChain])
                        apply fact+
                        apply(rule bisimTransitive[OF bisimResComm'])
                        apply fact
                        apply fact
                        apply(rule bisimResChainPres)
                        apply(rule bisimSymmetric)
                        apply(rule bisimTransitive[OF bisimSymmetric[OF bisimScopeExt]])
                        apply fact
                        apply(rule bisimResPres)
                        apply(rule bisimTransitive[OF bisimSymmetric[OF bisimParAssoc]])
                        apply(rule bisimTransitive[OF bisimParComm])
                        apply(rule bisimReflexive)
                        by fact+
                      from Disj show ?thesis
                      proof(rule disjE)
                        assume "\<Psi> \<rhd> Q'' \<sim> P'''"
                        have bis: "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q' \<sim> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)"
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF rhs])
                          apply(rule bisimResChainPres)
                          apply(rule bisimParPres')
                          apply(rule bisimTransitive[OF bisimSymmetric[OF `\<Psi> \<rhd> P''' \<sim> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`]])
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF `\<Psi> \<rhd> Q' \<sim> Q''`])
                          apply(rule `\<Psi> \<rhd> Q'' \<sim> P'''`)
                          by fact
                        show ?thesis
                          apply(rule lol)
                          apply(rule refl)
                          apply(rule bis)
                          apply(rule disjI2)
                          by(rule bisimReflexive)+
                      next
                        assume "(\<Psi>,Q'',P''') \<in> ?X2"
                        then obtain yvec Q\<^isub>i a\<^isub>i M\<^isub>i \<Psi>\<^isub>i P\<^isub>i where p'''def: "P''' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)" and q''def: "Q'' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i" and "a\<^isub>i \<sharp> \<Psi>" and "a\<^isub>i \<sharp> P\<^isub>i" and "a\<^isub>i \<sharp> Q\<^isub>i" and "\<not> a\<^isub>i \<sharp> M\<^isub>i" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i\<Leftarrow>P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a\<^isub>i], \<Psi>\<^isub>i\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i \<Leftarrow> R \<longrightarrow> (R = (P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>)))" and "guarded P\<^isub>i" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>i\<Leftarrow>P \<longrightarrow> (supp M\<^isub>i) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>i\<Leftarrow>P = M\<^isub>i\<Leftarrow>P') = (P = P'))" and "((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i" and "((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i" and "yvec \<sharp>* \<Psi>" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>i \<Leftarrow> P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
                          by blast
                        have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))))"
                        proof -
                          have "xvec \<sharp>* ([(a,b)] \<bullet> \<Psi>\<^isub>c)" using `a \<sharp> xvec` `b \<sharp> xvec` `xvec \<sharp>* \<Psi>\<^isub>c`
                            by simp
                          have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>b\<rparr>(\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                            apply(rule bisimResPres)
                            apply(rule bisimTransitive[OF bisimParComm])
                            apply(rule bisimSymmetric)
                            apply(rule bisimTransitive[OF bisimResChainPres])
                            apply(rule bisimParComm)
                            apply fact
                            apply(rule bisimScopeExtChain)
                            by(fact|simp add: `xvec \<sharp>* ([(a,b)] \<bullet> \<Psi>\<^isub>c)`)+
                          moreover have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>(\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>b\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                            by(rule bisimResComm') fact+
                          moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>b\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>(P'\<parallel> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                            apply(rule bisimResChainPres)
                            apply(rule bisimTransitive[OF bisimResPres[OF bisimParAssoc]])
                            apply fact
                            apply(rule bisimScopeExt)
                            by fact+
                          moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(P'\<parallel> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> (\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))))"
                            apply(rule bisimResChainPres)
                            apply(rule bisimTransitive[OF bisimParComm])
                            apply(rule bisimSymmetric[OF bisimTransitive[OF bisimParComm]])
                            apply(rule bisimParPres)
                            apply(rule bisimSymmetric)
                            apply(rule bisimTransitive[OF bisimSymmetric[OF `\<Psi> \<rhd> P''' \<sim> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`]])
                            apply(subst p'''def)
                            by(rule bisimReflexive) fact
                          ultimately show ?thesis
                            by(auto dest: bisimTransitive)
                        qed
                        obtain p::"name prm" where "set p \<subseteq> set yvec \<times> set(p\<bullet>yvec)" and "(p\<bullet>yvec) \<sharp>* a\<^isub>i" and "distinctPerm p" and "(p\<bullet>yvec) \<sharp>* M\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>\<^isub>i" and "(p\<bullet>yvec) \<sharp>* P\<^isub>i" and "(p\<bullet>yvec) \<sharp>* Q\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>" and "(p\<bullet>yvec) \<sharp>* P'" and "(p\<bullet>yvec) \<sharp>* xvec" and "(p\<bullet>yvec) \<sharp>* A\<^isub>Q"
                          by(rule chain.name_list_avoiding[where xvec=yvec and c="(M\<^isub>i,\<Psi>\<^isub>i,a\<^isub>i,P\<^isub>i,Q\<^isub>i,\<Psi>,P',xvec,A\<^isub>Q)"]) auto
                        hence yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>(p\<bullet>a\<^isub>i)\<rparr>(\<langle>(p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" and yvrdef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)"
                          by(subst resChainAlpha[where p=p],simp,simp,simp)+
                        obtain b\<^isub>i::name where "b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)" and "b\<^isub>i \<sharp> \<Psi>" and "b\<^isub>i \<sharp> \<Psi>\<^isub>Q" and "b\<^isub>i \<sharp> P'" and "b\<^isub>i \<sharp> A\<^isub>Q"
                          apply(rule at_exists_fresh[OF at_name_inst, where x="(p\<bullet>M\<^isub>i,p\<bullet>\<Psi>\<^isub>i,p\<bullet>P\<^isub>i,p\<bullet>Q\<^isub>i,\<Psi>,\<Psi>\<^isub>Q,P',A\<^isub>Q)"])
                          unfolding supp_prod pt_perm_supp[OF pt_name_inst,OF at_name_inst, symmetric]
                          apply(rule finite_UnI, (finite_guess | (subst pt_set_finite_ineq[OF pt_name_inst,OF at_name_inst], finite_guess)))+
                          apply(finite_guess)
                          by force
                        have yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" unfolding yvldef
                          by(subst alphaRes[where x="(p\<bullet>a\<^isub>i)" and y="b\<^isub>i"]) (simp add: `b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)`)+
                        have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (\<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                        proof -
                          note `\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))`
                          moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                            unfolding yvldef
                            by(rule bisimReflexive)
                          moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)) \<sim> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>P'\<parallel>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)"
                            apply(subst resChainAppend)
                            apply(rule bisimResChainPres)
                            apply(rule bisimSymmetric)
                            apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                            apply fact
                            apply(rule bisimScopeExtChain)
                            by fact+
                          ultimately show ?thesis
                            by(auto dest: bisimTransitive)
                        qed
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q'"
                        proof -
                          have "\<Psi> \<rhd> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i))" unfolding resChainAppend
                            by(rule bisimReflexive)
                          moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i)) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)))"
                            apply(rule bisimResChainPres)
                            apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                            apply fact
                            apply(rule bisimScopeExtChain)
                            by fact+
                          moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i))) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> Q')" using `\<Psi> \<rhd> Q' \<sim> Q''` `xvec \<sharp>* \<Psi>`
                            unfolding yvrdef[symmetric] q''def[symmetric]
                            by(metis bisimResChainPres bisimTransitive bisimParComm bisimSymmetric bisimParPres)
                          ultimately show ?thesis
                            by(auto dest: bisimTransitive)
                        qed
                        moreover have "(\<Psi>, \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i),(\<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))) \<in> ?X2"
                        proof -
                          have "(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)" using `a\<^isub>i \<sharp> P\<^isub>i` by(subst fresh_bij)
                          note `b\<^isub>i \<sharp> \<Psi>`
                          moreover note `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`
                          moreover have "b\<^isub>i \<sharp> (P' \<parallel> (p\<bullet>Q\<^isub>i))" using `b\<^isub>i \<sharp> P'` `b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)`
                            by simp
                          moreover have "guarded(p\<bullet>P\<^isub>i)" using `guarded P\<^isub>i` by simp
                          moreover have "\<not> b\<^isub>i \<sharp> ([(p\<bullet>a\<^isub>i,b\<^isub>i)] \<bullet> (p\<bullet>M\<^isub>i))" using `\<not> a\<^isub>i \<sharp> M\<^isub>i`
                            apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                            apply(subst (asm) perm_bool[where pi="p",symmetric])
                            by(simp add: eqvts swap_simps)
                          moreover have "\<And> \<Psi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>" using EntCl
                            apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                            apply(subst perm_bool[where pi="rev p",symmetric])
                            by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`)
                          moreover have "\<forall>F. (F \<otimes>\<^sub>F (\<langle>[b\<^isub>i],[(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)\<rangle>) \<simeq>\<^sub>F F)"
                            using Equiv1
                            apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                            apply(subst perm_bool[where pi="rev p",symmetric])
                            by(simp add: eqvts swap_simps)
                          moreover have "(\<forall>\<Psi> R. b\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i)) \<Leftarrow> R \<longrightarrow> (R = ((p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>)))" using Equiv2
                            apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                            apply(subst (asm) perm_bool[where pi="p",symmetric])
                            by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i` swap_simps)
                          moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P \<longrightarrow> (supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))) \<subseteq> ((supp \<Psi>)::name set))" using EntSupp
                            apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                            apply(subst (asm) perm_bool[where pi="p",symmetric])
                            by(simp add: eqvts swap_simps)
                          moreover have "\<forall> P P'. ((([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P') = (P = P'))" using Inj
                            apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                            apply(subst (asm) perm_bool[where pi="p",symmetric])
                            by(simp add: eqvts swap_simps)
                          moreover have "((supp(p\<bullet>P\<^isub>i))::name set) \<subseteq> supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i`
                            apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                            apply(subst (asm) perm_bool[where pi="p",symmetric])
                            by(simp add: eqvts swap_simps `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                          moreover have "((supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)))::name set) = supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i`
                            apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                            apply(subst (asm) perm_bool[where pi="p",symmetric])
                            by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                          moreover have "(xvec@(p\<bullet>yvec)) \<sharp>* \<Psi>" using `xvec \<sharp>* \<Psi>` `(p\<bullet>yvec) \<sharp>* \<Psi>`
                            by simp
                          moreover from EntEnt have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i)\<parallel>\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                            apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                            apply(subst perm_bool[where pi="p",symmetric])
                            apply(simp add: eqvts swap_simps `distinctPerm p` `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                            done
                          ultimately show ?thesis
                            by blast
                        qed
                        ultimately show ?thesis
                          apply(rule_tac lol)
                          apply(rule refl)
                          apply(rule bisimSymmetric)
                          apply assumption
                          apply(rule disjI1)
                          apply assumption
                          by(rule bisimSymmetric)
                      qed
                    qed
                    ultimately show ?case
                      by(rule_tac exI) (rule conjI)
                  next
                    case(cInv \<alpha> P')
                    show ?case
                      by(rule cInv) fact+
                  next
                    case(cComm2 M xvec N P' K P'')
                    have "b \<sharp> N" and "b \<sharp> P'" using `\<Psi>' \<rhd> P \<longmapsto> M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'` `xvec \<sharp>* b` `b \<sharp> P` `distinct xvec` `xvec \<sharp>* M`
                      by(auto intro: outputFreshDerivative)
                    have "a \<sharp> xvec" and "b \<sharp> xvec" using `xvec \<sharp>* a` `xvec \<sharp>* b`
                      by simp+
                    have "\<exists> P'. \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K\<lparr>N\<rparr> \<prec> P' \<and> (\<Psi>, P', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                      by(rule_tac cComm2) (fact|simp add: `b \<sharp> N` `b \<sharp> xvec` `a \<sharp> xvec`)+
                    then obtain Q' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K \<lparr>N\<rparr> \<prec> Q' \<and> (\<Psi>, Q', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                      by(rule exE)
                    hence "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K \<lparr>N\<rparr> \<prec> Q'" and "(\<Psi>, Q', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                      by simp+
                    have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<rhd> P \<longmapsto> M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'" using `\<Psi>' \<rhd> P \<longmapsto> M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'` by(subst (asm) cComm2)
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<one> \<rhd> P \<longmapsto> M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'"
                    apply(rule_tac statEqTransition[where \<Psi>="\<Psi>\<otimes>\<Psi>\<^isub>Q"])
                    apply(rule replTrans)
                    by(fact|simp|rule freshComp|rule AssertionStatEqSym[OF Identity])+
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<one> \<rhd> !P \<longmapsto> K\<lparr>N\<rparr> \<prec> Q'"
                    apply(rule statEqTransition)
                    by(fact) (rule AssertionStatEqSym[OF Identity])
                  obtain A\<^isub>P \<Psi>\<^isub>P where "extractFrame P = \<langle>A\<^isub>P,\<Psi>\<^isub>P\<rangle>" and "A\<^isub>P \<sharp>* \<Psi>" and "A\<^isub>P \<sharp>* \<Psi>\<^isub>Q" and "A\<^isub>P \<sharp>* !P" and "A\<^isub>P \<sharp>* P" and "A\<^isub>P \<sharp>* M" and "distinct A\<^isub>P"
                    by(rule freshFrame[where F="extractFrame P" and C="(\<Psi>,\<Psi>\<^isub>Q,P,M)"]) force
                  have "\<Psi>\<^isub>P \<simeq> \<one>" and "supp \<Psi>\<^isub>P = ({}::name set)" using `guarded P` `extractFrame P = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>`
                    by(auto dest: guardedStatEq)
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>P \<rhd> !P \<longmapsto> K\<lparr>N\<rparr> \<prec> Q'"
                    apply(rule statEqTransition)
                    apply fact
                    apply(rule compositionSym)
                    apply(rule AssertionStatEqSym)
                    by fact
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>P \<otimes> \<one> \<turnstile> M \<leftrightarrow> K"
                  proof -
                    from `\<Psi>' \<turnstile> M \<leftrightarrow> K` have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<turnstile> M \<leftrightarrow> K" using DisjointCCl[where M=M and N=K]
                      unfolding cComm2
                      apply(drule_tac cComm2)
                      apply(erule disjE)
                      by blast
                    thus ?thesis
                      apply(rule_tac statEqEnt)
                      apply(rule AssertionStatEqSym[OF AssertionStatEqTrans[OF compositionSym[OF Identity]]])
                      apply(rule AssertionStatEqTrans[OF compositionSym[OF `\<Psi>\<^isub>P \<simeq> \<one>`]])
                      by(rule Identity)
                  qed
                  have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<tau> \<prec> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q'"
                    apply(rule Bang)
                    apply(rule Comm2[where M=M and K=K and N=N and A\<^isub>P="A\<^isub>P" and \<Psi>\<^isub>P="\<Psi>\<^isub>P" and A\<^isub>Q="[]" and \<Psi>\<^isub>Q="\<one>"])
                    by(fact|rule freshCompChain|simp)+
                  moreover have "(\<Psi>, \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q', \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                  proof -
                    from `(\<Psi>, Q', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2` obtain Q'' P''' where "\<Psi> \<rhd> Q' \<sim> Q''" and Disj: "(\<Psi>,Q'',P''') \<in> ?X2 \<or> \<Psi> \<rhd> Q'' \<sim> P'''" and "\<Psi> \<rhd> P''' \<sim> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                      by auto blast
                    have "xvec \<sharp>* \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>" using `xvec \<sharp>* a` `xvec \<sharp>* b` `xvec \<sharp>* \<Psi>\<^isub>c`
                      by simp
                    have rhs: "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>))"
                      apply(rule bisimTransitive[OF bisimResPres])
                      apply(rule bisimTransitive[OF bisimParComm])
                      apply(rule bisimSymmetric[OF bisimScopeExtChain])
                      apply fact+
                      apply(rule bisimTransitive[OF bisimResComm'])
                      apply fact
                      apply fact
                      apply(rule bisimResChainPres)
                      apply(rule bisimSymmetric)
                      apply(rule bisimTransitive[OF bisimSymmetric[OF bisimScopeExt]])
                      apply fact
                      apply(rule bisimResPres)
                      apply(rule bisimTransitive[OF bisimSymmetric[OF bisimParAssoc]])
                      apply(rule bisimTransitive[OF bisimParComm])
                      apply(rule bisimReflexive)
                      by fact+
                    from Disj show ?thesis
                    proof(rule disjE)
                      assume "\<Psi> \<rhd> Q'' \<sim> P'''"
                      have bis: "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q' \<sim> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)"
                        apply(rule bisimSymmetric)
                        apply(rule bisimTransitive[OF rhs])
                        apply(rule bisimResChainPres)
                        apply(rule bisimParPres')
                        apply(rule bisimTransitive[OF bisimSymmetric[OF `\<Psi> \<rhd> P''' \<sim> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`]])
                        apply(rule bisimSymmetric)
                        apply(rule bisimTransitive[OF `\<Psi> \<rhd> Q' \<sim> Q''`])
                        apply(rule `\<Psi> \<rhd> Q'' \<sim> P'''`)
                        by fact
                      show ?thesis
                        apply(rule lol)
                        apply(rule refl)
                        apply(rule bis)
                        apply(rule disjI2)
                        by(rule bisimReflexive)+
                    next
                      assume "(\<Psi>,Q'',P''') \<in> ?X2"
                      then obtain yvec Q\<^isub>i a\<^isub>i M\<^isub>i \<Psi>\<^isub>i P\<^isub>i where p'''def: "P''' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)" and q''def: "Q'' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i" and "a\<^isub>i \<sharp> \<Psi>" and "a\<^isub>i \<sharp> P\<^isub>i" and "a\<^isub>i \<sharp> Q\<^isub>i" and "\<not> a\<^isub>i \<sharp> M\<^isub>i" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i\<Leftarrow>P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a\<^isub>i], \<Psi>\<^isub>i\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i \<Leftarrow> R \<longrightarrow> (R = (P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>)))" and "guarded P\<^isub>i" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>i\<Leftarrow>P \<longrightarrow> (supp M\<^isub>i) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>i\<Leftarrow>P = M\<^isub>i\<Leftarrow>P') = (P = P'))" and "((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i" and "((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i" and "yvec \<sharp>* \<Psi>" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>i \<Leftarrow> P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
                        by blast
                      have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))))"
                      proof -
                        have "xvec \<sharp>* ([(a,b)] \<bullet> \<Psi>\<^isub>c)" using `a \<sharp> xvec` `b \<sharp> xvec` `xvec \<sharp>* \<Psi>\<^isub>c`
                          by simp
                        have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>b\<rparr>(\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                          apply(rule bisimResPres)
                          apply(rule bisimTransitive[OF bisimParComm])
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimResChainPres])
                          apply(rule bisimParComm)
                          apply fact
                          apply(rule bisimScopeExtChain)
                          by(fact|simp add: `xvec \<sharp>* ([(a,b)] \<bullet> \<Psi>\<^isub>c)`)+
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>(\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>b\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                          by(rule bisimResComm') fact+
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>b\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>(P'\<parallel> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                          apply(rule bisimResChainPres)
                          apply(rule bisimTransitive[OF bisimResPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExt)
                          by fact+
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(P'\<parallel> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> (\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))))"
                          apply(rule bisimResChainPres)
                          apply(rule bisimTransitive[OF bisimParComm])
                          apply(rule bisimSymmetric[OF bisimTransitive[OF bisimParComm]])
                          apply(rule bisimParPres)
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimSymmetric[OF `\<Psi> \<rhd> P''' \<sim> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`]])
                          apply(subst p'''def)
                          by(rule bisimReflexive) fact
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      obtain p::"name prm" where "set p \<subseteq> set yvec \<times> set(p\<bullet>yvec)" and "(p\<bullet>yvec) \<sharp>* a\<^isub>i" and "distinctPerm p" and "(p\<bullet>yvec) \<sharp>* M\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>\<^isub>i" and "(p\<bullet>yvec) \<sharp>* P\<^isub>i" and "(p\<bullet>yvec) \<sharp>* Q\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>" and "(p\<bullet>yvec) \<sharp>* P'" and "(p\<bullet>yvec) \<sharp>* xvec" and "(p\<bullet>yvec) \<sharp>* A\<^isub>Q"
                        by(rule chain.name_list_avoiding[where xvec=yvec and c="(M\<^isub>i,\<Psi>\<^isub>i,a\<^isub>i,P\<^isub>i,Q\<^isub>i,\<Psi>,P',xvec,A\<^isub>Q)"]) auto
                      hence yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>(p\<bullet>a\<^isub>i)\<rparr>(\<langle>(p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" and yvrdef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)"
                        by(subst resChainAlpha[where p=p],simp,simp,simp)+
                      obtain b\<^isub>i::name where "b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)" and "b\<^isub>i \<sharp> \<Psi>" and "b\<^isub>i \<sharp> \<Psi>\<^isub>Q" and "b\<^isub>i \<sharp> P'" and "b\<^isub>i \<sharp> A\<^isub>Q"
                        apply(rule at_exists_fresh[OF at_name_inst, where x="(p\<bullet>M\<^isub>i,p\<bullet>\<Psi>\<^isub>i,p\<bullet>P\<^isub>i,p\<bullet>Q\<^isub>i,\<Psi>,\<Psi>\<^isub>Q,P',A\<^isub>Q)"])
                        unfolding supp_prod pt_perm_supp[OF pt_name_inst,OF at_name_inst, symmetric]
                        apply(rule finite_UnI, (finite_guess | (subst pt_set_finite_ineq[OF pt_name_inst,OF at_name_inst], finite_guess)))+
                        apply(finite_guess)
                        by force
                      have yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" unfolding yvldef
                        by(subst alphaRes[where x="(p\<bullet>a\<^isub>i)" and y="b\<^isub>i"]) (simp add: `b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)`)+
                      have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (\<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                      proof -
                        note `\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))`
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                          unfolding yvldef
                          by(rule bisimReflexive)
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)) \<sim> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>P'\<parallel>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)"
                          apply(subst resChainAppend)
                          apply(rule bisimResChainPres)
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExtChain)
                          by fact+
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      moreover have "\<Psi> \<rhd> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q'"
                      proof -
                        have "\<Psi> \<rhd> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i))" unfolding resChainAppend
                          by(rule bisimReflexive)
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i)) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)))"
                          apply(rule bisimResChainPres)
                          apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExtChain)
                          by fact+
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i))) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> Q')" using `\<Psi> \<rhd> Q' \<sim> Q''` `xvec \<sharp>* \<Psi>`
                          unfolding yvrdef[symmetric] q''def[symmetric]
                          by(metis bisimResChainPres bisimTransitive bisimParComm bisimSymmetric bisimParPres)
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      moreover have "(\<Psi>, \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i),(\<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))) \<in> ?X2"
                      proof -
                        have "(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)" using `a\<^isub>i \<sharp> P\<^isub>i` by(subst fresh_bij)
                        note `b\<^isub>i \<sharp> \<Psi>`
                        moreover note `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`
                        moreover have "b\<^isub>i \<sharp> (P' \<parallel> (p\<bullet>Q\<^isub>i))" using `b\<^isub>i \<sharp> P'` `b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)`
                          by simp
                        moreover have "guarded(p\<bullet>P\<^isub>i)" using `guarded P\<^isub>i` by simp
                        moreover have "\<not> b\<^isub>i \<sharp> ([(p\<bullet>a\<^isub>i,b\<^isub>i)] \<bullet> (p\<bullet>M\<^isub>i))" using `\<not> a\<^isub>i \<sharp> M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "\<And> \<Psi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>" using EntCl
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="rev p",symmetric])
                          by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`)
                        moreover have "\<forall>F. (F \<otimes>\<^sub>F (\<langle>[b\<^isub>i],[(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)\<rangle>) \<simeq>\<^sub>F F)"
                          using Equiv1
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="rev p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "(\<forall>\<Psi> R. b\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i)) \<Leftarrow> R \<longrightarrow> (R = ((p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>)))" using Equiv2
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i` swap_simps)
                        moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P \<longrightarrow> (supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))) \<subseteq> ((supp \<Psi>)::name set))" using EntSupp
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "\<forall> P P'. ((([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P') = (P = P'))" using Inj
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "((supp(p\<bullet>P\<^isub>i))::name set) \<subseteq> supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                        moreover have "((supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)))::name set) = supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                        moreover have "(xvec@(p\<bullet>yvec)) \<sharp>* \<Psi>" using `xvec \<sharp>* \<Psi>` `(p\<bullet>yvec) \<sharp>* \<Psi>`
                          by simp
                        moreover from EntEnt have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i)\<parallel>\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="p",symmetric])
                          apply(simp add: eqvts swap_simps `distinctPerm p` `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                          done
                        ultimately show ?thesis
                          by blast
                      qed
                      ultimately show ?thesis
                        apply(rule_tac lol)
                        apply(rule refl)
                        apply(rule bisimSymmetric)
                        apply assumption
                        apply(rule disjI1)
                        apply assumption
                        by(rule bisimSymmetric)
                    qed
                  qed
                  ultimately show ?case
                    by(rule_tac exI) (rule conjI)
                next
                  case(cPar2 \<alpha> Q')
                  have "\<exists> P'. \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P' \<and> (\<Psi>,P',\<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by(rule cPar2)fact+
                  then obtain P' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P' \<and> (\<Psi>,P',\<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by(rule_tac exE)
                  hence "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P'" and "(\<Psi>,P',\<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by simp+
                  obtain A\<^isub>P \<Psi>\<^isub>P where "extractFrame P = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>" and "distinct A\<^isub>P" and "A\<^isub>P \<sharp>* (\<Psi>\<otimes>\<Psi>\<^isub>Q)" and "A\<^isub>P \<sharp>* !P" and "A\<^isub>P \<sharp>* \<alpha>"
                    apply(rule_tac C="(\<Psi>\<otimes>\<Psi>\<^isub>Q,!P,\<alpha>)" in freshFrame)
                    by auto
                  from `extractFrame P = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>` `guarded P` have "\<Psi>\<^isub>P \<simeq> \<one>" and "supp(\<Psi>\<^isub>P) = ({}::name set)"
                    by(auto dest: guardedStatEq)
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<simeq> ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>P)"
                    apply(rule AssertionStatEqTrans[OF AssertionStatEqSym[OF Identity]])
                    apply(rule compositionSym)
                    apply(rule AssertionStatEqSym)
                    by fact
                  have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P \<parallel> P'"
                    apply(rule Bang)
                    apply(rule Par2[where A\<^isub>P=A\<^isub>P and \<Psi>\<^isub>P=\<Psi>\<^isub>P])
                    apply(rule statEqTransition)
                    by fact+
                  moreover have "(\<Psi>,P\<parallel>P',\<lparr>\<nu>b\<rparr>(P \<parallel> Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                  proof -
                    from `(\<Psi>,P',\<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2` obtain P'' Q'' where "\<Psi> \<rhd> P' \<sim> P''" and "(\<Psi>,P'',Q'') \<in> ?X2 \<or> \<Psi> \<rhd> P'' \<sim> Q''" and "\<Psi> \<rhd> Q'' \<sim>\<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                      by auto blast
                    from `(\<Psi>,P'',Q'') \<in> ?X2 \<or> \<Psi> \<rhd> P'' \<sim> Q''` show ?thesis
                    proof(rule disjE)
                      assume "\<Psi> \<rhd> P'' \<sim> Q''"
                      have "\<Psi> \<rhd> P\<parallel>P' \<sim> \<lparr>\<nu>b\<rparr>(P \<parallel> Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                      proof -
                        have "\<Psi> \<rhd> P\<parallel>P' \<sim> P\<parallel>Q''"
                          apply(rule bisimParPres')
                          by(rule bisimTransitive) fact+
                        moreover have "\<Psi> \<rhd> P\<parallel>Q'' \<sim> P \<parallel> \<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                          apply(rule bisimParPres')
                          by fact
                        moreover have "\<Psi> \<rhd> P \<parallel> \<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>b\<rparr>(P \<parallel> Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimResPres])
                          apply(rule bisimParAssoc)
                          apply fact
                          apply(rule bisimScopeExt)
                          apply fact
                          done
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      thus ?thesis
                        apply(rule_tac lol)
                        apply(rule refl)
                        apply assumption
                        apply(rule disjI2)
                        by(rule bisimReflexive)+
                    next
                      assume "(\<Psi>,P'',Q'') \<in> ?X2"
                      then obtain yvec Q\<^isub>i a\<^isub>i M\<^isub>i \<Psi>\<^isub>i P\<^isub>i where q''def: "Q'' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)" and p''def: "P'' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i" and "a\<^isub>i \<sharp> \<Psi>" and "a\<^isub>i \<sharp> P\<^isub>i" and "a\<^isub>i \<sharp> Q\<^isub>i" and "\<not> a\<^isub>i \<sharp> M\<^isub>i" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i\<Leftarrow>P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a\<^isub>i], \<Psi>\<^isub>i\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i \<Leftarrow> R \<longrightarrow> (R = (P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>)))" and "guarded P\<^isub>i" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>i\<Leftarrow>P \<longrightarrow> (supp M\<^isub>i) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>i\<Leftarrow>P = M\<^isub>i\<Leftarrow>P') = (P = P'))" and "((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i" and "((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i" and "yvec \<sharp>* \<Psi>" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>i \<Leftarrow> P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
                        by blast
                      have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((P\<parallel>Q')\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (P\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)))"
                      proof -
                        have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>(P\<parallel>Q'\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> P\<parallel> \<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                          apply(rule bisimTransitive[OF bisimResPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExt)
                          by fact
                        moreover have "\<Psi> \<rhd> P\<parallel> \<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> P\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))"
                          apply(rule bisimTransitive[OF bisimParComm])
                          apply(rule bisimSymmetric[OF bisimTransitive[OF bisimParComm]])
                          apply(rule bisimParPres)
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimSymmetric[OF `\<Psi> \<rhd> Q'' \<sim> \<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`]])
                          apply(subst q''def)
                          by(rule bisimReflexive)
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      obtain p::"name prm" where "set p \<subseteq> set yvec \<times> set(p\<bullet>yvec)" and "(p\<bullet>yvec) \<sharp>* a\<^isub>i" and "distinctPerm p" and "(p\<bullet>yvec) \<sharp>* M\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>\<^isub>i" and "(p\<bullet>yvec) \<sharp>* P\<^isub>i" and "(p\<bullet>yvec) \<sharp>* Q\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>" and "(p\<bullet>yvec) \<sharp>* P'" and "(p\<bullet>yvec) \<sharp>* P" and "(p\<bullet>yvec) \<sharp>* A\<^isub>Q"
                        by(rule chain.name_list_avoiding[where xvec=yvec and c="(M\<^isub>i,\<Psi>\<^isub>i,a\<^isub>i,P\<^isub>i,Q\<^isub>i,\<Psi>,P',P,A\<^isub>Q)"]) auto
                      hence yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>(p\<bullet>a\<^isub>i)\<rparr>(\<langle>(p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" and yvrdef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)"
                        by(subst resChainAlpha[where p=p],simp,simp,simp)+
                      obtain b\<^isub>i::name where "b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)" and "b\<^isub>i \<sharp> \<Psi>" and "b\<^isub>i \<sharp> \<Psi>\<^isub>Q" and "b\<^isub>i \<sharp> P" and "b\<^isub>i \<sharp> A\<^isub>Q"
                        apply(rule at_exists_fresh[OF at_name_inst, where x="(p\<bullet>M\<^isub>i,p\<bullet>\<Psi>\<^isub>i,p\<bullet>P\<^isub>i,p\<bullet>Q\<^isub>i,\<Psi>,\<Psi>\<^isub>Q,P,A\<^isub>Q)"])
                        unfolding supp_prod pt_perm_supp[OF pt_name_inst,OF at_name_inst, symmetric]
                        apply(rule finite_UnI, (finite_guess | (subst pt_set_finite_ineq[OF pt_name_inst,OF at_name_inst], finite_guess)))+
                        apply(finite_guess)
                        by force
                      have yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" unfolding yvldef
                        by(subst alphaRes[where x="(p\<bullet>a\<^isub>i)" and y="b\<^isub>i"]) (simp add: `b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)`)+
                      have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((P\<parallel>Q')\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                      proof -
                        note `\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((P\<parallel>Q')\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> P\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))`
                        moreover have "\<Psi> \<rhd> P\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)) \<sim> P\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                          unfolding yvldef
                          by(rule bisimReflexive)
                        moreover have "\<Psi> \<rhd> P\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)) \<sim> \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>P\<parallel>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)"
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExtChain)
                          by fact+
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      moreover have "\<Psi> \<rhd> \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> P \<parallel> P'"
                      proof -
                        have "\<Psi> \<rhd> \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> P \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i))"
                          apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExtChain)
                          by fact+
                        moreover have "\<Psi> \<rhd> P \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)) \<sim> P \<parallel> P'" using `\<Psi> \<rhd> P' \<sim> P''`
                          unfolding yvrdef[symmetric] p''def[symmetric]
                          by(rule_tac bisimParPres'[OF bisimSymmetric])
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      moreover have "(\<Psi>, \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i),(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))) \<in> ?X2"
                      proof -
                        have "(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)" using `a\<^isub>i \<sharp> P\<^isub>i` by(subst fresh_bij)
                        note `b\<^isub>i \<sharp> \<Psi>`
                        moreover note `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`
                        moreover have "b\<^isub>i \<sharp> (P \<parallel> (p\<bullet>Q\<^isub>i))" using `b\<^isub>i \<sharp> P` `b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)`
                          by simp
                        moreover have "guarded(p\<bullet>P\<^isub>i)" using `guarded P\<^isub>i` by simp
                        moreover have "\<not> b\<^isub>i \<sharp> ([(p\<bullet>a\<^isub>i,b\<^isub>i)] \<bullet> (p\<bullet>M\<^isub>i))" using `\<not> a\<^isub>i \<sharp> M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "\<And> \<Psi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>" using EntCl
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="rev p",symmetric])
                          by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`)
                        moreover have "\<forall>F. (F \<otimes>\<^sub>F (\<langle>[b\<^isub>i],[(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)\<rangle>) \<simeq>\<^sub>F F)"
                          using Equiv1
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="rev p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "(\<forall>\<Psi> R. b\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i)) \<Leftarrow> R \<longrightarrow> (R = ((p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>)))" using Equiv2
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i` swap_simps)
                        moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P \<longrightarrow> (supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))) \<subseteq> ((supp \<Psi>)::name set))" using EntSupp
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "\<forall> P P'. ((([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P') = (P = P'))" using Inj
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "((supp(p\<bullet>P\<^isub>i))::name set) \<subseteq> supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                        moreover have "((supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)))::name set) = supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                        moreover note `(p\<bullet>yvec) \<sharp>* \<Psi>`
                        moreover from EntEnt have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i)\<parallel>\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="p",symmetric])
                          apply(simp add: eqvts swap_simps `distinctPerm p` `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                          done
                        ultimately show ?thesis
                          by blast
                      qed
                      ultimately show ?thesis
                        apply(rule_tac lol)
                        apply(rule refl)
                        apply(rule bisimSymmetric)
                        apply assumption
                        apply(rule disjI1)
                        apply assumption
                        by(rule bisimSymmetric)
                    qed
                  qed
                  ultimately show ?case
                    by(rule_tac exI) (rule conjI)
                qed
                then obtain P' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P' \<and> (\<Psi>, P', \<lparr>\<nu>b\<rparr>(Q'\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                  by(rule exE)
                hence "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P'" and "(\<Psi>, P', \<lparr>\<nu>b\<rparr>(Q'\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                  by simp+
                from `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P'` have "\<Psi> \<rhd> Q \<parallel> !P \<longmapsto> \<alpha> \<prec> Q \<parallel> P'"
                  by(rule Par2) (fact|simp add: `A\<^isub>Q \<sharp>* P`)+
                moreover have "(\<Psi>, Q \<parallel> P', Q \<parallel> \<lparr>\<nu>b\<rparr>(Q'\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y"
                proof -
                  from `(\<Psi>, P', \<lparr>\<nu>b\<rparr>(Q'\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2` obtain P'' Q'' where "\<Psi> \<rhd> P' \<sim> P''" and "(\<Psi>,P'',Q'') \<in> ?X2 \<or> \<Psi> \<rhd> P'' \<sim> Q''" and "\<Psi> \<rhd> Q'' \<sim> \<lparr>\<nu>b\<rparr>(Q'\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                    by auto blast
                  from `(\<Psi>,P'',Q'') \<in> ?X2 \<or> \<Psi> \<rhd> P'' \<sim> Q''` have "(\<Psi>, Q \<parallel> P', Q \<parallel> \<lparr>\<nu>b\<rparr>(Q'\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                  proof(rule disjE)
                    assume "\<Psi> \<rhd> P'' \<sim> Q''"
                    have "\<Psi> \<rhd> Q \<parallel> P' \<sim> Q \<parallel> \<lparr>\<nu>b\<rparr>(Q'\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                      apply(rule bisimParPres')
                      apply(rule bisimTransitive,fact)+
                      by(rule bisimReflexive)
                    thus ?thesis
                      apply(rule_tac lol)
                      apply(rule refl)
                      apply assumption
                      apply(rule disjI2)
                      by(rule bisimReflexive)+
                  next
                    assume "(\<Psi>,P'',Q'') \<in> ?X2"
                    then obtain yvec Q\<^isub>i a\<^isub>i M\<^isub>i \<Psi>\<^isub>i P\<^isub>i where q''def: "Q'' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)" and p''def: "P'' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i" and "a\<^isub>i \<sharp> \<Psi>" and "a\<^isub>i \<sharp> P\<^isub>i" and "a\<^isub>i \<sharp> Q\<^isub>i" and "\<not> a\<^isub>i \<sharp> M\<^isub>i" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i\<Leftarrow>P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a\<^isub>i], \<Psi>\<^isub>i\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i \<Leftarrow> R \<longrightarrow> (R = (P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>)))" and "guarded P\<^isub>i" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>i\<Leftarrow>P \<longrightarrow> (supp M\<^isub>i) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>i\<Leftarrow>P = M\<^isub>i\<Leftarrow>P') = (P = P'))" and "((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i" and "((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i" and "yvec \<sharp>* \<Psi>" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>i \<Leftarrow> P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
                      by blast
                    have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((Q\<parallel>Q')\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (Q\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)))"
                    proof -
                      have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>(Q\<parallel>Q'\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> Q\<parallel> \<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                        apply(rule bisimTransitive[OF bisimResPres[OF bisimParAssoc]])
                        apply fact
                        apply(rule bisimScopeExt)
                        by fact
                      moreover have "\<Psi> \<rhd> Q\<parallel> \<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> Q\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))"
                        apply(rule bisimTransitive[OF bisimParComm])
                        apply(rule bisimSymmetric[OF bisimTransitive[OF bisimParComm]])
                        apply(rule bisimParPres)
                        apply(rule bisimSymmetric)
                        apply(rule bisimTransitive[OF bisimSymmetric[OF `\<Psi> \<rhd> Q'' \<sim> \<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`]])
                        apply(subst q''def)
                        by(rule bisimReflexive)
                      ultimately show ?thesis
                        by(auto dest: bisimTransitive)
                    qed
                    obtain p::"name prm" where "set p \<subseteq> set yvec \<times> set(p\<bullet>yvec)" and "(p\<bullet>yvec) \<sharp>* a\<^isub>i" and "distinctPerm p" and "(p\<bullet>yvec) \<sharp>* M\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>\<^isub>i" and "(p\<bullet>yvec) \<sharp>* P\<^isub>i" and "(p\<bullet>yvec) \<sharp>* Q\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>" and "(p\<bullet>yvec) \<sharp>* P'" and "(p\<bullet>yvec) \<sharp>* Q" and "(p\<bullet>yvec) \<sharp>* A\<^isub>Q"
                      by(rule chain.name_list_avoiding[where xvec=yvec and c="(M\<^isub>i,\<Psi>\<^isub>i,a\<^isub>i,P\<^isub>i,Q\<^isub>i,\<Psi>,P',Q,A\<^isub>Q)"]) auto
                    hence yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>(p\<bullet>a\<^isub>i)\<rparr>(\<langle>(p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" and yvrdef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)"
                      by(subst resChainAlpha[where p=p],simp,simp,simp)+
                    obtain b\<^isub>i::name where "b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)" and "b\<^isub>i \<sharp> \<Psi>" and "b\<^isub>i \<sharp> \<Psi>\<^isub>Q" and "b\<^isub>i \<sharp> Q" and "b\<^isub>i \<sharp> A\<^isub>Q"
                      apply(rule at_exists_fresh[OF at_name_inst, where x="(p\<bullet>M\<^isub>i,p\<bullet>\<Psi>\<^isub>i,p\<bullet>P\<^isub>i,p\<bullet>Q\<^isub>i,\<Psi>,\<Psi>\<^isub>Q,Q,A\<^isub>Q)"])
                      unfolding supp_prod pt_perm_supp[OF pt_name_inst,OF at_name_inst, symmetric]
                      apply(rule finite_UnI, (finite_guess | (subst pt_set_finite_ineq[OF pt_name_inst,OF at_name_inst], finite_guess)))+
                      apply(finite_guess)
                      by force
                    have yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" unfolding yvldef
                      by(subst alphaRes[where x="(p\<bullet>a\<^isub>i)" and y="b\<^isub>i"]) (simp add: `b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)`)+
                    have "\<Psi> \<rhd> Q\<parallel>\<lparr>\<nu>b\<rparr>(Q'\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(Q\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                    proof -
                      have "\<Psi> \<rhd> Q\<parallel>\<lparr>\<nu>b\<rparr>(Q'\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>b\<rparr>((Q\<parallel>Q')\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                        apply(rule bisimTransitive[OF bisimSymmetric[OF bisimScopeExt]])
                        apply fact
                        apply(rule bisimResPres)
                        apply(rule bisimSymmetric[OF bisimParAssoc])
                        by fact
                      moreover note `\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((Q\<parallel>Q')\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> Q\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))`
                      moreover have "\<Psi> \<rhd> Q\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)) \<sim> Q\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                        unfolding yvldef
                        by(rule bisimReflexive)
                      moreover have "\<Psi> \<rhd> Q\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)) \<sim> \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>Q\<parallel>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)"
                        apply(rule bisimSymmetric)
                        apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                        apply fact
                        apply(rule bisimScopeExtChain)
                        by fact+
                      ultimately show ?thesis
                        by(auto dest: bisimTransitive)
                    qed
                    moreover have "\<Psi> \<rhd> \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(Q \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> Q \<parallel> P'"
                    proof -
                      have "\<Psi> \<rhd> \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(Q \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> Q \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i))"
                        apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                        apply fact
                        apply(rule bisimScopeExtChain)
                        by fact+
                      moreover have "\<Psi> \<rhd> Q \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)) \<sim> Q \<parallel> P'" using `\<Psi> \<rhd> P' \<sim> P''`
                        unfolding yvrdef[symmetric] p''def[symmetric]
                        by(rule_tac bisimParPres'[OF bisimSymmetric])
                      ultimately show ?thesis
                        by(auto dest: bisimTransitive)
                    qed
                    moreover have "(\<Psi>, \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(Q \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i),(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(Q\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))) \<in> ?X2"
                    proof -
                      have "(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)" using `a\<^isub>i \<sharp> P\<^isub>i` by(subst fresh_bij)
                      note `b\<^isub>i \<sharp> \<Psi>`
                      moreover note `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`
                      moreover have "b\<^isub>i \<sharp> (Q \<parallel> (p\<bullet>Q\<^isub>i))" using `b\<^isub>i \<sharp> Q` `b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)`
                        by simp
                      moreover have "guarded(p\<bullet>P\<^isub>i)" using `guarded P\<^isub>i` by simp
                      moreover have "\<not> b\<^isub>i \<sharp> ([(p\<bullet>a\<^isub>i,b\<^isub>i)] \<bullet> (p\<bullet>M\<^isub>i))" using `\<not> a\<^isub>i \<sharp> M\<^isub>i`
                        apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                        apply(subst (asm) perm_bool[where pi="p",symmetric])
                        by(simp add: eqvts swap_simps)
                      moreover have "\<And> \<Psi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>" using EntCl
                        apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                        apply(subst perm_bool[where pi="rev p",symmetric])
                        by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`)
                      moreover have "\<forall>F. (F \<otimes>\<^sub>F (\<langle>[b\<^isub>i],[(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)\<rangle>) \<simeq>\<^sub>F F)"
                        using Equiv1
                        apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                        apply(subst perm_bool[where pi="rev p",symmetric])
                        by(simp add: eqvts swap_simps)
                      moreover have "(\<forall>\<Psi> R. b\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i)) \<Leftarrow> R \<longrightarrow> (R = ((p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>)))" using Equiv2
                        apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                        apply(subst (asm) perm_bool[where pi="p",symmetric])
                        by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i` swap_simps)
                      moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P \<longrightarrow> (supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))) \<subseteq> ((supp \<Psi>)::name set))" using EntSupp
                        apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                        apply(subst (asm) perm_bool[where pi="p",symmetric])
                        by(simp add: eqvts swap_simps)
                      moreover have "\<forall> P P'. ((([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P') = (P = P'))" using Inj
                        apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                        apply(subst (asm) perm_bool[where pi="p",symmetric])
                        by(simp add: eqvts swap_simps)
                      moreover have "((supp(p\<bullet>P\<^isub>i))::name set) \<subseteq> supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i`
                        apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                        apply(subst (asm) perm_bool[where pi="p",symmetric])
                        by(simp add: eqvts swap_simps `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                      moreover have "((supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)))::name set) = supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i`
                        apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                        apply(subst (asm) perm_bool[where pi="p",symmetric])
                        by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                      moreover note `(p\<bullet>yvec) \<sharp>* \<Psi>`
                      moreover from EntEnt have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i)\<parallel>\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                        apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                        apply(subst perm_bool[where pi="p",symmetric])
                        apply(simp add: eqvts swap_simps `distinctPerm p` `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                        done
                      ultimately show ?thesis
                        by blast
                    qed
                    ultimately show ?thesis
                      apply(rule_tac lol)
                      apply(rule refl)
                      apply(rule bisimSymmetric)
                      apply assumption
                      apply(rule disjI1)
                      apply assumption
                      by(rule bisimSymmetric)
                  qed
                  thus ?thesis by(rule Y2inY)
                qed
                ultimately show ?case
                  by(rule_tac exI) (rule conjI)
              qed
            qed
          next
            case(cComm1 \<Psi>\<^isub>P M N Q' A\<^isub>Q \<Psi>\<^isub>Q K xvec P' A\<^isub>P)
            have "extractFrame(\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<langle>[a],\<one> \<otimes> \<Psi>\<^isub>c\<rangle>"
              by simp
            thm cComm1
            with `extractFrame(\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>` have "\<langle>[a],\<one> \<otimes> \<Psi>\<^isub>c\<rangle> = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>" by simp
            obtain b where "A\<^isub>P=[b]"
            proof -
              assume "\<And>b. A\<^isub>P = [b] \<Longrightarrow> thesis"
              moreover from `extractFrame (\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>`
              have "\<exists> b. A\<^isub>P=[b]"
              proof(induct A\<^isub>P)
                case Nil thus ?case by(simp)
              next
                case(Cons a\<^isub>Q A\<^isub>Q)
                have "[(a, a\<^isub>Q)] \<bullet> A\<^isub>Q = [] \<Longrightarrow> A\<^isub>Q = []"
                  by(induct A\<^isub>Q) (auto simp add: eqvts)
                with Cons(2) have "A\<^isub>Q=[]"
                  by(auto simp add: abs_fun_eq[OF pt_name_inst, OF at_name_inst] frame.inject eqvts)
                thus ?case by simp
              qed
              ultimately show ?thesis
                by auto
            qed
            from cComm1 have "b \<sharp> \<Psi>" and "b \<sharp> \<Psi>\<^isub>Q" and "b \<sharp> Q" and "b \<sharp> K" and "b \<sharp> N" and "b \<sharp> Q'" and "b \<sharp> xvec" and "b \<sharp> P'" and "a \<sharp> b" and "b \<sharp> M\<^isub>a" and "b \<sharp> \<Psi>\<^isub>c" and "b \<sharp> P" unfolding `A\<^isub>P = [b]` by(simp)+
            have "\<Psi>\<^isub>P = \<one> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)" using `extractFrame (\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>` unfolding `A\<^isub>P = [b]`
              by(auto simp add: frame.inject abs_fun_eq'[OF pt_name_inst, OF at_name_inst] eqvts pt3[OF pt_name_inst, OF at_ds5, OF at_name_inst])
            have Equiv1': "\<And>F. F \<otimes>\<^sub>F (\<langle>[b], [(a, b)] \<bullet> \<Psi>\<^isub>c\<rangle>) \<simeq>\<^sub>F F" using `\<And>F. F \<otimes>\<^sub>F (\<langle>[a], \<Psi>\<^isub>c\<rangle>) \<simeq>\<^sub>F F`
              by(subst (asm) frameChainAlpha [where p="[(a,b)]"]) (simp add: swap_simps `b \<sharp> \<Psi>\<^isub>c`)+
            have Equiv2': "\<And>\<Psi> R. \<lbrakk>b \<sharp> \<Psi>; \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a) \<Leftarrow> R\<rbrakk> \<Longrightarrow> R = P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>"
            proof -
              fix \<Psi> R
              assume "b \<sharp> \<Psi>" and "\<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a) \<Leftarrow> R"
              moreover have "\<forall> \<Psi> R. (b \<sharp> \<Psi> \<and> \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a) \<Leftarrow> R) \<longrightarrow> R = P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>" using Equiv2
                by(subst perm_bool[where pi="[(a,b)]",symmetric]) (force simp add: eqvts swap_simps `a \<sharp> P` `b \<sharp> P`)
              ultimately show "R = P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>"
                by auto
            qed
            have "((supp P)::name set) \<subseteq> supp([(a,b)]\<bullet>M\<^isub>a)" using `((supp P)::name set) \<subseteq> supp M\<^isub>a`
              apply(subst perm_bool[where pi="[(a,b)]",symmetric])
              by(simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
            from `((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a` have "((supp([(a,b)]\<bullet>\<Psi>\<^isub>c))::name set) = supp([(a,b)]\<bullet>M\<^isub>a)"
              apply(subst (asm) perm_bool[where pi="[(a,b)]",symmetric])
              by(simp only: eqvts)
            have EntCl': "\<And>\<Psi>. \<Psi> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>" using `\<And>\<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>`
              by(subst perm_bool[where pi="[(a,b)]",symmetric]) (simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
            have EntEnt': "\<And>\<Psi> \<phi>. \<Psi> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> \<phi> \<Longrightarrow>
              \<phi> = ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
            proof -
              fix \<Psi> \<phi>
              assume "\<Psi> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> \<phi>"
              moreover have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<or> \<Psi> \<turnstile> \<phi>" using `\<And>\<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> \<phi> \<Longrightarrow> \<phi> = M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> \<phi>`
                apply(subst perm_bool[where pi="[(a,b)]",symmetric])
                by(simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
              ultimately show "\<phi> = ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                by simp
            qed
            {
              fix \<Psi> P
              assume "\<Psi> \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P"
              moreover have "\<forall> \<Psi> P. \<Psi> \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<longrightarrow> supp ([(a, b)] \<bullet> M\<^isub>a) \<subseteq> (supp \<Psi>::name set)" using `\<forall>\<Psi> P. \<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<longrightarrow> supp M\<^isub>a \<subseteq> supp \<Psi>`
                apply(subst perm_bool[where pi="[(a,b)]",symmetric])
                by(simp add: eqvts)
              ultimately have "supp ([(a, b)] \<bullet> M\<^isub>a) \<subseteq> (supp \<Psi>::name set)"
                by auto
            }
            note EntSupp' = this
            have Inj': "\<And>P P'. (([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P = ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P') = (P = P')" using `\<forall>P P'. (M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P')`
              by(subst (asm) perm_bool[where pi="[(a,b)]",symmetric]) (simp add: eqvts)
            have "\<not> b \<sharp> [(a, b)] \<bullet> M\<^isub>a" using `\<not> a \<sharp> M\<^isub>a`
              by(subst (asm) perm_bool [where pi="[(a,b)]",symmetric]) (simp add: eqvts swap_simps)
            have "b \<sharp> \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
              by(simp add: fresh_abs_fun_iff[OF pt_name_inst, OF at_name_inst, OF fin_supp])
            have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'" using `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'`
              apply(subst (asm) alphaRes[where y=b])
              apply(simp add: `b \<sharp> M\<^isub>a` `b \<sharp> \<Psi>\<^isub>c`)
              by(simp only: eqvts)
            moreover have "b \<sharp> \<Psi> \<otimes> \<Psi>\<^isub>Q" using `b \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q` by force
            moreover have "b \<sharp> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>" using `b \<sharp> K` `b \<sharp> xvec` `b \<sharp> N` by simp
            moreover have "b \<sharp> P'" using `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'` `b \<sharp> xvec` `b \<sharp> \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)` `distinct xvec` `xvec \<sharp>* K`
              by(rule_tac outputFreshDerivative)
            moreover have "bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)" using `xvec \<sharp>* \<Psi>` `xvec \<sharp>* \<Psi>\<^isub>Q` by force
            moreover have "bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* ([(a,b)]\<bullet>M\<^isub>a)" using `xvec \<sharp>* (a, M\<^isub>a, \<Psi>\<^isub>c, P)` `b \<sharp> xvec`
              by simp
            moreover have "bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* ([(a,b)]\<bullet>\<Psi>\<^isub>c)" using `xvec \<sharp>* (a, M\<^isub>a, \<Psi>\<^isub>c, P)` `b \<sharp> xvec`
              by simp
            moreover have "bn(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>) \<sharp>* subject(K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>)" using `xvec \<sharp>* K`
              by simp
            moreover note `b \<sharp> P`
            moreover note `guarded P`
            moreover note `\<not> b \<sharp> ([(a,b)]\<bullet>M\<^isub>a)`
            moreover have "\<forall> \<Psi>'. \<Psi>' \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a) \<Leftarrow> P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>" using EntCl' by simp
            moreover have "\<forall>\<Psi>' Q. b \<sharp> \<Psi>' \<and> \<Psi>' \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>Q \<longrightarrow> Q = P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>" using Equiv2'
              by force
            ultimately have "\<exists> P''. \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'' \<and> (\<Psi>,P'',P') \<in> ?Y2"
            proof(induct rule: resReplCases)
              case(cRes P')
              moreover have "xvec \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)" using `xvec \<sharp>* \<Psi>` `xvec \<sharp>* \<Psi>\<^isub>Q`
                by force
              thm parOutputCases
              moreover have "xvec \<sharp>* \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>" using `xvec \<sharp>* (a, M\<^isub>a, \<Psi>\<^isub>c, P)` `b \<sharp> xvec`
                by simp
              moreover have "xvec \<sharp>* \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>" using `xvec \<sharp>* (a, M\<^isub>a, \<Psi>\<^isub>c, P)` `b \<sharp> xvec`
                by simp
              moreover note `xvec \<sharp>* K`
              ultimately show ?case
              proof(induct rule: parOutputCases[where C="()"])
                case cPar2 thus ?case by(drule_tac nilTrans) (rule FalseE)
              next
                case(cPar1 Q' A\<^isub>R \<Psi>\<^isub>R)
                obtain \<alpha> where alphadef: "\<alpha> = K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle>" by simp
                from `extractFrame (\<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>) = \<langle>A\<^isub>R, \<Psi>\<^isub>R\<rangle>` have "A\<^isub>R = []" and psirdef: "\<Psi>\<^isub>R = ([(a,b)]\<bullet>\<Psi>\<^isub>c)"
                  by auto
                obtain \<Psi>' where psi'def: "\<Psi>'=(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)"
                  by simp
                obtain M\<^isub>b \<Psi>\<^isub>b where mbdef: "M\<^isub>b = [(a,b)]\<bullet>M\<^isub>a" and psibdef: "\<Psi>\<^isub>b = [(a,b)]\<bullet>\<Psi>\<^isub>c"
                  by auto
                have "b \<sharp> bn \<alpha>" and "b \<sharp> object \<alpha>" unfolding alphadef using `b \<sharp> xvec` `b \<sharp> N` by auto
                have "a \<sharp> \<Psi>\<^isub>Q" using `a \<sharp> Q` `extractFrame Q = \<langle>A\<^isub>Q,\<Psi>\<^isub>Q\<rangle>` `A\<^isub>Q \<sharp>* (a, M\<^isub>a, \<Psi>\<^isub>c, P)`
                  by(drule_tac extractFrameFresh) simp
                have "bn \<alpha> \<sharp>* \<Psi>\<^isub>Q" unfolding alphadef using `xvec \<sharp>* Q` `extractFrame Q = \<langle>A\<^isub>Q,\<Psi>\<^isub>Q\<rangle>` `A\<^isub>Q \<sharp>* xvec`
                  by(drule_tac extractFrameFreshChain) simp
                have "bn \<alpha> \<sharp>* a" "b \<sharp> bn \<alpha>" "b \<sharp> object \<alpha>" "bn \<alpha> \<sharp>* M\<^isub>a" "bn \<alpha> \<sharp>* P" "bn \<alpha> \<sharp>* \<Psi>\<^isub>c" "bn \<alpha> \<sharp>* \<Psi>" unfolding alphadef using `xvec \<sharp>* K` `xvec \<sharp>* (a, M\<^isub>a, \<Psi>\<^isub>c, P)` `b \<sharp> xvec` `xvec \<sharp>* \<Psi>` `b \<sharp> N`
                  by auto
                have "\<Psi>' \<rhd> \<langle>M\<^isub>b\<rangle> \<longmapsto> \<alpha> \<prec> Q'" unfolding psi'def mbdef psirdef[symmetric] alphadef by fact
                moreover have "bn \<alpha> \<sharp>* subject \<alpha>" unfolding alphadef by fact
                moreover note `guarded P`
                moreover have "\<forall> \<Psi>''. \<Psi>' \<otimes> \<Psi>'' \<turnstile> M\<^isub>b \<Leftarrow> P \<parallel> \<langle>M\<^isub>b\<rangle>"
                proof -
                  {
                    fix \<Psi>''
                    have "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>'' \<turnstile> M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle>" using EntCl[where \<Psi>="\<Psi>\<otimes>\<Psi>\<^isub>Q\<otimes>\<Psi>''"]
                      by(rule_tac statEqEnt[where \<Psi>="(\<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<Psi>'') \<otimes> \<Psi>\<^isub>c"]) (metis AssertionStatEqTrans Composition AssertionStatEqSym Associativity Commutativity)
                  }
                  thus ?thesis unfolding psi'def psirdef mbdef
                    by(subst perm_bool[where pi="[(a,b)]",symmetric]) (simp add: eqvts `a \<sharp> \<Psi>` `a \<sharp> \<Psi>\<^isub>Q` `b \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q` `a \<sharp> P` `b \<sharp> P`)
                qed
                moreover have "\<forall> \<Psi>'' Q. b \<sharp> \<Psi>'' \<and> \<Psi>' \<otimes> \<Psi>'' \<turnstile> M\<^isub>b \<Leftarrow> Q \<longrightarrow> Q = P \<parallel> \<langle>M\<^isub>b\<rangle>"
                proof -
                  {
                    fix \<Psi>'' Q
                    assume "a \<sharp> \<Psi>''"
                      and    "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>'' \<turnstile> M\<^isub>a \<Leftarrow> Q"
                    hence "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>'') \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> Q"
                      by(rule_tac statEqEnt[where \<Psi>="((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>''"]) (metis AssertionStatEqTrans Composition AssertionStatEqSym Associativity Commutativity)
                    moreover have "a \<sharp> (\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>''" using `a \<sharp> \<Psi>` `a \<sharp> \<Psi>\<^isub>Q` `a \<sharp> \<Psi>''`
                      by auto
                    ultimately have "Q = P \<parallel> \<langle>M\<^isub>a\<rangle>"
                      by(rule_tac Equiv2)
                  }
                  thus ?thesis unfolding psi'def psirdef mbdef
                    by(subst perm_bool[where pi="[(a,b)]",symmetric]) (auto simp add: eqvts `a \<sharp> \<Psi>` `a \<sharp> \<Psi>\<^isub>Q` `b \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q` `a \<sharp> P` `b \<sharp> P` swap_simps)
                qed
                ultimately have "\<exists> P'. \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P' \<and> (\<Psi>, P', \<lparr>\<nu>b\<rparr>(Q'\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2" using mbdef psi'def `bn \<alpha> \<sharp>* a` `b \<sharp> bn \<alpha>` `b \<sharp> object \<alpha>` `bn \<alpha> \<sharp>* M\<^isub>a` `bn \<alpha> \<sharp>* P` `bn \<alpha> \<sharp>* \<Psi>\<^isub>c` `bn \<alpha> \<sharp>* \<Psi>` `bn \<alpha> \<sharp>* \<Psi>\<^isub>Q` `b \<sharp> P` Equiv1' EntCl' EntSupp' Equiv2' Inj' EntEnt' `a \<sharp> P` `\<not>(b \<sharp> ([(a,b)]\<bullet>M\<^isub>a))` `((supp P)::name set) \<subseteq> supp([(a,b)]\<bullet>M\<^isub>a)` `((supp([(a,b)]\<bullet>\<Psi>\<^isub>c))::name set) = supp([(a,b)]\<bullet>M\<^isub>a)` `b \<sharp> M\<^isub>a` `b \<sharp> \<Psi>\<^isub>c` `b \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q`
                proof(nominal_induct avoiding: a b M\<^isub>a \<Psi> \<Psi>\<^isub>Q \<Psi>\<^isub>c rule: replInduct')
                  case(cAlpha \<alpha> P' p)
                  from `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` `b \<sharp> bn(p \<bullet> \<alpha>)` `bn \<alpha> \<sharp>* b` have "p \<bullet> b = b"
                    by simp
                  from `bn \<alpha> \<sharp>* b` have "b \<sharp> bn \<alpha>" by simp
                  from `b \<sharp> object(p \<bullet> \<alpha>)` have "b \<sharp> object \<alpha>"
                    by(subst perm_bool[where pi=p,symmetric]) (simp only: eqvts `p\<bullet>b=b`)
                  have "\<exists> P''. \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P'' \<and> (\<Psi>, P'', \<lparr>\<nu>b\<rparr>(P' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by(rule cAlpha) fact+
                  then obtain P'' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P'' \<and> (\<Psi>, P'', \<lparr>\<nu>b\<rparr>(P' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by(rule exE)
                  hence "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P''" and "(\<Psi>, P'', \<lparr>\<nu>b\<rparr>(P' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by simp+
                  from `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` `bn (p \<bullet> \<alpha>) \<sharp>* a` `bn \<alpha> \<sharp>* a` have "p \<bullet> a = a"
                    by simp
                  from `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` `bn (p \<bullet> \<alpha>) \<sharp>* M\<^isub>a` `bn \<alpha> \<sharp>* M\<^isub>a` have "p \<bullet> M\<^isub>a = M\<^isub>a"
                    by simp
                  from `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` `bn (p \<bullet> \<alpha>) \<sharp>* \<Psi>\<^isub>c` `bn \<alpha> \<sharp>* \<Psi>\<^isub>c` have "p \<bullet> \<Psi>\<^isub>c = \<Psi>\<^isub>c"
                    by simp
                  from `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` `bn (p \<bullet> \<alpha>) \<sharp>* P` `bn \<alpha> \<sharp>* P` have "p \<bullet> P = P"
                    by simp
                  from `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` `bn (p \<bullet> \<alpha>) \<sharp>* \<Psi>\<^isub>Q` `bn \<alpha> \<sharp>* \<Psi>\<^isub>Q` have "p \<bullet> \<Psi>\<^isub>Q = \<Psi>\<^isub>Q"
                    by simp
                  from `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` `bn (p \<bullet> \<alpha>) \<sharp>* \<Psi>` `bn \<alpha> \<sharp>* \<Psi>` have "p \<bullet> \<Psi> = \<Psi>"
                    by simp
                  from `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P''` have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> (p\<bullet>\<alpha>) \<prec> (p\<bullet>P'')"
                    by(subst (asm) perm_bool[where pi=p,symmetric])(simp only: eqvts `p\<bullet>\<Psi>=\<Psi>` `p\<bullet>\<Psi>\<^isub>Q=\<Psi>\<^isub>Q` `p\<bullet>P=P`)
                  moreover have "(\<Psi>, p\<bullet>P'', \<lparr>\<nu>b\<rparr>((p \<bullet> P') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                  proof -
                    from `(\<Psi>, P'', \<lparr>\<nu>b\<rparr>(P' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2` `eqvt ?Y2` have "(p\<bullet>(\<Psi>, P'', \<lparr>\<nu>b\<rparr>(P' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>))) \<in> ?Y2"
                      by(rule_tac eqvtI)
                    moreover have "(p\<bullet>(\<Psi>, P'', \<lparr>\<nu>b\<rparr>(P' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>))) = (\<Psi>, p\<bullet>P'', \<lparr>\<nu>b\<rparr>((p \<bullet> P') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>))"
                      by(simp add: eqvts  `p\<bullet>\<Psi>=\<Psi>` `p\<bullet>\<Psi>\<^isub>Q=\<Psi>\<^isub>Q` `p\<bullet>P=P` `p\<bullet>b=b` `p\<bullet>a=a` `p\<bullet>\<Psi>\<^isub>c=\<Psi>\<^isub>c` cp1[OF cp_pt_inst, OF pt_name_inst, OF at_name_inst, of p "[(a,b)]" \<Psi>\<^isub>c])
                    ultimately show ?thesis
                      by simp
                  qed
                  ultimately show ?case
                    by(rule_tac exI) (rule conjI)
                next
                  case(cPar1 \<alpha> P')
                  have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> P \<longmapsto> \<alpha> \<prec> P'"
                  proof -
                    have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<rhd> P \<longmapsto> \<alpha> \<prec> P'"
                      apply(rule statEqTransition)
                      apply fact
                      apply(subst cPar1)
                      apply(rule AssertionStatEqRefl)
                      done
                    thus ?thesis
                      apply(rule replTrans[where a="b" and M\<^isub>a="[(a,b)]\<bullet>M\<^isub>a"])
                      by(fact | rule freshComp)+
                  qed
                  hence "b \<sharp> P'"
                    by(rule freeFreshDerivative') fact+
                  have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P' \<parallel> !P"
                    apply(rule_tac Bang)
                    apply(rule Par1[where A\<^isub>Q="[]" and \<Psi>\<^isub>Q=\<one>])
                    apply(rule statEqTransition[OF `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> P \<longmapsto> \<alpha> \<prec> P'`])
                    apply(rule AssertionStatEqSym[OF Identity])
                    by(fact|simp add: `bn \<alpha> \<sharp>* P`)+
                  moreover have "(\<Psi>,P'\<parallel>!P,\<lparr>\<nu>b\<rparr>(P' \<parallel> \<langle>M\<^isub>b\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                  proof -
                    have "\<Psi> \<rhd> P'\<parallel>!P \<sim> P' \<parallel> !P" by(rule bisimReflexive)
                    moreover have "\<Psi> \<rhd> P' \<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>b\<rparr>(P' \<parallel> \<langle>M\<^isub>b\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)"
                      apply(subst `M\<^isub>b = [(a,b)]\<bullet>M\<^isub>a`)
                      apply(rule bisimSymmetric)
                      apply(rule bisimTransitive[OF bisimResPres, OF bisimParAssoc, OF `b \<sharp> \<Psi>`])
                      apply(rule bisimScopeExt[OF `b \<sharp> P'`])
                      done
                    moreover have "(\<Psi>,P' \<parallel> !P, P' \<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?X2"
                    proof -
                      note `b \<sharp> \<Psi>` `b \<sharp> P` `b \<sharp> P'` `\<not> b \<sharp> ([(a,b)]\<bullet>M\<^isub>a)`
                      moreover have "(\<forall>\<Psi>. \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)" using cPar1
                        by simp
                      moreover have "\<forall> F. F \<otimes>\<^sub>F (\<langle>[b],([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rangle>) \<simeq>\<^sub>F F" using cPar1 by simp
                      moreover have "(\<forall>\<Psi> R. b \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>R \<longrightarrow> R = P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)" using cPar1
                        by simp
                      moreover note `guarded P` `supp P \<subseteq> supp([(a,b)]\<bullet>M\<^isub>a)` `supp([(a,b)]\<bullet>\<Psi>\<^isub>c) = (supp([(a,b)]\<bullet>M\<^isub>a)::name set)`
                      moreover have "(\<forall>P P'. (([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P = ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P') = (P = P'))" using cPar1
                        by simp
                      moreover have "(\<forall>\<Psi> P. \<Psi> \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<longrightarrow> supp([(a,b)]\<bullet>M\<^isub>a) \<subseteq> (supp \<Psi>::name set))"
                        using cPar1 by auto
                      moreover have "([]::name list) \<sharp>* \<Psi>" by simp
                      moreover have "(\<forall>\<Psi> \<phi>. \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<or> \<Psi> \<turnstile> \<phi>)" using cPar1 by simp
                      ultimately have "(\<Psi>,\<lparr>\<nu>*[]\<rparr>P' \<parallel> !P, \<lparr>\<nu>*[]\<rparr>(P' \<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>))) \<in> ?X2"
                        by blast
                      thus ?thesis by simp
                    qed                        
                    ultimately show ?thesis
                      apply(rule_tac lol)
                      apply(rule refl)
                      apply(assumption)
                      apply(rule disjI1)
                      by -
                  qed
                  ultimately show ?case
                    by(rule_tac exI) (rule conjI)
                next
                  case(cComm1 M N P' K xvec P'')
                  have "b \<sharp> N"
                  proof -
                    note `\<Psi>' \<rhd> \<langle>M\<^isub>b\<rangle> \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P''`
                    moreover note `guarded P`
                    moreover have "\<forall> \<Psi>''. \<Psi>' \<otimes> \<Psi>'' \<turnstile> M\<^isub>b \<Leftarrow> P \<parallel> \<langle>M\<^isub>b\<rangle>"
                    proof -
                      {
                        fix \<Psi>''
                        have seq: "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<Psi>'' \<simeq>  (\<Psi>'' \<otimes> ((\<Psi> \<otimes> \<Psi>\<^isub>Q))) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)"
                          by(metis AssertionStatEqTrans Commutativity AssertionStatEqSym Associativity AssertionStatEqRefl)
                        have "\<Psi>' \<otimes> \<Psi>'' \<turnstile> M\<^isub>b \<Leftarrow> P \<parallel> \<langle>M\<^isub>b\<rangle>"
                          unfolding cComm1
                          apply(rule statEqEnt[OF AssertionStatEqSym, OF seq])
                          by fact
                      }
                      thus ?thesis by simp
                    qed
                    moreover have "\<forall> \<Psi>'' Q. b \<sharp> \<Psi>'' \<and> \<Psi>' \<otimes> \<Psi>'' \<turnstile> M\<^isub>b \<Leftarrow> Q \<longrightarrow> Q = P \<parallel> \<langle>M\<^isub>b\<rangle>"
                    proof -
                      {
                        fix \<Psi>'' Q
                        assume "b \<sharp> \<Psi>''" and "\<Psi>' \<otimes> \<Psi>'' \<turnstile> M\<^isub>b \<Leftarrow> Q"
                        have seq: "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<Psi>'' \<simeq>  (\<Psi>'' \<otimes> ((\<Psi> \<otimes> \<Psi>\<^isub>Q))) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)"
                          by(metis AssertionStatEqTrans Commutativity AssertionStatEqSym Associativity AssertionStatEqRefl)
                        have "(\<Psi>'' \<otimes> ((\<Psi> \<otimes> \<Psi>\<^isub>Q))) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> M\<^isub>b \<Leftarrow> Q" using `\<Psi>' \<otimes> \<Psi>'' \<turnstile> M\<^isub>b \<Leftarrow> Q` unfolding cComm1
                          by(rule statEqEnt[OF seq])
                        moreover have "b \<sharp> (\<Psi>'' \<otimes> ((\<Psi> \<otimes> \<Psi>\<^isub>Q)))" using `b \<sharp> \<Psi>` `b \<sharp> \<Psi>''` `b \<sharp> \<Psi>\<^isub>Q` by force
                        ultimately have "Q = P \<parallel> \<langle>M\<^isub>b\<rangle>"
                          unfolding cComm1
                          by(rule_tac cComm1)
                      }
                      thus ?thesis by auto
                    qed
                    ultimately have "b \<sharp> bnlNames(\<lparr>\<nu>*xvec\<rparr>N \<prec>' P'')" using `b \<sharp> P`
                    proof(nominal_induct avoiding: b rule: replOutputInduct)
                      case cInv thus ?case by -
                    next
                      case(cPar1 xvec N P')
                      from `\<Psi>' \<rhd> P \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'`  `xvec \<sharp>* b` `b \<sharp> P` `distinct xvec` `xvec \<sharp>* K`
                      have "b \<sharp> N"
                        by(rule_tac outputFreshDerivative) auto
                      thus ?case unfolding bnlNamesJudge fresh_def
                        by(auto simp add: name_set_supp[OF finite_Diff, OF fin_supp])
                    next
                      case(cPar2 xvec N P')
                      have "b \<sharp> bnlNames (\<lparr>\<nu>*xvec\<rparr>N \<prec>' P')" by(rule cPar2) fact
                      thus ?case unfolding bnlNamesJudge by -
                    qed
                    moreover have "b \<sharp> xvec" using `xvec \<sharp>* b` by simp
                    ultimately show "b \<sharp> N" unfolding bnlNamesJudge
                      by(auto simp add: name_set_supp[OF finite_Diff, OF fin_supp] fresh_def)
                  qed
                  hence "b \<sharp> P'" using `\<Psi>' \<rhd> P \<longmapsto> M\<lparr>N\<rparr> \<prec> P'` `b \<sharp> P`
                    by(rule_tac inputFreshDerivative)
                  have "a \<sharp> xvec" and "b \<sharp> xvec" using `xvec \<sharp>* a` `xvec \<sharp>* b`
                    by simp+
                  have "\<exists> P'. \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K \<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P' \<and> (\<Psi>, P', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by(rule_tac cComm1) (fact|simp add: `b \<sharp> N` `b \<sharp> xvec` `a \<sharp> xvec`)+
                  then obtain Q' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K \<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q' \<and> (\<Psi>, Q', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by(rule exE)
                  hence "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K \<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q'" and "(\<Psi>, Q', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by simp+
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<rhd> P \<longmapsto> M\<lparr>N\<rparr> \<prec> P'" using `\<Psi>' \<rhd> P \<longmapsto> M\<lparr>N\<rparr> \<prec> P'` by(subst (asm) cComm1)
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<one> \<rhd> P \<longmapsto> M\<lparr>N\<rparr> \<prec> P'"
                    apply(rule_tac statEqTransition[where \<Psi>="\<Psi>\<otimes>\<Psi>\<^isub>Q"])
                    apply(rule replTrans)
                    by(fact|simp|rule freshComp|rule AssertionStatEqSym[OF Identity])+
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<one> \<rhd> !P \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q'"
                    apply(rule statEqTransition)
                    by(fact) (rule AssertionStatEqSym[OF Identity])
                  obtain A\<^isub>P \<Psi>\<^isub>P where "extractFrame P = \<langle>A\<^isub>P,\<Psi>\<^isub>P\<rangle>" and "A\<^isub>P \<sharp>* \<Psi>" and "A\<^isub>P \<sharp>* \<Psi>\<^isub>Q" and "A\<^isub>P \<sharp>* !P" and "A\<^isub>P \<sharp>* P" and "A\<^isub>P \<sharp>* M" and "distinct A\<^isub>P"
                    by(rule freshFrame[where F="extractFrame P" and C="(\<Psi>,\<Psi>\<^isub>Q,P,M)"]) force
                  have "\<Psi>\<^isub>P \<simeq> \<one>" and "supp \<Psi>\<^isub>P = ({}::name set)" using `guarded P` `extractFrame P = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>`
                    by(auto dest: guardedStatEq)
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>P \<rhd> !P \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q'"
                    apply(rule statEqTransition)
                    apply fact
                    apply(rule compositionSym)
                    apply(rule AssertionStatEqSym)
                    by fact
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>P \<otimes> \<one> \<turnstile> M \<leftrightarrow> K"
                  proof -
                    from `\<Psi>' \<turnstile> M \<leftrightarrow> K` have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<turnstile> M \<leftrightarrow> K" using DisjointCCl[where M=M and N=K]
                      unfolding cComm1
                      apply(drule_tac cComm1)
                      apply(erule disjE)
                      by blast
                    thus ?thesis
                      apply(rule_tac statEqEnt)
                      apply(rule AssertionStatEqSym[OF AssertionStatEqTrans[OF compositionSym[OF Identity]]])
                      apply(rule AssertionStatEqTrans[OF compositionSym[OF `\<Psi>\<^isub>P \<simeq> \<one>`]])
                      by(rule Identity)
                  qed
                  have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<tau> \<prec> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q'"
                    apply(rule Bang)
                    apply(rule Comm1[where M=M and K=K and N=N and A\<^isub>P="A\<^isub>P" and \<Psi>\<^isub>P="\<Psi>\<^isub>P" and A\<^isub>Q="[]" and \<Psi>\<^isub>Q="\<one>"])
                    by(fact|rule freshCompChain|simp)+
                  moreover have "(\<Psi>, \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q', \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                  proof -
                    from `(\<Psi>, Q', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2` obtain Q'' P''' where "\<Psi> \<rhd> Q' \<sim> Q''" and Disj: "(\<Psi>,Q'',P''') \<in> ?X2 \<or> \<Psi> \<rhd> Q'' \<sim> P'''" and "\<Psi> \<rhd> P''' \<sim> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                      by auto blast
                    have "xvec \<sharp>* \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>" using `xvec \<sharp>* a` `xvec \<sharp>* b` `xvec \<sharp>* \<Psi>\<^isub>c`
                      by simp
                    have rhs: "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>))"
                      apply(rule bisimTransitive[OF bisimResPres])
                      apply(rule bisimTransitive[OF bisimParComm])
                      apply(rule bisimSymmetric[OF bisimScopeExtChain])
                      apply fact+
                      apply(rule bisimTransitive[OF bisimResComm'])
                      apply fact
                      apply fact
                      apply(rule bisimResChainPres)
                      apply(rule bisimSymmetric)
                      apply(rule bisimTransitive[OF bisimSymmetric[OF bisimScopeExt]])
                      apply fact
                      apply(rule bisimResPres)
                      apply(rule bisimTransitive[OF bisimSymmetric[OF bisimParAssoc]])
                      apply(rule bisimTransitive[OF bisimParComm])
                      apply(rule bisimReflexive)
                      by fact+
                    from Disj show ?thesis
                    proof(rule disjE)
                      assume "\<Psi> \<rhd> Q'' \<sim> P'''"
                      have bis: "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q' \<sim> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)"
                        apply(rule bisimSymmetric)
                        apply(rule bisimTransitive[OF rhs])
                        apply(rule bisimResChainPres)
                        apply(rule bisimParPres')
                        apply(rule bisimTransitive[OF bisimSymmetric[OF `\<Psi> \<rhd> P''' \<sim> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`]])
                        apply(rule bisimSymmetric)
                        apply(rule bisimTransitive[OF `\<Psi> \<rhd> Q' \<sim> Q''`])
                        apply(rule `\<Psi> \<rhd> Q'' \<sim> P'''`)
                        by fact
                      show ?thesis
                        apply(rule lol)
                        apply(rule refl)
                        apply(rule bis)
                        apply(rule disjI2)
                        by(rule bisimReflexive)+
                    next
                      assume "(\<Psi>,Q'',P''') \<in> ?X2"
                      then obtain yvec Q\<^isub>i a\<^isub>i M\<^isub>i \<Psi>\<^isub>i P\<^isub>i where p'''def: "P''' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)" and q''def: "Q'' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i" and "a\<^isub>i \<sharp> \<Psi>" and "a\<^isub>i \<sharp> P\<^isub>i" and "a\<^isub>i \<sharp> Q\<^isub>i" and "\<not> a\<^isub>i \<sharp> M\<^isub>i" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i\<Leftarrow>P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a\<^isub>i], \<Psi>\<^isub>i\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i \<Leftarrow> R \<longrightarrow> (R = (P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>)))" and "guarded P\<^isub>i" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>i\<Leftarrow>P \<longrightarrow> (supp M\<^isub>i) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>i\<Leftarrow>P = M\<^isub>i\<Leftarrow>P') = (P = P'))" and "((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i" and "((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i" and "yvec \<sharp>* \<Psi>" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>i \<Leftarrow> P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
                        by blast
                      have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))))"
                      proof -
                        have "xvec \<sharp>* ([(a,b)] \<bullet> \<Psi>\<^isub>c)" using `a \<sharp> xvec` `b \<sharp> xvec` `xvec \<sharp>* \<Psi>\<^isub>c`
                          by simp
                        have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>b\<rparr>(\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                          apply(rule bisimResPres)
                          apply(rule bisimTransitive[OF bisimParComm])
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimResChainPres])
                          apply(rule bisimParComm)
                          apply fact
                          apply(rule bisimScopeExtChain)
                          by(fact|simp add: `xvec \<sharp>* ([(a,b)] \<bullet> \<Psi>\<^isub>c)`)+
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>(\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>b\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                          by(rule bisimResComm') fact+
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>b\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>(P'\<parallel> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                          apply(rule bisimResChainPres)
                          apply(rule bisimTransitive[OF bisimResPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExt)
                          by fact+
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(P'\<parallel> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> (\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))))"
                          apply(rule bisimResChainPres)
                          apply(rule bisimTransitive[OF bisimParComm])
                          apply(rule bisimSymmetric[OF bisimTransitive[OF bisimParComm]])
                          apply(rule bisimParPres)
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimSymmetric[OF `\<Psi> \<rhd> P''' \<sim> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`]])
                          apply(subst p'''def)
                          by(rule bisimReflexive) fact
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      obtain p::"name prm" where "set p \<subseteq> set yvec \<times> set(p\<bullet>yvec)" and "(p\<bullet>yvec) \<sharp>* a\<^isub>i" and "distinctPerm p" and "(p\<bullet>yvec) \<sharp>* M\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>\<^isub>i" and "(p\<bullet>yvec) \<sharp>* P\<^isub>i" and "(p\<bullet>yvec) \<sharp>* Q\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>" and "(p\<bullet>yvec) \<sharp>* P'" and "(p\<bullet>yvec) \<sharp>* xvec" and "(p\<bullet>yvec) \<sharp>* A\<^isub>Q"
                        by(rule chain.name_list_avoiding[where xvec=yvec and c="(M\<^isub>i,\<Psi>\<^isub>i,a\<^isub>i,P\<^isub>i,Q\<^isub>i,\<Psi>,P',xvec,A\<^isub>Q)"]) auto
                      hence yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>(p\<bullet>a\<^isub>i)\<rparr>(\<langle>(p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" and yvrdef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)"
                        by(subst resChainAlpha[where p=p],simp,simp,simp)+
                      obtain b\<^isub>i::name where "b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)" and "b\<^isub>i \<sharp> \<Psi>" and "b\<^isub>i \<sharp> \<Psi>\<^isub>Q" and "b\<^isub>i \<sharp> P'" and "b\<^isub>i \<sharp> A\<^isub>Q"
                        apply(rule at_exists_fresh[OF at_name_inst, where x="(p\<bullet>M\<^isub>i,p\<bullet>\<Psi>\<^isub>i,p\<bullet>P\<^isub>i,p\<bullet>Q\<^isub>i,\<Psi>,\<Psi>\<^isub>Q,P',A\<^isub>Q)"])
                        unfolding supp_prod pt_perm_supp[OF pt_name_inst,OF at_name_inst, symmetric]
                        apply(rule finite_UnI, (finite_guess | (subst pt_set_finite_ineq[OF pt_name_inst,OF at_name_inst], finite_guess)))+
                        apply(finite_guess)
                        by force
                      have yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" unfolding yvldef
                        by(subst alphaRes[where x="(p\<bullet>a\<^isub>i)" and y="b\<^isub>i"]) (simp add: `b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)`)+
                      have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (\<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                      proof -
                        note `\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))`
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                          unfolding yvldef
                          by(rule bisimReflexive)
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)) \<sim> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>P'\<parallel>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)"
                          apply(subst resChainAppend)
                          apply(rule bisimResChainPres)
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExtChain)
                          by fact+
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      moreover have "\<Psi> \<rhd> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q'"
                      proof -
                        have "\<Psi> \<rhd> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i))" unfolding resChainAppend
                          by(rule bisimReflexive)
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i)) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)))"
                          apply(rule bisimResChainPres)
                          apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExtChain)
                          by fact+
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i))) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> Q')" using `\<Psi> \<rhd> Q' \<sim> Q''` `xvec \<sharp>* \<Psi>`
                          unfolding yvrdef[symmetric] q''def[symmetric]
                          by(metis bisimResChainPres bisimTransitive bisimParComm bisimSymmetric bisimParPres)
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      moreover have "(\<Psi>, \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i),(\<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))) \<in> ?X2"
                      proof -
                        have "(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)" using `a\<^isub>i \<sharp> P\<^isub>i` by(subst fresh_bij)
                        note `b\<^isub>i \<sharp> \<Psi>`
                        moreover note `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`
                        moreover have "b\<^isub>i \<sharp> (P' \<parallel> (p\<bullet>Q\<^isub>i))" using `b\<^isub>i \<sharp> P'` `b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)`
                          by simp
                        moreover have "guarded(p\<bullet>P\<^isub>i)" using `guarded P\<^isub>i` by simp
                        moreover have "\<not> b\<^isub>i \<sharp> ([(p\<bullet>a\<^isub>i,b\<^isub>i)] \<bullet> (p\<bullet>M\<^isub>i))" using `\<not> a\<^isub>i \<sharp> M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "\<And> \<Psi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>" using EntCl
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="rev p",symmetric])
                          by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`)
                        moreover have "\<forall>F. (F \<otimes>\<^sub>F (\<langle>[b\<^isub>i],[(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)\<rangle>) \<simeq>\<^sub>F F)"
                          using Equiv1
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="rev p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "(\<forall>\<Psi> R. b\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i)) \<Leftarrow> R \<longrightarrow> (R = ((p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>)))" using Equiv2
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i` swap_simps)
                        moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P \<longrightarrow> (supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))) \<subseteq> ((supp \<Psi>)::name set))" using EntSupp
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "\<forall> P P'. ((([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P') = (P = P'))" using Inj
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "((supp(p\<bullet>P\<^isub>i))::name set) \<subseteq> supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                        moreover have "((supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)))::name set) = supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                        moreover have "(xvec@(p\<bullet>yvec)) \<sharp>* \<Psi>" using `xvec \<sharp>* \<Psi>` `(p\<bullet>yvec) \<sharp>* \<Psi>`
                          by simp
                        moreover from EntEnt have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i)\<parallel>\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="p",symmetric])
                          apply(simp add: eqvts swap_simps `distinctPerm p` `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                          done
                        ultimately show ?thesis
                          by blast
                      qed
                      ultimately show ?thesis
                        apply(rule_tac lol)
                        apply(rule refl)
                        apply(rule bisimSymmetric)
                        apply assumption
                        apply(rule disjI1)
                        apply assumption
                        by(rule bisimSymmetric)
                    qed
                  qed
                  ultimately show ?case
                    by(rule_tac exI) (rule conjI)
                next
                  case(cInv \<alpha> P')
                  show ?case
                    by(rule cInv) fact+
                next
                  case(cComm2 M xvec N P' K P'')
                  have "b \<sharp> N" and "b \<sharp> P'" using `\<Psi>' \<rhd> P \<longmapsto> M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'` `xvec \<sharp>* b` `b \<sharp> P` `distinct xvec` `xvec \<sharp>* M`
                    by(auto intro: outputFreshDerivative)
                  have "a \<sharp> xvec" and "b \<sharp> xvec" using `xvec \<sharp>* a` `xvec \<sharp>* b`
                    by simp+
                  have "\<exists> P'. \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K\<lparr>N\<rparr> \<prec> P' \<and> (\<Psi>, P', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by(rule_tac cComm2) (fact|simp add: `b \<sharp> N` `b \<sharp> xvec` `a \<sharp> xvec`)+
                  then obtain Q' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K \<lparr>N\<rparr> \<prec> Q' \<and> (\<Psi>, Q', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by(rule exE)
                  hence "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K \<lparr>N\<rparr> \<prec> Q'" and "(\<Psi>, Q', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by simp+
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<rhd> P \<longmapsto> M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'" using `\<Psi>' \<rhd> P \<longmapsto> M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'` by(subst (asm) cComm2)
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<one> \<rhd> P \<longmapsto> M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'"
                    apply(rule_tac statEqTransition[where \<Psi>="\<Psi>\<otimes>\<Psi>\<^isub>Q"])
                    apply(rule replTrans)
                    by(fact|simp|rule freshComp|rule AssertionStatEqSym[OF Identity])+
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<one> \<rhd> !P \<longmapsto> K\<lparr>N\<rparr> \<prec> Q'"
                    apply(rule statEqTransition)
                    by(fact) (rule AssertionStatEqSym[OF Identity])
                  obtain A\<^isub>P \<Psi>\<^isub>P where "extractFrame P = \<langle>A\<^isub>P,\<Psi>\<^isub>P\<rangle>" and "A\<^isub>P \<sharp>* \<Psi>" and "A\<^isub>P \<sharp>* \<Psi>\<^isub>Q" and "A\<^isub>P \<sharp>* !P" and "A\<^isub>P \<sharp>* P" and "A\<^isub>P \<sharp>* M" and "distinct A\<^isub>P"
                    by(rule freshFrame[where F="extractFrame P" and C="(\<Psi>,\<Psi>\<^isub>Q,P,M)"]) force
                  have "\<Psi>\<^isub>P \<simeq> \<one>" and "supp \<Psi>\<^isub>P = ({}::name set)" using `guarded P` `extractFrame P = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>`
                    by(auto dest: guardedStatEq)
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>P \<rhd> !P \<longmapsto> K\<lparr>N\<rparr> \<prec> Q'"
                    apply(rule statEqTransition)
                    apply fact
                    apply(rule compositionSym)
                    apply(rule AssertionStatEqSym)
                    by fact
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>P \<otimes> \<one> \<turnstile> M \<leftrightarrow> K"
                  proof -
                    from `\<Psi>' \<turnstile> M \<leftrightarrow> K` have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<turnstile> M \<leftrightarrow> K" using DisjointCCl[where M=M and N=K]
                      unfolding cComm2
                      apply(drule_tac cComm2)
                      apply(erule disjE)
                      by blast
                    thus ?thesis
                      apply(rule_tac statEqEnt)
                      apply(rule AssertionStatEqSym[OF AssertionStatEqTrans[OF compositionSym[OF Identity]]])
                      apply(rule AssertionStatEqTrans[OF compositionSym[OF `\<Psi>\<^isub>P \<simeq> \<one>`]])
                      by(rule Identity)
                  qed
                  have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<tau> \<prec> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q'"
                    apply(rule Bang)
                    apply(rule Comm2[where M=M and K=K and N=N and A\<^isub>P="A\<^isub>P" and \<Psi>\<^isub>P="\<Psi>\<^isub>P" and A\<^isub>Q="[]" and \<Psi>\<^isub>Q="\<one>"])
                    by(fact|rule freshCompChain|simp)+
                  moreover have "(\<Psi>, \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q', \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                  proof -
                    from `(\<Psi>, Q', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2` obtain Q'' P''' where "\<Psi> \<rhd> Q' \<sim> Q''" and Disj: "(\<Psi>,Q'',P''') \<in> ?X2 \<or> \<Psi> \<rhd> Q'' \<sim> P'''" and "\<Psi> \<rhd> P''' \<sim> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                      by auto blast
                    have "xvec \<sharp>* \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>" using `xvec \<sharp>* a` `xvec \<sharp>* b` `xvec \<sharp>* \<Psi>\<^isub>c`
                      by simp
                    have rhs: "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>))"
                      apply(rule bisimTransitive[OF bisimResPres])
                      apply(rule bisimTransitive[OF bisimParComm])
                      apply(rule bisimSymmetric[OF bisimScopeExtChain])
                      apply fact+
                      apply(rule bisimTransitive[OF bisimResComm'])
                      apply fact
                      apply fact
                      apply(rule bisimResChainPres)
                      apply(rule bisimSymmetric)
                      apply(rule bisimTransitive[OF bisimSymmetric[OF bisimScopeExt]])
                      apply fact
                      apply(rule bisimResPres)
                      apply(rule bisimTransitive[OF bisimSymmetric[OF bisimParAssoc]])
                      apply(rule bisimTransitive[OF bisimParComm])
                      apply(rule bisimReflexive)
                      by fact+
                    from Disj show ?thesis
                    proof(rule disjE)
                      assume "\<Psi> \<rhd> Q'' \<sim> P'''"
                      have bis: "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q' \<sim> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)"
                        apply(rule bisimSymmetric)
                        apply(rule bisimTransitive[OF rhs])
                        apply(rule bisimResChainPres)
                        apply(rule bisimParPres')
                        apply(rule bisimTransitive[OF bisimSymmetric[OF `\<Psi> \<rhd> P''' \<sim> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`]])
                        apply(rule bisimSymmetric)
                        apply(rule bisimTransitive[OF `\<Psi> \<rhd> Q' \<sim> Q''`])
                        apply(rule `\<Psi> \<rhd> Q'' \<sim> P'''`)
                        by fact
                      show ?thesis
                        apply(rule lol)
                        apply(rule refl)
                        apply(rule bis)
                        apply(rule disjI2)
                        by(rule bisimReflexive)+
                    next
                      assume "(\<Psi>,Q'',P''') \<in> ?X2"
                      then obtain yvec Q\<^isub>i a\<^isub>i M\<^isub>i \<Psi>\<^isub>i P\<^isub>i where p'''def: "P''' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)" and q''def: "Q'' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i" and "a\<^isub>i \<sharp> \<Psi>" and "a\<^isub>i \<sharp> P\<^isub>i" and "a\<^isub>i \<sharp> Q\<^isub>i" and "\<not> a\<^isub>i \<sharp> M\<^isub>i" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i\<Leftarrow>P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a\<^isub>i], \<Psi>\<^isub>i\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i \<Leftarrow> R \<longrightarrow> (R = (P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>)))" and "guarded P\<^isub>i" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>i\<Leftarrow>P \<longrightarrow> (supp M\<^isub>i) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>i\<Leftarrow>P = M\<^isub>i\<Leftarrow>P') = (P = P'))" and "((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i" and "((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i" and "yvec \<sharp>* \<Psi>" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>i \<Leftarrow> P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
                        by blast
                      have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))))"
                      proof -
                        have "xvec \<sharp>* ([(a,b)] \<bullet> \<Psi>\<^isub>c)" using `a \<sharp> xvec` `b \<sharp> xvec` `xvec \<sharp>* \<Psi>\<^isub>c`
                          by simp
                        have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>b\<rparr>(\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                          apply(rule bisimResPres)
                          apply(rule bisimTransitive[OF bisimParComm])
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimResChainPres])
                          apply(rule bisimParComm)
                          apply fact
                          apply(rule bisimScopeExtChain)
                          by(fact|simp add: `xvec \<sharp>* ([(a,b)] \<bullet> \<Psi>\<^isub>c)`)+
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>(\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>b\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                          by(rule bisimResComm') fact+
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>b\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>(P'\<parallel> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                          apply(rule bisimResChainPres)
                          apply(rule bisimTransitive[OF bisimResPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExt)
                          by fact+
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(P'\<parallel> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> (\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))))"
                          apply(rule bisimResChainPres)
                          apply(rule bisimTransitive[OF bisimParComm])
                          apply(rule bisimSymmetric[OF bisimTransitive[OF bisimParComm]])
                          apply(rule bisimParPres)
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimSymmetric[OF `\<Psi> \<rhd> P''' \<sim> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`]])
                          apply(subst p'''def)
                          by(rule bisimReflexive) fact
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      obtain p::"name prm" where "set p \<subseteq> set yvec \<times> set(p\<bullet>yvec)" and "(p\<bullet>yvec) \<sharp>* a\<^isub>i" and "distinctPerm p" and "(p\<bullet>yvec) \<sharp>* M\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>\<^isub>i" and "(p\<bullet>yvec) \<sharp>* P\<^isub>i" and "(p\<bullet>yvec) \<sharp>* Q\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>" and "(p\<bullet>yvec) \<sharp>* P'" and "(p\<bullet>yvec) \<sharp>* xvec" and "(p\<bullet>yvec) \<sharp>* A\<^isub>Q"
                        by(rule chain.name_list_avoiding[where xvec=yvec and c="(M\<^isub>i,\<Psi>\<^isub>i,a\<^isub>i,P\<^isub>i,Q\<^isub>i,\<Psi>,P',xvec,A\<^isub>Q)"]) auto
                      hence yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>(p\<bullet>a\<^isub>i)\<rparr>(\<langle>(p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" and yvrdef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)"
                        by(subst resChainAlpha[where p=p],simp,simp,simp)+
                      obtain b\<^isub>i::name where "b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)" and "b\<^isub>i \<sharp> \<Psi>" and "b\<^isub>i \<sharp> \<Psi>\<^isub>Q" and "b\<^isub>i \<sharp> P'" and "b\<^isub>i \<sharp> A\<^isub>Q"
                        apply(rule at_exists_fresh[OF at_name_inst, where x="(p\<bullet>M\<^isub>i,p\<bullet>\<Psi>\<^isub>i,p\<bullet>P\<^isub>i,p\<bullet>Q\<^isub>i,\<Psi>,\<Psi>\<^isub>Q,P',A\<^isub>Q)"])
                        unfolding supp_prod pt_perm_supp[OF pt_name_inst,OF at_name_inst, symmetric]
                        apply(rule finite_UnI, (finite_guess | (subst pt_set_finite_ineq[OF pt_name_inst,OF at_name_inst], finite_guess)))+
                        apply(finite_guess)
                        by force
                      have yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" unfolding yvldef
                        by(subst alphaRes[where x="(p\<bullet>a\<^isub>i)" and y="b\<^isub>i"]) (simp add: `b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)`)+
                      have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (\<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                      proof -
                        note `\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))`
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                          unfolding yvldef
                          by(rule bisimReflexive)
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)) \<sim> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>P'\<parallel>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)"
                          apply(subst resChainAppend)
                          apply(rule bisimResChainPres)
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExtChain)
                          by fact+
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      moreover have "\<Psi> \<rhd> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q'"
                      proof -
                        have "\<Psi> \<rhd> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i))" unfolding resChainAppend
                          by(rule bisimReflexive)
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i)) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)))"
                          apply(rule bisimResChainPres)
                          apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExtChain)
                          by fact+
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i))) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> Q')" using `\<Psi> \<rhd> Q' \<sim> Q''` `xvec \<sharp>* \<Psi>`
                          unfolding yvrdef[symmetric] q''def[symmetric]
                          by(metis bisimResChainPres bisimTransitive bisimParComm bisimSymmetric bisimParPres)
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      moreover have "(\<Psi>, \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i),(\<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))) \<in> ?X2"
                      proof -
                        have "(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)" using `a\<^isub>i \<sharp> P\<^isub>i` by(subst fresh_bij)
                        note `b\<^isub>i \<sharp> \<Psi>`
                        moreover note `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`
                        moreover have "b\<^isub>i \<sharp> (P' \<parallel> (p\<bullet>Q\<^isub>i))" using `b\<^isub>i \<sharp> P'` `b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)`
                          by simp
                        moreover have "guarded(p\<bullet>P\<^isub>i)" using `guarded P\<^isub>i` by simp
                        moreover have "\<not> b\<^isub>i \<sharp> ([(p\<bullet>a\<^isub>i,b\<^isub>i)] \<bullet> (p\<bullet>M\<^isub>i))" using `\<not> a\<^isub>i \<sharp> M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "\<And> \<Psi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>" using EntCl
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="rev p",symmetric])
                          by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`)
                        moreover have "\<forall>F. (F \<otimes>\<^sub>F (\<langle>[b\<^isub>i],[(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)\<rangle>) \<simeq>\<^sub>F F)"
                          using Equiv1
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="rev p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "(\<forall>\<Psi> R. b\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i)) \<Leftarrow> R \<longrightarrow> (R = ((p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>)))" using Equiv2
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i` swap_simps)
                        moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P \<longrightarrow> (supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))) \<subseteq> ((supp \<Psi>)::name set))" using EntSupp
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "\<forall> P P'. ((([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P') = (P = P'))" using Inj
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "((supp(p\<bullet>P\<^isub>i))::name set) \<subseteq> supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                        moreover have "((supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)))::name set) = supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                        moreover have "(xvec@(p\<bullet>yvec)) \<sharp>* \<Psi>" using `xvec \<sharp>* \<Psi>` `(p\<bullet>yvec) \<sharp>* \<Psi>`
                          by simp
                        moreover from EntEnt have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i)\<parallel>\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="p",symmetric])
                          apply(simp add: eqvts swap_simps `distinctPerm p` `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                          done
                        ultimately show ?thesis
                          by blast
                      qed
                      ultimately show ?thesis
                        apply(rule_tac lol)
                        apply(rule refl)
                        apply(rule bisimSymmetric)
                        apply assumption
                        apply(rule disjI1)
                        apply assumption
                        by(rule bisimSymmetric)
                    qed
                  qed
                  ultimately show ?case
                    by(rule_tac exI) (rule conjI)
                next
                  case(cPar2 \<alpha> Q')
                  have "\<exists> P'. \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P' \<and> (\<Psi>,P',\<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by(rule cPar2)fact+
                  then obtain P' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P' \<and> (\<Psi>,P',\<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by(rule_tac exE)
                  hence "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P'" and "(\<Psi>,P',\<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by simp+
                  obtain A\<^isub>P \<Psi>\<^isub>P where "extractFrame P = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>" and "distinct A\<^isub>P" and "A\<^isub>P \<sharp>* (\<Psi>\<otimes>\<Psi>\<^isub>Q)" and "A\<^isub>P \<sharp>* !P" and "A\<^isub>P \<sharp>* \<alpha>"
                    apply(rule_tac C="(\<Psi>\<otimes>\<Psi>\<^isub>Q,!P,\<alpha>)" in freshFrame)
                    by auto
                  from `extractFrame P = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>` `guarded P` have "\<Psi>\<^isub>P \<simeq> \<one>" and "supp(\<Psi>\<^isub>P) = ({}::name set)"
                    by(auto dest: guardedStatEq)
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<simeq> ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>P)"
                    apply(rule AssertionStatEqTrans[OF AssertionStatEqSym[OF Identity]])
                    apply(rule compositionSym)
                    apply(rule AssertionStatEqSym)
                    by fact
                  have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P \<parallel> P'"
                    apply(rule Bang)
                    apply(rule Par2[where A\<^isub>P=A\<^isub>P and \<Psi>\<^isub>P=\<Psi>\<^isub>P])
                    apply(rule statEqTransition)
                    by fact+
                  moreover have "(\<Psi>,P\<parallel>P',\<lparr>\<nu>b\<rparr>(P \<parallel> Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                  proof -
                    from `(\<Psi>,P',\<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2` obtain P'' Q'' where "\<Psi> \<rhd> P' \<sim> P''" and "(\<Psi>,P'',Q'') \<in> ?X2 \<or> \<Psi> \<rhd> P'' \<sim> Q''" and "\<Psi> \<rhd> Q'' \<sim>\<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                      by auto blast
                    from `(\<Psi>,P'',Q'') \<in> ?X2 \<or> \<Psi> \<rhd> P'' \<sim> Q''` show ?thesis
                    proof(rule disjE)
                      assume "\<Psi> \<rhd> P'' \<sim> Q''"
                      have "\<Psi> \<rhd> P\<parallel>P' \<sim> \<lparr>\<nu>b\<rparr>(P \<parallel> Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                      proof -
                        have "\<Psi> \<rhd> P\<parallel>P' \<sim> P\<parallel>Q''"
                          apply(rule bisimParPres')
                          by(rule bisimTransitive) fact+
                        moreover have "\<Psi> \<rhd> P\<parallel>Q'' \<sim> P \<parallel> \<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                          apply(rule bisimParPres')
                          by fact
                        moreover have "\<Psi> \<rhd> P \<parallel> \<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>b\<rparr>(P \<parallel> Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimResPres])
                          apply(rule bisimParAssoc)
                          apply fact
                          apply(rule bisimScopeExt)
                          apply fact
                          done
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      thus ?thesis
                        apply(rule_tac lol)
                        apply(rule refl)
                        apply assumption
                        apply(rule disjI2)
                        by(rule bisimReflexive)+
                    next
                      assume "(\<Psi>,P'',Q'') \<in> ?X2"
                      then obtain yvec Q\<^isub>i a\<^isub>i M\<^isub>i \<Psi>\<^isub>i P\<^isub>i where q''def: "Q'' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)" and p''def: "P'' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i" and "a\<^isub>i \<sharp> \<Psi>" and "a\<^isub>i \<sharp> P\<^isub>i" and "a\<^isub>i \<sharp> Q\<^isub>i" and "\<not> a\<^isub>i \<sharp> M\<^isub>i" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i\<Leftarrow>P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a\<^isub>i], \<Psi>\<^isub>i\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i \<Leftarrow> R \<longrightarrow> (R = (P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>)))" and "guarded P\<^isub>i" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>i\<Leftarrow>P \<longrightarrow> (supp M\<^isub>i) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>i\<Leftarrow>P = M\<^isub>i\<Leftarrow>P') = (P = P'))" and "((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i" and "((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i" and "yvec \<sharp>* \<Psi>" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>i \<Leftarrow> P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
                        by blast
                      have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((P\<parallel>Q')\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (P\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)))"
                      proof -
                        have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>(P\<parallel>Q'\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> P\<parallel> \<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                          apply(rule bisimTransitive[OF bisimResPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExt)
                          by fact
                        moreover have "\<Psi> \<rhd> P\<parallel> \<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> P\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))"
                          apply(rule bisimTransitive[OF bisimParComm])
                          apply(rule bisimSymmetric[OF bisimTransitive[OF bisimParComm]])
                          apply(rule bisimParPres)
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimSymmetric[OF `\<Psi> \<rhd> Q'' \<sim> \<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`]])
                          apply(subst q''def)
                          by(rule bisimReflexive)
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      obtain p::"name prm" where "set p \<subseteq> set yvec \<times> set(p\<bullet>yvec)" and "(p\<bullet>yvec) \<sharp>* a\<^isub>i" and "distinctPerm p" and "(p\<bullet>yvec) \<sharp>* M\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>\<^isub>i" and "(p\<bullet>yvec) \<sharp>* P\<^isub>i" and "(p\<bullet>yvec) \<sharp>* Q\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>" and "(p\<bullet>yvec) \<sharp>* P'" and "(p\<bullet>yvec) \<sharp>* P" and "(p\<bullet>yvec) \<sharp>* A\<^isub>Q"
                        by(rule chain.name_list_avoiding[where xvec=yvec and c="(M\<^isub>i,\<Psi>\<^isub>i,a\<^isub>i,P\<^isub>i,Q\<^isub>i,\<Psi>,P',P,A\<^isub>Q)"]) auto
                      hence yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>(p\<bullet>a\<^isub>i)\<rparr>(\<langle>(p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" and yvrdef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)"
                        by(subst resChainAlpha[where p=p],simp,simp,simp)+
                      obtain b\<^isub>i::name where "b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)" and "b\<^isub>i \<sharp> \<Psi>" and "b\<^isub>i \<sharp> \<Psi>\<^isub>Q" and "b\<^isub>i \<sharp> P" and "b\<^isub>i \<sharp> A\<^isub>Q"
                        apply(rule at_exists_fresh[OF at_name_inst, where x="(p\<bullet>M\<^isub>i,p\<bullet>\<Psi>\<^isub>i,p\<bullet>P\<^isub>i,p\<bullet>Q\<^isub>i,\<Psi>,\<Psi>\<^isub>Q,P,A\<^isub>Q)"])
                        unfolding supp_prod pt_perm_supp[OF pt_name_inst,OF at_name_inst, symmetric]
                        apply(rule finite_UnI, (finite_guess | (subst pt_set_finite_ineq[OF pt_name_inst,OF at_name_inst], finite_guess)))+
                        apply(finite_guess)
                        by force
                      have yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" unfolding yvldef
                        by(subst alphaRes[where x="(p\<bullet>a\<^isub>i)" and y="b\<^isub>i"]) (simp add: `b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)`)+
                      have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((P\<parallel>Q')\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                      proof -
                        note `\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((P\<parallel>Q')\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> P\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))`
                        moreover have "\<Psi> \<rhd> P\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)) \<sim> P\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                          unfolding yvldef
                          by(rule bisimReflexive)
                        moreover have "\<Psi> \<rhd> P\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)) \<sim> \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>P\<parallel>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)"
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExtChain)
                          by fact+
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      moreover have "\<Psi> \<rhd> \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> P \<parallel> P'"
                      proof -
                        have "\<Psi> \<rhd> \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> P \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i))"
                          apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExtChain)
                          by fact+
                        moreover have "\<Psi> \<rhd> P \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)) \<sim> P \<parallel> P'" using `\<Psi> \<rhd> P' \<sim> P''`
                          unfolding yvrdef[symmetric] p''def[symmetric]
                          by(rule_tac bisimParPres'[OF bisimSymmetric])
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      moreover have "(\<Psi>, \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i),(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))) \<in> ?X2"
                      proof -
                        have "(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)" using `a\<^isub>i \<sharp> P\<^isub>i` by(subst fresh_bij)
                        note `b\<^isub>i \<sharp> \<Psi>`
                        moreover note `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`
                        moreover have "b\<^isub>i \<sharp> (P \<parallel> (p\<bullet>Q\<^isub>i))" using `b\<^isub>i \<sharp> P` `b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)`
                          by simp
                        moreover have "guarded(p\<bullet>P\<^isub>i)" using `guarded P\<^isub>i` by simp
                        moreover have "\<not> b\<^isub>i \<sharp> ([(p\<bullet>a\<^isub>i,b\<^isub>i)] \<bullet> (p\<bullet>M\<^isub>i))" using `\<not> a\<^isub>i \<sharp> M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "\<And> \<Psi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>" using EntCl
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="rev p",symmetric])
                          by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`)
                        moreover have "\<forall>F. (F \<otimes>\<^sub>F (\<langle>[b\<^isub>i],[(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)\<rangle>) \<simeq>\<^sub>F F)"
                          using Equiv1
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="rev p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "(\<forall>\<Psi> R. b\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i)) \<Leftarrow> R \<longrightarrow> (R = ((p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>)))" using Equiv2
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i` swap_simps)
                        moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P \<longrightarrow> (supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))) \<subseteq> ((supp \<Psi>)::name set))" using EntSupp
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "\<forall> P P'. ((([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P') = (P = P'))" using Inj
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "((supp(p\<bullet>P\<^isub>i))::name set) \<subseteq> supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                        moreover have "((supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)))::name set) = supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                        moreover note `(p\<bullet>yvec) \<sharp>* \<Psi>`
                        moreover from EntEnt have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i)\<parallel>\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="p",symmetric])
                          apply(simp add: eqvts swap_simps `distinctPerm p` `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                          done
                        ultimately show ?thesis
                          by blast
                      qed
                      ultimately show ?thesis
                        apply(rule_tac lol)
                        apply(rule refl)
                        apply(rule bisimSymmetric)
                        apply assumption
                        apply(rule disjI1)
                        apply assumption
                        by(rule bisimSymmetric)
                    qed
                  qed
                  ultimately show ?case
                    by(rule_tac exI) (rule conjI)
                qed
                thus ?case unfolding alphadef
                  by -
              qed
            qed
            then obtain P'' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'' \<and> (\<Psi>,P'',P') \<in> ?Y2"
              by(rule_tac exE)
            hence "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P''" and "(\<Psi>,P'',P') \<in> ?Y2"
              by simp+
            have "\<Psi> \<otimes> \<one> \<rhd> Q \<longmapsto> M\<lparr>N\<rparr> \<prec> Q'"
            proof -
              have "b \<sharp> \<Psi> \<otimes> \<one>" using `b \<sharp> \<Psi>` by force
              from `\<Psi> \<otimes> \<Psi>\<^isub>P \<rhd> Q \<longmapsto> M\<lparr>N\<rparr> \<prec> Q'` have "(\<Psi> \<otimes> \<one>) \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rhd> Q \<longmapsto> M\<lparr>N\<rparr> \<prec> Q'" unfolding `\<Psi>\<^isub>P = \<one> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)`
                by(rule statEqTransition) (rule AssertionStatEqSym[OF Associativity])
              thus ?thesis
                by(rule replTrans) (fact|simp)+
            qed
            have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<one> \<turnstile> M \<leftrightarrow> K"
            proof -
              have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<one> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<simeq> ((\<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<one>) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c))"
                apply(rule AssertionStatEqTrans[OF AssertionStatEqSym[OF Associativity]])
                apply(rule AssertionStatEqTrans[OF AssertionStatEqSym[OF Associativity]])
                apply(rule Composition)
                by(rule Associativity)
              with `\<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<Psi>\<^isub>P \<turnstile> M \<leftrightarrow> K` have "(\<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<one>) \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> M \<leftrightarrow> K" unfolding `\<Psi>\<^isub>P = \<one> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)`
                by(rule_tac statEqEnt)
              thus ?thesis using DisjointCCl
                by(auto dest: EntEnt')
            qed
            have "A\<^isub>Q \<sharp>* !P" using `A\<^isub>Q \<sharp>* (a, M\<^isub>a, \<Psi>\<^isub>c, P)` by simp
            have "\<Psi> \<rhd> Q \<parallel> !P \<longmapsto> \<tau> \<prec> \<lparr>\<nu>*xvec\<rparr>(Q' \<parallel> P'')"
              apply(rule Comm1)
              apply(rule `\<Psi> \<otimes> \<one> \<rhd> Q \<longmapsto> M\<lparr>N\<rparr> \<prec> Q'`)
              by(fact|simp)+
            moreover have "(\<Psi>, \<lparr>\<nu>*xvec\<rparr>Q' \<parallel> P'',\<lparr>\<nu>*xvec\<rparr>Q'\<parallel>P') \<in> ?Y"
            proof -
              from `(\<Psi>, P'', P') \<in> ?Y2` obtain P''1 P'1 where "\<Psi> \<rhd> P'' \<sim> P''1" and "(\<Psi>,P''1,P'1) \<in> ?X2 \<or> \<Psi> \<rhd> P''1 \<sim> P'1" and "\<Psi> \<rhd> P'1 \<sim> P'"
                by auto blast
              from `(\<Psi>,P''1,P'1) \<in> ?X2 \<or> \<Psi> \<rhd> P''1 \<sim> P'1` have "(\<Psi>, Q' \<parallel> P'', Q'\<parallel> P') \<in> ?Y2"
              proof(rule disjE)
                assume "\<Psi> \<rhd> P''1 \<sim> P'1"
                hence "\<Psi> \<rhd> Q' \<parallel> P''1 \<sim> Q' \<parallel> P'1"
                  by(rule bisimParPres')
                thus ?thesis
                  apply(rule_tac lol)
                  apply(rule refl)
                  apply(rule bisimParPres'[OF `\<Psi> \<rhd> P'' \<sim> P''1`])
                  apply(rule disjI2)
                  apply assumption
                  apply(rule bisimParPres'[OF `\<Psi> \<rhd> P'1 \<sim> P'`])
                  done
              next
                assume "(\<Psi>,P''1,P'1) \<in> ?X2"
                then obtain yvec Q\<^isub>i a\<^isub>i M\<^isub>i \<Psi>\<^isub>i P\<^isub>i where p'1def: "P'1 = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)" and p''1def: "P''1 = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i" and "a\<^isub>i \<sharp> \<Psi>" and "a\<^isub>i \<sharp> P\<^isub>i" and "a\<^isub>i \<sharp> Q\<^isub>i" and "\<not> a\<^isub>i \<sharp> M\<^isub>i" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i\<Leftarrow>P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a\<^isub>i], \<Psi>\<^isub>i\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i \<Leftarrow> R \<longrightarrow> (R = (P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>)))" and "guarded P\<^isub>i" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>i\<Leftarrow>P \<longrightarrow> (supp M\<^isub>i) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>i\<Leftarrow>P = M\<^isub>i\<Leftarrow>P') = (P = P'))" and "((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i" and "((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i" and "yvec \<sharp>* \<Psi>" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>i \<Leftarrow> P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
                  by blast
                have "\<Psi> \<rhd> Q' \<parallel> P' \<sim> (Q'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)))"
                  apply(rule bisimParPres')
                  apply(rule bisimTransitive[OF bisimSymmetric])
                  apply fact
                  unfolding p'1def
                  by(rule bisimReflexive)
                obtain p::"name prm" where "set p \<subseteq> set yvec \<times> set(p\<bullet>yvec)" and "(p\<bullet>yvec) \<sharp>* a\<^isub>i" and "distinctPerm p" and "(p\<bullet>yvec) \<sharp>* M\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>\<^isub>i" and "(p\<bullet>yvec) \<sharp>* P\<^isub>i" and "(p\<bullet>yvec) \<sharp>* Q\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>" and "(p\<bullet>yvec) \<sharp>* P'" and "(p\<bullet>yvec) \<sharp>* Q'" and "(p\<bullet>yvec) \<sharp>* P''" and "(p\<bullet>yvec) \<sharp>* A\<^isub>Q"
                  by(rule chain.name_list_avoiding[where xvec=yvec and c="(M\<^isub>i,\<Psi>\<^isub>i,a\<^isub>i,P\<^isub>i,Q\<^isub>i,\<Psi>,P',Q',P'',A\<^isub>Q)"]) auto
                hence yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>(p\<bullet>a\<^isub>i)\<rparr>(\<langle>(p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" and yvrdef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)"
                  by(subst resChainAlpha[where p=p],simp,simp,simp)+
                obtain b\<^isub>i::name where "b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)" and "b\<^isub>i \<sharp> \<Psi>" and "b\<^isub>i \<sharp> \<Psi>\<^isub>Q" and "b\<^isub>i \<sharp> Q'" and "b\<^isub>i \<sharp> A\<^isub>Q"
                  apply(rule at_exists_fresh[OF at_name_inst, where x="(p\<bullet>M\<^isub>i,p\<bullet>\<Psi>\<^isub>i,p\<bullet>P\<^isub>i,p\<bullet>Q\<^isub>i,\<Psi>,\<Psi>\<^isub>Q,Q',A\<^isub>Q)"])
                  unfolding supp_prod pt_perm_supp[OF pt_name_inst,OF at_name_inst, symmetric]
                  apply(rule finite_UnI, (finite_guess | (subst pt_set_finite_ineq[OF pt_name_inst,OF at_name_inst], finite_guess)))+
                  apply(finite_guess)
                  by force
                have yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" unfolding yvldef
                  by(subst alphaRes[where x="(p\<bullet>a\<^isub>i)" and y="b\<^isub>i"]) (simp add: `b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)`)+
                have "\<Psi> \<rhd> Q' \<parallel> P' \<sim> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(Q'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                proof -
                  note `\<Psi> \<rhd> Q' \<parallel> P' \<sim> (Q'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)))`
                  hence "\<Psi> \<rhd> Q' \<parallel> P' \<sim> (Q'\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)))"
                    unfolding yvldef
                    by -
                  moreover have "\<Psi> \<rhd> (Q'\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))) \<sim> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(Q'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                    apply(rule bisimTransitive[OF bisimSymmetric[OF bisimScopeExtChain]])
                    apply fact
                    apply fact
                    apply(rule bisimResChainPres)
                    apply(rule bisimSymmetric[OF bisimParAssoc])
                    by fact
                  ultimately show ?thesis
                    by(auto dest: bisimTransitive)
                qed
                moreover have "\<Psi> \<rhd> \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(Q' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> Q' \<parallel> P''"
                  apply(rule bisimSymmetric)
                  apply(rule bisimTransitive[OF bisimParPres'])
                  apply fact
                  unfolding p''1def yvrdef
                  apply(rule bisimTransitive[OF bisimSymmetric[OF bisimScopeExtChain]])
                  apply fact
                  apply fact
                  apply(rule bisimResChainPres)
                  apply(rule bisimSymmetric[OF bisimParAssoc])
                  by fact
                moreover have "(\<Psi>, \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(Q' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i),(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(Q'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))) \<in> ?X2"
                proof -
                  have "(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)" using `a\<^isub>i \<sharp> P\<^isub>i` by(subst fresh_bij)
                  note `b\<^isub>i \<sharp> \<Psi>`
                  moreover note `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`
                  moreover have "b\<^isub>i \<sharp> (Q' \<parallel> (p\<bullet>Q\<^isub>i))" using `b\<^isub>i \<sharp> Q'` `b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)`
                    by simp
                  moreover have "guarded(p\<bullet>P\<^isub>i)" using `guarded P\<^isub>i` by simp
                  moreover have "\<not> b\<^isub>i \<sharp> ([(p\<bullet>a\<^isub>i,b\<^isub>i)] \<bullet> (p\<bullet>M\<^isub>i))" using `\<not> a\<^isub>i \<sharp> M\<^isub>i`
                    apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps)
                  moreover have "\<And> \<Psi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>" using EntCl
                    apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                    apply(subst perm_bool[where pi="rev p",symmetric])
                    by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`)
                  moreover have "\<forall>F. (F \<otimes>\<^sub>F (\<langle>[b\<^isub>i],[(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)\<rangle>) \<simeq>\<^sub>F F)"
                    using Equiv1
                    apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                    apply(subst perm_bool[where pi="rev p",symmetric])
                    by(simp add: eqvts swap_simps)
                  moreover have "(\<forall>\<Psi> R. b\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i)) \<Leftarrow> R \<longrightarrow> (R = ((p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>)))" using Equiv2
                    apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i` swap_simps)
                  moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P \<longrightarrow> (supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))) \<subseteq> ((supp \<Psi>)::name set))" using EntSupp
                    apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps)
                  moreover have "\<forall> P P'. ((([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P') = (P = P'))" using Inj
                    apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps)
                  moreover have "((supp(p\<bullet>P\<^isub>i))::name set) \<subseteq> supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i`
                    apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                  moreover have "((supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)))::name set) = supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i`
                    apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                  moreover note `(p\<bullet>yvec) \<sharp>* \<Psi>`
                  moreover from EntEnt have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i)\<parallel>\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                    apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                    apply(subst perm_bool[where pi="p",symmetric])
                    apply(simp add: eqvts swap_simps `distinctPerm p` `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                    done
                  ultimately show ?thesis
                    by blast
                qed
                ultimately show ?thesis
                  apply(rule_tac lol)
                  apply(rule refl)
                  apply(rule bisimSymmetric)
                  apply assumption
                  apply(rule disjI1)
                  apply assumption
                  by(rule bisimSymmetric)
              qed
              hence "(\<Psi>, Q' \<parallel> P'', Q' \<parallel> P') \<in> ?Y" by(rule Y2inY)
              thus ?thesis using `xvec \<sharp>* \<Psi>` by(rule YresChainClose')
            qed
            ultimately show ?case
              by(rule_tac exI) (rule conjI)
          next
            case(cComm2 \<Psi>\<^isub>P M xvec N Q' A\<^isub>Q \<Psi>\<^isub>Q K P' A\<^isub>P)
            have "extractFrame(\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<langle>[a],\<one> \<otimes> \<Psi>\<^isub>c\<rangle>"
              by simp
            with `extractFrame(\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>` have "\<langle>[a],\<one> \<otimes> \<Psi>\<^isub>c\<rangle> = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>" by simp
            obtain b where "A\<^isub>P=[b]"
            proof -
              assume "\<And>b. A\<^isub>P = [b] \<Longrightarrow> thesis"
              moreover from `extractFrame (\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>`
              have "\<exists> b. A\<^isub>P=[b]"
              proof(induct A\<^isub>P)
                case Nil thus ?case by(simp)
              next
                case(Cons a\<^isub>Q A\<^isub>Q)
                have "[(a, a\<^isub>Q)] \<bullet> A\<^isub>Q = [] \<Longrightarrow> A\<^isub>Q = []"
                  by(induct A\<^isub>Q) (auto simp add: eqvts)
                with Cons(2) have "A\<^isub>Q=[]"
                  by(auto simp add: abs_fun_eq[OF pt_name_inst, OF at_name_inst] frame.inject eqvts)
                thus ?case by simp
              qed
              ultimately show ?thesis
                by auto
            qed
            from cComm2 have "b \<sharp> \<Psi>" and "b \<sharp> \<Psi>\<^isub>Q" and "b \<sharp> Q" and "b \<sharp> K" and "b \<sharp> N" and "b \<sharp> Q'" and "b \<sharp> xvec" and "b \<sharp> P'" and "a \<sharp> b" and "b \<sharp> M\<^isub>a" and "b \<sharp> \<Psi>\<^isub>c" and "b \<sharp> P" unfolding `A\<^isub>P = [b]` by(simp)+
            have "\<Psi>\<^isub>P = \<one> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)" using `extractFrame (\<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>)) = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>` unfolding `A\<^isub>P = [b]`
              by(auto simp add: frame.inject abs_fun_eq'[OF pt_name_inst, OF at_name_inst] eqvts pt3[OF pt_name_inst, OF at_ds5, OF at_name_inst])
            have Equiv1': "\<And>F. F \<otimes>\<^sub>F (\<langle>[b], [(a, b)] \<bullet> \<Psi>\<^isub>c\<rangle>) \<simeq>\<^sub>F F" using `\<And>F. F \<otimes>\<^sub>F (\<langle>[a], \<Psi>\<^isub>c\<rangle>) \<simeq>\<^sub>F F`
              by(subst (asm) frameChainAlpha [where p="[(a,b)]"]) (simp add: swap_simps `b \<sharp> \<Psi>\<^isub>c`)+
            have Equiv2': "\<And>\<Psi> R. \<lbrakk>b \<sharp> \<Psi>; \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a) \<Leftarrow> R\<rbrakk> \<Longrightarrow> R = P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>"
            proof -
              fix \<Psi> R
              assume "b \<sharp> \<Psi>" and "\<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a) \<Leftarrow> R"
              moreover have "\<forall> \<Psi> R. (b \<sharp> \<Psi> \<and> \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a) \<Leftarrow> R) \<longrightarrow> R = P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>" using Equiv2
                by(subst perm_bool[where pi="[(a,b)]",symmetric]) (force simp add: eqvts swap_simps `a \<sharp> P` `b \<sharp> P`)
              ultimately show "R = P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>"
                by auto
            qed
            have "((supp P)::name set) \<subseteq> supp([(a,b)]\<bullet>M\<^isub>a)" using `((supp P)::name set) \<subseteq> supp M\<^isub>a`
              apply(subst perm_bool[where pi="[(a,b)]",symmetric])
              by(simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
            from `((supp \<Psi>\<^isub>c)::name set) = supp M\<^isub>a` have "((supp([(a,b)]\<bullet>\<Psi>\<^isub>c))::name set) = supp([(a,b)]\<bullet>M\<^isub>a)"
              apply(subst (asm) perm_bool[where pi="[(a,b)]",symmetric])
              by(simp only: eqvts)
            have EntCl': "\<And>\<Psi>. \<Psi> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>" using `\<And>\<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle>`
              by(subst perm_bool[where pi="[(a,b)]",symmetric]) (simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
            have EntEnt': "\<And>\<Psi> \<phi>. \<Psi> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> \<phi> \<Longrightarrow>
              \<phi> = ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
            proof -
              fix \<Psi> \<phi>
              assume "\<Psi> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> \<phi>"
              moreover have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<or> \<Psi> \<turnstile> \<phi>" using `\<And>\<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>c \<turnstile> \<phi> \<Longrightarrow> \<phi> = M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> \<Psi> \<turnstile> \<phi>`
                apply(subst perm_bool[where pi="[(a,b)]",symmetric])
                by(simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
              ultimately show "\<phi> = ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                by simp
            qed
            {
              fix \<Psi> P
              assume "\<Psi> \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P"
              moreover have "\<forall> \<Psi> P. \<Psi> \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<longrightarrow> supp ([(a, b)] \<bullet> M\<^isub>a) \<subseteq> (supp \<Psi>::name set)" using `\<forall>\<Psi> P. \<Psi> \<turnstile> M\<^isub>a\<Leftarrow>P \<longrightarrow> supp M\<^isub>a \<subseteq> supp \<Psi>`
                apply(subst perm_bool[where pi="[(a,b)]",symmetric])
                by(simp add: eqvts)
              ultimately have "supp ([(a, b)] \<bullet> M\<^isub>a) \<subseteq> (supp \<Psi>::name set)"
                by auto
            }
            note EntSupp' = this
            have Inj': "\<And>P P'. (([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P = ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P') = (P = P')" using `\<forall>P P'. (M\<^isub>a\<Leftarrow>P = M\<^isub>a\<Leftarrow>P') = (P = P')`
              by(subst (asm) perm_bool[where pi="[(a,b)]",symmetric]) (simp add: eqvts)
            have "\<not> b \<sharp> [(a, b)] \<bullet> M\<^isub>a" using `\<not> a \<sharp> M\<^isub>a`
              by(subst (asm) perm_bool [where pi="[(a,b)]",symmetric]) (simp add: eqvts swap_simps)
            have "b \<sharp> \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
              by(simp add: fresh_abs_fun_iff[OF pt_name_inst, OF at_name_inst, OF fin_supp])
            have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<longmapsto> K\<lparr>N\<rparr> \<prec> P'" using `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<longmapsto> K\<lparr>N\<rparr> \<prec> P'`
              apply(subst (asm) alphaRes[where y=b])
              apply(simp add: `b \<sharp> M\<^isub>a` `b \<sharp> \<Psi>\<^isub>c`)
              by(simp only: eqvts)
            moreover have "b \<sharp> \<Psi> \<otimes> \<Psi>\<^isub>Q" using `b \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q` by force
            moreover have "b \<sharp> K\<lparr>N\<rparr>" using `b \<sharp> K` `b \<sharp> xvec` `b \<sharp> N` by simp
            moreover have "b \<sharp> P'" using `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<longmapsto> K\<lparr>N\<rparr> \<prec> P'` `b \<sharp> N` `b \<sharp> \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)`
              by(rule_tac inputFreshDerivative)
            moreover have "bn(K\<lparr>N\<rparr>) \<sharp>* (\<Psi> \<otimes> \<Psi>\<^isub>Q)" using `xvec \<sharp>* \<Psi>` `xvec \<sharp>* \<Psi>\<^isub>Q` by force
            moreover have "bn(K\<lparr>N\<rparr>) \<sharp>* ([(a,b)]\<bullet>M\<^isub>a)" using `xvec \<sharp>* (a, M\<^isub>a, \<Psi>\<^isub>c, P)` `b \<sharp> xvec`
              by simp
            moreover have "bn(K\<lparr>N\<rparr>) \<sharp>* ([(a,b)]\<bullet>\<Psi>\<^isub>c)" using `xvec \<sharp>* (a, M\<^isub>a, \<Psi>\<^isub>c, P)` `b \<sharp> xvec`
              by simp
            moreover have "bn(K\<lparr>N\<rparr>) \<sharp>* subject(K\<lparr>N\<rparr>)" using `xvec \<sharp>* K`
              by simp
            moreover note `b \<sharp> P`
            moreover note `guarded P`
            moreover note `\<not> b \<sharp> ([(a,b)]\<bullet>M\<^isub>a)`
            moreover have "\<forall> \<Psi>'. \<Psi>' \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a) \<Leftarrow> P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>" using EntCl' by simp
            moreover have "\<forall>\<Psi>' Q. b \<sharp> \<Psi>' \<and> \<Psi>' \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>Q \<longrightarrow> Q = P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>" using Equiv2'
              by force
            ultimately have "\<exists> P''. \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K\<lparr>N\<rparr> \<prec> P'' \<and> (\<Psi>,P'',P') \<in> ?Y2"
            proof(induct rule: resReplCases)
              case(cRes P')
              thus ?case
              proof(induct rule: parInputCases[where C="()"])
                case cPar2 thus ?case by(drule_tac nilTrans) (rule FalseE)
              next
                case(cPar1 Q' A\<^isub>R \<Psi>\<^isub>R)
                obtain \<alpha> where alphadef: "\<alpha> = K\<lparr>N\<rparr>" by simp
                from `extractFrame (\<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>) = \<langle>A\<^isub>R, \<Psi>\<^isub>R\<rangle>` have "A\<^isub>R = []" and psirdef: "\<Psi>\<^isub>R = ([(a,b)]\<bullet>\<Psi>\<^isub>c)"
                  by auto
                obtain \<Psi>' where psi'def: "\<Psi>'=(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)"
                  by simp
                obtain M\<^isub>b \<Psi>\<^isub>b where mbdef: "M\<^isub>b = [(a,b)]\<bullet>M\<^isub>a" and psibdef: "\<Psi>\<^isub>b = [(a,b)]\<bullet>\<Psi>\<^isub>c"
                  by auto
                have "b \<sharp> bn \<alpha>" and "b \<sharp> object \<alpha>" unfolding alphadef using `b \<sharp> xvec` `b \<sharp> N` by auto
                have "a \<sharp> \<Psi>\<^isub>Q" using `a \<sharp> Q` `extractFrame Q = \<langle>A\<^isub>Q,\<Psi>\<^isub>Q\<rangle>` `A\<^isub>Q \<sharp>* (a, M\<^isub>a, \<Psi>\<^isub>c, P)`
                  by(drule_tac extractFrameFresh) simp
                have "bn \<alpha> \<sharp>* \<Psi>\<^isub>Q" unfolding alphadef using `xvec \<sharp>* Q` `extractFrame Q = \<langle>A\<^isub>Q,\<Psi>\<^isub>Q\<rangle>` `A\<^isub>Q \<sharp>* xvec`
                  by(drule_tac extractFrameFreshChain) simp
                have "bn \<alpha> \<sharp>* a" "b \<sharp> bn \<alpha>" "b \<sharp> object \<alpha>" "bn \<alpha> \<sharp>* M\<^isub>a" "bn \<alpha> \<sharp>* P" "bn \<alpha> \<sharp>* \<Psi>\<^isub>c" "bn \<alpha> \<sharp>* \<Psi>" unfolding alphadef using `xvec \<sharp>* K` `xvec \<sharp>* (a, M\<^isub>a, \<Psi>\<^isub>c, P)` `b \<sharp> xvec` `xvec \<sharp>* \<Psi>` `b \<sharp> N`
                  by auto
                have "\<Psi>' \<rhd> \<langle>M\<^isub>b\<rangle> \<longmapsto> \<alpha> \<prec> Q'" unfolding psi'def mbdef psirdef[symmetric] alphadef by fact
                moreover have "bn \<alpha> \<sharp>* subject \<alpha>" unfolding alphadef by fact
                moreover note `guarded P`
                moreover have "\<forall> \<Psi>''. \<Psi>' \<otimes> \<Psi>'' \<turnstile> M\<^isub>b \<Leftarrow> P \<parallel> \<langle>M\<^isub>b\<rangle>"
                proof -
                  {
                    fix \<Psi>''
                    have "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>'' \<turnstile> M\<^isub>a \<Leftarrow> P \<parallel> \<langle>M\<^isub>a\<rangle>" using EntCl[where \<Psi>="\<Psi>\<otimes>\<Psi>\<^isub>Q\<otimes>\<Psi>''"]
                      by(rule_tac statEqEnt[where \<Psi>="(\<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<Psi>'') \<otimes> \<Psi>\<^isub>c"]) (metis AssertionStatEqTrans Composition AssertionStatEqSym Associativity Commutativity)
                  }
                  thus ?thesis unfolding psi'def psirdef mbdef
                    by(subst perm_bool[where pi="[(a,b)]",symmetric]) (simp add: eqvts `a \<sharp> \<Psi>` `a \<sharp> \<Psi>\<^isub>Q` `b \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q` `a \<sharp> P` `b \<sharp> P`)
                qed
                moreover have "\<forall> \<Psi>'' Q. b \<sharp> \<Psi>'' \<and> \<Psi>' \<otimes> \<Psi>'' \<turnstile> M\<^isub>b \<Leftarrow> Q \<longrightarrow> Q = P \<parallel> \<langle>M\<^isub>b\<rangle>"
                proof -
                  {
                    fix \<Psi>'' Q
                    assume "a \<sharp> \<Psi>''"
                      and    "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>'' \<turnstile> M\<^isub>a \<Leftarrow> Q"
                    hence "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>'') \<otimes> \<Psi>\<^isub>c \<turnstile> M\<^isub>a \<Leftarrow> Q"
                      by(rule_tac statEqEnt[where \<Psi>="((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>c) \<otimes> \<Psi>''"]) (metis AssertionStatEqTrans Composition AssertionStatEqSym Associativity Commutativity)
                    moreover have "a \<sharp> (\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>''" using `a \<sharp> \<Psi>` `a \<sharp> \<Psi>\<^isub>Q` `a \<sharp> \<Psi>''`
                      by auto
                    ultimately have "Q = P \<parallel> \<langle>M\<^isub>a\<rangle>"
                      by(rule_tac Equiv2)
                  }
                  thus ?thesis unfolding psi'def psirdef mbdef
                    by(subst perm_bool[where pi="[(a,b)]",symmetric]) (auto simp add: eqvts `a \<sharp> \<Psi>` `a \<sharp> \<Psi>\<^isub>Q` `b \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q` `a \<sharp> P` `b \<sharp> P` swap_simps)
                qed
                ultimately have "\<exists> P'. \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P' \<and> (\<Psi>, P', \<lparr>\<nu>b\<rparr>(Q'\<parallel>\<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2" using mbdef psi'def `bn \<alpha> \<sharp>* a` `b \<sharp> bn \<alpha>` `b \<sharp> object \<alpha>` `bn \<alpha> \<sharp>* M\<^isub>a` `bn \<alpha> \<sharp>* P` `bn \<alpha> \<sharp>* \<Psi>\<^isub>c` `bn \<alpha> \<sharp>* \<Psi>` `bn \<alpha> \<sharp>* \<Psi>\<^isub>Q` `b \<sharp> P` Equiv1' EntCl' EntSupp' Equiv2' Inj' EntEnt' `a \<sharp> P` `\<not>(b \<sharp> ([(a,b)]\<bullet>M\<^isub>a))` `((supp P)::name set) \<subseteq> supp([(a,b)]\<bullet>M\<^isub>a)` `((supp([(a,b)]\<bullet>\<Psi>\<^isub>c))::name set) = supp([(a,b)]\<bullet>M\<^isub>a)` `b \<sharp> M\<^isub>a` `b \<sharp> \<Psi>\<^isub>c` `b \<sharp> \<Psi>` `b \<sharp> \<Psi>\<^isub>Q`
                proof(nominal_induct avoiding: a b M\<^isub>a \<Psi> \<Psi>\<^isub>Q \<Psi>\<^isub>c rule: replInduct')
                  case(cAlpha \<alpha> P' p)
                  from `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` `b \<sharp> bn(p \<bullet> \<alpha>)` `bn \<alpha> \<sharp>* b` have "p \<bullet> b = b"
                    by simp
                  from `bn \<alpha> \<sharp>* b` have "b \<sharp> bn \<alpha>" by simp
                  from `b \<sharp> object(p \<bullet> \<alpha>)` have "b \<sharp> object \<alpha>"
                    by(subst perm_bool[where pi=p,symmetric]) (simp only: eqvts `p\<bullet>b=b`)
                  have "\<exists> P''. \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P'' \<and> (\<Psi>, P'', \<lparr>\<nu>b\<rparr>(P' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by(rule cAlpha) fact+
                  then obtain P'' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P'' \<and> (\<Psi>, P'', \<lparr>\<nu>b\<rparr>(P' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by(rule exE)
                  hence "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P''" and "(\<Psi>, P'', \<lparr>\<nu>b\<rparr>(P' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by simp+
                  from `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` `bn (p \<bullet> \<alpha>) \<sharp>* a` `bn \<alpha> \<sharp>* a` have "p \<bullet> a = a"
                    by simp
                  from `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` `bn (p \<bullet> \<alpha>) \<sharp>* M\<^isub>a` `bn \<alpha> \<sharp>* M\<^isub>a` have "p \<bullet> M\<^isub>a = M\<^isub>a"
                    by simp
                  from `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` `bn (p \<bullet> \<alpha>) \<sharp>* \<Psi>\<^isub>c` `bn \<alpha> \<sharp>* \<Psi>\<^isub>c` have "p \<bullet> \<Psi>\<^isub>c = \<Psi>\<^isub>c"
                    by simp
                  from `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` `bn (p \<bullet> \<alpha>) \<sharp>* P` `bn \<alpha> \<sharp>* P` have "p \<bullet> P = P"
                    by simp
                  from `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` `bn (p \<bullet> \<alpha>) \<sharp>* \<Psi>\<^isub>Q` `bn \<alpha> \<sharp>* \<Psi>\<^isub>Q` have "p \<bullet> \<Psi>\<^isub>Q = \<Psi>\<^isub>Q"
                    by simp
                  from `set p \<subseteq> set (bn \<alpha>) \<times> set (bn (p \<bullet> \<alpha>))` `bn (p \<bullet> \<alpha>) \<sharp>* \<Psi>` `bn \<alpha> \<sharp>* \<Psi>` have "p \<bullet> \<Psi> = \<Psi>"
                    by simp
                  from `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P''` have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> (p\<bullet>\<alpha>) \<prec> (p\<bullet>P'')"
                    by(subst (asm) perm_bool[where pi=p,symmetric])(simp only: eqvts `p\<bullet>\<Psi>=\<Psi>` `p\<bullet>\<Psi>\<^isub>Q=\<Psi>\<^isub>Q` `p\<bullet>P=P`)
                  moreover have "(\<Psi>, p\<bullet>P'', \<lparr>\<nu>b\<rparr>((p \<bullet> P') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                  proof -
                    from `(\<Psi>, P'', \<lparr>\<nu>b\<rparr>(P' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2` `eqvt ?Y2` have "(p\<bullet>(\<Psi>, P'', \<lparr>\<nu>b\<rparr>(P' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>))) \<in> ?Y2"
                      by(rule_tac eqvtI)
                    moreover have "(p\<bullet>(\<Psi>, P'', \<lparr>\<nu>b\<rparr>(P' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>))) = (\<Psi>, p\<bullet>P'', \<lparr>\<nu>b\<rparr>((p \<bullet> P') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>))"
                      by(simp add: eqvts  `p\<bullet>\<Psi>=\<Psi>` `p\<bullet>\<Psi>\<^isub>Q=\<Psi>\<^isub>Q` `p\<bullet>P=P` `p\<bullet>b=b` `p\<bullet>a=a` `p\<bullet>\<Psi>\<^isub>c=\<Psi>\<^isub>c` cp1[OF cp_pt_inst, OF pt_name_inst, OF at_name_inst, of p "[(a,b)]" \<Psi>\<^isub>c])
                    ultimately show ?thesis
                      by simp
                  qed
                  ultimately show ?case
                    by(rule_tac exI) (rule conjI)
                next
                  case(cPar1 \<alpha> P')
                  have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> P \<longmapsto> \<alpha> \<prec> P'"
                  proof -
                    have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<rhd> P \<longmapsto> \<alpha> \<prec> P'"
                      apply(rule statEqTransition)
                      apply fact
                      apply(subst cPar1)
                      apply(rule AssertionStatEqRefl)
                      done
                    thus ?thesis
                      apply(rule replTrans[where a="b" and M\<^isub>a="[(a,b)]\<bullet>M\<^isub>a"])
                      by(fact | rule freshComp)+
                  qed
                  hence "b \<sharp> P'"
                    by(rule freeFreshDerivative') fact+
                  have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P' \<parallel> !P"
                    apply(rule_tac Bang)
                    apply(rule Par1[where A\<^isub>Q="[]" and \<Psi>\<^isub>Q=\<one>])
                    apply(rule statEqTransition[OF `\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> P \<longmapsto> \<alpha> \<prec> P'`])
                    apply(rule AssertionStatEqSym[OF Identity])
                    by(fact|simp add: `bn \<alpha> \<sharp>* P`)+
                  moreover have "(\<Psi>,P'\<parallel>!P,\<lparr>\<nu>b\<rparr>(P' \<parallel> \<langle>M\<^isub>b\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                  proof -
                    have "\<Psi> \<rhd> P'\<parallel>!P \<sim> P' \<parallel> !P" by(rule bisimReflexive)
                    moreover have "\<Psi> \<rhd> P' \<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>b\<rparr>(P' \<parallel> \<langle>M\<^isub>b\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)"
                      apply(subst `M\<^isub>b = [(a,b)]\<bullet>M\<^isub>a`)
                      apply(rule bisimSymmetric)
                      apply(rule bisimTransitive[OF bisimResPres, OF bisimParAssoc, OF `b \<sharp> \<Psi>`])
                      apply(rule bisimScopeExt[OF `b \<sharp> P'`])
                      done
                    moreover have "(\<Psi>,P' \<parallel> !P, P' \<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?X2"
                    proof -
                      note `b \<sharp> \<Psi>` `b \<sharp> P` `b \<sharp> P'` `\<not> b \<sharp> ([(a,b)]\<bullet>M\<^isub>a)`
                      moreover have "(\<forall>\<Psi>. \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)" using cPar1
                        by simp
                      moreover have "\<forall> F. F \<otimes>\<^sub>F (\<langle>[b],([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rangle>) \<simeq>\<^sub>F F" using cPar1 by simp
                      moreover have "(\<forall>\<Psi> R. b \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>R \<longrightarrow> R = P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle>)" using cPar1
                        by simp
                      moreover note `guarded P` `supp P \<subseteq> supp([(a,b)]\<bullet>M\<^isub>a)` `supp([(a,b)]\<bullet>\<Psi>\<^isub>c) = (supp([(a,b)]\<bullet>M\<^isub>a)::name set)`
                      moreover have "(\<forall>P P'. (([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P = ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P') = (P = P'))" using cPar1
                        by simp
                      moreover have "(\<forall>\<Psi> P. \<Psi> \<turnstile> ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<longrightarrow> supp([(a,b)]\<bullet>M\<^isub>a) \<subseteq> (supp \<Psi>::name set))"
                        using cPar1 by auto
                      moreover have "([]::name list) \<sharp>* \<Psi>" by simp
                      moreover have "(\<forall>\<Psi> \<phi>. \<Psi> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(a,b)]\<bullet>M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<or> \<Psi> \<turnstile> \<phi>)" using cPar1 by simp
                      ultimately have "(\<Psi>,\<lparr>\<nu>*[]\<rparr>P' \<parallel> !P, \<lparr>\<nu>*[]\<rparr>(P' \<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>))) \<in> ?X2"
                        by blast
                      thus ?thesis by simp
                    qed                        
                    ultimately show ?thesis
                      apply(rule_tac lol)
                      apply(rule refl)
                      apply(assumption)
                      apply(rule disjI1)
                      by -
                  qed
                  ultimately show ?case
                    by(rule_tac exI) (rule conjI)
                next
                  case(cComm1 M N P' K xvec P'')
                  have "b \<sharp> N"
                  proof -
                    note `\<Psi>' \<rhd> \<langle>M\<^isub>b\<rangle> \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P''`
                    moreover note `guarded P`
                    moreover have "\<forall> \<Psi>''. \<Psi>' \<otimes> \<Psi>'' \<turnstile> M\<^isub>b \<Leftarrow> P \<parallel> \<langle>M\<^isub>b\<rangle>"
                    proof -
                      {
                        fix \<Psi>''
                        have seq: "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<Psi>'' \<simeq>  (\<Psi>'' \<otimes> ((\<Psi> \<otimes> \<Psi>\<^isub>Q))) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)"
                          by(metis AssertionStatEqTrans Commutativity AssertionStatEqSym Associativity AssertionStatEqRefl)
                        have "\<Psi>' \<otimes> \<Psi>'' \<turnstile> M\<^isub>b \<Leftarrow> P \<parallel> \<langle>M\<^isub>b\<rangle>"
                          unfolding cComm1
                          apply(rule statEqEnt[OF AssertionStatEqSym, OF seq])
                          by fact
                      }
                      thus ?thesis by simp
                    qed
                    moreover have "\<forall> \<Psi>'' Q. b \<sharp> \<Psi>'' \<and> \<Psi>' \<otimes> \<Psi>'' \<turnstile> M\<^isub>b \<Leftarrow> Q \<longrightarrow> Q = P \<parallel> \<langle>M\<^isub>b\<rangle>"
                    proof -
                      {
                        fix \<Psi>'' Q
                        assume "b \<sharp> \<Psi>''" and "\<Psi>' \<otimes> \<Psi>'' \<turnstile> M\<^isub>b \<Leftarrow> Q"
                        have seq: "((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)) \<otimes> \<Psi>'' \<simeq>  (\<Psi>'' \<otimes> ((\<Psi> \<otimes> \<Psi>\<^isub>Q))) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c)"
                          by(metis AssertionStatEqTrans Commutativity AssertionStatEqSym Associativity AssertionStatEqRefl)
                        have "(\<Psi>'' \<otimes> ((\<Psi> \<otimes> \<Psi>\<^isub>Q))) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> M\<^isub>b \<Leftarrow> Q" using `\<Psi>' \<otimes> \<Psi>'' \<turnstile> M\<^isub>b \<Leftarrow> Q` unfolding cComm1
                          by(rule statEqEnt[OF seq])
                        moreover have "b \<sharp> (\<Psi>'' \<otimes> ((\<Psi> \<otimes> \<Psi>\<^isub>Q)))" using `b \<sharp> \<Psi>` `b \<sharp> \<Psi>''` `b \<sharp> \<Psi>\<^isub>Q` by force
                        ultimately have "Q = P \<parallel> \<langle>M\<^isub>b\<rangle>"
                          unfolding cComm1
                          by(rule_tac cComm1)
                      }
                      thus ?thesis by auto
                    qed
                    ultimately have "b \<sharp> bnlNames(\<lparr>\<nu>*xvec\<rparr>N \<prec>' P'')" using `b \<sharp> P`
                    proof(nominal_induct avoiding: b rule: replOutputInduct)
                      case cInv thus ?case by -
                    next
                      case(cPar1 xvec N P')
                      from `\<Psi>' \<rhd> P \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'`  `xvec \<sharp>* b` `b \<sharp> P` `distinct xvec` `xvec \<sharp>* K`
                      have "b \<sharp> N"
                        by(rule_tac outputFreshDerivative) auto
                      thus ?case unfolding bnlNamesJudge fresh_def
                        by(auto simp add: name_set_supp[OF finite_Diff, OF fin_supp])
                    next
                      case(cPar2 xvec N P')
                      have "b \<sharp> bnlNames (\<lparr>\<nu>*xvec\<rparr>N \<prec>' P')" by(rule cPar2) fact
                      thus ?case unfolding bnlNamesJudge by -
                    qed
                    moreover have "b \<sharp> xvec" using `xvec \<sharp>* b` by simp
                    ultimately show "b \<sharp> N" unfolding bnlNamesJudge
                      by(auto simp add: name_set_supp[OF finite_Diff, OF fin_supp] fresh_def)
                  qed
                  hence "b \<sharp> P'" using `\<Psi>' \<rhd> P \<longmapsto> M\<lparr>N\<rparr> \<prec> P'` `b \<sharp> P`
                    by(rule_tac inputFreshDerivative)
                  have "a \<sharp> xvec" and "b \<sharp> xvec" using `xvec \<sharp>* a` `xvec \<sharp>* b`
                    by simp+
                  have "\<exists> P'. \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K \<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P' \<and> (\<Psi>, P', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by(rule_tac cComm1) (fact|simp add: `b \<sharp> N` `b \<sharp> xvec` `a \<sharp> xvec`)+
                  then obtain Q' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K \<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q' \<and> (\<Psi>, Q', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by(rule exE)
                  hence "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K \<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q'" and "(\<Psi>, Q', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by simp+
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<rhd> P \<longmapsto> M\<lparr>N\<rparr> \<prec> P'" using `\<Psi>' \<rhd> P \<longmapsto> M\<lparr>N\<rparr> \<prec> P'` by(subst (asm) cComm1)
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<one> \<rhd> P \<longmapsto> M\<lparr>N\<rparr> \<prec> P'"
                    apply(rule_tac statEqTransition[where \<Psi>="\<Psi>\<otimes>\<Psi>\<^isub>Q"])
                    apply(rule replTrans)
                    by(fact|simp|rule freshComp|rule AssertionStatEqSym[OF Identity])+
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<one> \<rhd> !P \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q'"
                    apply(rule statEqTransition)
                    by(fact) (rule AssertionStatEqSym[OF Identity])
                  obtain A\<^isub>P \<Psi>\<^isub>P where "extractFrame P = \<langle>A\<^isub>P,\<Psi>\<^isub>P\<rangle>" and "A\<^isub>P \<sharp>* \<Psi>" and "A\<^isub>P \<sharp>* \<Psi>\<^isub>Q" and "A\<^isub>P \<sharp>* !P" and "A\<^isub>P \<sharp>* P" and "A\<^isub>P \<sharp>* M" and "distinct A\<^isub>P"
                    by(rule freshFrame[where F="extractFrame P" and C="(\<Psi>,\<Psi>\<^isub>Q,P,M)"]) force
                  have "\<Psi>\<^isub>P \<simeq> \<one>" and "supp \<Psi>\<^isub>P = ({}::name set)" using `guarded P` `extractFrame P = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>`
                    by(auto dest: guardedStatEq)
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>P \<rhd> !P \<longmapsto> K\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q'"
                    apply(rule statEqTransition)
                    apply fact
                    apply(rule compositionSym)
                    apply(rule AssertionStatEqSym)
                    by fact
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>P \<otimes> \<one> \<turnstile> M \<leftrightarrow> K"
                  proof -
                    from `\<Psi>' \<turnstile> M \<leftrightarrow> K` have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<turnstile> M \<leftrightarrow> K" using DisjointCCl[where M=M and N=K]
                      unfolding cComm1
                      apply(drule_tac cComm1)
                      apply(erule disjE)
                      by blast
                    thus ?thesis
                      apply(rule_tac statEqEnt)
                      apply(rule AssertionStatEqSym[OF AssertionStatEqTrans[OF compositionSym[OF Identity]]])
                      apply(rule AssertionStatEqTrans[OF compositionSym[OF `\<Psi>\<^isub>P \<simeq> \<one>`]])
                      by(rule Identity)
                  qed
                  have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<tau> \<prec> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q'"
                    apply(rule Bang)
                    apply(rule Comm1[where M=M and K=K and N=N and A\<^isub>P="A\<^isub>P" and \<Psi>\<^isub>P="\<Psi>\<^isub>P" and A\<^isub>Q="[]" and \<Psi>\<^isub>Q="\<one>"])
                    by(fact|rule freshCompChain|simp)+
                  moreover have "(\<Psi>, \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q', \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                  proof -
                    from `(\<Psi>, Q', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2` obtain Q'' P''' where "\<Psi> \<rhd> Q' \<sim> Q''" and Disj: "(\<Psi>,Q'',P''') \<in> ?X2 \<or> \<Psi> \<rhd> Q'' \<sim> P'''" and "\<Psi> \<rhd> P''' \<sim> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                      by auto blast
                    have "xvec \<sharp>* \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>" using `xvec \<sharp>* a` `xvec \<sharp>* b` `xvec \<sharp>* \<Psi>\<^isub>c`
                      by simp
                    have rhs: "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>))"
                      apply(rule bisimTransitive[OF bisimResPres])
                      apply(rule bisimTransitive[OF bisimParComm])
                      apply(rule bisimSymmetric[OF bisimScopeExtChain])
                      apply fact+
                      apply(rule bisimTransitive[OF bisimResComm'])
                      apply fact
                      apply fact
                      apply(rule bisimResChainPres)
                      apply(rule bisimSymmetric)
                      apply(rule bisimTransitive[OF bisimSymmetric[OF bisimScopeExt]])
                      apply fact
                      apply(rule bisimResPres)
                      apply(rule bisimTransitive[OF bisimSymmetric[OF bisimParAssoc]])
                      apply(rule bisimTransitive[OF bisimParComm])
                      apply(rule bisimReflexive)
                      by fact+
                    from Disj show ?thesis
                    proof(rule disjE)
                      assume "\<Psi> \<rhd> Q'' \<sim> P'''"
                      have bis: "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q' \<sim> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)"
                        apply(rule bisimSymmetric)
                        apply(rule bisimTransitive[OF rhs])
                        apply(rule bisimResChainPres)
                        apply(rule bisimParPres')
                        apply(rule bisimTransitive[OF bisimSymmetric[OF `\<Psi> \<rhd> P''' \<sim> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`]])
                        apply(rule bisimSymmetric)
                        apply(rule bisimTransitive[OF `\<Psi> \<rhd> Q' \<sim> Q''`])
                        apply(rule `\<Psi> \<rhd> Q'' \<sim> P'''`)
                        by fact
                      show ?thesis
                        apply(rule lol)
                        apply(rule refl)
                        apply(rule bis)
                        apply(rule disjI2)
                        by(rule bisimReflexive)+
                    next
                      assume "(\<Psi>,Q'',P''') \<in> ?X2"
                      then obtain yvec Q\<^isub>i a\<^isub>i M\<^isub>i \<Psi>\<^isub>i P\<^isub>i where p'''def: "P''' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)" and q''def: "Q'' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i" and "a\<^isub>i \<sharp> \<Psi>" and "a\<^isub>i \<sharp> P\<^isub>i" and "a\<^isub>i \<sharp> Q\<^isub>i" and "\<not> a\<^isub>i \<sharp> M\<^isub>i" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i\<Leftarrow>P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a\<^isub>i], \<Psi>\<^isub>i\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i \<Leftarrow> R \<longrightarrow> (R = (P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>)))" and "guarded P\<^isub>i" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>i\<Leftarrow>P \<longrightarrow> (supp M\<^isub>i) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>i\<Leftarrow>P = M\<^isub>i\<Leftarrow>P') = (P = P'))" and "((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i" and "((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i" and "yvec \<sharp>* \<Psi>" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>i \<Leftarrow> P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
                        by blast
                      have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))))"
                      proof -
                        have "xvec \<sharp>* ([(a,b)] \<bullet> \<Psi>\<^isub>c)" using `a \<sharp> xvec` `b \<sharp> xvec` `xvec \<sharp>* \<Psi>\<^isub>c`
                          by simp
                        have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>b\<rparr>(\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                          apply(rule bisimResPres)
                          apply(rule bisimTransitive[OF bisimParComm])
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimResChainPres])
                          apply(rule bisimParComm)
                          apply fact
                          apply(rule bisimScopeExtChain)
                          by(fact|simp add: `xvec \<sharp>* ([(a,b)] \<bullet> \<Psi>\<^isub>c)`)+
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>(\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>b\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                          by(rule bisimResComm') fact+
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>b\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>(P'\<parallel> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                          apply(rule bisimResChainPres)
                          apply(rule bisimTransitive[OF bisimResPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExt)
                          by fact+
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(P'\<parallel> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> (\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))))"
                          apply(rule bisimResChainPres)
                          apply(rule bisimTransitive[OF bisimParComm])
                          apply(rule bisimSymmetric[OF bisimTransitive[OF bisimParComm]])
                          apply(rule bisimParPres)
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimSymmetric[OF `\<Psi> \<rhd> P''' \<sim> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`]])
                          apply(subst p'''def)
                          by(rule bisimReflexive) fact
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      obtain p::"name prm" where "set p \<subseteq> set yvec \<times> set(p\<bullet>yvec)" and "(p\<bullet>yvec) \<sharp>* a\<^isub>i" and "distinctPerm p" and "(p\<bullet>yvec) \<sharp>* M\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>\<^isub>i" and "(p\<bullet>yvec) \<sharp>* P\<^isub>i" and "(p\<bullet>yvec) \<sharp>* Q\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>" and "(p\<bullet>yvec) \<sharp>* P'" and "(p\<bullet>yvec) \<sharp>* xvec" and "(p\<bullet>yvec) \<sharp>* A\<^isub>Q"
                        by(rule chain.name_list_avoiding[where xvec=yvec and c="(M\<^isub>i,\<Psi>\<^isub>i,a\<^isub>i,P\<^isub>i,Q\<^isub>i,\<Psi>,P',xvec,A\<^isub>Q)"]) auto
                      hence yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>(p\<bullet>a\<^isub>i)\<rparr>(\<langle>(p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" and yvrdef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)"
                        by(subst resChainAlpha[where p=p],simp,simp,simp)+
                      obtain b\<^isub>i::name where "b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)" and "b\<^isub>i \<sharp> \<Psi>" and "b\<^isub>i \<sharp> \<Psi>\<^isub>Q" and "b\<^isub>i \<sharp> P'" and "b\<^isub>i \<sharp> A\<^isub>Q"
                        apply(rule at_exists_fresh[OF at_name_inst, where x="(p\<bullet>M\<^isub>i,p\<bullet>\<Psi>\<^isub>i,p\<bullet>P\<^isub>i,p\<bullet>Q\<^isub>i,\<Psi>,\<Psi>\<^isub>Q,P',A\<^isub>Q)"])
                        unfolding supp_prod pt_perm_supp[OF pt_name_inst,OF at_name_inst, symmetric]
                        apply(rule finite_UnI, (finite_guess | (subst pt_set_finite_ineq[OF pt_name_inst,OF at_name_inst], finite_guess)))+
                        apply(finite_guess)
                        by force
                      have yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" unfolding yvldef
                        by(subst alphaRes[where x="(p\<bullet>a\<^isub>i)" and y="b\<^isub>i"]) (simp add: `b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)`)+
                      have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (\<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                      proof -
                        note `\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))`
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                          unfolding yvldef
                          by(rule bisimReflexive)
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)) \<sim> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>P'\<parallel>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)"
                          apply(subst resChainAppend)
                          apply(rule bisimResChainPres)
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExtChain)
                          by fact+
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      moreover have "\<Psi> \<rhd> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q'"
                      proof -
                        have "\<Psi> \<rhd> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i))" unfolding resChainAppend
                          by(rule bisimReflexive)
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i)) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)))"
                          apply(rule bisimResChainPres)
                          apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExtChain)
                          by fact+
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i))) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> Q')" using `\<Psi> \<rhd> Q' \<sim> Q''` `xvec \<sharp>* \<Psi>`
                          unfolding yvrdef[symmetric] q''def[symmetric]
                          by(metis bisimResChainPres bisimTransitive bisimParComm bisimSymmetric bisimParPres)
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      moreover have "(\<Psi>, \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i),(\<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))) \<in> ?X2"
                      proof -
                        have "(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)" using `a\<^isub>i \<sharp> P\<^isub>i` by(subst fresh_bij)
                        note `b\<^isub>i \<sharp> \<Psi>`
                        moreover note `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`
                        moreover have "b\<^isub>i \<sharp> (P' \<parallel> (p\<bullet>Q\<^isub>i))" using `b\<^isub>i \<sharp> P'` `b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)`
                          by simp
                        moreover have "guarded(p\<bullet>P\<^isub>i)" using `guarded P\<^isub>i` by simp
                        moreover have "\<not> b\<^isub>i \<sharp> ([(p\<bullet>a\<^isub>i,b\<^isub>i)] \<bullet> (p\<bullet>M\<^isub>i))" using `\<not> a\<^isub>i \<sharp> M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "\<And> \<Psi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>" using EntCl
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="rev p",symmetric])
                          by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`)
                        moreover have "\<forall>F. (F \<otimes>\<^sub>F (\<langle>[b\<^isub>i],[(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)\<rangle>) \<simeq>\<^sub>F F)"
                          using Equiv1
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="rev p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "(\<forall>\<Psi> R. b\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i)) \<Leftarrow> R \<longrightarrow> (R = ((p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>)))" using Equiv2
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i` swap_simps)
                        moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P \<longrightarrow> (supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))) \<subseteq> ((supp \<Psi>)::name set))" using EntSupp
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "\<forall> P P'. ((([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P') = (P = P'))" using Inj
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "((supp(p\<bullet>P\<^isub>i))::name set) \<subseteq> supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                        moreover have "((supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)))::name set) = supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                        moreover have "(xvec@(p\<bullet>yvec)) \<sharp>* \<Psi>" using `xvec \<sharp>* \<Psi>` `(p\<bullet>yvec) \<sharp>* \<Psi>`
                          by simp
                        moreover from EntEnt have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i)\<parallel>\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="p",symmetric])
                          apply(simp add: eqvts swap_simps `distinctPerm p` `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                          done
                        ultimately show ?thesis
                          by blast
                      qed
                      ultimately show ?thesis
                        apply(rule_tac lol)
                        apply(rule refl)
                        apply(rule bisimSymmetric)
                        apply assumption
                        apply(rule disjI1)
                        apply assumption
                        by(rule bisimSymmetric)
                    qed
                  qed
                  ultimately show ?case
                    by(rule_tac exI) (rule conjI)
                next
                  case(cInv \<alpha> P')
                  show ?case
                    by(rule cInv) fact+
                next
                  case(cComm2 M xvec N P' K P'')
                  have "b \<sharp> N" and "b \<sharp> P'" using `\<Psi>' \<rhd> P \<longmapsto> M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'` `xvec \<sharp>* b` `b \<sharp> P` `distinct xvec` `xvec \<sharp>* M`
                    by(auto intro: outputFreshDerivative)
                  have "a \<sharp> xvec" and "b \<sharp> xvec" using `xvec \<sharp>* a` `xvec \<sharp>* b`
                    by simp+
                  have "\<exists> P'. \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K\<lparr>N\<rparr> \<prec> P' \<and> (\<Psi>, P', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by(rule_tac cComm2) (fact|simp add: `b \<sharp> N` `b \<sharp> xvec` `a \<sharp> xvec`)+
                  then obtain Q' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K \<lparr>N\<rparr> \<prec> Q' \<and> (\<Psi>, Q', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by(rule exE)
                  hence "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K \<lparr>N\<rparr> \<prec> Q'" and "(\<Psi>, Q', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by simp+
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<rhd> P \<longmapsto> M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'" using `\<Psi>' \<rhd> P \<longmapsto> M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'` by(subst (asm) cComm2)
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<one> \<rhd> P \<longmapsto> M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> P'"
                    apply(rule_tac statEqTransition[where \<Psi>="\<Psi>\<otimes>\<Psi>\<^isub>Q"])
                    apply(rule replTrans)
                    by(fact|simp|rule freshComp|rule AssertionStatEqSym[OF Identity])+
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<one> \<rhd> !P \<longmapsto> K\<lparr>N\<rparr> \<prec> Q'"
                    apply(rule statEqTransition)
                    by(fact) (rule AssertionStatEqSym[OF Identity])
                  obtain A\<^isub>P \<Psi>\<^isub>P where "extractFrame P = \<langle>A\<^isub>P,\<Psi>\<^isub>P\<rangle>" and "A\<^isub>P \<sharp>* \<Psi>" and "A\<^isub>P \<sharp>* \<Psi>\<^isub>Q" and "A\<^isub>P \<sharp>* !P" and "A\<^isub>P \<sharp>* P" and "A\<^isub>P \<sharp>* M" and "distinct A\<^isub>P"
                    by(rule freshFrame[where F="extractFrame P" and C="(\<Psi>,\<Psi>\<^isub>Q,P,M)"]) force
                  have "\<Psi>\<^isub>P \<simeq> \<one>" and "supp \<Psi>\<^isub>P = ({}::name set)" using `guarded P` `extractFrame P = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>`
                    by(auto dest: guardedStatEq)
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>P \<rhd> !P \<longmapsto> K\<lparr>N\<rparr> \<prec> Q'"
                    apply(rule statEqTransition)
                    apply fact
                    apply(rule compositionSym)
                    apply(rule AssertionStatEqSym)
                    by fact
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>P \<otimes> \<one> \<turnstile> M \<leftrightarrow> K"
                  proof -
                    from `\<Psi>' \<turnstile> M \<leftrightarrow> K` have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<turnstile> M \<leftrightarrow> K" using DisjointCCl[where M=M and N=K]
                      unfolding cComm2
                      apply(drule_tac cComm2)
                      apply(erule disjE)
                      by blast
                    thus ?thesis
                      apply(rule_tac statEqEnt)
                      apply(rule AssertionStatEqSym[OF AssertionStatEqTrans[OF compositionSym[OF Identity]]])
                      apply(rule AssertionStatEqTrans[OF compositionSym[OF `\<Psi>\<^isub>P \<simeq> \<one>`]])
                      by(rule Identity)
                  qed
                  have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<tau> \<prec> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q'"
                    apply(rule Bang)
                    apply(rule Comm2[where M=M and K=K and N=N and A\<^isub>P="A\<^isub>P" and \<Psi>\<^isub>P="\<Psi>\<^isub>P" and A\<^isub>Q="[]" and \<Psi>\<^isub>Q="\<one>"])
                    by(fact|rule freshCompChain|simp)+
                  moreover have "(\<Psi>, \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q', \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                  proof -
                    from `(\<Psi>, Q', \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2` obtain Q'' P''' where "\<Psi> \<rhd> Q' \<sim> Q''" and Disj: "(\<Psi>,Q'',P''') \<in> ?X2 \<or> \<Psi> \<rhd> Q'' \<sim> P'''" and "\<Psi> \<rhd> P''' \<sim> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                      by auto blast
                    have "xvec \<sharp>* \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>" using `xvec \<sharp>* a` `xvec \<sharp>* b` `xvec \<sharp>* \<Psi>\<^isub>c`
                      by simp
                    have rhs: "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>))"
                      apply(rule bisimTransitive[OF bisimResPres])
                      apply(rule bisimTransitive[OF bisimParComm])
                      apply(rule bisimSymmetric[OF bisimScopeExtChain])
                      apply fact+
                      apply(rule bisimTransitive[OF bisimResComm'])
                      apply fact
                      apply fact
                      apply(rule bisimResChainPres)
                      apply(rule bisimSymmetric)
                      apply(rule bisimTransitive[OF bisimSymmetric[OF bisimScopeExt]])
                      apply fact
                      apply(rule bisimResPres)
                      apply(rule bisimTransitive[OF bisimSymmetric[OF bisimParAssoc]])
                      apply(rule bisimTransitive[OF bisimParComm])
                      apply(rule bisimReflexive)
                      by fact+
                    from Disj show ?thesis
                    proof(rule disjE)
                      assume "\<Psi> \<rhd> Q'' \<sim> P'''"
                      have bis: "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q' \<sim> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>P' \<parallel> P'') \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)"
                        apply(rule bisimSymmetric)
                        apply(rule bisimTransitive[OF rhs])
                        apply(rule bisimResChainPres)
                        apply(rule bisimParPres')
                        apply(rule bisimTransitive[OF bisimSymmetric[OF `\<Psi> \<rhd> P''' \<sim> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`]])
                        apply(rule bisimSymmetric)
                        apply(rule bisimTransitive[OF `\<Psi> \<rhd> Q' \<sim> Q''`])
                        apply(rule `\<Psi> \<rhd> Q'' \<sim> P'''`)
                        by fact
                      show ?thesis
                        apply(rule lol)
                        apply(rule refl)
                        apply(rule bis)
                        apply(rule disjI2)
                        by(rule bisimReflexive)+
                    next
                      assume "(\<Psi>,Q'',P''') \<in> ?X2"
                      then obtain yvec Q\<^isub>i a\<^isub>i M\<^isub>i \<Psi>\<^isub>i P\<^isub>i where p'''def: "P''' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)" and q''def: "Q'' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i" and "a\<^isub>i \<sharp> \<Psi>" and "a\<^isub>i \<sharp> P\<^isub>i" and "a\<^isub>i \<sharp> Q\<^isub>i" and "\<not> a\<^isub>i \<sharp> M\<^isub>i" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i\<Leftarrow>P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a\<^isub>i], \<Psi>\<^isub>i\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i \<Leftarrow> R \<longrightarrow> (R = (P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>)))" and "guarded P\<^isub>i" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>i\<Leftarrow>P \<longrightarrow> (supp M\<^isub>i) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>i\<Leftarrow>P = M\<^isub>i\<Leftarrow>P') = (P = P'))" and "((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i" and "((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i" and "yvec \<sharp>* \<Psi>" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>i \<Leftarrow> P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
                        by blast
                      have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))))"
                      proof -
                        have "xvec \<sharp>* ([(a,b)] \<bullet> \<Psi>\<^isub>c)" using `a \<sharp> xvec` `b \<sharp> xvec` `xvec \<sharp>* \<Psi>\<^isub>c`
                          by simp
                        have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>b\<rparr>(\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                          apply(rule bisimResPres)
                          apply(rule bisimTransitive[OF bisimParComm])
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimResChainPres])
                          apply(rule bisimParComm)
                          apply fact
                          apply(rule bisimScopeExtChain)
                          by(fact|simp add: `xvec \<sharp>* ([(a,b)] \<bullet> \<Psi>\<^isub>c)`)+
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>(\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>b\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                          by(rule bisimResComm') fact+
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>b\<rparr>(P'\<parallel>P''\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>(P'\<parallel> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>))"
                          apply(rule bisimResChainPres)
                          apply(rule bisimTransitive[OF bisimResPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExt)
                          by fact+
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(P'\<parallel> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<sim> (\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))))"
                          apply(rule bisimResChainPres)
                          apply(rule bisimTransitive[OF bisimParComm])
                          apply(rule bisimSymmetric[OF bisimTransitive[OF bisimParComm]])
                          apply(rule bisimParPres)
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimSymmetric[OF `\<Psi> \<rhd> P''' \<sim> \<lparr>\<nu>b\<rparr>(P'' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`]])
                          apply(subst p'''def)
                          by(rule bisimReflexive) fact
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      obtain p::"name prm" where "set p \<subseteq> set yvec \<times> set(p\<bullet>yvec)" and "(p\<bullet>yvec) \<sharp>* a\<^isub>i" and "distinctPerm p" and "(p\<bullet>yvec) \<sharp>* M\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>\<^isub>i" and "(p\<bullet>yvec) \<sharp>* P\<^isub>i" and "(p\<bullet>yvec) \<sharp>* Q\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>" and "(p\<bullet>yvec) \<sharp>* P'" and "(p\<bullet>yvec) \<sharp>* xvec" and "(p\<bullet>yvec) \<sharp>* A\<^isub>Q"
                        by(rule chain.name_list_avoiding[where xvec=yvec and c="(M\<^isub>i,\<Psi>\<^isub>i,a\<^isub>i,P\<^isub>i,Q\<^isub>i,\<Psi>,P',xvec,A\<^isub>Q)"]) auto
                      hence yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>(p\<bullet>a\<^isub>i)\<rparr>(\<langle>(p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" and yvrdef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)"
                        by(subst resChainAlpha[where p=p],simp,simp,simp)+
                      obtain b\<^isub>i::name where "b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)" and "b\<^isub>i \<sharp> \<Psi>" and "b\<^isub>i \<sharp> \<Psi>\<^isub>Q" and "b\<^isub>i \<sharp> P'" and "b\<^isub>i \<sharp> A\<^isub>Q"
                        apply(rule at_exists_fresh[OF at_name_inst, where x="(p\<bullet>M\<^isub>i,p\<bullet>\<Psi>\<^isub>i,p\<bullet>P\<^isub>i,p\<bullet>Q\<^isub>i,\<Psi>,\<Psi>\<^isub>Q,P',A\<^isub>Q)"])
                        unfolding supp_prod pt_perm_supp[OF pt_name_inst,OF at_name_inst, symmetric]
                        apply(rule finite_UnI, (finite_guess | (subst pt_set_finite_ineq[OF pt_name_inst,OF at_name_inst], finite_guess)))+
                        apply(finite_guess)
                        by force
                      have yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" unfolding yvldef
                        by(subst alphaRes[where x="(p\<bullet>a\<^isub>i)" and y="b\<^isub>i"]) (simp add: `b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)`)+
                      have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (\<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                      proof -
                        note `\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((\<lparr>\<nu>*xvec\<rparr>(P'\<parallel>P''))\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))`
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)) \<sim> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                          unfolding yvldef
                          by(rule bisimReflexive)
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>P'\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)) \<sim> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>P'\<parallel>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)"
                          apply(subst resChainAppend)
                          apply(rule bisimResChainPres)
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExtChain)
                          by fact+
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      moreover have "\<Psi> \<rhd> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> \<lparr>\<nu>*xvec\<rparr>P' \<parallel> Q'"
                      proof -
                        have "\<Psi> \<rhd> \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i))" unfolding resChainAppend
                          by(rule bisimReflexive)
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i)) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)))"
                          apply(rule bisimResChainPres)
                          apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExtChain)
                          by fact+
                        moreover have "\<Psi> \<rhd> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i))) \<sim> \<lparr>\<nu>*xvec\<rparr>(P' \<parallel> Q')" using `\<Psi> \<rhd> Q' \<sim> Q''` `xvec \<sharp>* \<Psi>`
                          unfolding yvrdef[symmetric] q''def[symmetric]
                          by(metis bisimResChainPres bisimTransitive bisimParComm bisimSymmetric bisimParPres)
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      moreover have "(\<Psi>, \<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i),(\<lparr>\<nu>*(xvec@(p\<bullet>yvec))\<rparr>(P'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))) \<in> ?X2"
                      proof -
                        have "(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)" using `a\<^isub>i \<sharp> P\<^isub>i` by(subst fresh_bij)
                        note `b\<^isub>i \<sharp> \<Psi>`
                        moreover note `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`
                        moreover have "b\<^isub>i \<sharp> (P' \<parallel> (p\<bullet>Q\<^isub>i))" using `b\<^isub>i \<sharp> P'` `b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)`
                          by simp
                        moreover have "guarded(p\<bullet>P\<^isub>i)" using `guarded P\<^isub>i` by simp
                        moreover have "\<not> b\<^isub>i \<sharp> ([(p\<bullet>a\<^isub>i,b\<^isub>i)] \<bullet> (p\<bullet>M\<^isub>i))" using `\<not> a\<^isub>i \<sharp> M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "\<And> \<Psi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>" using EntCl
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="rev p",symmetric])
                          by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`)
                        moreover have "\<forall>F. (F \<otimes>\<^sub>F (\<langle>[b\<^isub>i],[(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)\<rangle>) \<simeq>\<^sub>F F)"
                          using Equiv1
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="rev p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "(\<forall>\<Psi> R. b\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i)) \<Leftarrow> R \<longrightarrow> (R = ((p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>)))" using Equiv2
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i` swap_simps)
                        moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P \<longrightarrow> (supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))) \<subseteq> ((supp \<Psi>)::name set))" using EntSupp
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "\<forall> P P'. ((([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P') = (P = P'))" using Inj
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "((supp(p\<bullet>P\<^isub>i))::name set) \<subseteq> supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                        moreover have "((supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)))::name set) = supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                        moreover have "(xvec@(p\<bullet>yvec)) \<sharp>* \<Psi>" using `xvec \<sharp>* \<Psi>` `(p\<bullet>yvec) \<sharp>* \<Psi>`
                          by simp
                        moreover from EntEnt have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i)\<parallel>\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="p",symmetric])
                          apply(simp add: eqvts swap_simps `distinctPerm p` `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                          done
                        ultimately show ?thesis
                          by blast
                      qed
                      ultimately show ?thesis
                        apply(rule_tac lol)
                        apply(rule refl)
                        apply(rule bisimSymmetric)
                        apply assumption
                        apply(rule disjI1)
                        apply assumption
                        by(rule bisimSymmetric)
                    qed
                  qed
                  ultimately show ?case
                    by(rule_tac exI) (rule conjI)
                next
                  case(cPar2 \<alpha> Q')
                  have "\<exists> P'. \<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P' \<and> (\<Psi>,P',\<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by(rule cPar2)fact+
                  then obtain P' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P' \<and> (\<Psi>,P',\<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by(rule_tac exE)
                  hence "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P'" and "(\<Psi>,P',\<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                    by simp+
                  obtain A\<^isub>P \<Psi>\<^isub>P where "extractFrame P = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>" and "distinct A\<^isub>P" and "A\<^isub>P \<sharp>* (\<Psi>\<otimes>\<Psi>\<^isub>Q)" and "A\<^isub>P \<sharp>* !P" and "A\<^isub>P \<sharp>* \<alpha>"
                    apply(rule_tac C="(\<Psi>\<otimes>\<Psi>\<^isub>Q,!P,\<alpha>)" in freshFrame)
                    by auto
                  from `extractFrame P = \<langle>A\<^isub>P, \<Psi>\<^isub>P\<rangle>` `guarded P` have "\<Psi>\<^isub>P \<simeq> \<one>" and "supp(\<Psi>\<^isub>P) = ({}::name set)"
                    by(auto dest: guardedStatEq)
                  have "(\<Psi> \<otimes> \<Psi>\<^isub>Q) \<simeq> ((\<Psi> \<otimes> \<Psi>\<^isub>Q) \<otimes> \<Psi>\<^isub>P)"
                    apply(rule AssertionStatEqTrans[OF AssertionStatEqSym[OF Identity]])
                    apply(rule compositionSym)
                    apply(rule AssertionStatEqSym)
                    by fact
                  have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> \<alpha> \<prec> P \<parallel> P'"
                    apply(rule Bang)
                    apply(rule Par2[where A\<^isub>P=A\<^isub>P and \<Psi>\<^isub>P=\<Psi>\<^isub>P])
                    apply(rule statEqTransition)
                    by fact+
                  moreover have "(\<Psi>,P\<parallel>P',\<lparr>\<nu>b\<rparr>(P \<parallel> Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2"
                  proof -
                    from `(\<Psi>,P',\<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)) \<in> ?Y2` obtain P'' Q'' where "\<Psi> \<rhd> P' \<sim> P''" and "(\<Psi>,P'',Q'') \<in> ?X2 \<or> \<Psi> \<rhd> P'' \<sim> Q''" and "\<Psi> \<rhd> Q'' \<sim>\<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                      by auto blast
                    from `(\<Psi>,P'',Q'') \<in> ?X2 \<or> \<Psi> \<rhd> P'' \<sim> Q''` show ?thesis
                    proof(rule disjE)
                      assume "\<Psi> \<rhd> P'' \<sim> Q''"
                      have "\<Psi> \<rhd> P\<parallel>P' \<sim> \<lparr>\<nu>b\<rparr>(P \<parallel> Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                      proof -
                        have "\<Psi> \<rhd> P\<parallel>P' \<sim> P\<parallel>Q''"
                          apply(rule bisimParPres')
                          by(rule bisimTransitive) fact+
                        moreover have "\<Psi> \<rhd> P\<parallel>Q'' \<sim> P \<parallel> \<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                          apply(rule bisimParPres')
                          by fact
                        moreover have "\<Psi> \<rhd> P \<parallel> \<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<lparr>\<nu>b\<rparr>(P \<parallel> Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimResPres])
                          apply(rule bisimParAssoc)
                          apply fact
                          apply(rule bisimScopeExt)
                          apply fact
                          done
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      thus ?thesis
                        apply(rule_tac lol)
                        apply(rule refl)
                        apply assumption
                        apply(rule disjI2)
                        by(rule bisimReflexive)+
                    next
                      assume "(\<Psi>,P'',Q'') \<in> ?X2"
                      then obtain yvec Q\<^isub>i a\<^isub>i M\<^isub>i \<Psi>\<^isub>i P\<^isub>i where q''def: "Q'' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)" and p''def: "P'' = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i" and "a\<^isub>i \<sharp> \<Psi>" and "a\<^isub>i \<sharp> P\<^isub>i" and "a\<^isub>i \<sharp> Q\<^isub>i" and "\<not> a\<^isub>i \<sharp> M\<^isub>i" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i\<Leftarrow>P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a\<^isub>i], \<Psi>\<^isub>i\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i \<Leftarrow> R \<longrightarrow> (R = (P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>)))" and "guarded P\<^isub>i" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>i\<Leftarrow>P \<longrightarrow> (supp M\<^isub>i) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>i\<Leftarrow>P = M\<^isub>i\<Leftarrow>P') = (P = P'))" and "((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i" and "((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i" and "yvec \<sharp>* \<Psi>" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>i \<Leftarrow> P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
                        by blast
                      have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((P\<parallel>Q')\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (P\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)))"
                      proof -
                        have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>(P\<parallel>Q'\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> P\<parallel> \<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>)"
                          apply(rule bisimTransitive[OF bisimResPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExt)
                          by fact
                        moreover have "\<Psi> \<rhd> P\<parallel> \<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> P\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))"
                          apply(rule bisimTransitive[OF bisimParComm])
                          apply(rule bisimSymmetric[OF bisimTransitive[OF bisimParComm]])
                          apply(rule bisimParPres)
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimSymmetric[OF `\<Psi> \<rhd> Q'' \<sim> \<lparr>\<nu>b\<rparr>(Q' \<parallel> \<lbrace>([(a, b)] \<bullet> \<Psi>\<^isub>c)\<rbrace>)`]])
                          apply(subst q''def)
                          by(rule bisimReflexive)
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      obtain p::"name prm" where "set p \<subseteq> set yvec \<times> set(p\<bullet>yvec)" and "(p\<bullet>yvec) \<sharp>* a\<^isub>i" and "distinctPerm p" and "(p\<bullet>yvec) \<sharp>* M\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>\<^isub>i" and "(p\<bullet>yvec) \<sharp>* P\<^isub>i" and "(p\<bullet>yvec) \<sharp>* Q\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>" and "(p\<bullet>yvec) \<sharp>* P'" and "(p\<bullet>yvec) \<sharp>* P" and "(p\<bullet>yvec) \<sharp>* A\<^isub>Q"
                        by(rule chain.name_list_avoiding[where xvec=yvec and c="(M\<^isub>i,\<Psi>\<^isub>i,a\<^isub>i,P\<^isub>i,Q\<^isub>i,\<Psi>,P',P,A\<^isub>Q)"]) auto
                      hence yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>(p\<bullet>a\<^isub>i)\<rparr>(\<langle>(p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" and yvrdef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)"
                        by(subst resChainAlpha[where p=p],simp,simp,simp)+
                      obtain b\<^isub>i::name where "b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)" and "b\<^isub>i \<sharp> \<Psi>" and "b\<^isub>i \<sharp> \<Psi>\<^isub>Q" and "b\<^isub>i \<sharp> P" and "b\<^isub>i \<sharp> A\<^isub>Q"
                        apply(rule at_exists_fresh[OF at_name_inst, where x="(p\<bullet>M\<^isub>i,p\<bullet>\<Psi>\<^isub>i,p\<bullet>P\<^isub>i,p\<bullet>Q\<^isub>i,\<Psi>,\<Psi>\<^isub>Q,P,A\<^isub>Q)"])
                        unfolding supp_prod pt_perm_supp[OF pt_name_inst,OF at_name_inst, symmetric]
                        apply(rule finite_UnI, (finite_guess | (subst pt_set_finite_ineq[OF pt_name_inst,OF at_name_inst], finite_guess)))+
                        apply(finite_guess)
                        by force
                      have yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" unfolding yvldef
                        by(subst alphaRes[where x="(p\<bullet>a\<^isub>i)" and y="b\<^isub>i"]) (simp add: `b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)`)+
                      have "\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((P\<parallel>Q')\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                      proof -
                        note `\<Psi> \<rhd> \<lparr>\<nu>b\<rparr>((P\<parallel>Q')\<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> P\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>))`
                        moreover have "\<Psi> \<rhd> P\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)) \<sim> P\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                          unfolding yvldef
                          by(rule bisimReflexive)
                        moreover have "\<Psi> \<rhd> P\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)) \<sim> \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>P\<parallel>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)"
                          apply(rule bisimSymmetric)
                          apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExtChain)
                          by fact+
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      moreover have "\<Psi> \<rhd> \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> P \<parallel> P'"
                      proof -
                        have "\<Psi> \<rhd> \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> P \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i))"
                          apply(rule bisimTransitive[OF bisimResChainPres[OF bisimParAssoc]])
                          apply fact
                          apply(rule bisimScopeExtChain)
                          by fact+
                        moreover have "\<Psi> \<rhd> P \<parallel> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)) \<sim> P \<parallel> P'" using `\<Psi> \<rhd> P' \<sim> P''`
                          unfolding yvrdef[symmetric] p''def[symmetric]
                          by(rule_tac bisimParPres'[OF bisimSymmetric])
                        ultimately show ?thesis
                          by(auto dest: bisimTransitive)
                      qed
                      moreover have "(\<Psi>, \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i),(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(P\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))) \<in> ?X2"
                      proof -
                        have "(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)" using `a\<^isub>i \<sharp> P\<^isub>i` by(subst fresh_bij)
                        note `b\<^isub>i \<sharp> \<Psi>`
                        moreover note `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`
                        moreover have "b\<^isub>i \<sharp> (P \<parallel> (p\<bullet>Q\<^isub>i))" using `b\<^isub>i \<sharp> P` `b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)`
                          by simp
                        moreover have "guarded(p\<bullet>P\<^isub>i)" using `guarded P\<^isub>i` by simp
                        moreover have "\<not> b\<^isub>i \<sharp> ([(p\<bullet>a\<^isub>i,b\<^isub>i)] \<bullet> (p\<bullet>M\<^isub>i))" using `\<not> a\<^isub>i \<sharp> M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "\<And> \<Psi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>" using EntCl
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="rev p",symmetric])
                          by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`)
                        moreover have "\<forall>F. (F \<otimes>\<^sub>F (\<langle>[b\<^isub>i],[(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)\<rangle>) \<simeq>\<^sub>F F)"
                          using Equiv1
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="rev p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "(\<forall>\<Psi> R. b\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i)) \<Leftarrow> R \<longrightarrow> (R = ((p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>)))" using Equiv2
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i` swap_simps)
                        moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P \<longrightarrow> (supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))) \<subseteq> ((supp \<Psi>)::name set))" using EntSupp
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "\<forall> P P'. ((([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P') = (P = P'))" using Inj
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps)
                        moreover have "((supp(p\<bullet>P\<^isub>i))::name set) \<subseteq> supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts swap_simps `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                        moreover have "((supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)))::name set) = supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i`
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst (asm) perm_bool[where pi="p",symmetric])
                          by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                        moreover note `(p\<bullet>yvec) \<sharp>* \<Psi>`
                        moreover from EntEnt have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i)\<parallel>\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                          apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                          apply(subst perm_bool[where pi="p",symmetric])
                          apply(simp add: eqvts swap_simps `distinctPerm p` `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                          done
                        ultimately show ?thesis
                          by blast
                      qed
                      ultimately show ?thesis
                        apply(rule_tac lol)
                        apply(rule refl)
                        apply(rule bisimSymmetric)
                        apply assumption
                        apply(rule disjI1)
                        apply assumption
                        by(rule bisimSymmetric)
                    qed
                  qed
                  ultimately show ?case
                    by(rule_tac exI) (rule conjI)
                qed
                thus ?case unfolding alphadef
                  by -
              qed
            qed
            then obtain P'' where "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K\<lparr>N\<rparr> \<prec> P'' \<and> (\<Psi>,P'',P') \<in> ?Y2"
              by(rule_tac exE)
            hence "\<Psi> \<otimes> \<Psi>\<^isub>Q \<rhd> !P \<longmapsto> K\<lparr>N\<rparr> \<prec> P''" and "(\<Psi>,P'',P') \<in> ?Y2"
              by simp+
            have "\<Psi> \<otimes> \<one> \<rhd> Q \<longmapsto> M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q'"
            proof -
              have "b \<sharp> \<Psi> \<otimes> \<one>" using `b \<sharp> \<Psi>` by force
              from `\<Psi> \<otimes> \<Psi>\<^isub>P \<rhd> Q \<longmapsto> M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q'` have "(\<Psi> \<otimes> \<one>) \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rhd> Q \<longmapsto> M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q'" unfolding `\<Psi>\<^isub>P = \<one> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)`
                by(rule statEqTransition) (rule AssertionStatEqSym[OF Associativity])
              thus ?thesis
                by(rule replTrans) (fact|simp)+
            qed
            have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<one> \<turnstile> M \<leftrightarrow> K"
            proof -
              have "\<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<one> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<simeq> ((\<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<one>) \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c))"
                apply(rule AssertionStatEqTrans[OF AssertionStatEqSym[OF Associativity]])
                apply(rule AssertionStatEqTrans[OF AssertionStatEqSym[OF Associativity]])
                apply(rule Composition)
                by(rule Associativity)
              with `\<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<Psi>\<^isub>P \<turnstile> M \<leftrightarrow> K` have "(\<Psi> \<otimes> \<Psi>\<^isub>Q \<otimes> \<one>) \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c) \<turnstile> M \<leftrightarrow> K" unfolding `\<Psi>\<^isub>P = \<one> \<otimes> ([(a,b)]\<bullet>\<Psi>\<^isub>c)`
                by(rule_tac statEqEnt)
              thus ?thesis using DisjointCCl
                by(auto dest: EntEnt')
            qed
            have "A\<^isub>Q \<sharp>* !P" using `A\<^isub>Q \<sharp>* (a, M\<^isub>a, \<Psi>\<^isub>c, P)` by simp
            have "xvec \<sharp>* !P" using `xvec \<sharp>* (a, M\<^isub>a, \<Psi>\<^isub>c, P)` by simp
            have "\<Psi> \<rhd> Q \<parallel> !P \<longmapsto> \<tau> \<prec> \<lparr>\<nu>*xvec\<rparr>(Q' \<parallel> P'')"
              apply(rule Comm2)
              apply(rule `\<Psi> \<otimes> \<one> \<rhd> Q \<longmapsto> M\<lparr>\<nu>*xvec\<rparr>\<langle>N\<rangle> \<prec> Q'`)
              by(fact|simp)+
            moreover have "(\<Psi>, \<lparr>\<nu>*xvec\<rparr>Q' \<parallel> P'',\<lparr>\<nu>*xvec\<rparr>Q'\<parallel>P') \<in> ?Y"
            proof -
              from `(\<Psi>, P'', P') \<in> ?Y2` obtain P''1 P'1 where "\<Psi> \<rhd> P'' \<sim> P''1" and "(\<Psi>,P''1,P'1) \<in> ?X2 \<or> \<Psi> \<rhd> P''1 \<sim> P'1" and "\<Psi> \<rhd> P'1 \<sim> P'"
                by auto blast
              from `(\<Psi>,P''1,P'1) \<in> ?X2 \<or> \<Psi> \<rhd> P''1 \<sim> P'1` have "(\<Psi>, Q' \<parallel> P'', Q'\<parallel> P') \<in> ?Y2"
              proof(rule disjE)
                assume "\<Psi> \<rhd> P''1 \<sim> P'1"
                hence "\<Psi> \<rhd> Q' \<parallel> P''1 \<sim> Q' \<parallel> P'1"
                  by(rule bisimParPres')
                thus ?thesis
                  apply(rule_tac lol)
                  apply(rule refl)
                  apply(rule bisimParPres'[OF `\<Psi> \<rhd> P'' \<sim> P''1`])
                  apply(rule disjI2)
                  apply assumption
                  apply(rule bisimParPres'[OF `\<Psi> \<rhd> P'1 \<sim> P'`])
                  done
              next
                assume "(\<Psi>,P''1,P'1) \<in> ?X2"
                then obtain yvec Q\<^isub>i a\<^isub>i M\<^isub>i \<Psi>\<^isub>i P\<^isub>i where p'1def: "P'1 = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)" and p''1def: "P''1 = \<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i" and "a\<^isub>i \<sharp> \<Psi>" and "a\<^isub>i \<sharp> P\<^isub>i" and "a\<^isub>i \<sharp> Q\<^isub>i" and "\<not> a\<^isub>i \<sharp> M\<^isub>i" and EntCl: "\<And> \<Psi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i\<Leftarrow>P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>" and Equiv1: "\<And> F. (F \<otimes>\<^sub>F \<langle>[a\<^isub>i], \<Psi>\<^isub>i\<rangle> \<simeq>\<^sub>F F)" and Equiv2: "(\<forall>\<Psi> R. a\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> M\<^isub>i \<Leftarrow> R \<longrightarrow> (R = (P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle>)))" and "guarded P\<^isub>i" and  EntSupp: "\<forall>\<Psi> P. (\<Psi> \<turnstile> M\<^isub>i\<Leftarrow>P \<longrightarrow> (supp M\<^isub>i) \<subseteq> ((supp \<Psi>)::name set))" and  Inj: "\<forall> P P'. ((M\<^isub>i\<Leftarrow>P = M\<^isub>i\<Leftarrow>P') = (P = P'))" and "((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i" and "((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i" and "yvec \<sharp>* \<Psi>" and EntEnt: "\<And> \<Psi> \<phi>. \<Psi> \<otimes> \<Psi>\<^isub>i \<turnstile> \<phi> \<Longrightarrow> (\<phi> = M\<^isub>i \<Leftarrow> P\<^isub>i \<parallel> \<langle>M\<^isub>i\<rangle> \<or> \<Psi> \<turnstile> \<phi>)"
                  by blast
                have "\<Psi> \<rhd> Q' \<parallel> P' \<sim> (Q'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)))"
                  apply(rule bisimParPres')
                  apply(rule bisimTransitive[OF bisimSymmetric])
                  apply fact
                  unfolding p'1def
                  by(rule bisimReflexive)
                obtain p::"name prm" where "set p \<subseteq> set yvec \<times> set(p\<bullet>yvec)" and "(p\<bullet>yvec) \<sharp>* a\<^isub>i" and "distinctPerm p" and "(p\<bullet>yvec) \<sharp>* M\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>\<^isub>i" and "(p\<bullet>yvec) \<sharp>* P\<^isub>i" and "(p\<bullet>yvec) \<sharp>* Q\<^isub>i" and "(p\<bullet>yvec) \<sharp>* \<Psi>" and "(p\<bullet>yvec) \<sharp>* P'" and "(p\<bullet>yvec) \<sharp>* Q'" and "(p\<bullet>yvec) \<sharp>* P''" and "(p\<bullet>yvec) \<sharp>* A\<^isub>Q"
                  by(rule chain.name_list_avoiding[where xvec=yvec and c="(M\<^isub>i,\<Psi>\<^isub>i,a\<^isub>i,P\<^isub>i,Q\<^isub>i,\<Psi>,P',Q',P'',A\<^isub>Q)"]) auto
                hence yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>(p\<bullet>a\<^isub>i)\<rparr>(\<langle>(p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>(p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" and yvrdef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> !P\<^isub>i = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> !(p\<bullet>P\<^isub>i)"
                  by(subst resChainAlpha[where p=p],simp,simp,simp)+
                obtain b\<^isub>i::name where "b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)" and "b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)" and "b\<^isub>i \<sharp> \<Psi>" and "b\<^isub>i \<sharp> \<Psi>\<^isub>Q" and "b\<^isub>i \<sharp> Q'" and "b\<^isub>i \<sharp> A\<^isub>Q"
                  apply(rule at_exists_fresh[OF at_name_inst, where x="(p\<bullet>M\<^isub>i,p\<bullet>\<Psi>\<^isub>i,p\<bullet>P\<^isub>i,p\<bullet>Q\<^isub>i,\<Psi>,\<Psi>\<^isub>Q,Q',A\<^isub>Q)"])
                  unfolding supp_prod pt_perm_supp[OF pt_name_inst,OF at_name_inst, symmetric]
                  apply(rule finite_UnI, (finite_guess | (subst pt_set_finite_ineq[OF pt_name_inst,OF at_name_inst], finite_guess)))+
                  apply(finite_guess)
                  by force
                have yvldef: "\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>) = \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)" unfolding yvldef
                  by(subst alphaRes[where x="(p\<bullet>a\<^isub>i)" and y="b\<^isub>i"]) (simp add: `b\<^isub>i \<sharp> (p\<bullet>M\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>\<Psi>\<^isub>i)`)+
                have "\<Psi> \<rhd> Q' \<parallel> P' \<sim> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(Q'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                proof -
                  note `\<Psi> \<rhd> Q' \<parallel> P' \<sim> (Q'\<parallel>(\<lparr>\<nu>*yvec\<rparr>Q\<^isub>i \<parallel> \<lparr>\<nu>a\<^isub>i\<rparr>(\<langle>M\<^isub>i\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>i\<rbrace>)))`
                  hence "\<Psi> \<rhd> Q' \<parallel> P' \<sim> (Q'\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>)))"
                    unfolding yvldef
                    by -
                  moreover have "\<Psi> \<rhd> (Q'\<parallel>(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(p\<bullet>Q\<^isub>i) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))) \<sim> (\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(Q'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))"
                    apply(rule bisimTransitive[OF bisimSymmetric[OF bisimScopeExtChain]])
                    apply fact
                    apply fact
                    apply(rule bisimResChainPres)
                    apply(rule bisimSymmetric[OF bisimParAssoc])
                    by fact
                  ultimately show ?thesis
                    by(auto dest: bisimTransitive)
                qed
                moreover have "\<Psi> \<rhd> \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(Q' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i) \<sim> Q' \<parallel> P''"
                  apply(rule bisimSymmetric)
                  apply(rule bisimTransitive[OF bisimParPres'])
                  apply fact
                  unfolding p''1def yvrdef
                  apply(rule bisimTransitive[OF bisimSymmetric[OF bisimScopeExtChain]])
                  apply fact
                  apply fact
                  apply(rule bisimResChainPres)
                  apply(rule bisimSymmetric[OF bisimParAssoc])
                  by fact
                moreover have "(\<Psi>, \<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(Q' \<parallel> (p\<bullet>Q\<^isub>i)) \<parallel> !(p\<bullet>P\<^isub>i),(\<lparr>\<nu>*(p\<bullet>yvec)\<rparr>(Q'\<parallel>(p\<bullet>Q\<^isub>i)) \<parallel> \<lparr>\<nu>b\<^isub>i\<rparr>(\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>M\<^isub>i)\<rangle> \<parallel> \<lbrace>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>p\<bullet>\<Psi>\<^isub>i)\<rbrace>))) \<in> ?X2"
                proof -
                  have "(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)" using `a\<^isub>i \<sharp> P\<^isub>i` by(subst fresh_bij)
                  note `b\<^isub>i \<sharp> \<Psi>`
                  moreover note `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`
                  moreover have "b\<^isub>i \<sharp> (Q' \<parallel> (p\<bullet>Q\<^isub>i))" using `b\<^isub>i \<sharp> Q'` `b\<^isub>i \<sharp> (p\<bullet>Q\<^isub>i)`
                    by simp
                  moreover have "guarded(p\<bullet>P\<^isub>i)" using `guarded P\<^isub>i` by simp
                  moreover have "\<not> b\<^isub>i \<sharp> ([(p\<bullet>a\<^isub>i,b\<^isub>i)] \<bullet> (p\<bullet>M\<^isub>i))" using `\<not> a\<^isub>i \<sharp> M\<^isub>i`
                    apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps)
                  moreover have "\<And> \<Psi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>" using EntCl
                    apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                    apply(subst perm_bool[where pi="rev p",symmetric])
                    by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> (p\<bullet>P\<^isub>i)` `b\<^isub>i \<sharp> (p\<bullet>P\<^isub>i)`)
                  moreover have "\<forall>F. (F \<otimes>\<^sub>F (\<langle>[b\<^isub>i],[(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)\<rangle>) \<simeq>\<^sub>F F)"
                    using Equiv1
                    apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                    apply(subst perm_bool[where pi="rev p",symmetric])
                    by(simp add: eqvts swap_simps)
                  moreover have "(\<forall>\<Psi> R. b\<^isub>i \<sharp> \<Psi> \<longrightarrow> \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i)) \<Leftarrow> R \<longrightarrow> (R = ((p\<bullet>P\<^isub>i) \<parallel> \<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle>)))" using Equiv2
                    apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i` swap_simps)
                  moreover have "\<forall>\<Psi> P. (\<Psi> \<turnstile> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P \<longrightarrow> (supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))) \<subseteq> ((supp \<Psi>)::name set))" using EntSupp
                    apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps)
                  moreover have "\<forall> P P'. ((([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>P') = (P = P'))" using Inj
                    apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps)
                  moreover have "((supp(p\<bullet>P\<^isub>i))::name set) \<subseteq> supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp P\<^isub>i)::name set) \<subseteq> supp M\<^isub>i`
                    apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts swap_simps `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                  moreover have "((supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)))::name set) = supp ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))" using `((supp \<Psi>\<^isub>i)::name set) = supp M\<^isub>i`
                    apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                    apply(subst (asm) perm_bool[where pi="p",symmetric])
                    by(simp add: eqvts at2[OF at_name_inst] at3[OF at_name_inst])
                  moreover note `(p\<bullet>yvec) \<sharp>* \<Psi>`
                  moreover from EntEnt have "\<forall> \<Psi> \<phi>. \<Psi> \<otimes> ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>\<Psi>\<^isub>i)) \<turnstile> \<phi> \<longrightarrow> \<phi> = ([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<Leftarrow>(p\<bullet>P\<^isub>i)\<parallel>\<langle>([(p\<bullet>a\<^isub>i,b\<^isub>i)]\<bullet>(p\<bullet>M\<^isub>i))\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
                    apply(subst perm_bool[where pi="[(p\<bullet>a\<^isub>i,b\<^isub>i)]", symmetric])
                    apply(subst perm_bool[where pi="p",symmetric])
                    apply(simp add: eqvts swap_simps `distinctPerm p` `(p\<bullet>a\<^isub>i) \<sharp> p\<bullet>P\<^isub>i` `b\<^isub>i \<sharp> p\<bullet>P\<^isub>i`)
                    done
                  ultimately show ?thesis
                    by blast
                qed
                ultimately show ?thesis
                  apply(rule_tac lol)
                  apply(rule refl)
                  apply(rule bisimSymmetric)
                  apply assumption
                  apply(rule disjI1)
                  apply assumption
                  by(rule bisimSymmetric)
              qed
              hence "(\<Psi>, Q' \<parallel> P'', Q' \<parallel> P') \<in> ?Y" by(rule Y2inY)
              thus ?thesis using `xvec \<sharp>* \<Psi>` by(rule YresChainClose')
            qed
            ultimately show ?case
              by(rule_tac exI) (rule conjI)
          qed
        qed
      qed
      thus ?thesis unfolding `S = \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> !P)` `T = \<lparr>\<nu>*xvec\<rparr>(Q \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>))`
        by -
    qed
  qed
}
note Goal = this
have "\<Psi> \<rhd> \<zero> \<parallel> \<lparr>\<nu>a\<rparr>(\<langle>M\<^isub>a\<rangle> \<parallel> \<lbrace>\<Psi>\<^isub>c\<rbrace>) \<sim> \<zero> \<parallel> !P"
proof -
  obtain b::name where "b \<sharp> M\<^isub>a" and "b \<sharp> \<Psi>\<^isub>c" and "b \<sharp> P" and "b \<sharp> \<Psi>"
    by(generate_fresh "name") force
  {
    fix \<Psi>
    have "\<Psi> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>" using EntCl[where \<Psi>="[(a,b)]\<bullet>\<Psi>"]
      apply(subst perm_bool[where pi="[(a,b)]",symmetric])
      by(simp add: eqvts `a \<sharp> P` `b \<sharp> P`)
  }
  {
    fix F
    have "F \<otimes>\<^sub>F \<langle>[b],([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rangle> \<simeq>\<^sub>F F" using Equiv1[where F="[(a,b)]\<bullet>F"]
      apply(subst perm_bool[where pi="[(a,b)]",symmetric])
      by(simp add: eqvts swap_simps)
  }
  {
    fix \<Psi> R
    assume "b \<sharp> \<Psi>"
      and  "\<Psi> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>R"
    hence "([(a,b)]\<bullet>R) = P \<parallel> \<langle>M\<^isub>a\<rangle>"
      apply(rule_tac Equiv2[where \<Psi>'2="[(a,b)]\<bullet>\<Psi>"])
      by(subst perm_bool[where pi="[(a,b)]",symmetric],simp add: eqvts swap_simps)+
    hence "R = P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle>"
      by(subst perm_bool[where pi="[(a,b)]",symmetric],simp add: eqvts swap_simps `a \<sharp> P` `b \<sharp> P`)
  }
  {
    fix \<Psi> P
    assume "\<Psi> \<turnstile> ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P"
    hence "(supp M\<^isub>a::name set) \<subseteq> supp([(a,b)]\<bullet>\<Psi>)"
      apply(rule_tac EntSupp[where P="[(a,b)]\<bullet>P"])
      by(subst perm_bool[where pi="[(a,b)]",symmetric],simp add: eqvts swap_simps)
    hence "supp ([(a, b)] \<bullet> M\<^isub>a) \<subseteq> (supp \<Psi>::name set)"
      by(subst perm_bool[where pi="[(a,b)]",symmetric],simp add: eqvts swap_simps)
  }
  {
    fix P P'
    have "(([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P = ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P') = (P = P')" using Inj[where P="[(a,b)]\<bullet>P" and P'="[(a,b)]\<bullet>P'"]
      by(subst perm_bool[where pi="[(a,b)]",symmetric], simp add: eqvts)
  }
  {
    fix \<Psi> \<phi>
    assume "\<Psi> \<otimes> ([(a, b)] \<bullet> \<Psi>\<^isub>c) \<turnstile> \<phi>"
    hence "([(a,b)]\<bullet>\<phi>) = M\<^isub>a\<Leftarrow>P \<parallel> \<langle>M\<^isub>a\<rangle> \<or> ([(a,b)]\<bullet>\<Psi>) \<turnstile> ([(a,b)]\<bullet>\<phi>)"
      apply(rule_tac EntDisj)
      by(subst perm_bool[where pi="[(a,b)]",symmetric],simp add: eqvts swap_simps `a \<sharp> P` `b \<sharp> P`)
    hence "\<phi> = ([(a, b)] \<bullet> M\<^isub>a)\<Leftarrow>P \<parallel> \<langle>([(a, b)] \<bullet> M\<^isub>a)\<rangle> \<or> \<Psi> \<turnstile> \<phi>"
      by(subst perm_bool[where pi="[(a,b)]",symmetric],simp add: eqvts swap_simps `a \<sharp> P` `b \<sharp> P`)
  }
 have "\<not> b \<sharp> [(a, b)] \<bullet> M\<^isub>a" using `\<not> a \<sharp> M\<^isub>a`
   by(subst perm_bool[where pi="[(a,b)]",symmetric],simp add: eqvts swap_simps `a \<sharp> P` `b \<sharp> P`)
 have "(supp P::name set) \<subseteq> supp ([(a, b)] \<bullet> M\<^isub>a)" using `supp P \<subseteq> supp M\<^isub>a`
   by(subst perm_bool[where pi="[(a,b)]",symmetric],simp add: eqvts swap_simps `a \<sharp> P` `b \<sharp> P`)
 have "(supp ([(a, b)] \<bullet> \<Psi>\<^isub>c)::name set) = supp ([(a, b)] \<bullet> M\<^isub>a)" using `supp \<Psi>\<^isub>c = supp M\<^isub>a`
   by(subst perm_bool[where pi="[(a,b)]",symmetric],simp add: eqvts swap_simps `a \<sharp> P` `b \<sharp> P`)
  have "\<Psi> \<rhd> \<zero> \<parallel> \<lparr>\<nu>b\<rparr>(\<langle>([(a,b)]\<bullet>M\<^isub>a)\<rangle> \<parallel> \<lbrace>([(a,b)]\<bullet>\<Psi>\<^isub>c)\<rbrace>) \<sim> \<zero> \<parallel> !P"
    by(rule Goal) fact+
  thus ?thesis
    apply(subst alphaRes[where y=b])
    apply(simp add: `b \<sharp> M\<^isub>a` `b \<sharp> \<Psi>\<^isub>c`)
    unfolding eqvts by -
qed
thus ?thesis
  by(metis bisimParNil bisimTransitive bisimSymmetric bisimParComm)
qed

end
end